#include "DataObj.h"

long DataObj::inputs_repeated_every = 0;
bool DataObj::binary_data_files = 0;

DataObj::DataObj(int argc, char *argv[])
{
   int i, j;
   if( argc < 2 )
   {
      cout << "Must specify data input filenames." << endl;
      cout << "Type \"" << argv[0] << " help\" for command line parameters." << endl;
      exit(1);
   }
   preliminary_DataObj_initialisation();
   for( i = 1; i < argc; i++ )
   {
      if ( strstr( argv[i], "help" ) )
      {
 	 cout << "Usage: " << argv[0] << " [options]" << endl;
 	 cout << "inputfile=" << InputFilename << endl;
         cout << "targetfile=" << TargetFilename << endl;
         cout << "inputrangefile" << InputRangeFilename << "   (If required, a file containing the minimum and maximum range values for each input dimension)" << endl;
         cout << "missingvaluesfile=" << MissingValuesFilename << endl;      
         cout << "missingyvaluesfile=" << MissingYValuesFilename << endl;
         cout << "pcafile=" << PCAFilename << endl;
         cout << "pcapartitionfile=" << PCAPartitionFilename << endl;
         cout << "pcaoutputcols=" << pca_output_cols << "   (The number of principal components allowed to be written to PCAFile)" << endl;
         cout << "datablocksizefile=" << DataBlockSizesFilename << "   (File holding the data_block_sizes for each experiment)" << endl;
         cout << "datablocksize=" << data_block_size << "   (The number of data points which should be regarded as one observation)" << endl;
         cout << "inputsrepeatedevery=" << inputs_repeated_every << endl;
         cout << "inputcolumns=" << p << endl;
         cout << "targetcolumns=" << p_y << endl;
         cout << "transposetargets" << transpose_targets << endl;
         cout << "inputcolumns=" << p << endl;
         cout <<  "normalisepredictors=" << normalise_predictors << "   (If 1, rescales the predictors to the range [0,1]; if 2, standardizes the predictors)" << endl;
         cout <<  "normalisetargets=" << normalise_targets << "   (If non zero, subtracts the mean of each column of the target matrix from the observed values, and divide by the range)" << endl;
         cout << "deletepoint=" << delete_point << "   (If non negative, remove the data point number delete_point, and if inputs_repeated_every>0, also remove delete_point + i * inputs_repeated_every, i=1,2,....  Useful for leave one out CV.)" << endl;
	 cout << "deletemissing" << delete_missing << "   (If true, delete all missing values from the data.)" << endl;
         cout << "binarydatafiles=" << binary_data_files << endl;
//         cout << "densityestimation=" << density_estimation <<"   (If true, we are doing density estimation and certain options should be switched off..)" << endl;
         cout << "printtoscreen=" << print_to_screen << "   (If not true, print minimal amount of information to screen)" << endl;
         return;
      }
      if ( InitialStringMatch( argv[i], "inputfile=" ) )
	 StringToString( InputFilename, argv[i] );
      if ( InitialStringMatch( argv[i], "targetfile=" ) )
	 StringToString( TargetFilename, argv[i] );
      if ( InitialStringMatch( argv[i], "inputrangefile=" ) )
	 StringToString( InputRangeFilename, argv[i] );
      if ( InitialStringMatch( argv[i], "missingvaluesfile=" ) )
	 StringToString( MissingValuesFilename, argv[i] );
      if ( InitialStringMatch( argv[i], "missingyvaluesfile=" ) )
	 StringToString( MissingYValuesFilename, argv[i] );
      if ( InitialStringMatch( argv[i], "pcafile=" ) )
	 StringToString( PCAFilename, argv[i] );
      if ( InitialStringMatch( argv[i], "pcapartitionfile=" ) )
	 StringToString( PCAPartitionFilename, argv[i] );
      if ( InitialStringMatch( argv[i], "pcaoutputcols=" ) )
         StringToLong( &pca_output_cols, argv[i] );
      if ( InitialStringMatch( argv[i], "datablocksizefile=" ) )
	 StringToString( DataBlockSizesFilename, argv[i] );
      if ( InitialStringMatch( argv[i], "datablocksize=" ) )
	StringToLong( &data_block_size, argv[i] );
      if ( InitialStringMatch( argv[i], "inputsrepeatedevery=" ) )
	StringToLong( &inputs_repeated_every, argv[i] );
      if ( InitialStringMatch( argv[i], "inputcolumns=" ) )
         StringToLong( &p, argv[i] );
      if ( InitialStringMatch( argv[i], "targetcolumns=" ) )
         StringToLong( &p_y, argv[i] );
      if ( InitialStringMatch( argv[i], "transposetargets=" ) )
	 StringToBool( &transpose_targets, argv[i] );
      if ( InitialStringMatch( argv[i], "inputcolumns=" ) )
         StringToLong( &p, argv[i] );
      if ( InitialStringMatch( argv[i], "normalisepredictors=" ) )
         StringToLong( &normalise_predictors, argv[i] );
      if ( InitialStringMatch( argv[i], "normalisetargets=" ) )
         StringToBool( &normalise_targets, argv[i] );
      if ( InitialStringMatch( argv[i], "deletepoint=" ) )
	StringToLong( &delete_point, argv[i] );
      if ( InitialStringMatch( argv[i], "deletemissing=" ) )
         StringToBool( &delete_missing, argv[i] );  
      if ( InitialStringMatch( argv[i], "binarydatafiles=" ) )
	 StringToBool( &binary_data_files, argv[i] );
      if( InitialStringMatch( argv[i], "densityestimation=" ) )
        StringToBool( &density_estimation, argv[i] );
      if ( InitialStringMatch( argv[i], "printtoscreen=" ) )
         StringToBool( &print_to_screen, argv[i] );
   }
   if( strlen( TargetFilename ) == 0 )
   {
     strcpy(TargetFilename, argv[1]);
     cout << "Must specify target filename and (if appropriate) predictor filename." << endl;
     cout << "Assuming first argument contains target matrix file" << endl;
     if(argc > 2 && !density_estimation)
       {
	 strcpy(InputFilename, argv[2]);
	 cout << "Assuming second argument contains input matrix file" << endl;
       }
   }

   Y = scan_data( TargetFilename, binary_data_files, transpose_targets, p_y );
   n = Y->size1;
   if( !p_y )
      p_y = Y->size2;

   bool data_block_sizes_repetitive = false;
   if( strlen( DataBlockSizesFilename ) )
   {
      ifstream DataBlockSizeFile( DataBlockSizesFilename, ios::in );
      if( DataBlockSizeFile )
      {
         gsl_matrix_int* dummy = scan_integer_data( DataBlockSizeFile );
         int dum_rows = dummy->size1;
         int dum_cols = dummy->size2;
         num_data_block_sizes_original = num_data_block_sizes = dum_rows * dum_cols;
         data_block_sizes = new int[ num_data_block_sizes ];
         data_block_sizes_sum = 0;
         data_block_sizes_repetitive = true;
         for( int i = 0; i < dum_rows; i++ )
            for( int j = 0; j < dum_cols; j ++ )
            {
               data_block_size = gsl_matrix_int_get( dummy, i, j );
               data_block_sizes[ i * dum_cols + j ] = data_block_size;
               data_block_sizes_sum += data_block_size;
               if( ( i != 0 || j != 0 ) && data_block_sizes_repetitive )
                  data_block_sizes_repetitive = data_block_size == data_block_sizes[ 0 ];
            }
         gsl_matrix_int_free( dummy );
         data_block_size = data_block_sizes[ 0 ];
         ave_data_block_size = (double)data_block_sizes_sum / num_data_block_sizes;
      }
   }
   if( strlen( MissingValuesFilename ) )
   {
      ifstream MissingValuesFile( MissingValuesFilename, ios::in );
      if( MissingValuesFile.is_open() )
      {
         num_missing_values = num_obsns_missing_values = 0;
         missing_values = new bool*[ n ]; 
         for( i = 0; i < n; i++ )
         {
            missing_values[ i ] = new bool[ p_y ];
	    bool missing_i = false;
            for( j = 0; j < p_y; j++ )
            {
               MissingValuesFile >> missing_values[ i ][ j ];
               if( missing_values[ i ][ j ] )
	       {
                  num_missing_values++;
		  if( !missing_i )
		     missing_i = true;
	       }
            }
	    if( missing_i )
	       num_obsns_missing_values++;
         }
         if( num_missing_values )
         {
             missing_Y = new int*[ num_missing_values ];
             int counter = 0;
             for( i = 0; i < n; i++ )
                for( j = 0; j < p_y; j++ )
                   if( missing_values[ i ][ j ] )
                   {
                       missing_Y[ counter ] = new int[ 2 ];
                       missing_Y[ counter ][ 0 ] = i;
                       missing_Y[ counter ][ 1 ] = j;
                       counter++;
                   }
         }
         else
         {
             delete [] missing_values;
             missing_values = 0;
         }
      }
   }
   else if( strlen( MissingYValuesFilename ) )
   {
      gsl_matrix_int* dummy  = scan_integer_data( MissingYValuesFilename );
      num_missing_values = dummy->size1;
      if( dummy->size2 !=2 )
      {
	 cout << "Missing values file should have two columns, one for row number and one for column" << endl;
	 exit(1);
      }
      missing_Y = new int*[ num_missing_values ];
      for( i = 0; i < num_missing_values; i++ )
         missing_Y[ i ] = new int[ 2 ];
      for( i = 0; i < num_missing_values; i++ )
         for( j = 0; j < 2; j++ )
            missing_Y[ i ][ j ] = gsl_matrix_int_get( dummy, i, j ) != 0;
      gsl_matrix_int_free( dummy );
   }
   if( !num_missing_values )
      delete_missing = false;

   if( strlen( InputFilename ) )
      X = scan_data( InputFilename, binary_data_files, false, p );
   if( !X && density_estimation )
   {
      X = gsl_matrix_alloc( n, p_y );
      gsl_matrix_memcpy( X, Y );
   }
   if( X )
   {
      if( !p )
         p = X->size2;
      int xrows = X->size1;
      if( xrows != n )
      {
         if( !(n % xrows) )
         {
            int copies = n / xrows;
            gsl_matrix *Xcopy = X;
            X = gsl_matrix_alloc( n, p );
            for( int ii = 0; ii < copies; ii++ )
               for( int i = 0; i < xrows; i++ )
                  for( int j = 0; j < p; j++ )
                     gsl_matrix_set( X, ii * xrows + i, j, gsl_matrix_get( Xcopy, i, j ) );
            if( data_block_sizes_repetitive )//check that Xcopy itself isn't repetitive
            {
               int i = 1;
               bool repetitive = true;
               while( repetitive && i < num_data_block_sizes )
               {
                  for( int j = 0; j < data_block_size; j++ )
                     repetitive = matrix_rows_equal( Xcopy, j, j + i * data_block_size );
                  i++;
               }
               if( repetitive )
               {
                  if( inputs_repeated_every > data_block_size )
                     inputs_repeated_every = xrows = data_block_size;
               }
            }
            if( !inputs_repeated_every )
               inputs_repeated_every = xrows;
            if( !data_block_sizes && data_block_size == 1 )
               data_block_size = xrows;
            gsl_matrix_free( Xcopy );
         }
         else
         {
            cout << "Input and Target matrices must have same or divisible number of observations (rows) " << xrows << ":" << n << endl;
            exit(1);
         }
      }
   }
   x_sorted = false;
   if( p == 1 )
   {
      x_sorted = true;
      int i = 1;
      while( x_sorted && i < n )
      {
	x_sorted = gsl_matrix_get( X, i, 0 ) >= gsl_matrix_get( X, i-1, 0 );
	i++;
      }
   }
   if( data_block_sizes_repetitive )
   {
      delete [] data_block_sizes;
      data_block_sizes = 0;
      data_block_sizes_sum = data_block_size;
      num_data_block_sizes = 1;
   }
   if( X && strlen( InputRangeFilename ) )
   {
       gsl_matrix* xranges = scan_data( InputRangeFilename, false, false, 2, p );
      if( xranges )
      {
         x_min = new double[ p ];
         x_max = new double[ p ];
         for( int j = 0; j < p; j++ )
         {
            x_min[ j ] = gsl_matrix_get( xranges, j, 0 );
            x_max[ j ] = gsl_matrix_get( xranges, j, 1 );
         }
         gsl_matrix_free( xranges );
      }
   }
   initialise();
}

DataObj::DataObj( gsl_matrix* X1, gsl_matrix* Y1 )
{
   preliminary_DataObj_initialisation();
   n = Y1->size1;
   if( (int)X1->size1 != n )
   {
      cout << "Input and Target matrices must have same number of observations (rows)" << endl;
      exit(1);
   }
   p = X1->size2;
   p_y = Y1->size2;
   X = gsl_matrix_alloc( n, p );
   gsl_matrix_memcpy( X, X1 );
   Y = gsl_matrix_alloc( n, p_y );
   gsl_matrix_memcpy( Y, Y1 );
   *PCAFilename = 0;
   *PCAPartitionFilename = 0;
   initialise();
}

DataObj::DataObj( gsl_matrix* Y1 )
{
   preliminary_DataObj_initialisation();
   n = Y1->size1;
   p_y = Y1->size2;
   Y = gsl_matrix_alloc( n, p_y );
   gsl_matrix_memcpy( Y, Y1 );
   X = 0;
   initialise();
}

DataObj::~DataObj()
{
   if( X_original && X_original != X && Y_original != X_original )
      gsl_matrix_free( X_original );
   if( Y_original && Y_original != Y )
      gsl_matrix_free( Y_original );
   if( X && Y != X)
      gsl_matrix_free( X );
   if( Y )
      gsl_matrix_free( Y );
   if( Y_hat )
      gsl_matrix_free( Y_hat );
   zap_vector( x_min );
   zap_vector( x_max );
   zap_vector( x_bar );
   zap_vector( x_sd );
   zap_vector( x_scale );
   zap_vector( y_min );
   zap_vector( y_max );
   zap_vector( y_bar );
   zap_vector( y_sd );
   zap_vector( y_scale );
   if( missing_values )
   {
      for( int i = 0; i < (n_original?n_original:n); i++ )
         delete [] missing_values[ i ];
   delete [] missing_values;
   }
   zap_vector( deleted_x_point );
   if( data_block_sizes_original != data_block_sizes )
      zap_vector( data_block_sizes_original );
   zap_vector( data_block_sizes );
   if( eigen_vectors )
      gsl_matrix_free( eigen_vectors );
   if( eigen_values )
      gsl_vector_free( eigen_values );
   if( pca_transformed_data )
      gsl_matrix_free( pca_transformed_data );
   if( missing_Y )
   {
      for( int i = 0; i < num_missing_values; i++)
	 delete missing_Y[ i ];
      delete missing_Y;
   }
   if( deleted_missing_points )
   {
      for( int i = 0; i < num_obsns_missing_values; i++)
         delete [] deleted_missing_points[ i ];
      delete [] deleted_missing_points;
   }
}

void DataObj::preliminary_DataObj_initialisation()
{
   X = Y = 0;
   Y_hat = 0;
   p = p_y = 0;
   delete_missing = false;
   missing_values = 0;
   Y_original = X_original = 0;
   n_original = 0;
   missing_Y = 0;
   transpose_targets = 0;
   x_min = x_max = x_bar = x_sd = x_scale = 0;
   y_min = y_max = y_bar = y_sd = y_scale = 0;
   global_x_min = 0;
   global_x_max = 1;
   global_x_range = 1;
   normalise_predictors = 1;
   normalise_targets = 0;
   log_ave_y_scale = 0;
   log_y_range_product = 0;
   print_to_screen = 1;
   delete_point = -1;
   deleted_x_point = 0;
   deleted_missing_points = 0;
   data_block_size_original = data_block_size = 1;
   data_block_sizes_original = data_block_sizes = NULL;
   data_block_sizes_sum = 1;
   data_block_sizes_sum_original = 1;
   num_data_block_sizes = 1;
   num_data_block_sizes_original = 1;
   pca_output_cols = INT_MAX;
   density_estimation = false;
   *InputFilename = 0;
   *TargetFilename = 0;
   *PCAFilename = 0;
   *DataBlockSizesFilename = 0;
   *InputRangeFilename = 0;
   *MissingValuesFilename = 0;
   *MissingYValuesFilename = 0;
}

void DataObj::initialise()
{
   if( !data_block_sizes )
   {
      num_data_block_sizes = 1;
      ave_data_block_size = data_block_sizes_sum = data_block_size;
   }
   if( n % data_block_sizes_sum )
   {
      cout << "Number of data points (" << n << ") not a multiple of data_block_sizes (" << data_block_sizes_sum << ")." << endl;
      exit(1);
   }
   num_data_blocks = n / data_block_sizes_sum * num_data_block_sizes;
   eigen_vectors = 0;
   eigen_values = 0;
   pca_transformed_data = 0;
   if( strlen( PCAFilename ) )
      calculate_principal_components();
}

gsl_matrix* DataObj::CalculateLinearRegressionResiduals( const int c1, const int c2 ) const
{
   if( c1 < 0 || c1 >= p_y )
   {
      cout << "First column number out of bounds" << endl;
      exit(1);
   }
   if( c2 < 0 || c2 >= p_y )
   {
      cout << "Second column number out of bounds" << endl;
      exit(1);
   }
   
   gsl_matrix* e = gsl_matrix_alloc( 1, n );
   double beta_0, beta_1, s_xx = 0, s_xy = 0, s_yy = 0, x_bar = 0, y_bar = 0;
   int i;
   for( i = 0; i < n; i++ )
   {
      x_bar += gsl_matrix_get( Y, i, c1 );
      y_bar += gsl_matrix_get( Y, i, c2 );
   }
   x_bar /= (double) n;
   y_bar /= (double) n;
   
   for( i = 0; i < n; i++ )
   {
      s_xx += ( gsl_matrix_get( Y, i, c1 ) - x_bar ) * ( gsl_matrix_get( Y, i, c1 ) - x_bar );
      s_xy += ( gsl_matrix_get( Y, i, c1 ) - x_bar ) * ( gsl_matrix_get( Y, i, c2 ) - y_bar );
      s_yy += ( gsl_matrix_get( Y, i, c2 ) - y_bar ) * ( gsl_matrix_get( Y, i, c2 ) - y_bar );
   }

   beta_1 = s_xx / s_xy;
   beta_0 = y_bar - beta_1 * x_bar;

   double rssq = 0;
   for( i = 0; i < n; i++ )
   {
      double e_i = gsl_matrix_get( Y, i, c2 ) - beta_0 - beta_1 * gsl_matrix_get( Y, i, c1 );
      gsl_matrix_set( e, 0, i, e_i );
      rssq += e_i * e_i;
   }
   cout << "R-squared value from linear regression = " << 1 - rssq / s_yy << endl;
   return e;
}

gsl_matrix* DataObj::CalculateLinearRegressionResiduals() const
{
   if( !X )
   {
      cout << "Need matrix of predictors to use this function" << endl;
      exit(1);
   }
   
   gsl_matrix* e = gsl_matrix_alloc( n, p_y );
   double* x_bar = new double[ p_y ];
   double* y_bar = new double[ p_y ];
   double* beta_0 = new double[ p_y ];
   double* beta_1 = new double[ p_y ];
   double* s_xx = new double[ p_y ];
   double* s_xy = new double[ p_y ];
   int i, j;
   for( j = 0; j < p_y; j++ )
   {
      x_bar[ j ] = 0;
      y_bar[ j ] = 0;
      for( int i = 0; i < n; i++ )
      {
         x_bar[ j ] += gsl_matrix_get( X, i, j );
         y_bar[ j ] += gsl_matrix_get( Y, i, j );
      }
      x_bar[ j ] /= (double) n;
      y_bar[ j ] /= (double) n;
   }

   double s_yy = 0;
   double rssq = 0;
   for( j = 0; j < p_y; j++ )
   {
      s_xx[ j ] = 0;
      s_xy[ j ] = 0;
      for( i = 0; i < n; i++ )
      {
         double x_ij = gsl_matrix_get( X, i, j ) - x_bar[ j ];
         double y_ij = gsl_matrix_get( Y, i, j ) - y_bar[ j ];
         s_xx[ j ] += x_ij * x_ij;
         s_xy[ j ] += x_ij * y_ij;
         s_yy += y_ij * y_ij;
      }

      beta_1[ j ] = s_xy[ j ] / s_xx[ j ];
      beta_0[ j ] = y_bar[ j ] - beta_1[ j ] * x_bar[ j ];

      for( i = 0; i < n; i++ )
      {
         double e_ij = gsl_matrix_get( Y, i, j ) - beta_0[ j ] - beta_1[ j ] * gsl_matrix_get( X, i, j );
         gsl_matrix_set( e, i, j, e_ij );
         rssq += e_ij * e_ij;
      }
   }

   cout << "R-squared value from linear regression = " << 1 - rssq / s_yy << endl;

   ofstream OutFile( "regparsout",ios::out );
   for( j = 0; j< p_y; j++ )
      OutFile << beta_0[ j ] << "  ";
   OutFile << "\n";
   for( j = 0; j< p_y; j++ )
      OutFile << beta_1[ j ] << "  ";

   delete [] x_bar;
   delete [] y_bar;
   delete [] beta_0;
   delete [] beta_1;
   delete [] s_xx;
   delete [] s_xy;
   return e;
}

int DataObj::InitialStringMatch( const char *s1, const char *s2 )
{
   int length = strlen( s2 );
   if ( !strncmp( s1, s2, length ) )
      return 1;
   return  0;
}

void DataObj::StringToString( char *StringOut, char *StringIn )
{
   char *p;
   /* Find string after = sign */
   p = strchr( StringIn, (int) '=' );
   p++;
   strncpy( StringOut, p, FILENAMELENGTH );
}

void StringToLong( long *Long, char *String )
{
   int base;
   char *stopstring;
   char *p;
   /* Find first long integer after = sign */	
   p = strchr( String, (int) '=' );
   p++;
   base = 10;
   *Long = strtol( p, &stopstring, base );
}

void StringToDouble( double *Double, char *String )
{
   char *stopstring;
   char *p;
   /* Find first double after = sign */	
   p = strchr( String, (int) '=' );
   p++;
   *Double = strtod( p, &stopstring );
}

void StringToBool( bool *Bool, char *String )
{
   long Long;
   StringToLong( &Long, String );
   if( Long )
      *Bool = 1;
   else
      *Bool = 0;
}

void IntToString( string *String, int *Int )
{
   std::stringstream ss;
   ss << *Int;
   ss >> *String;
}

void insert_in_string_before_dot( char *s1, const string *s2 )
{
   string s1_string( s1 );
   int pos = s1_string.rfind(".");
   if( pos > 0 )
      s1_string.insert( pos, *s2 );
   else
      s1_string += *s2;
   strcpy( s1, s1_string.c_str() );
}

void DataObj::delete_row_range( const int start, const int end )
{
   if( end > n - 1 )
   {
      cout << "rows to be deleted out of range, exiting." << endl;
      exit(1);
   }
   gsl_matrix *Xcopy = X;
   gsl_matrix *Ycopy = Y;
   n = n - ( end - start + 1 );
   X = gsl_matrix_alloc( n, p );
   Y = gsl_matrix_alloc( n, p_y );
   int i, j;
   for( i = 0; i < start; i++ )
   {
      for( j = 0; j < p; j++ )
         gsl_matrix_set( X, i, j, gsl_matrix_get( Xcopy, i, j ) );
      for( j = 0; j < p_y; j++ )
         gsl_matrix_set( Y, i, j, gsl_matrix_get( Ycopy, i, j ) );
   }
   for( i = start; i < n; i++ )
   {
      for( j = 0; j < p; j++ )
         gsl_matrix_set( X, i, j, gsl_matrix_get( Xcopy, i + ( end - start + 1 ), j ) );
      for( j = 0; j < p_y; j++ )
         gsl_matrix_set( Y, i, j, gsl_matrix_get( Ycopy, i + ( end - start + 1 ), j ) );
   }
   gsl_matrix_free( Xcopy );
   gsl_matrix_free( Ycopy );
}

void DataObj::delete_points()
{
   if( !deleted_x_point )
      deleted_x_point = new double[ p ];
   int i, j, k;
   for( j = 0; j < p; j++ )
      deleted_x_point[ j ] = gsl_matrix_get( X, delete_point, j );
   gsl_matrix *Xcopy = X;
   gsl_matrix *Ycopy = Y;
   if( !inputs_repeated_every )
      n--;
   else
      n -= n / inputs_repeated_every;
   Y = gsl_matrix_alloc( n, p_y );
   X = gsl_matrix_alloc( n, p );
   if( !inputs_repeated_every )
   {
      for( i = 0; i < delete_point; i++ )
      {
         for( j = 0; j < p_y; j++ )
            gsl_matrix_set( Y,i, j, gsl_matrix_get( Ycopy, i, j ) );
         for( j = 0; j < p; j++ )
            gsl_matrix_set( X, i, j, gsl_matrix_get( Xcopy, i, j ) );
      }
      for( int i = delete_point; i < n; i++ )
      {
         for( j = 0; j < p_y; j++ )
            gsl_matrix_set( Y, i, j, gsl_matrix_get( Ycopy, i + 1, j ) );
         for( j = 0; j < p; j++ )
            gsl_matrix_set( X, i, j, gsl_matrix_get( Xcopy, i + 1, j ) );
      }
   }
   else
   {
      inputs_repeated_every--;
      int num_repeats = n / inputs_repeated_every;
      data_block_size--;
      data_block_sizes_sum--;
      ave_data_block_size -= 1.0 / num_data_block_sizes;
      for( k = 0; k < delete_point; k++ )
         for( i = 0; i < num_repeats; i++ )
         {
            for( j = 0; j < p_y; j++ )
               gsl_matrix_set( Y, i * inputs_repeated_every + k, j, gsl_matrix_get( Ycopy, i * ( inputs_repeated_every + 1 ) + k, j ) );
            for( j = 0; j < p; j++ )
               gsl_matrix_set( X, i * inputs_repeated_every + k, j, gsl_matrix_get( Xcopy, i * ( inputs_repeated_every + 1 ) + k, j ) );
         }
      for( k = delete_point; k < inputs_repeated_every; k++ )
         for( i = 0; i < num_repeats; i++ )
         {
            for( j = 0; j < p_y; j++ )
               gsl_matrix_set( Y, i * inputs_repeated_every + k, j, gsl_matrix_get( Ycopy, i * ( inputs_repeated_every + 1 ) + k + 1, j ) );
            for( j = 0; j < p; j++ )
               gsl_matrix_set( X, i * inputs_repeated_every + k, j, gsl_matrix_get( Xcopy, i * ( inputs_repeated_every + 1 ) + k + 1, j ) );
         }
   }
   gsl_matrix_free( Xcopy );
   gsl_matrix_free( Ycopy );
}

void DataObj::update_missing_y_value( int index, double new_y_value )
{
  gsl_matrix_set( Y, missing_Y[ index ][ 0 ], missing_Y[ index ][ 1 ], new_y_value );
}

void DataObj::normalise_data_matrices( bool find_x_min_max )
{
   int i, j;
   if( X )
   {
      if( find_x_min_max && !x_min )
      {
         global_x_min = DBL_MAX;
         global_x_max = -DBL_MAX;
         x_min = new double[ p ];
         x_max = new double[ p ];
      }
      if( !x_bar )
      {
         x_bar = new double[ p ];
         x_sd = new double[ p ];
         x_scale = new double[ p ];
      }
      for( j = 0; j < p; j++ )
      {
         double x_min_j = n > 0 ? gsl_matrix_get( X, 0, j ) : DBL_MAX;
         double x_max_j = n > 0 ? gsl_matrix_get( X, 0, j ) : -DBL_MAX;
         double x_sum_j = 0;
         double x_ssq_j = 0;
         for( i = 0; i < n; i++ )
         {
            double x_ij = gsl_matrix_get( X, i, j );
            if( find_x_min_max )
	    {
               if( x_ij < x_min_j )
                  x_min_j = x_ij;
               else if( x_ij > x_max_j )
                  x_max_j = x_ij;
	    }
            x_sum_j += x_ij;
            x_ssq_j += x_ij * x_ij;
         }
         if( find_x_min_max )
         {
            x_min[ j ] = x_min_j;
            x_max[ j ] = x_max_j;
            if( x_min_j < global_x_min )
               global_x_min = x_min_j;
            if( x_max_j > global_x_max )
               global_x_max = x_max_j;
         }
         x_bar[ j ] = x_sum_j / n;
//         x_sd[ j ] = sqrt( ( x_ssq - n * x_bar[ j ] * x_bar[ j ] ) / ( n - 1 ) );
         x_sd[ j ] = sqrt( ( x_ssq_j - x_sum_j * x_sum_j / n ) / n );
         x_scale[ j ] = 1;
	 if( x_max[ j ] > x_min[ j ] )
	 {
	    if( normalise_predictors == 1 )
	       x_scale[ j ] = x_max[ j ] - x_min[ j ];
	    else if( normalise_predictors == 2 )
	       x_scale[ j ] = x_sd[ j ];
	 }
         if( normalise_predictors && x_scale[ j ] )
         {
            if( normalise_predictors == 1 )
               x_bar[ j ] = ( x_bar[ j ] - x_min[ j ] ) / x_scale[ j ];
            else if( normalise_predictors == 2 )
            {
               x_min[ j ] /= x_scale[ j ];
               x_max[ j ] /= x_scale[ j ];
               x_bar[ j ] /= x_scale[ j ];
            }
            for( i = 0; i < n; i++ )
            {
               if( normalise_predictors == 1 )
                  gsl_matrix_set( X, i, j, gsl_matrix_get( X, i, j ) - x_min[ j ] );
               gsl_matrix_set( X, i, j, gsl_matrix_get( X, i, j ) / x_scale[ j ] );
            }
         }
      }
      global_x_range = global_x_max - global_x_min;
   }
   if( !y_min )
   {
      y_min = new double[ p_y ];
      y_max = new double[ p_y ];
      y_bar = new double[ p_y ];
      y_sd = new double[ p_y ];
      y_scale = new double[ p_y ];
   }
   log_ave_y_scale = 0;
   log_y_range_product = 0;
   for( j = 0; j < p_y; j++ )
   {
      double y_min_j = n > 0 ? gsl_matrix_get( Y, 0, j ) : DBL_MAX;
      double y_max_j = n > 0 ? gsl_matrix_get( Y, 0, j ) : -DBL_MAX;
      double y_sum_j = 0;
      double y_ssq_j = 0;
      for( i = 0; i < n; i++ )
      {
         double y_ij = gsl_matrix_get( Y, i, j );
         if( y_ij < y_min_j )
            y_min_j = y_ij;
         else if( y_ij > y_max_j )
            y_max_j = y_ij;
         y_sum_j += y_ij;
         y_ssq_j += y_ij * y_ij;
      }
      y_min[ j ] = y_min_j;
      y_max[ j ] = y_max_j;
      if( y_max_j > y_min_j )
	 log_y_range_product += log( y_max_j - y_min_j );
      y_bar[ j ] = y_sum_j / n;
//      y_sd[ j ] = sqrt( ( y_ssq_j - n * y_bar[ j ] * y_bar[ j ] ) / ( n - 1 ) );
      y_sd[ j ] = sqrt( ( y_ssq_j - y_sum_j * y_sum_j / n ) / n );
//      y_scale[ j ] = y_max[ j ] - y_min[ j ];
      y_scale[ j ] = y_sd[ j ];
      if( y_scale[ j ] > 0 )
      {
         log_ave_y_scale += log( y_scale[ j ] );
         if( normalise_targets )
         {
            y_bar[ j ] /= y_scale[ j ];
            for( i = 0; i < n; i++ )
               gsl_matrix_set( Y, i, j, gsl_matrix_get( Y, i, j ) / y_scale[ j ] - y_bar[ j ] );
//               gsl_matrix_set( Y, i, j, (gsl_matrix_get( Y, i, j )-y_bar[j]) / y_scale[ j ] );
                }
      }
   }
   if( normalise_targets )
      log_y_range_product -= log_ave_y_scale;
   log_ave_y_scale /= p_y;
}

gsl_matrix* DataObj::calculate_principal_components( const gsl_matrix* m, const bool corr )
{
   gsl_matrix *local_m = NULL;
   int i, j;
   if( !m )
   {
      if( p_y == 1 && data_block_sizes_sum > 1 )
      {
         local_m = gsl_matrix_alloc( num_data_blocks / num_data_block_sizes, p_y * data_block_sizes_sum );
         for( i = 0; i < num_data_blocks / num_data_block_sizes; i++ )
            for( j = 0; j < p_y; j++ )
               for( int block_inc = 0; block_inc < data_block_sizes_sum; block_inc++ )
                  gsl_matrix_set( local_m,  i, j + block_inc * p_y, gsl_matrix_get( Y, i * data_block_sizes_sum + block_inc, j ) );
      }
      else
      {
         local_m = gsl_matrix_alloc( n, p_y );
         gsl_matrix_memcpy( local_m, Y );
      }
      m = local_m;
   }
   int num_rows = m->size1;
   int num_cols = m->size2;
   double* m_bar = new double[ num_cols ];
   double *m_sd = new double[ num_cols ];
   for( j = 0; j < num_cols; j++ )
   {
      m_bar[ j ] = 0;
      m_sd[ j ] = 0;
      for( i = 0; i < num_rows; i++ )
      {
         double m_ij = gsl_matrix_get( m, i, j );
         m_bar[ j ] += m_ij;
         m_sd[ j ] += m_ij * m_ij;
      }
      m_bar[ j ] /= num_rows;
      m_sd [ j ] = sqrt( ( m_sd[ j ] - num_rows * m_bar[ j ] * m_bar[ j ] ) / ( num_rows - 1 ) );
   }

   if( pca_transformed_data )
      gsl_matrix_free( pca_transformed_data );
   pca_transformed_data = gsl_matrix_alloc( num_rows, num_cols );
   for( i = 0; i < num_rows; i++ )
      for( j = 0; j < num_cols; j++ )
      {
         gsl_matrix_set( pca_transformed_data, i, j, gsl_matrix_get( m, i , j ) );// - m_bar[ j ] );
         if( corr )
            gsl_matrix_set( pca_transformed_data, i, j, gsl_matrix_get( pca_transformed_data, i, j ) / m_sd[ j ] );
      }

   delete [] m_bar;
   delete [] m_sd;
   singular_value_decomposition( pca_transformed_data );
   for( j = 0; j < num_cols; j++ )
   {
      double eigen_value_j = gsl_vector_get( eigen_values, j );
      for( i = 0; i < num_rows; i++ )
         gsl_matrix_set( pca_transformed_data, i, j, gsl_matrix_get( pca_transformed_data, i, j ) * eigen_value_j );
   }
   if( strlen( PCAFilename ) )
   {
      ofstream PCAFile( PCAFilename, ios::out );
      ofstream PCAPartitionFile( PCAPartitionFilename, ios::out );
      int num_output_cols = pca_output_cols < num_cols ? pca_output_cols : num_cols;
      for( i = 0; i < num_rows; i++ )
      {
         int max_pc_i = 0;
         double max_pc_coeff_i = gsl_matrix_get( pca_transformed_data, i, 0 );
         for( j = 0; j < num_output_cols; j++ )
         {
            double pc_ij = gsl_matrix_get( pca_transformed_data, i, j );
            PCAFile << pc_ij;
            if( j < (num_output_cols - 1) )
            {
//               if( local_m &&  p_y == 1 && data_block_sizes_sum > 1 )
//                  PCAFile << endl;
//               else
                  PCAFile << " ";
            }
            if( fabs(pc_ij) > fabs(max_pc_coeff_i) )
            {
               max_pc_i = j;
               max_pc_coeff_i = pc_ij;
            }
         }
         PCAFile << endl;
         if( PCAPartitionFile.is_open() )
         {
            if( max_pc_coeff_i > 0 )
               PCAPartitionFile << 2 * max_pc_i + 1 << endl;
            else
               PCAPartitionFile << 2 * max_pc_i + 2 << endl;
         }                
      }
   }
   if( print_to_screen )
   {
      int percent_cols_start = 1;
      int percent_cols_end = 3;
      if( percent_cols_end > num_cols - 1 )
	percent_cols_end = num_cols - 1;
      if( pca_output_cols < num_cols )
         percent_cols_start = percent_cols_end = pca_output_cols;
      int percent_cols = percent_cols_start;
      double eigen_val_j, eigen_val_sum = 0, total_eigen_val_sum = 0;
      for( j = 0; j < num_cols; j++ )
      {
	 eigen_val_j = gsl_vector_get( eigen_values, j );
         total_eigen_val_sum += eigen_val_j;
         if( j < percent_cols )
             eigen_val_sum += eigen_val_j;
      }
      double min_eigen, max_eigen;
      min_eigen = max_eigen = gsl_vector_get( eigen_values, 0 );
      while( percent_cols <= percent_cols_end )
      {
	 if( eigen_val_sum > 0 && num_cols >= 2 )
	 {
	    cout << "First ";
	    if( percent_cols == 1)
	       cout << "PC accounts";
	    else
	       cout << percent_cols << " PCs account";
	    cout << " for " << eigen_val_sum / total_eigen_val_sum * 100 << "% of the total variation." << endl;
	    for( i = 1; i < num_cols; i++ )
            {
	       double eigen_value_i = gsl_vector_get( eigen_values, i );
	       if( min_eigen > eigen_value_i )
		  min_eigen = eigen_value_i;
	       else if( max_eigen < eigen_value_i )
		  max_eigen = eigen_value_i;
	    }
	    if( percent_cols != percent_cols_end )
	       eigen_val_sum += gsl_vector_get( eigen_values, percent_cols );
	    percent_cols++;
	 }
      }
      cout << "Condition number = " << max_eigen / min_eigen << endl;
   }
   if( local_m )
      gsl_matrix_free( local_m );
   return pca_transformed_data;
}

void DataObj::singular_value_decomposition( gsl_matrix* m )
{
   int num_cols = m->size2;
   gsl_vector *work = gsl_vector_alloc( num_cols );
   if( eigen_vectors )
      gsl_matrix_free( eigen_vectors );
   eigen_vectors = gsl_matrix_alloc( num_cols, num_cols );
   if( eigen_values )
      gsl_vector_free( eigen_values );
   eigen_values = gsl_vector_alloc( num_cols );
   gsl_linalg_SV_decomp( m, eigen_vectors, eigen_values, work );
   gsl_vector_free( work );
}

double calculate_squared_euclidean_distance( const gsl_matrix* y1, const gsl_matrix* y2 )
{
   double yssq = 0;
   for( int j = 0; j < (int)y1->size2; j++ )
      for( int i = 0; i < (int)y1->size1; i++ )
      {
         double diff = gsl_matrix_get( y1, i, j ) - gsl_matrix_get( y2, i, j );
         yssq += diff * diff;
      }
   return yssq;
}

double** calculate_squared_euclidean_distances( const gsl_matrix* y1, const gsl_matrix* y2 )
{
   double** y3 = 0;
   y3 = new double*[ y1->size1 ];
   for( int i1 = 0; i1 < (int)y1->size1; i1++ )
   {
      y3[ i1 ] = new double[ y2->size1 ];
      for( int i2 = 0; i2 < (int)y2->size1; i2++ )
      {
         double yssq = 0;
         for( int j = 0; j < (int)y2->size2; j++ )
         {
            double diff = gsl_matrix_get( y1, i1, j ) - gsl_matrix_get( y2, i2, j );
            yssq += diff * diff;
         }
         y3[ i1 ][ i2 ]=yssq;
      }
   }
   return y3;
}

double** calculate_squared_euclidean_distances( double** y1, const gsl_matrix* y2, const int num_rows1 )
{
   double** y3 = 0;
   y3 = new double*[ num_rows1 ];
   for( int i1 = 0; i1 < num_rows1; i1++ )
   {
      y3[ i1 ] = new double[ y2->size1 ];
      for( int i2 = 0; i2 < (int)y2->size1; i2++ )
      {
         double yssq = 0;
         for( int j = 0; j < (int)y2->size2; j++ )
         {
            double diff = y1[ i1 ][ j ] - gsl_matrix_get( y2, i2, j );
            yssq += diff * diff;
         }
         y3[ i1 ][ i2 ]=yssq;
      }
   }
   return y3;
}

gsl_matrix* scan_data( const char* s, bool binary_files, bool read_in_transpose, int num_cols, int num_rows )
{
//   if( binary_files )
//      return scan_data_binary( s, read_in_transpose );
   ifstream DataFile( s, ios::in );
   if( !DataFile )
   {
      cout<< "Data file " << s << " could not be opened" << endl;
      exit(1);
   }
   return scan_data( DataFile, read_in_transpose, num_cols, num_rows );
}

gsl_matrix* scan_data( ifstream& DataFile, bool read_in_transpose, int num_cols, int num_rows )
{
   double dummy;
   list<double> elements;
   char st[20000];
   while( DataFile >> dummy )
      elements.push_back( dummy );
   if( !num_cols )
   {
      DataFile.clear();
      DataFile.seekg( 0 );
      while( DataFile >> dummy )
      {
         DataFile.getline(st,20000);
         num_rows++;
      }
      num_cols = elements.size() / num_rows;
      if( read_in_transpose )
         switch_values( num_cols, num_rows );
   }
   else if( !num_rows )
      num_rows = elements.size() / num_cols;
   list<double>::iterator iter = elements.begin();
   gsl_matrix* m = gsl_matrix_alloc( num_rows, num_cols );
   if( !read_in_transpose )
      for( int i = 0; i < num_rows; i++ )
         for( int j = 0; j < num_cols; j++ )
            gsl_matrix_set( m, i, j, *(iter++) );
   else
      for( int j = 0; j < num_cols; j++ )
         for( int i = 0; i < num_rows; i++ )
            gsl_matrix_set( m, i, j, *(iter++) );
   return m;
}

gsl_matrix_int* scan_integer_data( const char* s, bool binary_files, bool read_in_transpose, int num_cols, int num_rows )
{
//   if( binary_files )
//      return ScanIntegerData_Binary( s, read_in_transpose );
   ifstream DataFile( s, ios::in );
   if( !DataFile || binary_files )
   {
      cout<< "Data file " << s << " could not be opened" << endl;
      exit(1);
   }
   return scan_integer_data( DataFile, read_in_transpose, num_cols, num_rows );
}

gsl_matrix_int* scan_integer_data( ifstream& DataFile, bool read_in_transpose, int num_cols, int num_rows )
{
   int dummy;
   list<int> elements;
   char st[20000];
   
   while( DataFile >> dummy )
      elements.push_back( dummy );

   if( !num_cols )
   {
      DataFile.clear();
      DataFile.seekg( 0 );
      while( DataFile >> dummy )
      {
         DataFile.getline(st,20000);
         num_rows++;
      }
      num_cols = elements.size() / num_rows;
      if( read_in_transpose )
         switch_values( num_cols, num_rows );
   }
   else if( !num_rows )
      num_rows = elements.size() / num_cols;
//   cout << num_rows << " datapoints,  " << num_cols << " variables" <<endl;

   list<int>::iterator iter = elements.begin();
   gsl_matrix_int* M = gsl_matrix_int_alloc( num_rows, num_cols );
   if( !read_in_transpose )
      for( int i = 0; i < num_rows; i++ )
         for( int j = 0; j < num_cols; j++ )
            gsl_matrix_int_set( M, i, j, *(iter++) );
   else
      for( int j = 0; j < num_cols; j++ )
         for( int i = 0; i < num_rows; i++ )
            gsl_matrix_int_set( M, i, j, *(iter++) );
   return M;
}

gsl_matrix* calculate_correlations( const gsl_matrix* m, bool** missing_vals )
{
   int p1 = m->size2;
   gsl_matrix* mx = calculate_covariances( m, missing_vals );
   double* sd = new double[ p1 ];
   int i;
   for( i = 0; i < p1; i++ )
   {
      sd[ i ] = sqrt( gsl_matrix_get( mx, i, i ) );
      gsl_matrix_set( mx, i, i, 1.0 );
   }
   for( i = 0; i < p1; i++ )
      for( int j = 0; j < p1; j++ )
         if( i < j )
         {
            double sdi_sdj = sd[ i ] * sd[ j ];
            if( sdi_sdj > 0 )
               *gsl_matrix_ptr( mx, i, j ) /= sdi_sdj;
         }
         else if( i > j )
            gsl_matrix_set( mx, i, j, gsl_matrix_get( mx, j, i ) );
   
   delete [] sd;
   return mx;
}

double calculate_correlation( const gsl_matrix* m1, const gsl_matrix* m2 )
{
   int n1 = m1->size1;
   if( n1 != (int)m1->size1 )
   {
      cout << "matrices of differing dimensions, can't calculate correlation" << endl;
      exit(1);
   }
   if( m1->size2 > 1 || m2->size2 > 1 )
      cout << "Warning, only calculating correlation for first column of matrices" << endl;
   double mean1 = 0, mean2 = 0, s1 = 0, s2 = 0;
   int i;
   for( i = 0; i < n1; i++ )
   {
      double m1_i0 = gsl_matrix_get( m1, i, 0 );
      double m2_i0 = gsl_matrix_get( m2, i, 0 );
      mean1 += m1_i0;
      mean2 += m2_i0;
      s1 += m1_i0 * m1_i0;
      s2 += m2_i0 * m2_i0;
   }
   mean1 /= n1;
   mean2 /= n1;
   s1 -= n1 * mean1 * mean1;
   s2 -= n1 * mean2 * mean2;
   double s12 = 0;
   for( i = 0; i < n1; i++ )
      s12 += ( gsl_matrix_get( m1, i, 0 ) - mean1 ) * ( gsl_matrix_get( m2, i, 0 ) - mean2 );
   return s12 / ( sqrt( s1 ) * sqrt( s2 ) );
}

gsl_matrix* calculate_covariances( const gsl_matrix* m, bool** missing_vals )
{
   int n1 = m->size1;
   int p1 = m->size2;
   gsl_matrix* mx = gsl_matrix_alloc( p1, p1 );
   double* means = new double[ p1 ];
   for( int j = 0; j < p1; j++ )
   {
      int number_not_missing = n1;
      means[ j ] = 0;
      for( int i = 0; i < n1; i++ )
      {
         if( !missing_vals || !missing_vals[ i ][ j ] )
            means[ j ] += gsl_matrix_get( m, i, j );
         else
            number_not_missing--;
      }
      means[ j ] /= static_cast<double>( number_not_missing );
   }

   double mx_ij;
   for( int i = 0; i < p1; i++ )
      for( int j = 0; j < p1; j++ )
      {
         if( j >= i )
         {
            mx_ij = 0;
            int number_not_missing = n1;
            for( int k = 0; k < n1; k++ )
            {
               if( !missing_vals || (!missing_vals[ k ][ i ]  && !missing_vals[ k ][ j ]) )
                  mx_ij += ( gsl_matrix_get( m, k, i ) - means[ i ] ) * ( gsl_matrix_get( m, k, j ) - means[ j ] );
               else
                  number_not_missing--;
            }
            if( number_not_missing > 2 )
               mx_ij /= (number_not_missing - 1);
            else
               mx_ij = 0;
         }
         else
            mx_ij = gsl_matrix_get( mx, j, i );
         gsl_matrix_set( mx, i, j, mx_ij );
      }

   return mx;
}

void covariances_to_correlations( gsl_matrix* mx )
{
   int n1 = mx->size1;
   double* sdevs = new double[ n1 ];
   int i;
   for( i = 0; i < n1; i++ )
   {
      sdevs[ i ] = sqrt( gsl_matrix_get( mx, i, i ) );
      gsl_matrix_set( mx, i, i, 1 );
   }
   for( i = 1; i < n1; i++ )
      for( int j = 0; j < i; j++ )
      {
         double corr_ij = gsl_matrix_get( mx, i, j ) / ( sdevs[ i ] * sdevs[ j ] );
         gsl_matrix_set( mx, i, j, corr_ij );
         gsl_matrix_set( mx, j, i, corr_ij );
      }
   delete [] sdevs;
}

void gsl_matrix_multiply( gsl_matrix*& a, const gsl_matrix* b, const gsl_matrix* c, const bool transpose )
{
   unsigned int b1 = b->size1;
   unsigned int b2 = b->size2;
   unsigned int c1 = c->size1;
   unsigned int c2 = c->size2;
   if( transpose )
      switch_values( c1, c2 );
   if( c1 != b2 )
   {
      cout << "matrix sizes non-conformant for multiplication" << endl;
      exit( 1 );
   }
   if( a && ( a->size1 != b1 || a->size2 != c2 ) )
   {
      gsl_matrix_free( a );
      a = 0;
   }
   if( !a )
      a = gsl_matrix_alloc( b1, c2 );
   for( unsigned int i = 0; i < b1; i++ )
      for( unsigned int j = 0; j < c2; j++ )
      {
         double a_ij = 0;
         for( unsigned int k = 0; k < b2; k++ )
            if( !transpose )
               a_ij += gsl_matrix_get( b, i, k ) * gsl_matrix_get( c, k, j );
            else
               a_ij += gsl_matrix_get( b, i, k ) * gsl_matrix_get( c, j, k );
         gsl_matrix_set( a, i, j, a_ij );
      }
}

double scalar_result_gsl_matrix_multiply( const gsl_matrix* b, const gsl_matrix* c, const bool transpose )
{
   gsl_matrix* a = NULL;
   gsl_matrix_multiply( a, b, c, transpose );
   double asum = 0;
   for( unsigned int i = 0; i < a->size1; i++ )
      for( unsigned int j = 0; j < a->size2; j++ )
         asum += gsl_matrix_get( a, i, j );
   gsl_matrix_free( a );
   return asum;
}

void gsl_matrix_multiply( gsl_matrix*& a, const gsl_matrix* b, const gsl_matrix* c, const gsl_matrix* d, const bool transpose )
{
   gsl_matrix* cd = NULL;
   gsl_matrix_multiply( cd, c, d, transpose );
   gsl_matrix_multiply( a, b, cd );
   gsl_matrix_free( cd );
}

double scalar_result_gsl_matrix_multiply( const gsl_matrix* b, const gsl_matrix* c, const gsl_matrix* d, const bool transpose )
{
   gsl_matrix* a = NULL;
   gsl_matrix_multiply( a, b, c, d, transpose );
   double asum = 0;
   for( unsigned int i = 0; i < a->size1; i++ )
      for( unsigned int j = 0; j < a->size2; j++ )
         asum += gsl_matrix_get( a, i, j );
   gsl_matrix_free( a );
   return asum;
}

bool significantly_different( const double d1, const double d2, const double precision )
{
   double diff = d1 - d2;
   if( diff > precision || diff < -precision )
      return true;
   return false;
}

double zero_threshold( double d, const double precision )
{
   if( fabs( d  ) < precision )
      return 0;
   return d;
}

void print_gsl_matrix( const gsl_matrix* m )
{
   for( unsigned int i = 0; i < m->size1; i++ )
   {
      for( unsigned int j = 0; j < m->size2; j++)
         cout << gsl_matrix_get( m, i, j ) << "  ";
      cout << endl;
   }
}

void write_gsl_matrix( const gsl_matrix* m, const char*& s )
{
   ofstream Stream( s, ios::out );
   write_gsl_matrix( m, Stream );
}

void write_gsl_matrix( const gsl_matrix* m, ofstream& OutputFile )
{
   for( unsigned int i = 0; i < m->size1; i++ )
   {
      for( unsigned int j = 0; j < m->size2; j++)
         OutputFile << gsl_matrix_get( m, i, j ) << "  ";
      OutputFile << endl;
   }
}

bool matrix_rows_equal( const gsl_matrix* m, const int row1, const int row2 )
{
   for( unsigned int j = 0; j < m->size2; j++ )
      if( gsl_matrix_get( m, row1, j ) != gsl_matrix_get( m, row2, j ) )
         return false;
   return true;
}

bool matrix_cols_equal( const gsl_matrix* m, const int col1, const int col2 )
{
   for( unsigned int i = 0; i < m->size1; i++ )
      if( gsl_matrix_get( m, i, col1 ) != gsl_matrix_get( m, i, col2 ) )
         return false;
   return true;
}

