// DataObj1.h: interface for the DataObj class.
//
//////////////////////////////////////////////////////////////////////

#if !defined( _data_obj_h )
#define _data_obj_h

#include <cmath>
#include <cstdlib>
#include <stdlib.h>
#include <stdio.h>
#include <cstring>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <list>
#include <gsl/gsl_version.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_matrix.h>
using std::cout;
using std::endl;
using std::ios;
using std::ifstream;
using std::ofstream;
using std::string;

#define zap(x) if(x){delete x; x=0;}
#define zap_vector(x) if(x){delete [] x; x=0;}
#define GSL_RANGE_CHECK_OFF
#define FILENAMELENGTH 100
using namespace std;


class DataObj  
{
public:
   DataObj(int argc, char *argv[]);
   DataObj( gsl_matrix* X1, gsl_matrix* Y1 );// creates a new data object with inputs X1 and targets Y1.
   DataObj( gsl_matrix* Y1 );// creates a new data object with just targets Y1.
   virtual ~DataObj();
   void preliminary_DataObj_initialisation();
   void initialise();
   static long get_inputs_repeated_every(){ return inputs_repeated_every; }
   static void set_inputs_repeated_every( long new_inputs_repeated_every ){ inputs_repeated_every = new_inputs_repeated_every; }
   gsl_matrix* CalculateLinearRegressionResiduals() const;
   // returns matrix of residuals when columns of Y are regressed on columns X
   gsl_matrix* CalculateLinearRegressionResiduals(const int c1, const int c2) const;
   // returns matrix of residuals when column c2 is regressed on c1
   
   int InitialStringMatch( const char *s1, const char *s2 );// Returns true if s2 matches start of s1
   
   void StringToString( char *StringOut, char *StringIn );

   ///
   void delete_row_range( const int start, const int end );// deletes all rows between start and end.
   void delete_points();// deletes the points from the matrices according to delete_point and inputs_repeated_every.
   virtual void update_missing_y_value( int index, double new_y_value );//replace the index missing value in Y with new_y_value.
   void normalise_data_matrices( bool find_x_min_max = true );// normalises X and Y; if find_min_max=false, assume we have already found the min and max of X.
   gsl_matrix* calculate_principal_components( const gsl_matrix* m = NULL, const bool corr = false );//if corr = true use correlation matrix, o/w use covariance matrix.
   void singular_value_decomposition( gsl_matrix* m );

protected:
//	int i, j, k; // counters
   gsl_matrix* Y;// Data matrix.
   gsl_matrix* Y_hat;// Matrix of predicted values for the data Y.
   gsl_matrix* X;// Matrix of predictors if necessary.
   int** missing_Y;// Two column array where each row gives a row and column number pair of a data point in Y that should be treated as missing.
   double *x_min, *x_max, *x_bar, *x_sd, *x_scale, *y_min, *y_max, *y_bar, *y_sd, *y_scale;// gsl::vector of minimum, maximum and mean x and y values for each component, and s.d. of y.
   double global_x_min, global_x_max;//min and max values of X across all p variables.
   double global_x_range;//global_x_max-global_x_min
   bool** missing_values;//a matrix of indicators for whether the ijth target is missing.
   gsl_matrix* Y_original;// Copy of original Y matrix if missing values are then deleted.
   gsl_matrix* X_original;// Copy of original X matrix if missing values are then deleted.
   gsl_matrix *eigen_vectors;
   gsl_vector *eigen_values;
   gsl_matrix *pca_transformed_data;
   long pca_output_cols;//the number of principal components allowed to be written to PCAFile.
   int n;
   int n_original;//Copy of original n if missing values are then deleted.
   long p_y;
   long p;
   bool delete_missing;//if true, delete all missing values from the data.
   int num_missing_values;//the number of non-zero entries in missing_values.
   int num_obsns_missing_values;//the number of obsns having a missing value somehwere in thier p_y entries.
   int num_data_blocks;
   long data_block_size;//every data_block_size observations are assumed to be from the same experimental unit ( e.g observed data are a time series of length data_block_series ).
   long data_block_size_original;//copy of original data_block_size.
   int* data_block_sizes;//every data_block_sizes[i] observations in experiment i are assumed to be from the same individual ( e.g observed data are a time series of length data_block_series ).
   int* data_block_sizes_original;//Copy of original data_block_sizes array if missing values are then deleted.
   int num_data_block_sizes;//the dimension of data_block_sizes.
   int num_data_block_sizes_original;//Copy of num_data_block_sizes value if missing values are then deleted.
   int data_block_sizes_sum;//the sum of the elements of data_block_sizes.
   int data_block_sizes_sum_original;//Copy of original data_block_sizes_sum value if missing values are then deleted.
   double ave_data_block_size;//the average of the elements of data_block_sizes.
   ///
   bool print_to_screen;//if zero, print minimal information to screen.
   static bool binary_data_files; //if non zero, input data files are taken to be binary
   bool transpose_targets;//if true, uses the transpose of the targets read in.
   static long inputs_repeated_every; // if non-zero, input & therfore design matrix is repetitive every inputs_repeated_every rows;
   long normalise_predictors;//if 1, rescales the predictors to the range [0,1]; if 2, standardises the predictors.
   bool normalise_targets;//if true, standardises the targets
   long delete_point;// if >= 0, remove the data point number delete_point, and if inputs_repeated_every>0, also remove delete_point + i * inputs_repeated_every, i=1,2,....  Useful for leave one out CV.
   double* deleted_x_point;// takes the value of the deleted x point if delete_point >= 0.
   double** deleted_missing_points;// takes the value of the deleted x points if deleting missing values.
   bool x_sorted;//true if X is 1 dimensional and its values are non-decreasing.
   bool density_estimation;//If true, we are doing density estimation and certain options should be switched off.
      
   
   double log_ave_y_scale;//log average scale by which targets normalised. 

   double log_y_range_product;//the log product of the ranges of the components of the target variable.

   char InputFilename[FILENAMELENGTH];
   ///
   char TargetFilename[FILENAMELENGTH];
   ///
   char InputRangeFilename[FILENAMELENGTH];//If required, a file containing the minimum and maximum range values for each input dimension.
   ///
   char MissingValuesFilename[FILENAMELENGTH];
   ///
   char MissingYValuesFilename[FILENAMELENGTH];
   ///
   char PCAFilename[FILENAMELENGTH];
   ///
   char PCAPartitionFilename[FILENAMELENGTH];
   ///
   char DataBlockSizesFilename[FILENAMELENGTH];//File holding the data_block_sizes for each experiment.

};

template <class T> void switch_values( T& val1, T& val2 )
{
   T store = val1;
   val1 = val2;
   val2 = store;
}

void StringToLong( long *Long, char *String );
void StringToDouble( double *Double, char *String );
void StringToBool( bool *Bool, char *String );
void IntToString( string *String, int *Int );
void insert_in_string_before_dot( char *s1, const string *s2 );//creates a new string, a copy of s1 with s2 inserted before the first dot in s1.
extern gsl_matrix* scan_data( const char* s, bool binary_files = 0, bool read_in_transpose = 0, int num_cols = 0, int num_rows = 0 );// scans data from "s" and returns gsl_matrix of data
extern gsl_matrix* scan_data( ifstream& DataFile, bool read_in_transpose = 0, int num_cols = 0, int num_rows = 0 );//scans double data from stream DataFile and returns gsl_matrix of data
extern gsl_matrix_int* scan_integer_data( const char* s, bool binary_files = 0, bool read_in_transpose = 0, int num_cols = 0, int num_rows = 0 );// scans data from "s" and returns integer matrix of data
extern gsl_matrix_int* scan_integer_data( ifstream& DataFile, bool read_in_transpose = 0, int num_cols = 0, int num_rows = 0 );// scans integer data from stream DataFile and returns matrix of data
extern gsl_matrix* calculate_correlations( const gsl_matrix* m, bool** missing_vals = NULL );
extern double calculate_correlation( const gsl_matrix* m1, const gsl_matrix* m2 );
// calculate corr matrix for the columns of supplied matrix
extern gsl_matrix* calculate_covariances( const gsl_matrix* m, bool** missing_vals = NULL );
extern void covariances_to_correlations( gsl_matrix* );
extern void gsl_matrix_multiply( gsl_matrix*& a, const gsl_matrix* b, const gsl_matrix* c, const bool transpose = false );//a = b * c. (or a = b * c' if transpose=true)
extern double scalar_result_gsl_matrix_multiply( const gsl_matrix* b, const gsl_matrix* c, const bool transpose = false );//returns a = b * c. (or a = b * c' if transpose=true)
extern void gsl_matrix_multiply( gsl_matrix*& a, const gsl_matrix* b, const gsl_matrix* c, const gsl_matrix* d, const bool transpose = false );//a = b * c * d. (or a = b * c * d' if transpose=true)
extern double scalar_result_gsl_matrix_multiply( const gsl_matrix* b, const gsl_matrix* c, const gsl_matrix* d, const bool transpose = false );//returns a = b * c * d. (or a = b * c * d' if transpose=true)
bool significantly_different( const double d1, const double d2, const double precision = DBL_EPSILON );// a function for finding if two doubles are not even approximately equal to some precision.
double zero_threshold( double d, const double precision = DBL_EPSILON );//a thresholding function which returns 0 if |d|<precision, and d otherwise.
double calculate_squared_euclidean_distance( const gsl_matrix* y1, const gsl_matrix* y2 );//finds the squared euclidean distance between the two matrices.
double** calculate_squared_euclidean_distances( const gsl_matrix* y1, const gsl_matrix* y2 );//finds the squared euclidean distances between the rows of the two matrices.
double** calculate_squared_euclidean_distances( double** y1, const gsl_matrix* y2, const int num_rows1 );//as above, finds the squared euclidean distances between the rows of the array and matrix.

void print_gsl_matrix( const gsl_matrix* m );//prints the matrix m to screen.
void write_gsl_matrix( const gsl_matrix* m, const char*& s );//write matrix m to file named s.
void write_gsl_matrix( const gsl_matrix* m, ofstream& OutputFile );//write matrix m to file named s.
bool matrix_rows_equal( const gsl_matrix* m, const int row1, const int row2 );// returns true iff the two rows of m are equal.
bool matrix_cols_equal( const gsl_matrix* m, const int col1, const int col2 );// returns true iff the two cols of m are equal.

template <class T> void StringToNumber( T* t, char* MyString )
{
   int base = 10;
   char *stopstring = 0;
   char *p = strchr( MyString, (int) '=' );//Find first numberer after = sign
   p++;
   *t = strtol( p, &stopstring, base );
}

template <class T> void scan_array_data( ifstream& DataFile, T*& T_array, int dim1 )
{
   if( !T_array )
      T_array = new T[ dim1 ];
   for( int i = 0; i < dim1; i++ )
      DataFile >> T_array[ i ];
}

template <class T> void scan_array_data( const char* s, T*& T_array, int dim1 )
{
   ifstream DataFile( s, ios::in );
   if( !DataFile )
   {
      cout<< "Data file " << s << " could not be opened" << endl;
      exit(1);
   }
   scan_array_data( DataFile, T_array, dim1 );
}

template <class T> void scan_array_data( ifstream& DataFile, T**& T_array, int dim1, int dim2 )
{
   if( !T_array )
      T_array = new T*[ dim1 ];
   for( int i = 0; i < dim1; i++ )
   {
      if( !T_array[ i ] )
	 T_array[ i ] = new T[ dim2 ];
      for( int j = 0; j < dim2; j++ )
	 DataFile >> T_array[ i ][ j ];
   }
}


template <class T> void scan_array_data( const char* s, T**& T_array, int dim1, int dim2 )
{
   ifstream DataFile( s, ios::in );
   if( !DataFile )
   {
      cout<< "Data file " << s << " could not be opened" << endl;
      exit(1);
   }
   scan_array_data( DataFile, T_array, dim1, dim2 );
}

template <class T> void write_matrix( T** m, unsigned int num_rows, unsigned int num_cols, const char*& s, T* shift = NULL )
{
   ofstream Stream( s, ios::out );
   write_matrix( m, num_rows, num_cols, Stream, shift );
}

template <class T> void write_matrix( T** m, unsigned int num_rows, unsigned int num_cols, ofstream& OutputFile, T* shift = NULL  )
{
   for( unsigned int i = 0; i < num_rows; i++ )
   {
      for( unsigned int j = 0; j < num_cols; j++)
      {
         T m_ij =  m[i][j];
         if( shift )
            m_ij += *shift;
         OutputFile << m_ij << "  ";
      }
      OutputFile << endl;
   }
}

template <class T> void write_lower_triangle_matrix( T** m, unsigned int num_rows, const char*& s, T* shift = NULL )
{
   ofstream Stream( s, ios::out );
   write_lower_triangle_matrix( m, num_rows, Stream, shift  );
}

template <class T> void write_lower_triangle_matrix( T** m, unsigned int num_rows, ofstream& OutputFile, T* shift = NULL )
{
   for( unsigned int i = 1; i < num_rows; i++ )
      if( m[i] )
      {
         for( unsigned int j = 0; j < i; j++)
            if( j==0 || m[j] )
            {
               T m_ij =  m[i][j];
               if( shift )
                  m_ij += *shift;
               OutputFile << m_ij << "  ";
            }
         OutputFile << endl;
      }
}

template <class T> void write_row_of_lower_triangle_matrix( T** m, unsigned int num_rows, unsigned int row, ofstream& OutputFile, T* shift = NULL )
{

   if( row==0 || m[row] )
   {
      for( unsigned int j = 0; j < row; j++)
         if( j==0 || m[j] )
         {
            T m_rowj =  m[row][j];
            if( shift )
               m_rowj += *shift;
            OutputFile << m_rowj << "  ";
         }            
      for( unsigned int i = row+1; i < num_rows; i++ )
         if( m[i] )
         {
            T m_irow =  m[i][row];
            if( shift )
               m_irow += *shift;
            OutputFile << m_irow << "  ";
         }
      OutputFile << endl;
   }
}


#endif
