#include "basis_fn.h"

int Basis_Fn::Basis_Fn_count = 0;
int Basis_Fn::dim_component_Basis_Fn_counts = 0;
int* Basis_Fn::component_Basis_Fn_counts = NULL;

Basis_Fn::Basis_Fn()
{
   defaults();
   increment_count( component[ 0 ] );
}

Basis_Fn::Basis_Fn( double* v, int terms, int* cmp, int* pw, int* sgn, int* actv_dim )
{
   num_interactions = terms;
   set_dimensions();
   for( int i = 0; i < num_interactions; i++ )
   {
      t[ i ] = v[ i ];
      component[ i ] = cmp[ i ];
      power[ i ] = pw[ i ];
      sign[ i ] = sgn[ i ];
   }
   if( actv_dim )
   {
      active_dimension = new int;
      *active_dimension = *actv_dim;
   }
   else
      active_dimension = NULL;
   coeff = 0;
   scale = 1;
   mean = 0;
   sd = 1;
   basis_standardized = false;
   increment_count( component[ 0 ] );
}

Basis_Fn::Basis_Fn( double d, int cmp, int pw, int sgn, int* actv_dim )
{
   num_interactions = 1;
   set_dimensions();
   t[ 0 ] = d;
   component[ 0 ] = cmp;
   power[ 0 ] = pw;
   sign[ 0 ] = sgn;
   if( actv_dim )
   {
      active_dimension = new int;
      *active_dimension = *actv_dim;
   }
   else
      active_dimension = NULL;
   coeff = 0;
   scale = 1;
   mean = 0;
   sd = 1;
   basis_standardized = false;
   increment_count( component[ 0 ] );
}

Basis_Fn::~Basis_Fn()
{
   decrement_count( component[ 0 ] );
   delete [] t;
   delete [] component;
   delete [] power;
   delete [] sign;
   if( active_dimension )
      delete active_dimension;
}

void Basis_Fn::increment_count( int component )
{
   Basis_Fn_count++;
   if( component_Basis_Fn_counts )
      component_Basis_Fn_counts[ component ]++;
}

void Basis_Fn::decrement_count( int component )
{
   Basis_Fn_count--;
   if( component_Basis_Fn_counts )
      component_Basis_Fn_counts[ component ]--;
}

void Basis_Fn::initialize_component_Basis_Fn_counts( int num_component_dimensions )
{
   dim_component_Basis_Fn_counts = num_component_dimensions;
   if( !component_Basis_Fn_counts )
      component_Basis_Fn_counts = new int[ dim_component_Basis_Fn_counts ];
   for( int i = 0; i < dim_component_Basis_Fn_counts; i++ )
      component_Basis_Fn_counts[ i ] = 0;
}

void Basis_Fn::destroy_Basis_Fn_counts()
{
   if( component_Basis_Fn_counts )
      delete [] component_Basis_Fn_counts;
}

int Basis_Fn::get_component_Basis_Fn_count( int comp )
{
   return component_Basis_Fn_counts[ comp ];
}

int Basis_Fn::get_component_Basis_Fn_count_products_up_until( int comp )
{
   int prod = 1;
   for( int i = 0; i < comp; i++ )
      prod *= component_Basis_Fn_counts[ i ];
   return prod;
}

int Basis_Fn::get_component_Basis_Fn_count_products_after( int comp )
{
   double prod = 1;
   for( int i = comp + 1; i < dim_component_Basis_Fn_counts; i++ )
      prod *= component_Basis_Fn_counts[ i ];
   return prod;
}

void Basis_Fn::defaults()
{
   num_interactions = 1;
   set_dimensions();
   for( int i = 0; i < num_interactions; i++ )
   {
      t[ i ] = 0;
      component[ i ] = 0;
      power[ i ] = 1;
      sign[ i ] = 0;
   }
   active_dimension = NULL;
   coeff = 0;
   scale = 1;
   mean = 0;
   sd = 1;
   basis_standardized = false;
}

void Basis_Fn::set_dimensions()
{
   t = new double[ num_interactions ];
   component = new int[ num_interactions ];
   power= new int[ num_interactions ];
   sign= new int[ num_interactions ];
}

int Basis_Fn::get_Basis_Fn_count()
{
   return Basis_Fn_count;
}

