#include "basis_fn_regression.h"

long Basis_Fn_Regression::power_min = INT_MAX;
long Basis_Fn_Regression::power_max = 1;
long Basis_Fn_Regression::max_fourier_basis = 1;

Basis_Fn_Regression::Basis_Fn_Regression( int argc, char **argv)
      : MCMC( argc, argv )
{
   preliminary_basis_fn_regression_initialisation();
   for( int i = 1; i < argc; i++ )
   {
      if( strstr( argv[i], "help" ) )
      {
	 cout << "minpower=" << ( power_min < power_max ? power_min : power_max ) << endl;
         cout << "maxpower=" << power_max << "   (Default spline model will create truncated power series splines with powers ranging between these values)" << endl;
         cout << "maxfourier=" << max_fourier_basis << "   (If using Fourier bases, this is the maximum order of bases that will be used)" << endl;
         cout << "splinebases=" << include_spline_bases << "   (If true, include spline basis functions in constructing the design matrix)" << endl;
         cout << "fourierbases=" << include_fourier_bases << "   (If true, include spline basis functions in constructing the design matrix)" << endl;
         cout << "twosidedsplines" << two_sided_splines << "   (If true, allow splines to be two sided, as in Denison et al 2002)" << endl;
         cout << "fullmodelsign" << full_1way_model_sign << "   (If -1/0/1, use left/double/right sided splines respectively when creating a full model)" << endl;
	 cout << "discreteknots=" << discrete_knot_points << "   (If non zero, then knot points can only be chosen from a discrete range.)" << endl;
         cout << "splinegrid=" << knot_point_grid << "   (If zero, default basis function model will put candidate knot point locations at each distinct data point, otherwise will put them at that many equally spaced grid points)" << endl;
         cout << "maxnumsplines" << max_num_splines <<  "   (The maximum number of splines allowed in the model)" << endl;
	 cout << "priorprecision=" << coeff_prior_precision << "   (prior precision for spline coefficients, (may be overwritten lower down in hierarchy))" << endl;
	 cout << "priorprecision1=" << coeff_prior_precision1 << "   (alternative prior precision for spline coefficients)" << endl;
         cout << "startwithfullmodel=" << start_with_full_model << "(If true, start with all possible splines in model)" << endl;
         cout << "standardizedesignmatrix=" << standardize_design_matrix << "(If true, standardize all columns in the design matrix except the intercept)" << endl;
	 cout << "geometricp" << geometric_p << "   (Parameter for geometric prior on # basis fns.)" << endl;
	 cout << "poissonmu" << poisson_mu << "   (Rate parameter for Poisson prior on # basis fns.)" << endl;
         cout << "basispowersfile=" << Basis_Powers_Filename << "   (File where each row contains minpower, maxpower, maxfourier for each x-dimension.)" << endl;
	 cout << "basisseparation" << basis_separation << "   (The minimum distance between two basis functions.)" << endl;
         cout << "basisproposalwidth" << basis_proposal_width << "   (If proposing a new basis centred around a deleted one, use this proposal width uniformly around it.)" << endl;
         return;
      }
      if( InitialStringMatch( argv[i], "minpower=" ) )
         StringToLong( &power_min, argv[i] );
      if( InitialStringMatch( argv[i], "maxpower=" ) )
         StringToLong( &power_max, argv[i] );
      if( InitialStringMatch( argv[i], "maxfourier=" ) )
         StringToLong( &max_fourier_basis, argv[i] );
      if( InitialStringMatch( argv[i], "splinebases=" ) )
         StringToBool( &include_spline_bases, argv[i] );
      if( InitialStringMatch( argv[i], "fourierbases=" ) )
         StringToBool( &include_fourier_bases, argv[i] );
      if( InitialStringMatch( argv[i], "twosidedsplines=" ) )
         StringToBool( &two_sided_splines, argv[i] );
      if( InitialStringMatch( argv[i], "fullmodelsign=" ) )
         StringToLong( &full_1way_model_sign, argv[i] );
      if( InitialStringMatch( argv[i], "startwithfullmodel=" ) )
         StringToBool( &start_with_full_model, argv[i] );
      if( InitialStringMatch( argv[i], "standardizedesignmatrix=" ) )
         StringToBool( &standardize_design_matrix, argv[i] );
      if( InitialStringMatch( argv[i], "discreteknots=" ) )
         StringToBool( &discrete_knot_points, argv[i] );
      if( InitialStringMatch( argv[i], "splinegrid=" ) )
         StringToLong( &knot_point_grid, argv[i] );
      if( InitialStringMatch( argv[i], "maxnumsplines=" ) )
          StringToLong( &max_num_splines, argv[i] );
      if( InitialStringMatch( argv[i], "priorprecision=" ) )
         StringToDouble( &coeff_prior_precision, argv[i] );
      if( InitialStringMatch( argv[i], "priorprecision1=" ) )
         StringToDouble( &coeff_prior_precision1, argv[i] );
      if( InitialStringMatch( argv[i], "geometricp=" ) )
         StringToDouble( &geometric_p, argv[i] );
      if( InitialStringMatch( argv[i], "poissonmu=" ) )
         StringToDouble( &poisson_mu, argv[i] );
      if( InitialStringMatch( argv[i], "basispowersfile=" ) )
         StringToString( Basis_Powers_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "basisseparation=" ) )
	 StringToDouble( &basis_separation, argv[i] );
      if( InitialStringMatch( argv[i], "basisproposalwidth=" ) )
         StringToDouble( &basis_proposal_width, argv[i] );
   }

   if( basis_separation > 0 )
   {
      double range = normalise_predictors ? 1 : x_max[0] - x_min[0];
      if( normalise_predictors )
	 basis_separation /= x_max[0] - x_min[0];
      int implied_max_num_splines = (int)(range / basis_separation);
      if( max_num_splines > implied_max_num_splines )
	 max_num_splines = implied_max_num_splines;
   }
   if( treat_experiments_as_different_dimensions && strlen( Basis_Powers_Filename ) )
   {
      ifstream BasisPowersFile( Basis_Powers_Filename, ios::in );
      if( !BasisPowersFile )
      {
         cout<< "Basis powers file " << Basis_Powers_Filename << " could not be opened" << endl;
         exit(1);
      }
      create_zeroed_array( basis_powers, p, 3);
      for( int row = 0; row < p; row++ )
         for( int col = 0; col < 3; col++ )
            BasisPowersFile >> basis_powers[ row ][ col ];
   }
   if( power_max < power_min )
      power_min = power_max;
//   if( !knot_point_grid && data_block_sizes )
//      knot_point_grid = 10;//if different design points for each expt, need a spline grid to correlate the experiments.
   if( power_max < 0 )
      knot_point_grid = 0;
   if( gram_schmidt )
     //      if( !inputs_repeated_every )
	 //         gram_schmidt = false;
     //      else
      {
         standardize_design_matrix = false;
         if( knot_point_grid )//can't have more basis fns than x values if orthonormalizing
         {
            if( data_block_sizes )
            {
               create_zeroed_array( knot_point_grids, num_data_block_sizes );
               for( int j = 0; j < num_data_block_sizes; j++ )
               {
                  set_active_dimension( j );
		  int max_num_points_j = data_block_sizes[ j ];
		  if( power_max > 0 )
		    max_num_points_j -= (power_max - 1);
                  if( include_fourier_bases )
                     max_num_points_j -= 2 * max_fourier_basis;
                  if( knot_point_grid > max_num_points_j )
                     knot_point_grids[ j ] = max_num_points_j;
                  else
                     knot_point_grids[ j ] = knot_point_grid;
               }
            }
            else
            {
               int max_num_points = data_block_size;
               if( include_fourier_bases )
                  max_num_points -= 2 * max_fourier_basis;
               if( knot_point_grid > max_num_points )
                   knot_point_grid = max_num_points;
            }
         }
      }
   coeff_prior_st_dev = 1 / sqrt( coeff_prior_precision );
   number_unremovable_basis_functions = 0;
   Removed_Basis = Moved_Basis = NULL;
   create_design_matrix();
   max_num_splines_possible = number_unremovable_basis_functions + total_num_knot_point_candidates * ( power_max - power_min + 1 );
   if( !discrete_knot_points )
      max_num_splines_possible = 1000;
   if( !max_num_splines || max_num_splines > max_num_splines_possible )
      max_num_splines = max_num_splines_possible;
   if( gram_schmidt )
   {
      if( print_to_screen )
         cout << "Performing Gram-Schmidt orthonomalization process on design matrix columns." << endl;
      create_array( design_matrix_basis_fn_coeffs, num_design_cols, num_design_cols );
//       create_array( basis_fn_design_matrix_coeffs, num_design_cols, num_design_cols );
      gram_schmidt_process( D, design_matrix_basis_fn_coeffs, true, (unsigned int)inputs_repeated_every, p, num_1way_basis_fns, regression_num_subjects );
//      print_array( design_matrix_basis_fn_coeffs, num_design_cols, num_design_cols );
//      print_gsl_matrix(&(gsl_matrix_submatrix(D,0,0,num_design_cols, num_design_cols).matrix));
   }
   if( geometric_p > 0 && geometric_p < 1 )
   {
      log_geometric_p = log( geometric_p );
      log_1minusgeometric_p = log( 1- geometric_p );
      if( print_to_screen )
         cout << "Number of basis functions ~ Geom(" << geometric_p << ")" << endl;
      poisson_mu = 0;
   }
   else
      geometric_p = 0;
   if( poisson_mu > 0 )
   {
      poisson_mu *= global_x_range;
      log_poisson_mu = log( poisson_mu );
      if( print_to_screen )
         cout << "Number of basis functions ~ Poisson(" << poisson_mu << ")" << endl;
   }
   if( test_grid_points )
   {
      create_zeroed_array( knot_point_frequencies, test_grid_size - 1, p );
      if( !strlen( Knot_Point_Frequency_Filename ) )
      {
	 char* new_file = Knot_Point_Frequency_Filename;
	 strcpy( new_file, TargetFilename );
	 string insert = "_knot_point_frequencies_";
	 insert_in_string_before_dot( new_file, &insert );
      }
      create_zeroed_array( distances_to_last_knot, test_grid_size, p );
   }
}

Basis_Fn_Regression::Basis_Fn_Regression( gsl_matrix* X1, gsl_matrix* Y1 ): MCMC( X1, Y1 )
{
   preliminary_basis_fn_regression_initialisation();
   create_design_matrix();
}

Basis_Fn_Regression::Basis_Fn_Regression( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 ): MCMC( X1, Y1, D1 )
{
   preliminary_basis_fn_regression_initialisation();
}

Basis_Fn_Regression::~Basis_Fn_Regression()
{
   for( Basis_Fn_numtoptr::iterator iter = Basis_Fn_number_to_pointer.begin(); iter != Basis_Fn_number_to_pointer.end(); iter++ )
      zap( iter->second );
   zap_array( knot_point_candidates, p, total_num_knot_point_candidates );
   zap_array( design_matrix_basis_fn_coeffs, num_design_cols, num_design_cols );
   zap_array( knot_point_frequencies, test_grid_size - 1, p );
   zap_array( distances_to_last_knot, test_grid_size, p );
   if( knot_point_grids )
      delete [] knot_point_grids;
   if( num_1way_basis_fns )
      delete [] num_1way_basis_fns;
   Basis_Fn::destroy_Basis_Fn_counts();
}

void Basis_Fn_Regression::preliminary_basis_fn_regression_initialisation()
{
   current_basis_type = SPLINE_BASIS;//default to spline bases.
   two_sided_splines = false;
   knot_point_grid = 0;
   knot_point_grids = NULL;
   max_num_splines = 0;
   basis_separation = 0;
   full_1way_model_sign = 1;
   storing_basis_function_coefficients = 0;
//   coeff_prior_precision = pow(0.001,power_max + 1);
   coeff_prior_precision1 = coeff_prior_precision = .01;
   start_with_full_model = 1;
   knot_point_candidates = 0;
   create_zeroed_array( num_knot_point_candidates, p );
   total_num_knot_point_candidates = 0;
   standardize_design_matrix = !gram_schmidt;
   include_spline_bases = true;
   include_fourier_bases = false;
   basis_powers = NULL;
   design_matrix_basis_fn_coeffs = NULL;
   basis_fn_design_matrix_coeffs = NULL;
   create_null_array( first_unremovable_basis_fns, p );
   create_zeroed_array( num_1way_basis_fns, p );
   total_num_1way_basis_fns = 0;
   knot_point_frequencies = NULL;
   distances_to_last_knot = NULL;
   num_knot_point_frequency_samples = 0;
   discrete_knot_points = true;
   geometric_p = log_geometric_p = log_1minusgeometric_p = 0;
   poisson_mu = log_poisson_mu = 0;
   basis_proposal_width = 0.1;
   New_Basis = Basis_to_the_left = Basis_to_the_right = NULL;
   check_data_between_bases = false;
   if( density_estimation )
   {
      discrete_knot_points = false;
      start_with_full_model = false;
      standardize_design_matrix = false;
      power_max = power_min = 0;
      check_data_between_bases = true;
      Basis_Fn::initialize_component_Basis_Fn_counts( p );
   }
   *Basis_Powers_Filename = 0;
   *Knot_Point_Frequency_Filename = 0;
}

void Basis_Fn_Regression::create_design_matrix()
{
  //  cout << "calling derived class function" << endl;
   remove_all_basis_functions();
   if( discrete_knot_points )
      find_knot_point_candidates();
   else
      start_with_full_model = false;
   if( start_with_full_model )
      create_full_1way_model();
   else
      for( int cmp = 0; cmp < p; cmp++ )
         if( cmp == 0 || intercepts_for_each_dimension )
         {
            set_active_dimension( cmp );
            first_unremovable_basis_fns[ cmp ] = add_basis_function( normalise_predictors == 1 ? 0: x_min[cmp], cmp, 0, 0 );
            number_unremovable_basis_functions++;
         }
}

void Basis_Fn_Regression::add_basis_function( Basis_Fn* basis_fn_pointer, int* position )
{
   int i, j;
   int num_bases = Basis_Fn_number_to_pointer.size();
   int col = num_bases;
   if( position )
      col = *position;
   for( i = num_bases - 1; i >= col; i-- )
      Basis_Fn_number_to_pointer[ i + 1 ] = Basis_Fn_number_to_pointer[ i ];
   Basis_Fn_number_to_pointer[ col ] = basis_fn_pointer;
   number_removable_basis_fns++;
   if( basis_fn_pointer == Removed_Basis )
      Removed_Basis = NULL;
   if( basis_fn_pointer->get_num_interactions() == 1 )
   {
      int cmp = basis_fn_pointer->get_basis_fn_component()[ 0 ];
      num_1way_basis_fns[ cmp ]++;
      total_num_1way_basis_fns++;
   }
   if( density_estimation )
      return;
   double* D_new_column = new double[ n ];
   double* x_row_i = new double[ p ];
   int d_n = inputs_repeated_every ? inputs_repeated_every : n;
   int num_repeats = regression_n / d_n;
   int local_active_dimension = -1;
   int db_sum = 0;
   int block_counter = 0;
   for( i = 0; i < d_n; i++ )
   {
      for( j = 0; j < p; j++ )
         x_row_i[ j ] = gsl_matrix_get( X, i, j );
      if( seemingly_unrelated_regressions && i == db_sum )
      {
	 int exp_num = get_experiment_number( block_counter );
	 if( exp_num != local_active_dimension )
	    local_active_dimension = exp_num;//( local_active_dimension + 1 ) % p;
	 if( data_block_sizes )
	    data_block_size = data_block_sizes[ local_active_dimension ];
	 db_sum += data_block_size;
	 block_counter++;
      }
      double di = basis_fn_pointer->evaluate( x_row_i, seemingly_unrelated_regressions ? &local_active_dimension : NULL );
      for( int rep = 0; rep < num_repeats; rep++ )
         D_new_column[ rep * inputs_repeated_every + i ] = di;
   }
   if( n > regression_n )
      for( i = 0; i < num_test_points; i++ )
         D_new_column[ regression_n + i ] = basis_fn_pointer->evaluate( test_grid_points[ i ], NULL );
   delete [] x_row_i;
   if( standardize_design_matrix && d_n > 1 && !basis_fn_pointer->is_basis_standardized() )
   {
      double d_mean = 0, d_ssq = 0, d_sd;
      for( i = 0; i < d_n; i++ )
      {
         d_mean += D_new_column[ i ];
         d_ssq += D_new_column[ i ] * D_new_column[ i ];
      }
      d_mean /= d_n;
      d_ssq *= n / d_n;
      d_sd = sqrt( ( d_ssq - n * d_mean * d_mean ) / n );//( n - 1 ) );
      if( d_sd )
         for( i = 0; i < n; i++ )
            D_new_column[ i ] -= d_mean;
      else
      {
         d_mean = 0;
         d_sd = sqrt( d_ssq / n );//( n - 1 ) );
      }
      for( i = 0; i < n; i++ )
         D_new_column[ i ] /= d_sd;
      basis_fn_pointer->set_mean_sd( d_mean, d_sd );
//       if( d_sd )
//       {
//          for( i = 0; i < n; i++ )
//          {
//             D_new_column[ i ] -= d_mean;
//             D_new_column[ i ] /= d_sd;
//          }
//          basis_fn_pointer->set_mean_sd( d_mean, d_sd );
//       }
   }
   gsl_matrix *Dcopy = D;
   num_design_cols = D ? D->size2 : 0;
   col = num_design_cols;
   if( position )
      col = *position;
   D = gsl_matrix_alloc( n, num_design_cols + 1 );
   for( i = 0; i < n; i++ )
      gsl_matrix_set( D, i, col, D_new_column[ i ] );

   if( Dcopy )
      for( i = 0; i < n; i++ )
      {
         for( j = 0; j < col; j++ )
            gsl_matrix_set( D, i, j, gsl_matrix_get( Dcopy, i, j ) );
         for( j = col; j < num_design_cols; j++ )
            gsl_matrix_set( D, i, j + 1, gsl_matrix_get( Dcopy, i, j ) );
      }
   if( Dcopy )
      gsl_matrix_free( Dcopy );
   num_design_cols++;
   delete [] D_new_column;
   if( random_effects_model )
      calculate_design_matrix_column_sums();
   if( storing_basis_function_coefficients )
      draw_basis_coefficient_from_prior( basis_fn_pointer );
}

Basis_Fn* Basis_Fn_Regression::add_basis_function( double* v, int terms, int* cmp, int* pw, int* sgn, int* actv_dim )
{
   New_Basis = NULL;
   if( current_basis_type == SPLINE_BASIS )
      New_Basis = new Spline_Basis_Fn( v, terms, cmp, pw, sgn, actv_dim );
   else if( current_basis_type == FOURIER_BASIS )
   {
      New_Basis = new Fourier_Basis_Fn( v, terms, cmp, pw, sgn, actv_dim );
      if( !normalise_predictors )
         New_Basis->set_scale( x_max[cmp[0]] - x_min[cmp[0]] );
   }
   add_basis_function( New_Basis );
   return New_Basis;
}


Basis_Fn* Basis_Fn_Regression::add_basis_function( double d, int cmp, int pw, int sgn, int* actv_dim )
{
   double v[ 1 ];
   int cmp_vec[ 1 ], pw_vec[ 1 ], sgn_vec[ 1 ];
   v[ 0 ] = d;
   cmp_vec[ 0 ] = cmp;
   pw_vec[ 0 ] = pw;
   sgn_vec[ 0 ] = sgn;
   return add_basis_function( v, 1, cmp_vec, pw_vec, sgn_vec, actv_dim );
}

void Basis_Fn_Regression::add_random_basis_function()
{
   double v[ 1 ];
   int cmp_vec[ 1 ], pw_vec[ 1 ], sgn_vec[ 1 ];
   pw_vec[ 0 ] = power_min;
   sgn_vec[ 0 ] = full_1way_model_sign;   
   double v0;
   int cmp0 = 0;
   bool replica = true;
   while( replica )
   {
      replica = false;
      if( p > 0 )
         cmp0 = gsl_rng_uniform_int( rn, p );
      if( two_sided_splines )
         sgn_vec[ 0 ] = gsl_rng_uniform_int( rn, 3 ) - 1;
      if( power_max > power_min )
         pw_vec[ 0 ] = power_min + gsl_rng_uniform_int( rn, power_max - power_min + 1 );
      if( knot_point_candidates && knot_point_candidates[ cmp0 ] )
	 v0 = knot_point_candidates[ cmp0 ][ gsl_rng_uniform_int( rn, num_knot_point_candidates[ cmp0 ] ) ];
      else
         v0 = draw_continuous_basis_location( cmp0 );
      int i = 0;
      int num_bases = Basis_Fn_number_to_pointer.size();
      while( i < num_bases && !replica )
      {
         Basis_Fn* bas = Basis_Fn_number_to_pointer[ i ];
         if( bas->get_basis_fn_location()[ 0 ] == v0 && bas->get_basis_fn_power()[ 0 ] == pw_vec[ 0 ] && bas->get_basis_fn_component()[ 0 ] == cmp0 && bas->get_basis_fn_sign()[ 0 ] == sgn_vec[ 0 ] )
            replica = true;
         else
            i++;
      }
   }
   cmp_vec[ 0 ] = cmp0;
   v[ 0 ] = v0;
   add_basis_function( v, 1, cmp_vec, pw_vec, sgn_vec );
}

void Basis_Fn_Regression::remove_basis_function( int col, bool permanently )
{
   if( col < number_unremovable_basis_functions )
   {
      cout << "Can't remove this basis function, exiting" << endl;
      exit(1);
   }
   int new_num_bases = Basis_Fn_number_to_pointer.size() - 1;
   destroy_removed_basis();
   Removed_Basis = Basis_Fn_number_to_pointer[ col ];
   removed_basis_position = col;
   if( Removed_Basis->get_num_interactions() == 1 )
   {
      int cmp = Removed_Basis->get_basis_fn_component()[ 0 ];
      num_1way_basis_fns[ cmp ]--;
      total_num_1way_basis_fns--;
   }
   if( density_estimation )
   {
       Basis_to_the_left = get_first_basis_fn_to_the_left( Removed_Basis );
       Basis_to_the_right = get_first_basis_fn_to_the_right( Removed_Basis );
   }
   if( permanently )
   {
      delete Removed_Basis;
      Removed_Basis = NULL;
   }
   for( int i = col; i < new_num_bases; i++ )
      Basis_Fn_number_to_pointer[ i ] = Basis_Fn_number_to_pointer[ i + 1 ];
   Basis_Fn_number_to_pointer.erase( --Basis_Fn_number_to_pointer.end() ); // removes the extra "basis_fn number to pointer".
   number_removable_basis_fns--;
   if( density_estimation )
      return;
   matrix_delete_column( D, col );
   if( random_effects_model )
      calculate_design_matrix_column_sums();
}

void Basis_Fn_Regression::remove_all_basis_functions()
{
   int num_bases = Basis_Fn_number_to_pointer.size();
   if( num_bases )
   {
      for( int i = 0; i < num_bases; i++ )
      {
         delete Basis_Fn_number_to_pointer[ num_bases - 1 - i ];
         Basis_Fn_number_to_pointer.erase( --Basis_Fn_number_to_pointer.end() );
      }
      gsl_matrix_free( D );
   }
   D = 0;
   number_removable_basis_fns = 0;
}

void Basis_Fn_Regression::remove_random_basis_function()
{
   remove_basis_function( random_basis_fn_to_remove(), false );//does not delete the basis fn. permanently.
}

int Basis_Fn_Regression::random_basis_fn_to_remove()
{
   if( number_removable_basis_fns <= 0 )
   {
      cout << "No removable basis functions available" << endl;
      exit(1);
   }
   return number_unremovable_basis_functions + gsl_rng_uniform_int( rn, number_removable_basis_fns );
}

int Basis_Fn_Regression::get_number_of_basis_fn_deletions_possible()
{
   int num = number_removable_basis_fns;
   if( !number_unremovable_basis_functions )
      num--;
   return num;
}

void Basis_Fn_Regression::find_knot_point_candidates()
{
   double xval;
   int i, j;
   bool component_at_a_time = true;
   total_num_knot_point_candidates = 0;
   knot_point_candidates = new double*[ p ];
   if( knot_point_grid )
      for( j = 0; j < p; j++ )
      {
         if( knot_point_grids )
            knot_point_grid = knot_point_grids[ j ];
         if( power_min && !two_sided_splines )
            num_knot_point_candidates[ j ] = knot_point_grid - 1;
         else
            num_knot_point_candidates[ j ] = knot_point_grid;
         total_num_knot_point_candidates += num_knot_point_candidates[ j ];
         knot_point_candidates[ j ] = new double[ num_knot_point_candidates[ j ] ];
         for( i = 0; i < num_knot_point_candidates[ j ]; i++ )
         {
            if( two_sided_splines )
               xval = (double)i / (double)( num_knot_point_candidates[ j ] - 1 );
            else
               if( power_max < 0 )
                  xval = (double)( i + 1 ) / (double)( num_knot_point_candidates[ j ] );
               else
                  if( full_1way_model_sign != -1 )
                     xval = (double)i / (double)num_knot_point_candidates[ j ];
                  else
                     xval = (double)( i + 1 ) / (double)num_knot_point_candidates[ j ];
            if( normalise_predictors == 2 || !normalise_predictors )
               knot_point_candidates[ j ][ i ] = x_min[ j ] + ( x_max[ j ] - x_min[ j ] ) * xval;
            else
               knot_point_candidates[ j ][ i ] = xval;
         }
      }
   else
   {
      list<int> candidate_list;
      list<int> candidate_component_list;
      list<int>::iterator iter, iter_comp;
      int j_start = 0, j_end = p;
      if( component_at_a_time ) 
         j_end = 1;
      while( j_end <= p )
      {
//	 if( x_max[ j_start ] > x_min[ j_start ] )
            for( i = 0; i < (inputs_repeated_every ? inputs_repeated_every : regression_n) + ( deleted_x_point ? 1 : 0 ); i++ )
//	    for( i = 0; i < regression_n + ( deleted_x_point ? 1 : 0 ); i++ )
	    {
	       j = j_start;
	       set_active_dimension( j );
	       bool replica = false;
	       do{
		 replica = false;
		 if( i < regression_n )
		   xval = gsl_matrix_get( X, i, j );
		 else
		   xval = deleted_x_point[ j ];
		 bool not_smallest = normalise_predictors == 1 ? xval > 0 : ( xval - x_min[ j ] ) > 0;
		 bool not_largest = normalise_predictors == 1 ? ( 1 - xval ) != 0 : ( x_max[ j ] - xval ) != 0;
		 bool not_null = (normalise_predictors == 1) ? (xval >= 0) : (( xval - x_min[ j ] ) >= 0);
		 if( not_null && ( power_max < 0 || two_sided_splines || ( ( power_max > 0 || not_smallest ) && ( !power_min || ( ( full_1way_model_sign != -1 && not_largest ) || ( full_1way_model_sign == -1 && not_smallest ) ) ) ) ) )
		 {
		   iter = candidate_list.begin();
		   iter_comp = candidate_component_list.begin();
		   while( iter != candidate_list.end() && !replica )
                   {
		     if( gsl_matrix_get( X, *iter, j ) == xval  && *iter_comp == j )
		       replica = true;
                     ++iter;
                     ++iter_comp;
		   }
		 }
		 else
		   replica = true;
	       }while( replica && ++j < j_end );
            
	       if( !replica )
	       {
		 num_knot_point_candidates[ j_start ]++;
		 candidate_list.push_back( i );
		 candidate_component_list.push_back( j );
	       }
	    }
         j_start++;
         j_end++;
      }

      iter = candidate_list.begin();
      for( j = 0; j < p; j++ )
      {
         total_num_knot_point_candidates += num_knot_point_candidates[ j ];
         knot_point_candidates[ j ] = new double[ num_knot_point_candidates[ j ] ];
         for( i = 0; i < num_knot_point_candidates[ j ]; i++ )
         {
            int row_num = *iter;
            if( row_num < regression_n )
               knot_point_candidates[ j ][ i ] = gsl_matrix_get( X, row_num, j );
            else
	       knot_point_candidates[ j ][ i ] = deleted_x_point[ j ];
	    //	    cout << j << " " << i << " "<< knot_point_candidates[ j ][ i ] << endl;
            ++iter;
         }
      }
   }
}

void Basis_Fn_Regression::create_full_1way_model()
{
   for( int cmp = 0; cmp < p; cmp++ )
   {
      set_active_dimension( cmp );
      if( power_min >= 0 && ( cmp == 0 || intercepts_for_each_dimension ) )
      {
         first_unremovable_basis_fns[ cmp ] = add_basis_function( ( normalise_predictors == 1 ) ? 0 : x_min[ cmp ], cmp, 0, 0, intercepts_for_each_dimension ? &active_dimension : NULL );
         number_unremovable_basis_functions++;
      }
      if( !include_spline_bases && ( cmp == 0 || linear_term_for_each_dimension ) )
      {
         add_basis_function( ( normalise_predictors == 1 ) ? 0 : x_min[ cmp ], cmp, 1, 0, intercepts_for_each_dimension ? &active_dimension : NULL );
         number_unremovable_basis_functions++;
      }
      if( !max_num_splines || max_num_splines == max_num_splines_possible )
      {
         if( include_spline_bases )
            for( int i = 0; i < num_knot_point_candidates[ cmp ]; i++ )
            {
               for( int pw = power_min; pw <= power_max; pw++ )
                  add_basis_function( knot_point_candidates[ cmp ][ i ], cmp, pw, full_1way_model_sign, seemingly_unrelated_regressions ? &active_dimension : NULL );
            }
         if( include_fourier_bases )
         {
            int max_fourier_bases_cmp = max_fourier_basis;
            if( gram_schmidt )
            {
               int max_num_points_j = num_knot_point_candidates[ cmp ]; //data_block_sizes ? data_block_sizes[ cmp ] : data_block_size;
               if( max_num_points_j < 2 * max_fourier_basis + 1 )
                  max_fourier_bases_cmp = (max_num_points_j - 1) / 2;
            }
            current_basis_type = FOURIER_BASIS;
            for( int pw = 1; pw <= max_fourier_bases_cmp; pw++ )
            {
               add_basis_function( x_min[ cmp ], cmp, pw, 0, seemingly_unrelated_regressions ? &active_dimension : NULL );
               add_basis_function( x_min[ cmp ], cmp, pw, 1, seemingly_unrelated_regressions ? &active_dimension : NULL );
            }
            current_basis_type = SPLINE_BASIS;
         }
      }
      else
         for( int i = 0; i < max_num_splines - number_unremovable_basis_functions; i++ )
            add_random_basis_function();
   }
   set_basis_coefficients( 0.0 );
}

Basis_Fn* Basis_Fn_Regression::get_Basis_Fn_ptr( int basis_num )
{
   if( basis_num < 0 || basis_num > (int)Basis_Fn_number_to_pointer.size() )
   {
      cout << "no such basis fn, exiting" << endl;
      exit(1);
   }
   return Basis_Fn_number_to_pointer[ basis_num ];
}

double Basis_Fn_Regression::function_evaluate( double* v, int dim_v )
{
   int num_bases = Basis_Fn_number_to_pointer.size();
   Basis_Fn* bas;
   current_regression_fn_val = 0;
   for( int j = 0; j < num_bases; j++ )
   {
      bas = Basis_Fn_number_to_pointer[ j ];
      double basis_val = bas->evaluate( v, seemingly_unrelated_regressions ? &active_dimension : NULL ) * bas->get_coeff();
      current_regression_fn_val += basis_val;
   }
   return current_regression_fn_val;
}

double Basis_Fn_Regression::function_evaluate_bases( double* v )
{
   int num_bases = Basis_Fn_number_to_pointer.size();
   current_regression_fn_val = 0;
   for( int j = 0; j < num_bases; j++ )
      current_regression_fn_val +=  v[ j ] * Basis_Fn_number_to_pointer[ j ]->get_coeff();
   return current_regression_fn_val;
}

double Basis_Fn_Regression::integral_function_evaluate( double* v, int integral_order, int integral_component )
{
   double val = 0;
   int num_bases = Basis_Fn_number_to_pointer.size();
   Basis_Fn* bas;
   for( int j = 0; j < num_bases; j++ )
   {
      bas = Basis_Fn_number_to_pointer[ j ];
      val += bas->evaluate( v, seemingly_unrelated_regressions ? &active_dimension : NULL, integral_order, integral_component ) * bas->get_coeff();
   }
   if( normalise_predictors )
      val *= pow( x_scale[ 0 ], integral_order );
   return val;
}

double Basis_Fn_Regression::integral_absolute_function_evaluate( double* v, int integral_order )
{
   double val = 0;
   int num_bases = Basis_Fn_number_to_pointer.size();
   Basis_Fn *bas1, *bas2;
   double *t1, *t2, t1_0, t2_0;
   double f1;
   int basis_power = (Basis_Fn_number_to_pointer[ num_bases - 1 ]->get_basis_fn_power())[ 0 ];
   for( Basis_Fn_numtoptr::iterator iter = Basis_Fn_number_to_pointer.begin(); iter != Basis_Fn_number_to_pointer.end(); ++iter )
   {
      bas1 = iter->second;
      bas2 = get_first_basis_fn_to_the_right( bas1 );
      t1 = bas1->get_basis_fn_location();
      t2 = bas2->get_basis_fn_location();
      t1_0 = t1[ 0 ];
      t2_0 = t2[ 0 ];
      if( v[ 0 ] < t2_0 )
         t2_0 = v[ 0 ];
      if( t2_0 > t1_0 )
      {
         f1 = function_evaluate( t1, 1 );
         if( !basis_power )
         {
            val += abs( f1 ) * pow( ( x_max[ 0 ] - x_min[ 0 ] ) * ( t2_0 - t1_0 ), integral_order ) / gsl_sf_gamma( integral_order + 1 );
         }
         else if( basis_power == 1 )
         {
            double f2 = function_evaluate( t2, 1 );
            if( ( f2 > 0 && f1 > 0 ) || ( f2 < 0 && f1 < 0 ) )
            {
               double slope = ( abs( f2 ) - abs( f1 ) ) / ( ( x_max[ 0 ] - x_min[ 0 ] ) * ( t2_0 - t1_0 ) );
               val += abs( f1 ) * pow( ( x_max[ 0 ] - x_min[ 0 ] ) * ( t2_0 - t1_0 ), integral_order ) / gsl_sf_gamma( integral_order + 1 );
               val += slope * pow( ( x_max[ 0 ] - x_min[ 0 ] ) * ( t2_0 - t1_0 ), integral_order + 1 ) / gsl_sf_gamma( integral_order + 2 );
            }
            else
            {
               // first find intercept.
               double t3 = t1_0 + ( t2_0 - t1_0 ) * abs( f2 ) / (abs( f1 ) + abs( f2 ) );
               double slope1 = abs( f1 ) / ( ( x_max[ 0 ] - x_min[ 0 ] ) * ( t3 - t1_0 ) );
               double slope2 = abs( f2 ) / ( ( x_max[ 0 ] - x_min[ 0 ] ) * ( t2_0 - t3 ) );
               double denom = gsl_sf_gamma( integral_order + 2 );
               val += slope1 * pow( ( x_max[ 0 ] - x_min[ 0 ] ) * ( t3 - t1_0 ), integral_order + 1 ) / denom;
               val += slope2 * pow( ( x_max[ 0 ] - x_min[ 0 ] ) * ( t2_0 - t3 ), integral_order + 1 ) / denom;
            }
         }
      }
   }
   return val;
}

Basis_Fn* Basis_Fn_Regression::get_first_unremovable_basis_fn( int comp )
{
   if( !first_unremovable_basis_fns )
      return NULL;
   return first_unremovable_basis_fns[ comp ];
}

Basis_Fn* Basis_Fn_Regression::get_first_basis_fn_to_the_right( Basis_Fn* current_basis )
{
   double t = (current_basis->get_basis_fn_location())[ 0 ];
   int comp0 = (current_basis->get_basis_fn_component())[ 0 ];
   double min_dist = DBL_MAX;
   Basis_Fn *first_basis_to_the_right = current_basis, *basis_j;
   int num_bases = Basis_Fn_number_to_pointer.size();
   for( int j = 0; j < num_bases; j++ )
   {
      basis_j = Basis_Fn_number_to_pointer[ j ];
      if( basis_j != current_basis && (basis_j->get_basis_fn_component())[0] == comp0 )
      {
         double t_j = (basis_j->get_basis_fn_location())[ 0 ];
         if( t_j > t && ( t_j - t ) < min_dist )
         {
            min_dist = t_j - t;
            first_basis_to_the_right = basis_j;
         }
      }
   }
   return first_basis_to_the_right;
}

Basis_Fn* Basis_Fn_Regression::get_first_basis_fn_to_the_right( double t, int comp0 )
{
   double min_dist = DBL_MAX;
   Basis_to_the_right = NULL;
   Basis_Fn* basis_j;
   int num_bases = Basis_Fn_number_to_pointer.size();
   for( int j = 0; j < num_bases; j++ )
   {
      basis_j = Basis_Fn_number_to_pointer[ j ];
      if( (basis_j->get_basis_fn_component())[0] == comp0 )
      {
         double t_j = (basis_j->get_basis_fn_location())[ 0 ];
         if( t_j > t && ( t_j - t ) < min_dist )
         {
            min_dist = t_j - t;
            Basis_to_the_right = basis_j;
         }
      }
   }
   return Basis_to_the_right;
}

Basis_Fn* Basis_Fn_Regression::get_first_basis_fn_to_the_left( Basis_Fn* current_basis )
{
   double t = (current_basis->get_basis_fn_location())[ 0 ];
   double min_dist = DBL_MAX;
   Basis_Fn *first_basis_to_the_left = current_basis, *basis_j;
   int num_bases = Basis_Fn_number_to_pointer.size();
   for( int j = num_bases - 1; j >= 0; --j )
   {
      basis_j = Basis_Fn_number_to_pointer[ j ];
      if( basis_j != current_basis )
      {
         double t_j = (basis_j->get_basis_fn_location())[ 0 ];
         if( t_j < t && ( t - t_j ) < min_dist )
         {
            min_dist = t - t_j;
            first_basis_to_the_left = basis_j;
         }
      }
   }
   return first_basis_to_the_left;
}

Basis_Fn* Basis_Fn_Regression::get_first_basis_fn_to_the_left( double t )
{
   double min_dist = DBL_MAX;
   Basis_to_the_left = NULL;
   Basis_Fn* basis_j;
   int num_bases = Basis_Fn_number_to_pointer.size();
   for( int j = num_bases - 1; j >= 0; --j )
   {
      basis_j = Basis_Fn_number_to_pointer[ j ];
      double t_j = (basis_j->get_basis_fn_location())[ 0 ];
      if( t_j < t && ( t - t_j ) < min_dist )
      {
         min_dist = t - t_j;
         Basis_to_the_left = basis_j;
      }
   }
   return Basis_to_the_left;
}

Basis_Fn* Basis_Fn_Regression::get_leftmost_removable_basis_fn()
{
   if( !number_removable_basis_fns )
      return NULL;
   Basis_Fn *bas, *leftmost_basis = Basis_Fn_number_to_pointer[ number_unremovable_basis_functions ];
   double bas_location, leftmost_location = leftmost_basis->get_basis_fn_location()[ 0 ];
   int i = 1;
   while( i < number_removable_basis_fns )
   {
      bas = Basis_Fn_number_to_pointer[ number_unremovable_basis_functions + i ];
      bas_location = bas->get_basis_fn_location()[ 0 ];
      if( bas_location < leftmost_location )
      {
         leftmost_basis = bas;
         leftmost_location = bas_location;
      }
      i++;
   }
   return leftmost_basis;
}

int Basis_Fn_Regression::position_in_basis_number_to_pointer( Basis_Fn* bas )
{
   int num_bases = Basis_Fn_number_to_pointer.size();
   int position = 0;
   while( position < num_bases )
   {
      if( Basis_Fn_number_to_pointer[ position ] == bas )
         return position;
      position++;
   }
   return -1;
}
   
double Basis_Fn_Regression::average_function_squared_at_bases_evaluate()
{
   double val = 0;
   int unique_counter = 0;
   double *v = new double[ p ];
   for( int j = 0; j < n; j++ )
   {
      for( int j2 = 0; j2 < p; j2++ )
         v[ j2 ] = gsl_matrix_get( X, j, j2 );
      bool unique = true;
      int i = 0;
      while( unique && i < j )
      {
         unique = false;
         int j2 = 0;
         while( j2 < p && !unique )
         {
            if( v[ j2 ] == gsl_matrix_get( X, i, j2 ) )
               unique = true;
            else
               j2++;
         }
         i++;
      }
      if( unique )
      {
         unique_counter++;
         double d = function_evaluate( v, p );
         val += d * d;
      }
   }
   delete [] v;
   return val / unique_counter;
}

void Basis_Fn_Regression::set_basis_coefficients( gsl_matrix* m )
{
   int num_bases = Basis_Fn_number_to_pointer.size();
   if( (int)m->size1 != num_bases )
   {
      cout << "Number of coefficients provided different to the number of basis functions" << endl;
      exit(1);
   }
   
   for( int i = 0; i < num_bases; i++ )
   {
      double coeff_i = 0;
      if( !design_matrix_basis_fn_coeffs )
         coeff_i = gsl_matrix_get( m, i, 0 );
      else
         for( int j = 0; j < num_bases; j++ )
            coeff_i += design_matrix_basis_fn_coeffs[ j ][ i ] * gsl_matrix_get( m, j, 0 );
      get_Basis_Fn_ptr( i )->set_coeff( coeff_i );
   }
}

void Basis_Fn_Regression::set_basis_coefficients( double* v )
{
   int num_bases = Basis_Fn_number_to_pointer.size();
   for( int i = 0; i < num_bases; i++ )
   {
      double coeff_i = 0;
      if( !design_matrix_basis_fn_coeffs )
         coeff_i =  v[ i ];
      else
         for( int j = 0; j < num_bases; j++ )
	    coeff_i += design_matrix_basis_fn_coeffs[ j ][ i ] *  v[ j ];
      get_Basis_Fn_ptr( i )->set_coeff( coeff_i );
   }
}

void Basis_Fn_Regression::set_basis_coefficients( double d )
{
   int num_bases = Basis_Fn_number_to_pointer.size();
   for( int i = 0; i < num_bases; i++ )
      get_Basis_Fn_ptr( i )->set_coeff( d );
}

void Basis_Fn_Regression::set_basis_coefficients()
{
   int num_bases = Basis_Fn_number_to_pointer.size();
   for( int i = 0; i < num_bases; i++ )
      get_Basis_Fn_ptr( i )->set_coeff( gsl_matrix_get(beta, i, 0 ) );
}

void Basis_Fn_Regression::revert_basis_coefficients()
{
   int num_bases = Basis_Fn_number_to_pointer.size();
   for( int i = 0; i < num_bases; i++ )
      get_Basis_Fn_ptr( i )->revert_coeff();
}

double Basis_Fn_Regression::get_absolute_sum_basis_type_coefficients( BasisType basis_type, int* component, int power_upper_bound )
{
   int num_bases = Basis_Fn_number_to_pointer.size();
   double fourier_abs_sum = 0;
   for( int i = 0; i < num_bases; i++ )
   {
      Basis_Fn* bas = Basis_Fn_number_to_pointer[ i ];
      int bas_power = bas->get_basis_fn_power()[0];
      if( bas->get_basis_type() == basis_type && (!component || bas->get_basis_fn_component()[0] == *component) && bas_power != 0 && bas_power <= power_upper_bound )
      {
         double coeff = bas->get_coeff();
         if( coeff > 0 )
            fourier_abs_sum += coeff;
         else
            fourier_abs_sum -= coeff;
      }
   }
   return fourier_abs_sum;
}

void Basis_Fn_Regression::print_basis_fns()
{
   int num_bases = Basis_Fn_number_to_pointer.size();
   for( int i = 0; i < num_bases; i++ )
   {
      Basis_Fn* bas = Basis_Fn_number_to_pointer[ i ];
      double* v = bas->get_basis_fn_location();
      int size_v = bas->get_num_interactions();
      for( int j = 0; j < size_v; j++ )
         cout << v[ j ] << "  ";
      cout << ": ";
      cout << bas->get_coeff() << endl;
   }
}

void Basis_Fn_Regression::draw_random_basis_fn()
{
   int num_bases = Basis_Fn_number_to_pointer.size();
   moved_basis_position = gsl_rng_uniform_int( rn, num_bases );
   Moved_Basis = get_Basis_Fn_ptr( moved_basis_position );
}

void Basis_Fn_Regression::draw_random_basis_coefficient_from_prior()
{
   draw_random_basis_fn();
   draw_basis_coefficient_from_prior( Moved_Basis );
}

void Basis_Fn_Regression::draw_basis_coefficient_from_prior( Basis_Fn* bas )
{
   bas->set_coeff( gsl_ran_gaussian( rn, coeff_prior_st_dev ) );
}

double Basis_Fn_Regression::calculate_basis_coefficient_log_prior( Basis_Fn* bas )
{
   return calculate_basis_coefficient_log_prior( bas->get_coeff() );
}

double Basis_Fn_Regression::calculate_basis_coefficient_log_prior( double basis_coeff )
{
   return -.5 * ( LOG_PI + LOG_TWO ) - log( coeff_prior_st_dev ) - .5 * basis_coeff * basis_coeff * coeff_prior_precision;
}

double Basis_Fn_Regression::calculate_basis_coefficients_log_prior()
{
   double d = 0;
   for( map< int, Basis_Fn*>::iterator iter = Basis_Fn_number_to_pointer.begin(); iter != Basis_Fn_number_to_pointer.end(); iter++ )
      d += calculate_basis_coefficient_log_prior( iter->second );
   return d;
}

void Basis_Fn_Regression::create_design_point( double* v )
{
   int new_design_point_dim = Basis_Fn_number_to_pointer.size();
   if( design_point && design_point_dim != new_design_point_dim )
      zap_vector( design_point );
   design_point_dim = new_design_point_dim;
   if( !design_point )
      design_point = new double[ design_point_dim ];
   for( int j = 0; j < design_point_dim; j++ )
   {
      Basis_Fn* bas = Basis_Fn_number_to_pointer[ j ];
      design_point[ j ] =  bas->evaluate( v, seemingly_unrelated_regressions ? &active_dimension : NULL );
   }
}

void Basis_Fn_Regression::test_for_monotonicity()
{
   isotonic = antitonic = true;
   if( !number_removable_basis_fns )
      return;
   Basis_Fn* bas = NULL;
   double coeff = 0;
   int i = 0;
   while( i < number_removable_basis_fns && ( isotonic || antitonic ) )
   {
      if( !power_max )
      {
         bas = Basis_Fn_number_to_pointer[ number_unremovable_basis_functions + i ];
         coeff = bas->get_coeff();
      }
      else
      {
         if( i == 0 )
            bas = get_leftmost_removable_basis_fn();
         else
            bas = get_first_basis_fn_to_the_right( bas );
       coeff += bas->get_coeff();
      }
      if( coeff < 0 )
         isotonic = false;
      else if( coeff > 0 )
         antitonic = false;
      i++;
   }
}

bool Basis_Fn_Regression::room_for_new_basis_fn()
{
   if( basis_separation > 0 )
   {
      if( 2 * get_number_of_basis_fns() < max_num_splines )
	 return true;
      double left_edge = normalise_predictors ? 0 : x_min[0];
      double right_edge = normalise_predictors ? 1 : x_max[0];
      while( right_edge > left_edge )
      {
	 Basis_Fn* bas = get_first_basis_fn_to_the_left( right_edge );
	 double left_value = bas->get_basis_fn_location()[0];
	 if( right_edge - left_value > 2 * basis_separation )
	    return true;
	 right_edge = left_value;
      }
      return false;
   }
   else
      return true;
}

void Basis_Fn_Regression::set_active_dimension( int activ_dim )
{
   Regression::set_active_dimension( activ_dim );
   if( basis_powers )
   {
      power_min = basis_powers[ active_dimension ][ 0 ];
      power_max = basis_powers[ active_dimension ][ 1 ];
      max_fourier_basis = basis_powers[ active_dimension ][ 2 ];
      if( max_fourier_basis <= 0 )
      {
         include_fourier_bases = false;
         include_spline_bases = true;
      }
      else
         include_fourier_bases = true;
      if( power_min > power_max || power_max < -1 )
         include_spline_bases = false;
      if( ( include_fourier_bases || include_spline_bases ) == false )
      {
         include_spline_bases = true;
         power_min = power_max = -1;
      }
   }
}

double Basis_Fn_Regression::calculate_log_prior()
{
   proposed_log_prior = log_geometric_p;
   if( geometric_p > 0 )
      proposed_log_prior = (get_number_of_basis_fns() - number_unremovable_basis_functions) * log_1minusgeometric_p;
   else if( poisson_mu > 0 )
      proposed_log_prior = -poisson_mu + number_unremovable_basis_functions * log_poisson_mu - gsl_sf_lnfact( number_unremovable_basis_functions );
   return proposed_log_prior;
}

double Basis_Fn_Regression::draw_continuous_basis_location( int cmp0 )
{
   bool satisfactory_draw = false;
   double v0 = 0;
   double lb = 0, rb = 1;
   if( Basis_to_the_left )
   {
      lb = Basis_to_the_left->get_basis_fn_location()[0];
      if( Basis_to_the_right && Basis_to_the_right != Removed_Basis )
         rb = Basis_to_the_right->get_basis_fn_location()[0];
      else
         if( normalise_predictors != 1 )
            rb = x_max[cmp0];
   }
   else
      if( normalise_predictors != 1 )
      {
         lb = x_min[cmp0];
         rb = x_max[cmp0];
      }
   lb += basis_separation;
   rb -= basis_separation;

   if( Removed_Basis )//propose symetrically around the deleted basis function.
   {
      double proposal_centre = Removed_Basis->get_basis_fn_location()[0];
      double width = min( rb-proposal_centre , proposal_centre-lb );
      if( width > basis_proposal_width )
      {
         lb = proposal_centre - basis_proposal_width;
         rb = proposal_centre + basis_proposal_width;
      }
   }
   
   if( Basis_to_the_left )
   {
      v0 = lb + (rb - lb) * gsl_rng_uniform( rn );
      satisfactory_draw = true;
   }

   while( !satisfactory_draw )
   {
      v0 = lb + (rb - lb) * gsl_rng_uniform( rn );
      if( density_estimation || check_data_between_bases || basis_separation > 0 )
      {
          get_first_basis_fn_to_the_left( v0 );
      }
      if( !(check_data_between_bases || basis_separation > 0 ) )
         satisfactory_draw = true;
      else
      {
         double t_l = - DBL_MAX;
//         Basis_Fn* bas_l = get_first_basis_fn_to_the_left( v0 );
         if( Basis_to_the_left )
            t_l = (Basis_to_the_left->get_basis_fn_location())[0];
         double t_u = 1;//DBL_MAX;
//         Basis_Fn* bas_r = get_first_basis_fn_to_the_right( v0, cmp0 );
         get_first_basis_fn_to_the_right( v0, cmp0 );
         if( Basis_to_the_right )
            t_u = (Basis_to_the_right->get_basis_fn_location())[0];
	 else if( normalise_predictors != 1 )
            t_u = x_max[cmp0];
         int i = 0;
         bool check_left_next = true, check_right_next = true;
         while( i < regression_n && !satisfactory_draw )
         {
            if( gsl_matrix_get( X, i, cmp0 ) >= t_l && gsl_matrix_get( X, i, cmp0 ) < t_u )
	    {
               satisfactory_draw = true;
	       if( gsl_matrix_get( X, i, cmp0 ) < v0 )
		  check_left_next = false;//then don't need to check this one
	       else
		  check_right_next = false;
	    }
            else
               i++;
         }
         if( satisfactory_draw && check_left_next && Basis_to_the_left )
         {
             satisfactory_draw = false;
             double t_ll = - DBL_MAX;
             Basis_Fn* bas_ll = get_first_basis_fn_to_the_left( Basis_to_the_left );
             if( bas_ll )
                 t_ll = (bas_ll->get_basis_fn_location())[0];
             i = 0;
             while( i < regression_n && !satisfactory_draw )
             {
                 if( gsl_matrix_get( X, i, cmp0 ) >= t_ll && gsl_matrix_get( X, i, cmp0 ) < v0 )
                     satisfactory_draw = true;
                 else
                     i++;
             }
         }
         if( satisfactory_draw && check_right_next && Basis_to_the_right )
         {
             satisfactory_draw = false;
             double t_ru = DBL_MAX;
             Basis_Fn* bas_rr = get_first_basis_fn_to_the_right( Basis_to_the_right );
             if( bas_rr && bas_rr != Basis_to_the_right )
                 t_ru = (bas_rr->get_basis_fn_location())[0];
             i = 0;
             while( i < regression_n && !satisfactory_draw )
             {
                 if( gsl_matrix_get( X, i, cmp0 ) >= v0 && gsl_matrix_get( X, i, cmp0 ) < t_ru )
                     satisfactory_draw = true;
                 else
                     i++;
             }
         }
	 if( basis_separation > 0 && ( !significantly_different( v0, t_l, basis_separation ) || !significantly_different( v0, t_u, basis_separation ) ) )
	    satisfactory_draw = false;
      }
   }
   return v0;
}

void Basis_Fn_Regression::update_knot_point_frequencies()
{
   num_knot_point_frequency_samples++;
   Basis_Fn *old_bas, *bas, *leftmost_basis = get_leftmost_removable_basis_fn();
   if( !leftmost_basis )
     return;
   double distance_to_last_knot = 0;
   for( int j = 0; j < p; j++ )
   {
      bas = leftmost_basis;
      bool last_basis = false;
      double t_loc = bas->get_basis_fn_location()[ 0 ];
      double t_left = test_grid_points[ 0 ][ j ];
      int i = 1;
      while( i < (test_grid_size-1) && !last_basis )
      {
	  double t_right = test_grid_points[ i ][ j ];
	  if( t_loc < t_right && t_loc >= t_left )
	  {
	     knot_point_frequencies[ i ][ j ]++;
	     while( t_loc < t_right && !last_basis )
	     {
	        distance_to_last_knot = (t_right-t_loc);
	        old_bas = bas;
	        bas = get_first_basis_fn_to_the_right( t_loc, j );
		if( bas && bas != old_bas )
		   t_loc = bas->get_basis_fn_location()[ 0 ];
		else
		   last_basis = true;
		//		cout << bas << "  " << old_bas << " " << t_loc << " " << t_right << " " << i << endl;
	     }
	  }
	  else
	     distance_to_last_knot += (t_right-t_left);
	  if( distances_to_last_knot )
	     distances_to_last_knot[i][j] += distance_to_last_knot;
	  t_left = t_right;
	  i++;
      }
   }
}

void Basis_Fn_Regression::write_knot_point_frequencies( ofstream *any_stream )
{
   if( !knot_point_frequencies )
      return;
   if( !any_stream )
   {
      if( strlen( Knot_Point_Frequency_Filename ) )
      {
         ofstream KnotFreqStream( Knot_Point_Frequency_Filename, ios::out );
         write_knot_point_frequencies( &KnotFreqStream );
      }
      return;
   }
   for( int i = 0; i < test_grid_size - 1; i++ )
   {
      for( int j = 0; j < p; j++ )
      {
	*any_stream << (double)knot_point_frequencies[i ][ j ]/num_knot_point_frequency_samples << " ";
	if( distances_to_last_knot )
	  *any_stream << x_scale[ j ]*distances_to_last_knot[i ][ j ]/num_knot_point_frequency_samples << " ";
      }
      *any_stream << endl;
   }
}
