// cluster_glm_mvn.h

#if !defined( _cluster_glm_mvn_h )
#define _cluster_glm_mvn_h

#include "glm_mvn_clustering.h"
///
class GLM_MVN_Clustering;

class Cluster_GLM_MVN
{
  public:
   ///
   Cluster_GLM_MVN( gsl_matrix_int* num_obsns_new, gsl_matrix* yty_new, gsl_matrix* ytd_new, gsl_matrix* dtd_new, bool distinct_new, bool mergeable_new, list<int>& subjects_allocated_new, double closeness_contrib_new );
   ///
   Cluster_GLM_MVN( gsl_matrix_int* num_obsns_new, gsl_matrix* yty_new, gsl_matrix* ytd_new, gsl_matrix* dtd_new, bool distinct_new, bool mergeable_new, list<int>& subjects_allocated_new, double closeness_contrib_new, string label_new );
   ///
   Cluster_GLM_MVN( list<Cluster_GLM_MVN*> clusters_allocated_new );
   ///
   Cluster_GLM_MVN( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2, bool temporary = false );
   ///
   void initialise();//method called at start of all constructors in this class.
   
   ~Cluster_GLM_MVN();
   ///

   static void kill_static_variables();//function called to kill delete any static variable memory allocated.

   friend ostream& operator<< ( ostream& os, Cluster_GLM_MVN* & cl );

   void order_clusters_allocated();//when preserving tree, this orders the clusters_allocated lists so that similar clusters are next to each other when printing out the hierarchy.
   ///
   void order_clusters_allocated( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 );//when not preserving tree,this orders the subjects_allocated lists so that similar clusters are next to each other when printing out the hierarchy.
   ///
   int find_best_cluster_order( Cluster_GLM_MVN* cl11, Cluster_GLM_MVN* cl12, Cluster_GLM_MVN* cl21, Cluster_GLM_MVN* cl22, Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 );//used by above two order_clusters_allocated functions to find best arrangement of child clusters (first 4 arguments). parents are also given as arguments in function (last 2 arguments).
   ///
   void reverse_clusters_allocated();//works down through the tree, reversing the lists clusters_allocated.
   ///
   void switch_subjects_about_join();//split subjects_allocated at subject_join, switch the two pieces and glue together again.
   ///

   gsl_matrix_int* get_num_obsns( bool recursive_call = false );

   int get_total_num_obsns();

   int get_num_experiments_represented();//returns the number of experiments i for which *num_obsns[ i ] is non-zero.

   gsl_matrix* get_yty( bool recursive_call = false );

   gsl_matrix* get_ytd( bool recursive_call = false );

   gsl_matrix* get_dtd( bool recursive_call = false );

   double get_yty_sum() const{ return yty_sum; }

   void set_v_index( int new_v_index ){ v_index = new_v_index; }

   int get_v_index(){ return v_index; }

   bool get_distinct() const{ return distinct; }

   bool get_mergeable() const{ return mergeable; }

   void set_mergeable( bool new_mergeable ){ mergeable_store = mergeable; mergeable = new_mergeable; }

   void revert_mergeable(){ mergeable = mergeable_store; }

   bool get_splitable() const{ return splitable; }

   static void set_split_off_one_subject( bool new_split_off_one_subject ){ split_off_one_subject = new_split_off_one_subject; }

   static void set_two_means_separation( bool new_two_means_separation ){ two_means_separation = new_two_means_separation; }
   
   static bool get_split_off_one_subject(){ return split_off_one_subject; }

   string* get_label();

   void set_label( string new_label );

   void set_y_averages( gsl_matrix* new_y_averages = NULL );

   gsl_matrix* get_y_averages();

   list<int> get_subjects_allocated();

   list<Cluster_GLM_MVN*> get_clusters_allocated(){ return clusters_allocated; }

   int get_num_clusters_allocated(){ return clusters_allocated.size(); }

   map< Cluster_GLM_MVN*, double> get_single_subject_clusters_allocated( bool deleting_tree = false );//returns a map containing all the single subject clusters underneath this cluster, killing all parental clusters in between - if deleting_tree true, deletes all the parent clusters along the way.

   void get_single_subject_clusters_allocated( map<Cluster_GLM_MVN*, double>& map_of_clusters, bool deleting_tree = false );

   int get_num_single_subject_clusters_allocated( bool top = true );//returns the size of the map that would be created by the functions above

   Cluster_GLM_MVN* draw_random_single_subject_cluster_allocated();//draws a random single subject cluster from those allocated to this cluster.

   map<Cluster_GLM_MVN*, double> get_all_clusters_allocated();//returns a map holding all the nodes in the tree under this cluster.

   void get_all_clusters_allocated( map<Cluster_GLM_MVN*, double>& map_of_clusters );//used by the fn. above.

   Cluster_GLM_MVN* draw_random_cluster_from_all_allocated();//draws a random cluster from all the nodes under this cluster in the tree.

   void delete_down_to_subject_clusters();//deletes all the clusters between this one and the single subjects clusters at the bottom.
   
   void get_clusters_allocated( Cluster_GLM_MVN*& cl1, Cluster_GLM_MVN*& cl2 );//put pointers to the first two clusters allocated in the argument variables.

   double get_number_of_possible_clusterings();//returns the number of possible clusterings of the cluster's subjects according to the tree beneath it.

   void get_random_size_n_cluster_clustering( int size_n, map<Cluster_GLM_MVN*,double>& map_of_clusters );//returns a map containing a random partition of this cluster into size_n clusters.
      
   Cluster_GLM_MVN* get_sibling_cluster() const;//returns the first child (cluster allocated) of the parent cluster which is not this.

   bool cluster_contains_subject( int subject );

   bool cluster_has_subjects( list<int>& cl_subjects );//returns true if cluster's subjects allocated are precisely cl_subjects.
   
   bool cluster_has_child_with_subjects( list<int>& cl_subjects );//returns true if cluster has a child (cluster allocated) whose subjects allocated are precisely cl_subjects.
   
   Cluster_GLM_MVN* find_sub_cluster_containing_subjects( list<int> splitting_subjects );//returns the cluster pointer from clusters_allocated which holds the subjects splitting_subjects.
   
   void print_subjects_allocated();

   int get_num_subjects_allocated();

   int size(){ return get_num_subjects_allocated();}

   bool is_singleton_cluster(){ return num_subjects_allocated== 1; }

   void find_smallest_and_largest_subjects_allocated();//calculates the smallest & largest indices of subjects_allocated;
   
   int get_smallest_subject_allocated(){ return smallest_subject; }

   int get_largest_subject_allocated(){ return largest_subject; }

   int get_first_subject_allocated();//return the first element of the subjects allocated;

   int get_last_subject_allocated();//return the last element of the subjects allocated;

   void get_first_and_last_subjects_allocated( int& first_subject, int& last_subject );
   
   list<pair <int,int> > get_merger_details();// returns the merger details for subtree under this cluster.
   
   double get_closeness_contribution();

   void set_closeness_contribution( double closeness_contrib_new );

   void revert_closeness_contribution(){ closeness_contribution = *old_closeness_contribution; }

   void set_closest_cluster( Cluster_GLM_MVN* new_closest_cluster ){ closest_cluster = new_closest_cluster; }

   void set_second_closest_cluster( Cluster_GLM_MVN* new_second_closest_cluster ){ second_closest_cluster = new_second_closest_cluster; }

   Cluster_GLM_MVN* get_closest_cluster();

   Cluster_GLM_MVN* get_second_closest_cluster(){ return second_closest_cluster; }

   map< Cluster_GLM_MVN*, double, less <Cluster_GLM_MVN*> > get_closest_clusters( int k );// returns the sub-map of closeness_map containing the k nearest clusters.
   
   void set_cluster_closeness( Cluster_GLM_MVN* cl, const double d );// sets the closeness of this cluster to *cl as d.

   double get_cluster_closeness( Cluster_GLM_MVN* cl, bool look_up = true );// get the closeness of cluster* cl to this one. If look_up true, check to see if it is stored before calculating.

   double get_heller_prior_d();

   void erase_from_closeness_map( Cluster_GLM_MVN* cl );

   void clear_closeness_map(){ closeness_map.clear(); };

   double find_closest_cluster( map< Cluster_GLM_MVN*, double >* map_of_clusters_ptr = NULL );//finds the closes cluster to this one, makes it the closest_cluster and returns how close it is.

   Cluster_GLM_MVN* merge_with_nearest_cluster();//calls fn below for nearest neighbour.
      
   Cluster_GLM_MVN* merge_cluster( Cluster_GLM_MVN*& cl, bool calc_closeness = true );//merges this cluster with *cl, and deletes cl.

   Cluster_GLM_MVN* split_cluster( list<int>* cl_subjects_ptr = NULL );//randomly (or deterministically if pointer not null)  splits this cluster in two, returning a pointer to the newly formed cluster.

   Cluster_GLM_MVN* split_cluster( int cl_subject );//calls the above function with a list containing just cl_subject.

   int find_random_subject();//randomly pick one subject from the subjects_allocated.
   
   void find_two_random_subjects( int& subj1, int& subj2 );//randomly pick two subjects from the subjects_allocated.

   void find_most_distant_subjects( int& subj1, int& subj2 );//returns the two subject in subjects_allocated which are furthest apart according to euclidean distance or correlation.

   void initialise_two_means_clustering( list<int>& cl_subjects );//performs initialisation of k-means clustering with k=2;

   void initialise_two_means_clustering( list<int>& cl_subjects, list<int>& cl_subjects2 );//performs initialisation of k-means clustering with k=2, returns a two list format.

   void tempered_gibbs_sweep( list<int>& cl_subjects, list<int>& cl_subjects2, list<int>* cl_subjects_ptr = NULL, int number_of_scans = 1, bool calculate_probabilities = true );//starting with two subclusters defined by the two subject lists passed, iterate through the subjects, reallocating according to predictive probabilities, raised to a temperature.  Also calculates the log probability of the sweep.
   
   void add_subject( int subject, string* cl_label = NULL );// adds the subject to the cluster with option label *cl_label.

   void set_clusters_allocated( list<Cluster_GLM_MVN*> new_clusters_allocated );//set clusters_allocated to be equal to this new list.

   void set_clusters_allocated( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 );//allocates the two clusters to this cluster.

   void remove_allocated_cluster( Cluster_GLM_MVN* cl );// replaces cl in the clusters_allocated with cl_new.
   
   static int get_cluster_count() { return cluster_count; }// returns the current number of Cluster_GLM_MVN objects.

   static int get_number_mergeable(){ return number_mergeable; }

   static int get_number_splitable(){ return number_splitable; }

   void set_parent_cluster( Cluster_GLM_MVN* new_parent_cluster );

   Cluster_GLM_MVN* get_parent_cluster() const { return parent_cluster; }

   double get_sibling_closeness( Cluster_GLM_MVN* cl_sibling = NULL );//returns the closeness of this to the other cluster it shares its parent with.

   double get_sibling_merge_prob( Cluster_GLM_MVN* cl_sibling = NULL );//returns the closeness of this to the other cluster it shares its parent with, corrected by the change in prior on the cluster sizes if they were to merge.
   
   double get_childrens_closeness();//returns the closeness of the two clusters allocated to this cluster, if it has two, or -DBL_MAX otherwise.
   
   double get_childrens_merge_prob();//retruns the closeness of the two clusters allocated to this cluster, corrected by the change in prior on the cluster sizes if merged.
   
   void set_targets();//allocates to ys a gsl_matrix containing the entries of Y belonging to the subjects allocated.

   gsl_matrix* get_targets();//returns a gsl_matrix containing the entries of Y belonging to the subjects allocated (=ys if these are being stored).
   
   static void set_Rptr( GLM_MVN_Clustering* Rptr_new ){ Rptr = Rptr_new; }

   void set_my_root_cluster( Cluster_GLM_MVN* new_root_cluster );
   
   static void set_root_cluster( Cluster_GLM_MVN* new_root_cluster);

   static void set_root_cluster( string cl_label, Cluster_GLM_MVN* new_root_cluster);

   Cluster_GLM_MVN* get_my_root_cluster();
   
   static Cluster_GLM_MVN* get_root_cluster(){ return root_cluster; }

   static Cluster_GLM_MVN* get_root_cluster( string& cl_label );//returns the cluster at the top of the hierarch for label cl_label.

   static list<Cluster_GLM_MVN*> get_root_clusters();//returns all the root clusters in a list.
   static int get_number_of_root_clusters();

   bool is_root_cluster(){ return this == get_my_root_cluster(); }
   
   static void destroy_tree();//kills all clusters except those containing subjects.

   void destroy_clusters_allocated();//kills all clusters allocated except those containing subjects.

   static void empty_parent_data_matrices_from_top();//kills all the data matrices of all clusters except those containing subjects - these matrices would never change.

   static double get_gibbs_sweep_log_prob(){ return gibbs_sweep_log_prob; }

   void empty_parent_data_matrices();//kills all the data matrices of all clusters working down from the current cluster, except those containing subjects.

   void set_score( int new_score ){ score = new_score; }

   int get_score() const { return score; }
   
   bool up_to_date();//returns true if a MCMC move has not been accepted since this cluster's statistics were last updated.

   int get_type(){ if( !using_types ) return 0; return type; }

   void set_type( int new_type ){ type = new_type; }
   
   static void set_using_types( bool new_using_types ){ using_types = new_using_types; }
   static bool get_using_types(){ return using_types; }

   gsl_matrix* update_cluster_error_covariances( bool bayesian_estimate = true );//calculates an estimate of the covariance matrix of the errors for this cluster.

   gsl_matrix* get_cluster_error_sqrt_inverse_covariances();//return the square root inverse of the error covariance matrix.


   private:

   gsl_matrix_int* num_obsns;

   gsl_matrix* yty;

   gsl_matrix* ytd;

   gsl_matrix* dtd;

   gsl_matrix* y_averages;//useful for simple Euclidean or correlation clustering.

   gsl_matrix* ys;//for general clustering, a matrix to hold the targets for each cluster.

   gsl_matrix* cluster_error_covariances;//an estimate of the covariance matrix of the errors for this cluster.

   gsl_matrix* cluster_error_sqrt_inverse_covariances;//the inverse square root of the estimate of the covariance matrix of the errors for this cluster.

   int v_index;

   bool distinct;

   bool mergeable, mergeable_store;//if false, this cluster cannot be merged with ANY other cluster.

   bool splitable;//if false, this cluster cannot be split in two.

   string* label;

   list< int > subjects_allocated;

   int num_subjects_allocated;

   list< int >::iterator subject_join;//point in the subject list where the last merger started, zero if never had one or have split.

   list< Cluster_GLM_MVN* > clusters_allocated;// a list of pointers of actual clusters allocated to this cluster (if we want to nest clusters).

   double closeness_contribution, *old_closeness_contribution;

   double *heller_prior_d;

   double yty_sum;//sum of the yty matrix, used if !heteroskedastic and a uniform background model.

   static GLM_MVN_Clustering* Rptr;//needed for the mcmc clustering methods, a pointer to the underlying regression for recalculating YTY, etc.

   static Cluster_GLM_MVN* root_cluster;//when preserving tree and not clustering by labels, this is the cluster at the top of the tree. 

   static map< string, Cluster_GLM_MVN* >* label_to_root_cluster;//when preserving tree and clustering by labels, this gives the cluster at the top of the tree for each label. 

   typedef map< Cluster_GLM_MVN*, double, less <Cluster_GLM_MVN*> > cluster_ptrtodbl;

   cluster_ptrtodbl closeness_map;// map from cluster pointer :-> its closeness to this cluster (may be left incomplete to save space, exploiting symmetry).

   Cluster_GLM_MVN* closest_cluster;//pointer to closest cluster to this one.

   Cluster_GLM_MVN* second_closest_cluster;//pointer to next closest cluster to this one.

   Cluster_GLM_MVN* parent_cluster;//pointer to this cluster's parent cluster, if it has one.

   double sibling_closeness; // the closenss of this cluster to it's sibling, if it has one.

   int score;// the score given to the cluster by scoring tree distribution, depending on the quality of the tree beneath it.

   int last_iteration_updated;
      
   int smallest_subject, largest_subject;//the smallest and largest indices of the subjects allocated.

   int type;// varaible used to allow clusters to be (perhaps temporarily) considered as of different 'types', and so not mergeable.

   static int cluster_count;   // (Static) Number of cluster objects.

   static int number_mergeable; // (Static) Number of cluster objects which are mergeable.

   static int number_splitable; // (Static) Number of cluster objects which are splitable.

   static bool split_off_one_subject;// if true, split move splits off one subject only

   static bool two_means_separation;//if true, split move splits according to 2 means clustering using initialise_two_means_clustering().

   static gsl_matrix_int* temp_num_obsns;//these 4 matrices used as holders when using clusters_allocated hierarchy
   static gsl_matrix* temp_yty;
   static gsl_matrix* temp_ytd;
   static gsl_matrix* temp_dtd;
   static string* temp_label;
   static double gibbs_sweep_log_prob;//the combined log probability of the gibbs sweep.
   
 public:

   static bool using_types;//If false, calling get_type() will return 0 whatever the type acutally is.
};   

int count_non_zero_entries( gsl_matrix_int* m );//count the number of entries of mx which are non-zero.

inline int unequal_integer_compare( const void* s1, const void* s2 )
{
   int s_1 = *(int*)s1;
   int s_2 = *(int*)s2;
   return s_1 < s_2 ? -1 : 1;
}


#endif

