#include "fourier_basis_fn.h"
double* Fourier_Basis_Fn::argument_coefficients = NULL;
int Fourier_Basis_Fn::num_argument_coefficients = 0;

Fourier_Basis_Fn::Fourier_Basis_Fn()
      : Basis_Fn()
{
   basis_type = FOURIER_BASIS;
}

Fourier_Basis_Fn::Fourier_Basis_Fn( double* v, int terms, int* cmp, int* pw, int* sgn, int* actv_dim )
      : Basis_Fn( v, terms, cmp, pw, sgn, actv_dim )
{
   basis_type = FOURIER_BASIS;
}

Fourier_Basis_Fn::Fourier_Basis_Fn( double d, int cmp, int pw, int sgn, int* actv_dim )
      : Basis_Fn( d, cmp, pw, sgn, actv_dim )
{
   basis_type = FOURIER_BASIS;
}

Fourier_Basis_Fn::~Fourier_Basis_Fn()
{
}


void Fourier_Basis_Fn::set_argument_coefficients( double* new_arg_coeffs, int num_arg_coeffs )
{
    zap_argument_coefficients();
    num_argument_coefficients = num_arg_coeffs;
    argument_coefficients = new double[ num_argument_coefficients ];
    for( int i = 0; i < num_argument_coefficients; i++ )
       argument_coefficients[ i ] = new_arg_coeffs[ i ];
}

void Fourier_Basis_Fn::zap_argument_coefficients()
{
   if( argument_coefficients )
   {
      delete argument_coefficients;
      argument_coefficients = NULL;
   }
   num_argument_coefficients = 0;
}

double Fourier_Basis_Fn::evaluate( double* x, int* actv_dim, int order_integral, int integral_component )
{
   if( actv_dim && active_dimension && *active_dimension != *actv_dim )
      return 0;
   double x_i, x_val = 1, contribution = 1;
   bool positive = true, component_matched = false;
   int i = 0;
   double scale_i;
   while( positive && i < num_interactions )
   {
      x_i = (x[ component[ i ] ] - t[ i ])/scale;
      if( x_i < 0 )
         x_val = 0;
      else if( power[ i ] == 0 )
      {
         if( sign[ i ] == 1 )
         {
            x_val = 0;
            positive = false;
         }
      }
      else if( !order_integral || component[ i ] == integral_component )
      {
         component_matched = true;
         if( !argument_coefficients || power[ i ] >= num_argument_coefficients )
            scale_i = M_PI * power[ i ];
         else
            scale_i = M_PI * argument_coefficients[ power[ i ] ];
         bool sine_indicator = ( order_integral + sign[ i ] ) % 2 != 0;
         bool negative_indicator = ( order_integral + sign[ i ] ) % 4 > 1;
         if( sine_indicator )
            contribution *= sin( x_i * scale_i );
         else
            contribution *= cos( x_i * scale_i );
         if( negative_indicator )
            contribution = -contribution;
         if( order_integral > 0 )
            contribution /= pow( scale_i, order_integral );
         x_val *= contribution;
      }
      i++;
   }
   if( order_integral && !component_matched )
   {
      if( order_integral > 0 )
      {
         x_val *= pow( x[ integral_component ], order_integral );
         x_val /= gsl_sf_fact( order_integral );
      }
      else
         x_val = 0;
   }
   double mean_mult = 1;
   if( order_integral > 0 )
      mean_mult = pow( x[ integral_component ], order_integral ) / gsl_sf_fact( order_integral );
   return ( x_val - mean * mean_mult ) / sd;
}
