
#if !defined( _glm_h )
#define _glm_h

#include "basis_fn_regression.h"
class GLM : public Basis_Fn_Regression
{
  public:
   ///
   GLM( int argc, char **argv );
   ///
   GLM( gsl_matrix* X1, gsl_matrix* Y1 );// uses matrix,matrix constructor of base class
   ///
   GLM( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 );// uses matrix,matrix,matrix constructor of base class
   ///
   ~GLM();
   ///
   void preliminary_glm_initialisation();
   void initialise_glm();
   virtual double calculate_log_likelihood();
   virtual double calculate_non_conjugate_log_likelihood();
   virtual double likelihood_fn( int i, int j ); // returns likelihood of the ijth entry of *Yptr
   virtual void calculate_posterior_mean_parameters();
   double get_lambda();
   void set_lambda( double );
   void calculate_linear_predictor();
   gsl_matrix* inverse_link();
   gsl_matrix* link( gsl_matrix* );
   virtual double inverse_link_fn( double );
   virtual double link_fn( double );
   void draw_betas_from_prior();
   double draw_lambda_from_prior();
   double get_lambda_density( double lambda );
//   double draw_lambda_from_posterior();
   double draw_error();//draw a random error N(0,lambda) (for simulating data from the model).
   void calculate_empirical_V(); //uses beta_hat_matrix to get an empirical cov matrix for betas.
   void write_V_to_file();//writes V_ to V_Filename.
   virtual void calculate_beta_hat_matrix() = 0;//{;}; // method for creating a beta gsl::vector estimate for each observation (data block).
   gsl_matrix* get_m_star(){ return m_star;}
   double get_log_prior_v_multiplier();//if v_multiplier variable, returns its log prior density, o/w zero.
   virtual void set_v_index( int index );//change the matrix V_ being used to the index one from V_array, and change V_inv and vstar_map accordingly.
   void update_v_arrays( bool update_probabilities = false );//update each of the V_s in V_array, V_inv_array after a change to underlying V_.
   
 private:
   double* eta; // the linear predictor, beta * D.
   gsl_matrix* E_y; // g^(-1)(eta)   
   char Prior_V_Values_Filename[FILENAMELENGTH];// file for reading/writing the values of the prior on V_.
 protected:
   char V_Filename[FILENAMELENGTH];// file for reading in/writing out a matrix for V_.
   double lambda; // hyperparameter for beta.
   gsl_matrix* m_star;
   double a_, b_, b_g, b_h, a_1, b_1;
   static double delta1, delta2;//parameters for Gamma prior for V, if treated random.
   double a_star;
   gsl_matrix *b_star, *b_matrix;//need this to be a matrix in multidimensional, matrix normal case, where this acts as sigma_star.
   double e_sigma_sq;//posterior expectation of error variance, b*/(a*-1)
   static double v_multiplier;// multiply V by this to down/up weight it, set to 1 by default.
   double v_multiplier_store;//when proposing a new v_multiplier value, this is somewhere to store the old one.
   static double v_multiplier_gamma_shape_parameter;//Gamma distribution shape parameter for prior of V multiplier.
   static double v_multiplier_gamma_scale_parameter;//Gamma distribution scale parameter for prior of V multiplier.
   double lambda_prior_normalising_constant, lambda_prior_normalising_constant_store, lambda_prior_normalising_constant1;
   bool gauss_laguerre_v;//if true, integrate out V using gauss laguerre integration.

   long mixture_of_vs;//if 2, then at each likelihood calculation, choose V from the list of number_of_vs possible candidates. If 1, average over them.  Otherwise do clustering for fixed V, and repeat for the number_of_vs candidates.
   bool correlated_uncorrelated_mixture;//if true, then have additional mixture on V_ w.p. 1/2 of having experimental correlations.
   long number_of_vs;//the number of different V_ matrices to search over AT EACH LIKELIHOOD CALCULATION.
   long v_grid_size;//the number of different V_matrices on which to perform SEPARATE CLUSTERINGS when clustering.
   double v_range;//the maximum mulitplicative factor we use to create alternative V_s  in the arrays below.
   double* half_logdet_v_inv;//array holding 1/2*log(det(V_inv)) for the different choices of V_.
   double* log_prior_v;//array holding the log prior for the different choices of V_.
   double hldv;//holds the current value of half_log_det_vstar;
   int current_v_index, optimal_v_index;// the last value used in set_v_index, and an optimal v_index value.
   gsl_matrix **V_array, **V_inv_array;
   gsl_matrix* V_; // prior variance for betas if activated is lambda*V_, otherwise iid with variance lambda.
   gsl_matrix* V_inv;
   gsl_matrix* V_inv_sqrt;
   int dim_V;//dimension of above V.. matrices.
   bool v_matrices_differ;//set true if at least one of the number_of_vs V_ matrices are different (often they'll be initialised the same).
   gsl_matrix* beta_hat_mx;//( n /data_block size ) * D.get_cols() matrix where in Gaussian case column sums give (D'D)^{-1}D'Y.
   static bool calc_empir_v;// non-zero if want to use emprical correlation matrix for V.
   static bool g_prior;// non-zero if want to use g-prior for V.
   static bool diagonalise_empir_v;//if non-zero, then if calc_empir_v, set all the off diagonal elements of V to 0.
   static bool diagonal_v;//if true, then all off-diagonal elements of V_ and V_inv are 0.
   static bool diagonal_vstar;//if true, then all off-diagonal elements of V_star are 0 (implied by diagonal_v && gram_schmidt).
   bool variable_v;//if non-zero, allow changes to the matrix V
   gsl_matrix* error_covariances;
   gsl_matrix* global_error_covariances;
   gsl_matrix* scaled_error_covariances_prior_mean;//the prior mean of for the error covariance matrix.

  };

double ran_dble_exponential( gsl_rng *, double );

#endif
