
#if !defined( _glm_mvn_h )
#define _glm_mvn_h

#include "glm.h"

class GLM_MVN : public GLM
{
  public:
   ///
   GLM_MVN( int argc, char **argv );
   ///
   GLM_MVN( gsl_matrix* X1, gsl_matrix* Y1 );// uses matrix,matrix constructor of base class
   ///
   GLM_MVN( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 );// uses matrix,matrix,matrix constructor of base class
   ///
   ~GLM_MVN();
   ///
   virtual Regression* create_new_regression( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 );//needed for partition regression.

   void preliminary_glm_mvn_initialisation();
   
   void initialise_glm_mvn();

   void calculate_ml_const();//Calculate the 'constant' part of the marginal likelihood.

   void calculate_ytys();//finds the sum of squares.
   
   virtual double calculate_basis_coefficient_log_prior( double basis_coeff );//returns the log prior density for the basis coefficient, now assumed to be N(0,v_multiplier).

   virtual void create_posterior_parameter_matrices();

   virtual void calculate_posterior_mean_parameters();

   double get_e_sigma_sq(){ return e_sigma_sq; }

   gsl_matrix* get_b_star_correlations();//return the corresponding correlation matrix to b_star.
   
   double draw_lambda_from_posterior();

   void draw_betas_from_posterior();

   virtual double calculate_log_likelihood();
   //   virtual double likelihood_fn( int i, int j ); // returns likelihood of the ijth entry of *Yptr
   void calculate_closeness_matrix();// calculates correlation matrix if correlation_block non-zero.
   void calculate_C1_matrix();//calculates the C1 matrix.
   void calculate_D_sq();// calculates D' D (done quickly if D is block repetitive.

   void calculate_translated_D();// calculates D 'block multiplied' by C1.

   void calculate_translated_Y();// calculates Y 'block multiplied' by C1.

   void update_D_dependent_matrices();//recalculate matrices which depend on D, used when D has been altered.
  
   void calculate_beta_hat_matrix(); // Assumes p_y=1, calculates num_data_blocks * num_design_cols matrix, whose column sums give D'Y.

   virtual double function_variance_evaluate( double* v );//overwrites regression class fn, calculates Var(D(v).betas)

   virtual void output_residual_variogram( gsl_matrix* x_diffs, gsl_matrix* residual_diffs );// overwrites regression class fn, fitting a GLM_MVN to the variogram values.

   long number_monte_carlo_samples;

   double monte_carlo_estimate_l2_distance_from_zero();//takes a monte carlo sample of size number_monte_carlo_samples, and for each calculates the l2 distance of the sampled curve from zero. Returns the average.

   double monte_carlo_estimate_absolute_integral();//takes a monte carlo sample of size number_monte_carlo_samples, and for each calculates the integral over the full range for the absolute value of the sampled curve. Returns the average.

   void monte_carlo_estimate_l2_distance_from_zero_and_absolute_integral( double& ave_l2, double& ave_ia );//does both the above.
   
   void find_correlated_experiments_covariances();//calculates V_ and D_sq if experiments are correlated.

   void update_correlated_experiments_covariances( int exp_number );//updates V_ for change in location of experiment exp_number.
   
   void update_correlated_experiments_covariances();//updates V_ for change in location of all experiments.

   void calculate_global_error_covariances();//calculate an empirical prior mean based on all 
 protected:
   
   gsl_matrix* V_star;

   gsl_matrix* V_star_inv;// inverse of V_star.

   gsl_matrix* V_star_sqrt;// square root matrix of V_star.

   double** yty;// Y'Y - need this to be a matrix in multidimensional, matrix normal case.

   double ml_const;// the part of the marginal likelihood invariant to the design matrix

   double ml_const_store;//when V can be change, this stores the old value of ml_const;

   gsl_matrix* D_sq;//  D' * D (or C1D' * C1D if C1), calculated quickly if X (and C1) is block repetitive evrey inputs_repeated_every rows.

   gsl_matrix** D_sqs;//  If there are multiple data block sizes (for different experiments) this can hold D_sq for each data block type.

   gsl_matrix* C1;// square root of correlation matrix for Y observations.

   gsl_matrix* C1D;
   
   gsl_matrix* C1Y;//  D and Y matrices 'block multiplied' by C1.

   gsl_matrix* yd_vstar_dy;//matrix used in calculating ML, stored to prevent continually creating and deleting


//   matrix*C1D_sq;//  C1D' * C1D, calculated quickly if X is block repetitive evrey correlation_block rows.

   static long correlation_block;// if non-zero, blocks of length correlation are assumed correlated. (so the Y's have a block diagonal correlation matrix)
   static double closeness_coefficient;
   static double closeness_power;
   static double dirac_coefficient;

   bool calculate_error_covariances;//if true, calcluates a posterior estimate of the covariance matrix for the errors from the fitted model.
   gsl_matrix* global_error_covariances;//the sample error covariance matrix for all of the data.
   int error_cov_dim;//the dimension of the error covariance matrix

   char L2Distance_Filename[FILENAMELENGTH];//file for writing monte carlo sample of l2 distances of posterior mean curve samples from the zero function.

   char Integral_Absolute_Filename[FILENAMELENGTH];//file for writing monte carlo sample of integrals of the absolute value of posterior mean curve samples from the zero function.

};

#endif
