
#if !defined( _glm_mvn_clustering_h )
#define _glm_mvn_clustering_h

#include "glm_mvn.h"
#include "cluster_glm_mvn.h"
#include <deque>
//#include <vector>
//#include <algorithm>

class Cluster_GLM_MVN;

class GLM_MVN_Clustering : public GLM_MVN
{
  public:
   ///
   GLM_MVN_Clustering( int argc, char **argv );
   ///
   GLM_MVN_Clustering( gsl_matrix* X1, gsl_matrix* Y1 );// uses matrix,matrix constructor of base class
   ///
   GLM_MVN_Clustering( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 );// uses matrix,matrix,matrix constructor of base class
   ///
   ~GLM_MVN_Clustering();
   ///
   void initialise_glm_mvn_clustering();

   void create_closeness_matrix( double** copy_source = NULL );

   void create_working_sufficient_statistic_matrices();
      
   Cluster_GLM_MVN* make_cluster_containing_subject( int subject_num, bool temp_mergeable = true, string* temp_label = NULL );//create a new cluster containing subject subject_num with label.

   bool* get_mergeable_indices();//returns an array of bools saying whether each subject is mergeable or not.

   void restart_subjects_in_own_clusters();//clear the cluster map and put all subjects in their own clusters.
   
   void initialise_mergeable_cluster_deque( map< Cluster_GLM_MVN*, double >& map_of_clusters );//put all mergeable clusters into the mergeable_cluster_deque.

   void initialise_mergeable_cluster_deque();//performs the above action with cluster_map.

   void remove_from_all_clusters_deque( Cluster_GLM_MVN* cl );//removes cl from the deque all_clusters_in_tree.
   
   int get_number_mergeable_in_cluster_map();// returns the number of clusters in the cluster_map which are mergeable

   virtual double calculate_log_likelihood();//calls the fn below for cluster_map.

   double calculate_log_likelihood( map< Cluster_GLM_MVN*, double > & map_of_clusters );

   void print_cluster_sizes();//calls the fn below for cluster_map.

   void print_cluster_sizes( map< Cluster_GLM_MVN*, double >& map_of_clusters );//prints to screen the sizes of the mergeable clusters in the map.
      
   void calculate_inter_cluster_closeness( map< Cluster_GLM_MVN*, double >& map_of_clusters );//calculates the closeness of all of the clusters in map_of_clusters to each other, and finding the highest for each and entering that into the cluster_map.
  
   void calculate_inter_cluster_closeness();//calls the above function for cluster_map.

   double calculate_closest_cluster( Cluster_GLM_MVN* cl, map< Cluster_GLM_MVN*, double >& map_of_clusters );//calculates the closeness of cl to all other mergeable clusters in map_of_clusters, sets it as its closest cluster and returns the closeness value.

   double calculate_closest_cluster( Cluster_GLM_MVN* cl );//calls the above function for cluster_map.

   double calculate_closeness_contribution( gsl_matrix_int* nn, gsl_matrix* yy, gsl_matrix* yd, gsl_matrix* dd );// uses the overloaded fns below, picking off each entry of yy, etc.
   
   double calculate_closeness_contribution( const int nn, const gsl_matrix* yy, const gsl_matrix* yd, const double hlvs, const gsl_matrix_int* nn_obns = NULL );//ONLY CALLED BY OTHER VERSIONS OF THIS OVERLOADED FUNCTION. In treat_experiments_as_separate_dimensions case nn may be needed.
   
   double calculate_closeness_contribution( int nn, gsl_matrix* yy, gsl_matrix* yd );// calls the above fn, using (*vstar_map)[nn] for vv and (*half_log_det_vstars)[nn] for hlvs.ONLY CALLED BY OTHER VERSIONS OF THIS OVERLOADED FUNCTION.

   double calculate_closeness_contribution( int nn, gsl_matrix* yy, gsl_matrix* yd, gsl_matrix* dd, gsl_matrix_int* nn_obsns = NULL );// calls the above fn, using (v_ + dd )^{-1} for vv and log(det(vv)) for hlvs.ONLY CALLED BY OTHER VERSIONS OF THIS OVERLOADED FUNCTION.
   
   double calculate_closeness_contribution( Cluster_GLM_MVN* cl, bool v_optimise = true );//calculates the closeness contribution of cl
   
   virtual double calculate_closeness_contribution( Cluster_GLM_MVN* cl, Cluster_GLM_MVN* cl2, bool subtract = false, bool v_optimise = true );//calculates the closeness contribution of cl (or if cl2 non-null, of cl merged with cl2 (or with cl2 split off if subtract = true)).

   double calculate_non_conjugate_closeness_contribution( list<int> cl_subjects );//calculate closeness contribution for the subjects cl_subjects under the non-conjugate model.

   void calculate_closeness_contributions();//calculates the closeness contributions of all mergeable clusters in the cluster_map.

   void calculate_single_subject_cluster_closeness_contributions();//calculates the closeness contributions of all the single subject clusters (used when preserving tree).

   double calculate_closeness_contribution_with_subject( Cluster_GLM_MVN* cl1, int subject_number );//calculates the closeness contribution of cl with subject subject_number added to it.

   void revert_closeness_contributions();//for each mergeable cluster in the cluster_map, returns its closeness contribution to its stored *old_closeness_contribution.

   double calculate_closeness( Cluster_GLM_MVN* c1, Cluster_GLM_MVN* c2 );

   bool clusters_nearest_neighbours( Cluster_GLM_MVN*cl1, Cluster_GLM_MVN*cl2 );//finds whether cl1 and cl2 are nearest neighbours according to initial distances.

   bool check_triangle_inequality( Cluster_GLM_MVN*cl1, Cluster_GLM_MVN*cl2 );//finds whether closeness(cl1,cl2)>=closeness(cl1,cl3)+closeness(cl2,cl3) for all cl3.
   
   void calculate_posterior_mean_parameters( Cluster_GLM_MVN* c1 );    

   double calculate_corrected_closeness( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 = NULL );//as above, but corrected for change in prior on cluster sizes.

   double calculate_tree_corrected_closeness( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 = NULL );// as above, but also includes term for change in tree probability.
   
   double calculate_squared_euclidean_or_correlation_distance( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 );//finds the average target responses of cl1 and cl2, and returns the squared euclidean distance or correlation between them.

   double calculate_squared_euclidean_or_correlation_distance( list<int>& subj_list1, list<int>& subj_list2 );//as above for two clusters, one containing list of subjects subj_list1, the other subject list subj_list2.

   double calculate_squared_euclidean_or_correlation_distance( int subj1, int subj2 );//as above for two clusters, one containing subject subj1, the other subject subj2.

   double calculate_squared_euclidean_or_correlation_distance( gsl_matrix* y1, gsl_matrix* y2 );//returns the squared euclidean distance or correlation between the two matrices.
   
   double get_log_prior_cluster_size( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 = NULL );//return the prior prob of a cluster the size of cl1 merged with cl2.

   double get_change_log_prior_cluster_size( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2, bool tree_part = true );//returns the change in prior prob when clusters of size of cl1 and cl2 merged.

   double get_change_log_prior_cluster_size( int cl1s, int cl2s, bool tree_part = true );//returns the change in prior prob when clusters of size cl1s and cl2s merged.

   double get_uniform_tree_log_prior_change( int cl1s, int cl2s );//returns the change in prior of tree distribution when merging cl1s and cl2s subjects into one cluster, not accounting for change in number of clusters (as not necessary when part of a shift move, where there is a corresponding split).

   double get_uniform_tree_log_prior_change( int cl1s, int cl2s, int numclust );//returns the change in prior of tree distribution when merging cl1s and cl2s subjects into one cluster, accounting for change in number of clusters .

   double get_optimal_subtree_log_prob( int num_leaves );//returns the probability of a tree with numleaves leaves being the optimal tree T* under the model.
   
   double get_tree_branch_log_prior( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 );//When not using conditional tree prior, returns the part of the unconditional tree prior distribution for the branch starting at cl2 and working up when joined with cl1.

   double get_log_catalan_number( int m );//returns the log of the mth Catalan number, and puts it in a look up map if it's been calculated for the first time.

   double get_log_one_plus_alpha_catalan_number( int m );//returns the numerator for conditional_tree_distribution 4.

   double get_log_one_plus_alpha_catalan_number( Cluster_GLM_MVN* cl );//calls the above function for cl->get_num_subjects_allocated().

   double get_tree_prior_parameter();//returns the tree prior parameter for the proposed number of clusters.
   
   double get_number_of_possible_clusterings_for_tree_containging_cluster( Cluster_GLM_MVN* cl );//returns the number of possible clustering configurations for the tree containg cluster cl.
   
   double calculate_log_predictive_probability( Cluster_GLM_MVN* c1, Cluster_GLM_MVN* c2 );// returns log( p( c1 merged with c2 | c2 ) ).
      
   double calculate_predictive_probability( Cluster_GLM_MVN* c1, Cluster_GLM_MVN* c2 );// returns p( c1 merged with c2 | c2 )

   double calculate_predictive_probability( Cluster_GLM_MVN* c1, int subject );
// returns p( subject merged wih c1 | subject )

   double calculate_predictive_probability( Cluster_GLM_MVN* cl );// returns p( cl | mixture of clusters distn. )

   double calculate_log_conditional_probability( int subject, Cluster_GLM_MVN* c1 );// returns log( p( subject| c1 with subject removed ) )

   double calculate_log_conditional_probability( int subject, Cluster_GLM_MVN* c1, bool cl_contains_subject, Cluster_GLM_MVN* single_subject_cluster_ptr = NULL );// returns log( p( subject| c1 with subject removed ) ); the final argument optionally provides a pointer to a singleton cluster holid just this subject.

   double calculate_conditional_probability( int subject, Cluster_GLM_MVN* c1 );
// returns p( subject| c1 with subject removed )

   void calculate_cluster_label_probabilities();//calculate cluster_and_label_probability for the current clustering.

   void update_cluster_label_probabilities();//calls the fn. below for all entries in the map labels.
   
   void update_cluster_label_probabilities( string& s );//update the entries of cluster_and_label_probability which have the string s as label.

   void update_mean_label_probabilities();//

   void calculate_membership_probabilities();// list<Cluster_GLM_MVN*>* clusterlist );//calculate the probability of membership of a subject to its current cluster relative to the other clusters.

   void ac_through_current_partition();// Do agglomerative clustering constrained to pass through the current partition.

   int get_num_subjects_ever_reallocated();

   int find_subject_with_highest_new_singleton_cluster_probability();//using membership probabilities, find the subject which would most like to be in its own cluster.

//   int compare_new_cluster_membership_probs(const void * a, const void * b);//comparsion function used by qsort in the above function to find the order of subjects wanting to form their own cluster.
   bool compare_new_cluster_membership_probs(const int sub1, const int sub2);//comparsion function used by qsort in the above function to find the order of subjects wanting to form their own cluster.
       
   Cluster_GLM_MVN* find_cluster_containing_subject( int sub );

   Cluster_GLM_MVN* move_subject_to_a_singlteon_cluster( int sub );//moves the subject sub from it's current cluster and creates a new singleton cluster in the cluster_map.
   
   bool clusters_mergeable( Cluster_GLM_MVN* c1, Cluster_GLM_MVN* c2  );// returns true if c1 and c2 can be legitimately merged.

   bool closeness_preferable( Cluster_GLM_MVN* cl1, double d1, Cluster_GLM_MVN* cl2, double d2 );//for standard agglomerative clustering, simply checks if d1 > d2; but if worst_first switched on, then looks for smallest positive closeness.

   Cluster_GLM_MVN* find_next_input_merger();//if merger details have been inputted, this finds the next merger.
   
   Cluster_GLM_MVN* find_best_merger();//finds which cluster (pair) is best for merging.

   Cluster_GLM_MVN* find_random_best_merger();//finds a cluster pair randomly for mergin, with probability proportional to the posterior.

   void propose_best_merger();//proposes the merger of the best cluster pair.

   void best_merge_for_cluster( Cluster_GLM_MVN* cl = NULL );//merges cl with its closest cluster using merge_clusters(.).
   
   void merge_clusters( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2);//proposes then executes merger of the two clusters.
   
   void propose_merge_clusters( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 );// calculates new loglikelihood and bic for the merger of these two clusters.
   
   void execute_merge_clusters( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 );//performed in two parts, the functions below. 

   Cluster_GLM_MVN* execute_merge_part_one__merge_the_clusters( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2);//deletes cl2 from the cluster_map and all other clusters' closeness_maps, merges *cl with cl2, returns pointer to new merged cluster.

   void execute_merge_part_two__update_cluster_map( Cluster_GLM_MVN* cl1 );//,then calcultes the closeness of the newly formed cluster to all the other clusters, finds the closest and updates the cluster_map.

   void propose_move_up_tree( Cluster_GLM_MVN* cl1 );//finds the cluster which shares the parent of cl1, and proposes to shift frame of reference from these two to the parent.

   Cluster_GLM_MVN* cascade_down_tree( Cluster_GLM_MVN* cl1 );//goes to the root cluster of cl1, and then randomly drops down the tree until a merger is chosen.

   void calculate_path_probability( Cluster_GLM_MVN* cl0, Cluster_GLM_MVN* cl1 );//calculates the (reverse) probability of cascading back down the tree to retrieve our current model.

   double calculate_log_cascade_probability( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 );//returns the log probability of cluster cl1 taking the route of cl2 down the tree.
   
   double calculate_cascade_probability( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 );//returns the probability of cluster cl1 taking the route of cl2 down the tree.
      
   bool is_cluster_in_model( Cluster_GLM_MVN* cl );//returns TRUE if cl is in the cluster_map.

   int get_number_of_mergeable_clusters_in_my_model( Cluster_GLM_MVN* cl );//returns the number of model clusters in the tree containing cl.
   
   Cluster_GLM_MVN* find_parent_in_model( Cluster_GLM_MVN* cl );// when preserving tree, looks up tree from cl to find the parent/grandparent in the cluster_map. Othwrwise, returns cl.

   bool has_child_in_model( Cluster_GLM_MVN* cl );//when preserving tree, looks down to see if ther is a child/grandchild in the custer_map.
   
   Cluster_GLM_MVN* split_cluster( Cluster_GLM_MVN* cl1, list<int>* cl_subjects = NULL );//proposes then executes split of the cluster (if cl_subjects not null, splits cl1 removing these subjects deterministically).

   Cluster_GLM_MVN* split_cluster( Cluster_GLM_MVN* cl1, int cl_subject );////proposes then executes split of the cluster, splitting off a single subject.
   
   Cluster_GLM_MVN* propose_split_cluster( Cluster_GLM_MVN*& cl1, list<int>* cl_subjects = NULL );//forms a new cluster cl2 by dividing up cl1 (if cl_subjects not null, splits cl1 removing these subjects deterministically), calculates new loglikelihood and bic for this split.

   Cluster_GLM_MVN* propose_split_cluster( Cluster_GLM_MVN*& cl1, int cl_subject );//calls propose_split_cluster as above with a list containing a single subject.
      
   Cluster_GLM_MVN* execute_split_cluster( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 );// updates the cluster_map, etc. after forming revised cl1 and new cluster cl2 returns a ptr to the new cluster.

   void undo_split_cluster( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 );//merges the two clusters back together again.
   
   Cluster_GLM_MVN* split_off_branch( Cluster_GLM_MVN*& cl1, list<int>* cl_subjects = NULL );// splits the cluster cl1 into the two clusters it contains, or goes further down the tree if cl_subjects non-null and describes a cluster further down.
   
   void move_subjects( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 );// randomly removes some of the subects allocated to cl1 and allocates them to cl2.

   Cluster_GLM_MVN* propose_move_subjects( Cluster_GLM_MVN*& cl1, Cluster_GLM_MVN*& cl2, list<int>* cl1_subjects = NULL );//splits some subjects (cl1_subjects_ptr if non-null) off from cl1 into a holding cluster (to which a pointer is returned), calculates revised log_likelihood and bic.

   Cluster_GLM_MVN* propose_move_subject( Cluster_GLM_MVN*& cl1, Cluster_GLM_MVN*& cl2, int cl1_subject );//splits some subject cl1_subject off from cl1 into a holding cluster (to which a pointer is returned), calculates revised log_likelihood and bic.

   Cluster_GLM_MVN* execute_move_subjects( Cluster_GLM_MVN* cl0, Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 );//finishes up the split of cl1 and cl2, then merges the holding cluster cl0 with cl2, returning a pointer to the newly merged cluster.
   
   void undo_move_subjects( Cluster_GLM_MVN* cl0, Cluster_GLM_MVN* cl1 );//returns the subjects back to cl1 from the holding cluster cl0.
   
   void merge();// do the best possible merge of two clusters.

   void merge_until( int n_clust, bool write_merger_details = false );// do merge() until we have n_clust clusters, and write the merger details to file iff the second variable is true.

   void merge_until_desired_num_clusters( bool write_merger_details = true );//merge_until(desired_num_clusters).

   void check_allocations_after_merger();//after agglomerative merger, looks to reallocated subjects to optimal clusters.
   
   void increment_association_matrix( double* increment_ptr = NULL );// update the assocation matrix, incrementing by the double pointer if provided, otherwise by 1.

   void normalise_association_matrix();//normalise the association matrix (so diagonal has 1s).
                                      
   int* argmax_association_matrix();//returns the (row,column) pair maximsing the association matrix (off diagonal).
      
   int rebuild_agglomerative_tree_from_cluster( Cluster_GLM_MVN*& cl, bool permanent = false, bool optimal = true );//when preserving tree, this function destroys the tree from beneath cl and rebuilds it according to agglomerative clustering, or randomly if optimal is false.
   
   void sample_model_node_subtree_from_full_conditional( Cluster_GLM_MVN*& cl );//draws a new subtree under the model node cl, from its full conditional distribution.

   void update_current_clustering();//updates current_clustering array to current model.
   
   void update_optimal_clustering();//updates optimal_clustering array to current model.

   void split_to_optimal_clustering();//assumes the optimal model can be reached by splitting present model.

   void merge_to_single_cluster();//merges all the clusters in the cluster_map into one cluster.

   virtual void get_subject_cluster_types();//returns a vector showing thecluster type for each subject, or NULL if not using types.

   void give_clusters_different_types();//sets each cluster in the model to be of a differnt type.
   
   Cluster_GLM_MVN* create_temporary_cluster_for_subjects( list<int>& cl_subjects, bool find_closeness_contribution = true );//creates a (temporary) cluster with cl_subjects in it, for temprorary comparisons.

   Cluster_GLM_MVN* create_temporary_cluster_for_subjects( list<int>& cl_subjects, list<int>::iterator iter1, list<int>::iterator iter2, bool find_closeness_contribution = true );//creates a (temporary) cluster with the subjects between the two iterators in cl_subjects in it, for temprorary comparisons.

   Cluster_GLM_MVN* create_temporary_cluster_for_subject( int subject_number, bool find_closeness_contribution = true );//creates a (temporary) cluster with just that subject in it, for nearest neighbour comparisons.

   Cluster_GLM_MVN* create_temporary_cluster_without_subject( int subject_number, Cluster_GLM_MVN* cl1 );//creates a (temporary) cluster with all the subject of cl1 in it except subject_number.

   Cluster_GLM_MVN* create_temporary_cluster_with_subject( int subject_number, Cluster_GLM_MVN* cl1 );//creates a (temporary) cluster with all the subject of cl1 in it and also subject_number.

   virtual void update_any_stats(){;}//to be overwritten in any derived classes to update statistics after changes to the clusterings have been made by merges_until().

   void post_agglomerative_processing();//any processing after the initial agglomerative clustering.
   
   void write_clusters_to_file();

   void print_label_probabilities();//print the most probable label for each unmergeable cluster, if labelled count up number misclassified.

   void print_deque_cluster_sizes();//print to screen the number of subjects alloated to each cluster in the the mergeable_cluster_deque.
   
   void make_predictions();//calculates the predictive probs of each unmergeable cluster.

   int update_merger_details();//at each step of hierarchical clustering, enter in the next row down of merger_details the new nubmer of clusters, smallest subject number of each of merged clusters, and associated ML and BIC for the merge. Returns the smallest subject index involved in the merger.
   
   void update_merger_details_first_row();//initialises the first row of the merger details array, putting -1s where the merged subjects would go.
   
   void update_dendrogram();//writes the current ordering of the subjects after hierarchical clustering to the dendrogram array.

   void update_subjects_first_merge();//updates the matrix subjects_first_merge with the latest merger.

   void double_size_of_merger_details_array();//expands the merger details array to twice the number of rows, and puts the old details in the 1st half.

   void write_merger_details_to_file();//write the array merger_details to file.

   void write_dendrogram_to_file();//write the data to file in the order prescribed by agglomerative clustering - can be used to make a dendogram.

   void write_dendrogram_scores_to_file();//write the likelihood (or BIC) for each level of the final agglomerative hierarchy to file.

   void write_association_matrix_to_file();//write the association matrix to file.

   void write_subjects_first_merge_to_file();//write the subjects_first_merge matrix to file.

   void write_closeness_matrix_to_file() const;//write the closenesss matrix to file.

   void write_ammended_closeness_matrix_to_file( const unsigned int row ) const;//writes row for the (cluster containing the) row-th subject of the closenesss matrix to file.

   void write_single_subject_contributions_to_file();//write the closeness contributions for each of the individual subjects to file.
   
   void sample_clusters_from_prior();//set the Ymatrix to new values drawn from the prior.

   double half_log_det_vstar( gsl_matrix_int* m );//calculates the sum of the half_log_det_vstars according to whether we are treating experiments separately.

   void weight_the_experiments();//weights the different experiments according to their pooled sample variance, allows for different variances for each experiment but same mean function. *** Strictly, cannot be used currently, need to calculate V* differently,as now different experiments have different Ds. works ok only if approx same number of each experiment type in each cluster.

   void store_closeness_contributions();

   void recover_closeness_contributions();

   virtual void set_v_index( int index );//change the matrix V_ being used to the index one from V_array, and change V_inv and vstar_map accordingly.
   
   void propose_new_v();

   void execute_new_v();

   void undo_new_v();

   void propose_new_b();

   void undo_new_b();

   bool required_vstar_of_storable_kind( long int index, bool* full = NULL );//checks ot see whether a call to get_vstar satsifies the conditions where such a v_star would be/will have been stored.
   
   gsl_matrix* get_vstar( gsl_matrix_int* nn );// returns the gslobj of matrix vstar_map[ d ], calculating it if it hasn't already been. d is sum function of the indices, without subject_experiment_weights just the sum.

   gsl_matrix* get_vstar( long int index, gsl_matrix_int* nn = NULL, bool* full = NULL );// returns the gslobj of matrix vstar_map[ index ], calculating it if it hasn't already been.

   gsl_matrix* get_vstar_dbl( double d );// returns the gslobj of matrix vstar_map[ ] when d is a double, calculating it if it hasn't already been.

   void delete_vstar( long int index );//deletes the index'th entry fomr vstar_map.

   int find_position_in_cluster_map( Cluster_GLM_MVN* cl );//finds which position cl is in the cluster map.
   
   void empty_cluster_map();// kills all the stored clusters in cluster_map, and clears cluster_map.

   void empty_vstar_map();// kills all the stored matrices in vstar_map, and clears vstar_map and half_log_det_vstars.
   
   double get_half_log_det_vstar( long int index );// returns the double half_log_det_vstars[ d ], calculating it if it hasn't already been.

   double get_half_log_det_vstar_dbl( double d );// returns the double half_log_det_vstars[ ] when d is a double, calculating it if it hasn't already been.
   
   double get_half_lndet_invvinvplusdtd( gsl_matrix* dd );
   
   double get_half_lndet_invvinvplusdtd( Cluster_GLM_MVN* cl );

   gsl_matrix* get_non_conjugate_precision_matrix( double sigma_sq_val, int num_obs = 1, int *shift_ptr = NULL );//creates the design matrix for the non conjugate model with num_obs observations and variance paramter sigma_sq_val.

   double get_logdet_non_conjugate_precision_matrix( gsl_matrix* precision_mx, int num_obs = 1, int* shift_ptr = NULL );// returns the determinant of the above matrix.
   gsl_matrix_int* get_nob_for_subject( int subject_number );//returns the corresponding rows of *NOBs for that subject as a matrix.
   
   gsl_matrix* get_yty_for_subject( int subject_number );//returns the corresponding rows of *YTYs for that subject as a matrix pointer.
   
   gsl_matrix* get_ytd_for_subject( int subject_number );//returns the corresponding rows of *YTDs for that subject as a matrix pointer.

   gsl_matrix* get_dtd_for_subject( int subject_number );// returns the submatrix of *DTDs for that subject as a matrix pointer.

   gsl_matrix* get_design_points( list<int>& cl_subjects_allocated, int* exp_number = NULL );//same as get_targets but using *D in place of *Yptr.

   double get_criterion_maximum();// returns max_BIC if bic_criterion = 1, and max_log_likelihood otherwise.

   double calculate_gauss_laguerre_non_conjugate_ml( list<int> cl_subjects, int* exp_number = NULL );//returns the gauss-laguerre estimate of the log marginal likelihood when using non-conjugate model.

   double sum_cluster_configuration_probabilities();//sums the value of the function below for all possible numbers of clusters for num_subjects subjects.

   double sum_cluster_configuration_probabilities( int num_subs, int num_clust , int k = 0, int* allocations = 0 );//sums the marginal likelihoods for all the possible configurations with num_clust clusters.

   gsl_matrix* get_target_covariances( Cluster_GLM_MVN* cl = NULL, bool corr = false, bool residual_covariances = false );//finds the ocvariances (or correlations if corr=true) between the targets at each expt time point for the subjects in the cluster, discarding missing values.

   class ProxyGreater{
       GLM_MVN_Clustering& that;
     public:
     ProxyGreater(GLM_MVN_Clustering& gmc) : that(gmc) {}
       bool operator()(int leftID, int rightID) const {
           return that.compare_new_cluster_membership_probs(leftID,rightID);
       }
   };

 protected:

   int number_misclassified;

   int number_classified;

   long heteroskedastic;//0 if same variance for all, 1 if different for every observation type in every cluster, 2 if different only in every cluster (so same for each observation type).
   
   bool exclude_singleton_clusters;//if true, then when !heteroskedastic do not include singleton clusters in this shared variance calculation.
 
   long number_distinct;//the first number_distinct subjects cannot be merged with each other ever, set to 0 as default.

   long number_unmergeable;//the last number_unmergeable subjects cannot be merged any others, set to 0 as default.
   
   bool randomised_unmergeable;//if zero, the last number_unmergeable subjects are taken as unmergeable, otherwise a random set is taken (for random cross validation).
      
   long number_of_labels;//number of strings which are labelled from Cluster_Label_Filename.

   typedef map< string, int > str_to_int;
   
   str_to_int labels;//a map of all the different (mergeable) cluster labels used at the start to the number of obsns with that label.

   string* subject_labels;//array holding the label of each subject

   typedef map< pair< Cluster_GLM_MVN*, string >, double > cllabtoprob;

   cllabtoprob cluster_and_label_probability;//map taking an unmergeable cluster pointer and one of the cluster labels to the corresponding probability.

   cllabtoprob mean_cluster_and_label_probability;//map taking an unmergeable cluster pointer and one of the cluster labels to the corresponding probability.

   double** membership_probability;//the probability of membership of a subject to the different clusters.

   int membership_probability_dim;//the number of rows for the matrix membership_probability (either #cluster or #clusters+1).

   int membership_contribution_dim;//the number of rows for the matrix membership_contributions (either #cluster or #clusters+1).

   int* map_memberships;//the cluster with the highest memebership probability for each subject.

   double** membership_contributions;//array holding the closeness of each subject(row) to each cluster(column).

   double *max_log_probs;//array holding the highest allocation probability for each subject
   
   double *probsums;//array holding the sum of allocation probabilities for each subject
   
   int number_not_in_map_membership;//the nubmer of subjects not currently allocated to the cluster with highest membership probability.

   list<int> subjects_not_in_map_membership;//a list of those subjects not currently allocated to the cluster with highest membership probability.

   double starting_log_likelihood;//the loglikelihood for the original model.

   double baseline_configuration_log_likelihood;//used when summing the likelihoods over every possible configuration, this holds the likelihood for the model of all in one cluster, used to prevent exp(.) of something big or small.
   
   double model_weight;//the weight given to the current model when updating cllabtoprob mean_cluster_and_label_probability, defaults to 1.

   bool cluster_by_labels;//if non-zero, we give -DBL_MAX closeness to two clusters with identical, non-null labels (So you can only merge clusters with the same labels).

   bool maximal_probs;//if zero, marginal density given of type label is average of densities for each cluster, if 1 it is the maximum of those densities.

   bool weight_by_missing_values;//if missing values have been provided, weight subject according to them.
   
   bool subject_experiment_weights;//if non-zero, then each subject gets a different weight for each experiment (based on the number of missing values ).
   
   int** NOBs;//matrix whose jth column contains num_obsns for each subject which were in the jth experiment.

   double** YTYs;//matrix whose jth column contains yty for each subject for the jth experiment.

   double** Y_totals;//matrix whose i'th row is the Y totals for subject i across experiments.
   
   double** YTDs;//matrix who's ith row containins y'd for subject i.
   
   double** DTDs;//matrix who's ith row containins d'd for subject i, needed if inputs_repeated_every = 0.

   double** YTDVDTYs;//matrix whose ij'th element is Y_iDVDTY_j, used for non-conjugate model.

   double yty_total;//when heteroskedastic=0, this is all we need, b_+.5* (the sum of  *YTYs).

   double proposed_yty_total;//for varying yty_total when singleton clusters get excluded.

   int num_obsns_total;//the total number of observations.

   long int current_vstar_index, current_hldv_index;//the last values used in get_vstar(), get_half_log_det_vstar().

   double current_ytd_vstar_dty_sum;// when heteroskedastic=0, need to keep track of this as well as log_likelihood for quick recalcuation of log_likelihood (due to dependence on all clusters of local changes).

   double proposed_ytd_vstar_dty_sum;

   double second_proposed_ytd_vstar_dty_sum;

   double current_hlvs;//for storing half log det(V*) contribution of current cluster when !heteroskedastic.
   
   typedef map< long int, gsl_matrix*, less< long int > > mx_numtoptr;

   typedef map< long int, double, less< long int > > inttodbl;

//   struct LTclustptr
//   {
//      bool operator()(const Cluster_GLM_MVN* cptr1, const Cluster_GLM_MVN* cptr2 ) const
//         {
//            return *cptr1 < *cptr2 ;
//         }
//   };

   typedef map< Cluster_GLM_MVN*, double > cluster_ptrtodbl;

   cluster_ptrtodbl cluster_map;// map from each cluster pointer :-> its closeness to its closest_cluster.

   double** closeness_matrix;//a look up matrix, to be used for very large data sets when storing maps in each cluster requires too much memory.

   double** single_subject_closeness_matrix;//a copy of closeness matrix made in the special case where each subject has it's own cluster.

   int** subjects_first_merge;//ijth entry holds the number of agglomerative clustering merges that had occurred when i and j first share a cluster.

   bool store_when_subjects_first_merge;//decides whether or not to create and fill subjects_first_merge.
   
   int* single_subject_closest_subject;//a vector showing the closest subject to each subject.
   
   bool reduce_closeness_matrix, reduce_closeness_matrix_store;//if true, when agglomerative clustering delete row of matrix when cluster is merged.

   double* single_subject_closeness_contributions;//an array holding the closeness contribution for each subject considered as a singleton cluster.

   double* single_subject_hldvs;//an array holding the half log det(V*) contribution for each subject considered as a singleton cluster, when !heteroskedastic.

   bool uniform_background;//if true, then the data in any singleton clusters are assumed to be uniformly distributed over the data space.

   bool singleton_clusters_uniformly_distributed;//true if uniform background.

   double background_prob;//the probability that a subject is from the Uniform background model.

   double log_nonbackground_penalty;//the penalty in not being from the background model;

   double background_correction;//correction term for log_likelihood for a background cluster.

   double background_random_effect_correction;//correction term for random effects for log_likelihood for a background cluster.

   bool retain_singleton_clusters;//
   
   Cluster_GLM_MVN** single_subject_cluster_ptrs;//a map holding each of the single subject clusters when retain_singleton_clusters=true;

   deque< Cluster_GLM_MVN* > single_subject_clusters;//a deque holding the original single subject clusters, useful when performing mcmc when preserving tree hierarchy.

   deque< Cluster_GLM_MVN* > all_clusters_in_tree;//a deque holding all the clusters in the tree if preserving, useful when performing mcmc when preserving tree hierarchy and not moving just single subject clusters.

   typedef deque< Cluster_GLM_MVN* > cluster_ptrs;

   cluster_ptrs mergeable_cluster_deque;//a deque containing pointers to all mergeable clusters.

   cluster_ptrs unmergeable_cluster_deque;//a deque containing pointers to all unmergeable clusters.

   mx_numtoptr *vstar_map, **vstar_maps, *vstar_map_store, **vstar_maps_store;

   map<pair<double,long int>, long int> vstar_hash, vstar_hash_store;//hash function for vstar_map when calling with doubles.

   inttodbl *half_log_det_vstars, **half_log_det_vstar_maps, *half_log_det_vstars_store, **half_log_det_vstar_maps_store;

   int random_effect_rows;//the number of rows added to the covariance matrix for random effects

   bool currently_diagonal_vstar;//

   list<double> *closeness_contributions_store;//stores the closeness contributions of all mergeable clusters when required.

   int experiment_vote;//the number of experiments assigned to a_, b_ rather than a_1, b_1, and so driving the clustering.

   bool change_basis_variance_for_vote;//if true, change prior variance of basis coefficients rather than changing noise variances.
	
   int effective_num_exps;//the real number of experiments, possibly hidden if we are treating experiments as different dimesnions of just one experiment.

   gsl_matrix* vstar_working;//matrix used for working out in get_vstar(d), so we can call by reference.

   int dim_num_obsns;//the number of columns of the num_obsns matrices.

   gsl_matrix_int *num_obsns_working, *num_obsns_working2;

   gsl_matrix *yty_working, *ytd_working, *dtd_working;//Matrices for working out new sufficient statistics, saves having to continually create and delete.

   list<int> cl_subjects_working;//a globally viewable list of subjects, where applicable those in the cluster currently being considered.

   int yty_cols, ytd_cols, dtd_cols;//number of columns of above matrices.
   
   double* ln_gamma_a_star_lookup;// houses log(Gamma(a_ + n/2 * data_block_size )) for each n.

   double* a_star_lookup;// houses a_ + n/2 * data_block_size

   double *ln_gamma_a_star_lookup1, *a_star_lookup1;//ln_gamma_a_star_lookup, a_star_lookup, but using parameter a_1.

   double *a_star_vector, *b_star_vector;//vectors for holding a* and b* for each experiment, used when treat_experiments_as_different_dimensions.

   double *a_star_vector1, *b_star_vector1;//As a_star_vector, b_star_vector, but using parameters a_1, b_1.

   double **log_vote_probs;//A vector holding the prior probability of each experiment contributing strongly to the clustering procedure.

   bool *zero_vote_prob;//A vetor holding true/false values as to whether each experiment has zero probability of contributing strongly to the clustering procedure.

   double* error_prior_contribution_differences;//vector holding difference in closeness contribution for each expt using either a_,b_ or a_1,b_1

   int* error_prior_contribution_difference_ranks;//vector ranking the entries of error_prior_contribution_differences, with 1 indicating the largest

   double* log_prior_num_clusters;// ith entry gives the log prior for i clusters.

   double* log_prior_cluster_size;// ith entry gives the log prior for a cluster having i subjects allocated.

   double singleton_merge_prior_change;//the change in prior given by the previous array if two clusters of size 1 are merged, stored for speed.

   double* cluster_probs;//ith element is number of times there are i mergeable clusters in the mcmc.

   int cluster_probs_size;//length of cluster_probs array.
   
   double** association_matrix;//matrix holding the probability of association between subjects.

   double** m_star_exp;//matrix holding the current m_star for each experiment.

   double **merger_details, **best_merger_details;//arrays for holding the details of each merger in hierarchical clustering.

   int *dendrogram, *best_dendrogram, *inverse_dendrogram;//arrays for holding the ordering of the subjects in hierarchical clustering.

   double* dendrogram_scores;//holds the likelihood (or BIC) for each level of the final agglomerative hierarchy.
   
   int merger_details_rows, merger_details_cols, merger_details_rows_original;// number of rows and columns in merger_details arrays.
   
   Cluster_GLM_MVN* most_mergeable_cluster;// the cluster with the maximal closeness to its closest cluster.

   Cluster_GLM_MVN* most_mergeable_clusters_closest;// nearest cluster to most_mergeable_cluster.

   int most_mergeable_cluster_size, most_mergeable_clusters_closest_size;//the size of these two clusters.

   int most_mergeable_cluster_smallest_subject, most_mergeable_clusters_closest_smallest_subject;//the smallest subject index of each of these two clusters.

   Cluster_GLM_MVN *utility_cluster_pointer, *utility_cluster_pointer2;//two pointers for general use, for keeping track of a cluster.

   double utility_contribution;//a double for holding the contribution of utility_cluster_pointer;

   Cluster_GLM_MVN* proposed_parent_cl;//used for keeping track when preserving tree and performing split.
   
   Cluster_GLM_MVN* proposed_original_cl;//used for keeping track when preserving tree and performing split.

   Cluster_GLM_MVN* childs_parent_in_model;

   Cluster_GLM_MVN *node_in_model_1, *node_in_model_2;//used to hold the nodes in the model being affected by tree changes.

   int node_in_model_1_score, node_in_model_1_proposed_score, node_in_model_2_proposed_score;//variables used to hold the old and proposed scores for the above clusters.

   double max_closeness;// holds the maximal closeness value accross all clusters.

   double new_contribution, old_contributions;//when considering a merger, these hold the the new and old likelihood contributions respectively.
   
   int num_mergeable_subjects;//number of subjects allocated to clusters in mergeable_cluster_deque.

   int number_of_clusters, proposed_number_of_clusters;//Variables to hold the number of clusters in the model if needed, and the number of clusters in a proposed model.

   int number_of_singleton_clusters, proposed_number_of_singleton_clusters;//Variables to hold the number of clusters in the model which are singletons.

   int number_of_cluster_types;//Number of cluster types in the model
   
   int number_of_configurations;//the total number of possible cluster configurations - will only be calculated for small sample sizes.
   
   long desired_num_clusters; // number of clusters desired.

   int minimum_num_clusters;// minimum number of clusters allowed.

   int* nn_equivalence_relation_partition;//the partition implied by an equivalence relation using single subject closenesses.

   int nn_equivalence_relation_partition_size;//the size of nn_equivalence_relation_partition.

   bool no_optimal_clusters_totally_missing_experiments;//if true, do not allow optimal clusterings where a cluster has entirely missing values for 1 or more experiments.

   long minimum_cluster_size, maximum_cluster_size;// minimum/maximum size a cluster may be to qualify as an optimal clustering.
   
   long bic_criterion;// if 1, decide the number of clusters using BIC, if 2, decdie by maximising ML.

   long euclidean_or_correlation_clustering;//if non-zero, then if 1 just do standard euclidean hierarchical clustering, else if 2 just do correlation hierarchical clustering.

   long average_complete_single_link;//if non-zero, then if 1 do average-link clustering, if 2 do complete-link clustering, else if 3 do single-link clustering, all on the initial closeness matrix for the single subject clusters.

   bool worst_first;//if true, then rather than merging two closest clusters merge, merges the two with the smallest positive closeness.

   bool aim_for_positive_closeness;//if true and worst_first, at least looks for a positive closeness if available.

   long singletons_require_best_friend;//if non zero, then a singleton can only merge with the cluster which contains its nearest subject; if equal to 2 and both clusters are singletons, then each must be nearest neighbours to one another.

   bool nn_constraint;//if true, two clusters can merge only if they each have subjects where one is the nearest neighbour of one the other (not necessarily both ways).

   bool non_negative_nearest_neighbours;//if true, a subject has no nearest neighbour if its closest subject has negative closeness.

   bool require_outstanding_merge;//if true, then we take not the best merger but merge the cluster which has the greatest difference in score from its optimal merge against all the alternatives.

   long stochastic_agglomeration;//if non-zero, each merge in agglomerative clustering is stochastic, with probability proportional to posterior score; if equal to 1, consider mergers of any cluster with their nearest; if 2, consider any merger.
   
   bool heller_algorithm, originally_heller_algorithm;//if true, runs the clustering algorithm of Heller and Ghahramani.

   double heller_pi;//Value of pi probability if using the clustering algorithm of Heller and Ghahramani.

   double heller_log_odds, heller_odds, heller_reverse_odds;//Values of log(pi/(1-pi)), pi/(1-pi) and (1-pi)/pi if using the clustering algorithm of Heller and Ghahramani.

   double heller_dirichlet_alpha, log_heller_dirichlet_alpha;//the value of alpha (and its log) for the DP variatey of Heller and Ghahramani.

   double heller_d;//the value of d_k for the DP variatey of Heller and Ghahramani.

   double heller_log_r;//the value of log(r_k) for the DP variatey of Heller and Ghahramani.

   double heller_current_contribution;//for storing the Heller and Ghahramani likelihood contribution of the current cluster.

   bool store_targets_in_cluster;//set true for derived classes where we would like to store the targets in each cluster.

   bool euclidean_not_correlation;//if true, use euclidean distance rather than correlation.
   
   bool dirichlet_process;//if 1, cluster allocations follow a dirichlet process. otherwise uniform distn on number of allocations C, and dirichlet distn given C.

   bool hier_discrim;//if non zero, do discrimination by calculating allocation probs after each merger, then taking mixture of those probs weighted by the ML = exp(log_likelihood).

   long nearest_neighbour_discrim;//if non zero, do discrimination with probabilities calculated by finding the nearest neighbour subject for each cluster and ignoring all the others.

   bool simulate_data;//non-zero if from the start we wish to replace the target matrix values Y with some values simulated from the prior, assuming desired_num_clusters for each label.

   double dirichlet_parameter;// the parameter of the dirichlet process controlling the number of clusters/dirichlet distn.

   double log_dirichlet_parameter, log_gamma_dirichlet_parameter;

   double log_path_probability, log_proposed_path_probability;//variables for holding log of the accumulated probability as a cluster cascades down the tree.

   double prob_stop_cascading;//when cascading down tree, with this probability a priori we stop at the current level.

   double max_analytic_num_clusters;//The maximum number of starting clusters (usually subjects) for which we would attempt to find the marginal probabilities analytically, summing over all possible cluster configurations.

   double tree_score, proposed_tree_score, score_change;//overall score of the tree.

   double tree_log_normalising_constant, proposed_tree_log_normalising_constant;//variables for holding the normalising constant for the tree distribution.

   double tree_prior_parameter;//parameter for tree prior distribution.

   bool membership_probabilities_required;//if true, calculate probability of membership for each subject to its cluster, and perhaps reallocate according to these probabilities.
   
   long max_map_allocation_sweeps;//if > 0, do this many sweeps of iteratively allocating subjects to their most probable cluster, where after each sweep these probabilities are recalculated.

   long check_at_each_max_map_allocation_sweeps;//the number of sweeps to be done at each iteration of agglomerative clustering.

   int map_allocation_sweep;//sweep counter for allocating subjects to their most probable cluster.

   int total_number_of_reallocations;//Counts the number of times a subject has been reallocated by the algorithm.

   bool* subject_ever_reallocated;//a boolean vector indicating whether each subject has ever been reallocated by the algorithm.
   
   bool variable_dimension_allocation_sweeps;//if true, allow the reallocations sweeps to create new clusters containing a single subject.

   bool variable_dimension_allocation_sweeps_at_each_merger;//if true and doing iterative reallocation at each step of agglomerative clustering, allow the reallocations sweeps to create new clusters containing a single subject.
   
   bool one_reallocation_per_sweep;//if true, at each sweep only reallocate the first misallocated subject.

   bool merge_after_each_sweep;//if true, after each sweep do agglomerative clustering to see if any clusters should now merge.

   long reallocate_worst_each_sweep;//if one reallocation per sweep, then: if 1 (-1), at each sweep only reallocate the misallocated subject with highest (lowes) prob of belonging to another particular cluster; if 2 (-2), only reallocate the misallocated subject with lowest (highest) prob of belonging to its current cluster; if 0, reallocate the first misallocated subject.

   bool reallocate_subjects_at_random;//if true, randomly reallocate subject to clusters with prob proporitonal to their membership prob.

   bool check_allocations_at_each_merger, originally_check_allocations_at_each_merger;//if true, do iterative reallocation at each step of agglomerative clustering.

   bool checking_memberships_during_clustering;//internal variable which tells the function calculate_membership_probabilities() whether this is being done during the main agglomerative clustering procedure, and so should not do merging within this call.

   bool rebuild_tree_after_sweeps;//if true, then after reallocating subjects to different clusters, rebuild the dendrogram around the new clustering (and perhaps find a better new one at a different level of the hierarchy).

   bool trying_best_singleton;//internal variable  which tells the function calculate_membership_probabilities() whether this is being done after forcing a subject to become it's own singleton cluster.

   int subject_not_allowed_to_move, subject2_not_allowed_to_move;// for calculate_membership_probabilities(), these subjects can be prevented from moving.

   long try_best_singletons;//if non-zero, then in turn move the top (this many) most probable subjects to become a singleton cluster and then try reallocating.

   bool made_gain_from_best_singletons;//set as true if a gain is ever made from trying best singletons.

   double gibbs_sweep_power;//power by which predictive probs are raised when performing gibbs sweep when two_means_separation is true.

   long intermediate_gibbs_scans;//The number of intermediate gibbs scans that are taken before final gibbs scan if using gibbstwomeanseparation.

   bool deterministic_initial_split;//If true and using gibbstwomeanseparation, initially split a cluster by finding the two most distant subjects, and then successively allocate each of the remaining subjects to their nearest cluster, otherwise split totally randomly. 

   bool functional_tree_prior_parameter;//if zero, tree_prior_parameter the same for all numbers of clusters - otherwise adjust it to compensate when conditional_tree_distribution == 4.
   
   bool cascade_from_cluster_in_model;//if zero, then cascade a cluster down from root cluster, otherwise down from parent cluster in model.

   bool cascade_move_clusters;//if one, then under tree model shift cluster subjects by starting at the root and cascading down.

   bool stop_cascading_at_model;//if non zero, only cascade down to a node in the model, then resample the subtree below there from the full conditional.

   int* current_clustering;//array sending subjects to a cluster number.
   
   int* optimal_clustering;//array sending subjects to a cluster number.

   int* best_clustering;//array holding the best clustering found.

   int* subject_cluster_types;//array holding the cluster type which for the cluser each subject belongs too, if types are turned on. 

   map< int, string, less<int> > optimal_labels;// map containing the labels of the optimal clusters in the order of the map output above.

   bool find_parameter_values;//if true, calclulate m_star, b_star explicitly, otherwise perform contribution calculation more efficiently.
   
   int optimal_num_clusters;//size of optimal_clustering.

   bool non_conjugate_model;//if non zero, then using the non conjugate bayesian model where the covariance matrix for the spline coefficients is not multiplied by the error variance (sigma squared) of the observations. Y~N(Xb,sigma), b~N(0,V) => Y~N(0,sigma*I + XVX').

   int storing_non_conjugate_precision_matrices;//if greater than 0 and using non_conjugate_model, then for speed attempt for all numbers of observations up to this value to store in memory the corresponding precision matrices calculated.

   int keep_vstars;//if > 0, then for speed attempt for all numbers of observations to store in memory the corresponding precision matrices calculated. if 2 then delete the vstars of two clusters when they merge, in case this number isn't needed any more (to save space). if 3 then only store matrix when there are no missing values and no random effect rows.

   bool unimodal_bic_function;//if non zero, we assume BIC fn is unimodal, and so stop hierarchichal clustering when bic starts to drop.

   bool storing_cluster_closenesses, storing_cluster_closenesses_store;// if zero, do not store cluster closenesses - time consuming but saves memory

   bool order_cluster_subjects;// if non-zero, then within each cluster maintain a good order for abosrbed clusters so that neighbouring clusters in the dendrogram are close.

   bool merge_through_input_partition;//if true, then when reading in an initial partition, mark the subject in each cluster with a type and then immdediately split up the partition into singletons and do constrained agglomerative clustering passing through the input partition.

   bool output_all_partitions;//if true, write the partition at each level of AC to file.

   int** input_merger_details;//an array of merger details read in from Input_Merger_Details_Filename.

   bool preserve_tree;//if non-zero, do not delete clusters when merging, but instead create 'super-clusters' which contain clusters rather than subjects, and swap them into the closeness map.

   bool store_data_throughout_tree;//if non-zero and preserving tree, each parent cluster stores it's own data (eg. yty matrix).

   long conditional_tree_distribution;//non-zero if the distribution chosen for the tree is specified condtionally on all those compatible with the current model 1 - uniform, 2 - `score' distribution.

   long tree_shift_distribution;//If non zero and preserving tree, direction probabilities for a cluster cascading down tree will be proportional to closeness, otherwise uniform.

   bool preserve_model_with_tree_shifts;//If true, then if cluster picked to shift is somewhere below a cluster in the model in the tree, it will be shifted somewhere below that cluster.
   
   bool gibbs_two_mean_separation;//if true, split a cluster by first performing 2 means clustering, then running one probabilistic gibbs sweep.
   
   bool move_single_subjects_in_mcmc;//if non-zero, in tree-based mcmc only attempt to move clusters that contain a single subject, otherwise any in the tree but root clusters.

   bool splitting_to_optimal;//a book-keeping variable, to prevent random permutaions of clusters when we're deterministically splitting to the optimal number of clusters, will be 0 most of the time.

   bool recalculate_whilst_splitting_to_optimal;//determines whether we evaluate cluster contributions whilst we're deterministically splitting to the optimal number of clusters.

   bool dendro_output_format;//variable determining the dendrogram output - if zero just write the subject permutation, otherwise write out each cluster number and their subjects' responses.

   bool md_write_smallest_subjects;// when writing merger details, if non-zero, write the smallest subjects of the two clusters being merged to file. otherwise, write their first and last subjects.

   bool kill_the_parent;//if non-zero, deletes a parent cluster when it's split

   bool cluster_in_model, cluster_or_child_in_model, cluster_to_merge_in_model, clusters_parent_in_model;//if zero, then a cluster is in the tree but not the cluster map contributing directly to the likelihood.

   bool altering_current_model;//if zero, then the move being proposed will not alter the marginal likelihood.

   bool executing_move_subjects;//if non-zero, we are doing a move subject move and so need different properties when performing merge.

   bool timing_agglomerative_clustering;//if true, then record the time taken for each iteration of first run of agglomerative clustering.

   map<int,double> catalan_numbers;//map holding the catalan numbers - need to be doubles because of size.

   list<int>* proposed_subject_list;//pointer to a list of subjects proposed to be split off from a cluster.
   
   char Cluster_Output_Filename[FILENAMELENGTH];//file for writing clusters to.

   char Cluster_Means_Filename[FILENAMELENGTH];//file for writing cluster means to.

   char Cluster_Cov_Filename[FILENAMELENGTH];//file for writing cluster covariance matrices to.

   char Cluster_Variance_Filename[FILENAMELENGTH];//file for writing cluster variances to.

   char Cluster_X_Output_Filename[FILENAMELENGTH];//file for writing predictors of clusters to.

   char Cluster_Missing_Values_Filename[FILENAMELENGTH];//file for writing cluster missing values to.

   char Missing_Value_Predictions_Filename[FILENAMELENGTH];//file for writing missing value predictions to.

   char Cluster_Error_Covariances_Filename[FILENAMELENGTH];//file for writing cluster error covariances to.

   char Partition_Output_Filename[FILENAMELENGTH];//file for writing partition underlying clusters to.

   char All_Partitions_Output_Filename[FILENAMELENGTH];//file for writing all partitions visited.
  
   char Partition_Input_Filename[FILENAMELENGTH];//file for reading in a partition underlying initial clusters.
   
   char Cluster_Sizes_Filename[FILENAMELENGTH];//file containg the sizes of the clusters.

   char Subject_Labels_Filename[FILENAMELENGTH];//file containg labels for the first so many subjects and their clusters.

   char Cluster_Labels_Filename[FILENAMELENGTH];//file to which the final cluster labels will be written.

   char Cluster_Subjects_Filename[FILENAMELENGTH];//file to which the final cluster labels will be written.

   char Merger_Details_Filename[FILENAMELENGTH];//file for writing details of each merger in hierarchical clustering.

   char Bare_Merger_Details_Filename[FILENAMELENGTH];//file for writing bare details of each merger in hierarchical clustering (just 4 columsn, 2 for each cluster being merged).

   char Input_Merger_Details_Filename[FILENAMELENGTH];//file for reading in bare details of each merger in hierarchical clustering (just 4 columsn, 2 for each cluster being merged), to force all mergers.

   char Dendrogram_Output_Filename[FILENAMELENGTH];//file for writing final order of subjects after hierarchical clustering.
   
   char Dendrogram_Cutpoint_Filename[FILENAMELENGTH];//file for writing the order of the cutpoints in the Dendrogram_Output_Filename in the hierarchical clustering.

   char Dendrogram_Scores_Filename[FILENAMELENGTH];//file for writing the likelihood (or BIC) for each level of the final agglomerative hierarchy.

   char Subjects_First_Merge_Filename[FILENAMELENGTH];//file for writing the subjects_first_merger array.
   
   char Association_Matrix_Filename[FILENAMELENGTH];//file for writing the averaged subject association matrix.

   char Closeness_Matrix_Filename[FILENAMELENGTH];//file for writing the cluster closeness matrix.

   char Single_Subject_Contributions_Filename[FILENAMELENGTH];//file for writing the closeness contributions for each of the individual subjects.

   char Single_Subject_Nearest_Neighbour_Filename[FILENAMELENGTH];//file for writing the closest other subject for each of the individual subjects.

   char NN_equivalence_relation_partition_Filename[FILENAMELENGTH];//file for writing the partition implied by an equivalence relation using single subject closenesses.

   char Membership_Probability_Filename[FILENAMELENGTH];//file for writing the subject cluster membership probabilities to.

   char Misallocated_Subject_Filename[FILENAMELENGTH];//file for writing the subjects misallocated according to the cluster membership probabilities.

   char Cluster_Nearest_Neighbour_Filename[FILENAMELENGTH];//file for writing the closest other cluster for each of the reported clusters.

   char Cluster_Closeness_Filename[FILENAMELENGTH];//file for writing the closeness of the other clusters to each of the reported clusters.

   char Lift_Chart_Filename[FILENAMELENGTH];//file for writing lift chart showing performance of classifier.

   char Subject_Allocations_Filename[FILENAMELENGTH];//file for writing clusternumber and subjects allocated to those clusters.

   char Subject_Misallocation_History_Filename[FILENAMELENGTH];//file for writing the numbers of subjects not allocated to most probable clusters at each sweep of reallocation algorithm.

   char Votes_Filename[FILENAMELENGTH];//file for writing the votes as to whether an experiment is being (strongly) used in each of the clusters.

   char Vote_Probabilities_Filename[FILENAMELENGTH];//file for reading in the probability of each experiment having a strong say in the clustering.

   char Simulated_Data_Filename[FILENAMELENGTH];//file for writing simulated data to if simulate_data non zero.

   char Agglomerative_Timing_Filename[FILENAMELENGTH];//file for writing the time taken for each iteration of agglomerative clustering.

   ofstream TimingStream;//An output stream for Agglomerative_Timing_Filename.

   friend class Cluster_GLM_MVN;
};

#endif
