#include "glm_mvn_mcmc_clustering.h"

GLM_MVN_MCMC_Clustering::GLM_MVN_MCMC_Clustering( int argc, char **argv )
      : GLM_MVN_Clustering( argc, argv )
{
   starting_configuration = 2;
   random_number_of_clusters = true;
   subject_shift_move_included = 0;
   single_subject_shift_move_included = true;
   singleton_merge_split_moves_included = 0;
   changing_clustering_in_mcmc = false;
   v_shift_move_included = 0;
   prior_v_shift_move_included = 0;
   experiment_shift_move_included = true;
   tree_full_conditional_move_included = 0;
   prob_full_conditional = 1;
   shift_clusters_below_model = 0;
   metropolis_tree_changes = 1;
   agglomerative_based_split = 0;
   jump_distn = 0;
   v_shift_size = 1;
   prior_v_shift_size = .5;
   experiment_shift_size = .2;
   experiment_power_shift_size = .1;
   for( int i = 1; i < argc; i++ )
   {
      if( strstr( argv[i], "help" ) )
      {
         cout << "startingconfiguration=" << starting_configuration << "   (If =0, start MCMC with one cluster; if =1, start MCMC with every subject in its own cluster; if =2, start at the resulting configuration after hierarchical clustering)" << endl;
         cout << "birthdeath=" << random_number_of_clusters << "   (If =0, do not vary the number of clusters during the MCMC)" << endl;
//          cout << "includeshiftmove=" << subject_shift_move_included << endl;
//          cout << "includesingleshiftmove=" << single_subject_shift_move_included << endl;
//          cout <<  "shiftclustersbelowmodel=" << shift_clusters_below_model << endl;
//          cout << "singletonmergesplitmovesincluded=" << singleton_merge_split_moves_included << endl;
//          cout << "vshiftmoveincluded=" << v_shift_move_included << endl;
//          cout << "priorvshiftmoveincluded=" << prior_v_shift_move_included << endl;
//          cout << "experimentshiftmoveincluded=" << experiment_shift_move_included << endl;
//         cout << "experimentshiftsize=" << experiment_shift_size << endl;
//         cout << "experimentpowershiftsize=" << experiment_power_shift_size << endl;

//          cout << "treefullconditionalmove=" << tree_full_conditional_move_included << endl;
//          cout << "probfullconditional"  << prob_full_conditional << endl;
//          cout << "vshiftsize=" << v_shift_size << endl;
//          cout << "metropolistreechanges=" << metropolis_tree_changes << endl; 
//          cout << "jumpdistribution=" << jump_distn << endl;
         return;
      }
      if( InitialStringMatch( argv[i], "startingconfiguration=" ) )
         StringToLong( &starting_configuration, argv[i] );
      if( InitialStringMatch( argv[i], "birthdeath=" ) )
         StringToLong( &random_number_of_clusters, argv[i] );
      if( InitialStringMatch( argv[i], "includeshiftmove=" ) )
         StringToLong( &subject_shift_move_included, argv[i] );
      if( InitialStringMatch( argv[i], "includesingleshiftmove=" ) )
         StringToLong( &single_subject_shift_move_included, argv[i] );
      if( InitialStringMatch( argv[i], "shiftclustersbelowmodel=" ) )
         StringToBool( &shift_clusters_below_model, argv[i] );
      if( InitialStringMatch( argv[i], "singletonmergesplitmovesincluded=" ) )
         StringToLong( &singleton_merge_split_moves_included, argv[i] );
      if( InitialStringMatch( argv[i], "vshiftmoveincluded=" ) )
         StringToBool( &v_shift_move_included, argv[i] );
      if( InitialStringMatch( argv[i], "priorvshiftmoveincluded=" ) )
         StringToBool( &prior_v_shift_move_included, argv[i] );
      if( InitialStringMatch( argv[i], "vshiftsize" ) )
         StringToDouble( &v_shift_size, argv[i] );
      if( InitialStringMatch( argv[i], "experimentshiftmoveincluded=" ) )
         StringToBool( &experiment_shift_move_included, argv[i] );
      if( InitialStringMatch( argv[i], "experimentshiftsize=" ) )
         StringToDouble( &experiment_shift_size, argv[i] );
      if( InitialStringMatch( argv[i], "experimentpowershiftsize=" ) )
         StringToDouble( &experiment_power_shift_size, argv[i] );
      if( InitialStringMatch( argv[i], "treefullconditionalmove=" ) )
         StringToLong( &tree_full_conditional_move_included, argv[i] );
      if( InitialStringMatch( argv[i], "probfullconditional" ) )
         StringToDouble( &prob_full_conditional, argv[i] );
      if( InitialStringMatch( argv[i], "metropolistreechanges=" ) )
         StringToBool( &metropolis_tree_changes, argv[i] );
      if( InitialStringMatch( argv[i], "jumpdistribution=" ) )
         StringToLong( &jump_distn, argv[i] );
   }
   initialise_glm_mvn_mcmc_clustering();
}

GLM_MVN_MCMC_Clustering::GLM_MVN_MCMC_Clustering( gsl_matrix* X1, gsl_matrix* Y1 )
      : GLM_MVN_Clustering( X1, Y1 )
{
   initialise_glm_mvn_mcmc_clustering();
}

GLM_MVN_MCMC_Clustering::GLM_MVN_MCMC_Clustering( gsl_matrix* X1, gsl_matrix* Y1 , gsl_matrix* D1 )
      : GLM_MVN_Clustering( X1, Y1, D1 )
{
   initialise_glm_mvn_mcmc_clustering();
}

GLM_MVN_MCMC_Clustering::~GLM_MVN_MCMC_Clustering()
{
   zap_vector( stirling );
   if( stirling_mx )
   {
      for( int i = 0; i < dim_stirling_mx; i++ )
         zap_vector( stirling_mx[ i ] );
      delete [] stirling_mx;
   }
   zap_vector( experiment_location_store );
   zap_vector( experiment_location_parameter_store );
   zap_vector( experiment_location_sum_store );
   zap_vector( experiment_location_shrinkages );
   zap_vector( experiment_location_mean_shrinkages );
   zap_vector( experiment_location_posterior_precisions );
   zap_vector( experiment_location_sum_posterior_precisions );
   zap_vector( experiment_signs_store );
   zap_vector( ave_prior_v_probs );
   zap_vector( best_log_prior_v );
}

void GLM_MVN_MCMC_Clustering::initialise_glm_mvn_mcmc_clustering()
{
   kill_the_parent = 0;
   MERGE = -1;
   SPLIT = -2;
   SUBJECT_SHIFT = SINGLE_SUBJECT_SHIFT = SINGLETON_MERGE = SINGLETON_SPLIT = V_SHIFT = PRIOR_V_SHIFT = EXPERIMENT_SHIFT = EXPERIMENT_POWER_SHIFT = EXPERIMENT_SIGN_CHANGE = EXPERIMENT_MULTIPLIER_SHIFT = EXPERIMENT_PARAMETER_SHIFT = -INT_MAX;
   proposed_iter_label = 0;
   log_prior_v_store = log_prior_v_store2 = 0;
   ave_prior_v_probs = 0;
   best_log_prior_v = 0;
   experiment_location_store = 0;
   experiment_location_parameter_store = 0;
   integrate_out_experiment_parameters = true;
   experiment_location_sum_store = 0;
   experiment_location_shrinkages = 0;
   experiment_location_mean_shrinkages = 0;
   experiment_location_posterior_precisions = 0;
   experiment_location_sum_posterior_precisions = 0;
   experiment_signs_store = 0;
   experiment_variance_likelihood = proposed_experiment_variance_likelihood = 0;
   proposed_experiment_variance_likelihood_diff = proposed_experiment_variance_likelihood_diff2 = 0;
   if( experiment_shift_move_included )
   {
      if( !experiment_locations )
         experiment_shift_move_included = 0;
      else
      {
         experiment_location_store = new double[ experiment_location_dimension ];
         experiment_location_parameter_store = new double[ experiment_location_dimension ];
	 experiment_location_sum_store = new double[ experiment_location_dimension ];
         if( experiment_signs )
            experiment_signs_store = new int[ num_locations ];
         cout << "Experiment latent position variance = " << 1/experiment_location_precision << endl;
         if( kriging_experiment_correlation )
            cout << "Experiment latent mean position variance = " << 1/experiment_parameter_precision << endl;
      }
   }
   if( random_number_of_clusters )
   {
      MERGE = add_move( "Merge" );
      SPLIT = add_move( "Split" );
      trans_dimensional_move_types.push_back( MERGE );
      trans_dimensional_move_types.push_back( SPLIT );
      integer_parameter_pointers.push_back( &number_of_clusters );
   }
   if( subject_shift_move_included )
   {
      if( !preserve_tree )
         SUBJECT_SHIFT = add_move( "Subject shift" );
      else
         SUBJECT_SHIFT = add_move( "Tree cluster shift" );
   }
   if( single_subject_shift_move_included )
      SINGLE_SUBJECT_SHIFT = add_move( "Single subject shift" );
   if( singleton_merge_split_moves_included )
   {
      SINGLETON_MERGE = add_move( "Singleton merge" );
      SINGLETON_SPLIT = add_move( "Singleton split" );
      trans_dimensional_move_types.push_back( SINGLETON_MERGE );
      trans_dimensional_move_types.push_back( SINGLETON_SPLIT );
   }
   if( random_number_of_clusters || subject_shift_move_included || single_subject_shift_move_included || singleton_merge_split_moves_included )
      changing_clustering_in_mcmc = true;
   if( v_shift_move_included )
   {
      if( number_of_vs == 1 )
      {
         V_SHIFT = add_move( "V shift" );
         parameter_pointers.push_back( &v_multiplier );
      }
      else
         v_shift_move_included = 0;
   }
   if( prior_v_shift_move_included )
   {
      if( mixture_of_vs )
      {
         PRIOR_V_SHIFT = add_move( "p(V) shift" );
         if( run_ergodic_ave_normaliser && !simulated_annealing )
            ave_prior_v_probs = new double[ number_of_vs ];
      }
      else
         prior_v_shift_move_included = 0;
   }
   if( experiment_shift_move_included )
   {
      EXPERIMENT_SHIFT = add_move( "Expt. location shift" );
      if( experiment_signs && kriging_experiment_correlation )
         EXPERIMENT_SIGN_CHANGE = add_move( "Expt. correlation sign change" );
      if( experiment_var_multipliers && kriging_experiment_correlation )
         EXPERIMENT_MULTIPLIER_SHIFT = add_move( "Expt. variance shift" );
      if( kriging_experiment_correlation )
      {
         EXPERIMENT_POWER_SHIFT = add_move( "Expt. location power shift" );
	 if( !experiment_closeness_power )
	 {
	    experiment_closeness_power = new double;
	    *experiment_closeness_power = 1;
	 }
         best_experiment_closeness_power = *experiment_closeness_power;
         parameter_pointers.push_back( experiment_closeness_power );
      }
      if( experiment_location_parameters != experiment_location_means )
//         if( !integrate_out_experiment_parameters )
            EXPERIMENT_PARAMETER_SHIFT = add_move( "Expt. location mean shift" );
   }
   if( tree_full_conditional_move_included )
   {
      if( preserve_tree && ( conditional_tree_distribution == 3 || conditional_tree_distribution == 4 ) )
      {
         stop_cascading_at_model = 1;
         if( !simulated_annealing )//otherwise need the structure
            cascade_move_clusters = 0;
         if( tree_full_conditional_move_included == 2 )
            TREE_FULL_CONDITIONAL = add_move( "Tree full conditional" );
      }
      else
         tree_full_conditional_move_included = 0;
   }

   iter_distinct = 0;
   iter_splitable = 0;
   zap_vector( cluster_probs );
   cluster_probs_size = num_subjects + 1;
   cluster_probs = new double[ cluster_probs_size ];
   for( int i = 0; i < cluster_probs_size; i++ )
      cluster_probs[ i ] = 0;
   initialise_mergeable_cluster_deque();
   if( preserve_tree )
   {
      starting_configuration = 2;
//      bic_criterion = 0;
      desired_num_clusters = 1;
   }
   set_starting_clusters();
   stirling = 0;
   stirling_mx = 0;
   if( !preserve_tree )
      calculate_stirling_numbers();
   if( v_shift_move_included )
   {
      variable_v = 1;
      if( !strlen( V_Filename ) )
      {
         char* new_file = V_Filename;
         strcpy( new_file, TargetFilename );
         string insert = "_v_";
         insert_in_string_before_dot( new_file, &insert );
      }
   }
   if( shift_clusters_below_model )
      cascade_from_cluster_in_model = 1;
   if( preserve_tree && ( conditional_tree_distribution >= 2 && conditional_tree_distribution <= 4 ) )
      parameter_pointers.push_back( &tree_score );
   Cluster_GLM_MVN::set_two_means_separation( gibbs_two_mean_separation );
   if( random_number_of_clusters && gibbs_two_mean_separation )
   {
      cout << "Splitting clusters ";
      if( deterministic_initial_split )
         cout << "using 2-means clustering";
      else
         cout << "completely at random";
      cout << " followed by " << intermediate_gibbs_scans + 1 << " gibbs sweep(s), with probabilities raised to the power " << gibbs_sweep_power << endl;
   }
   zap_array( association_matrix, num_subjects, num_subjects );
   //we'll create it again after any agglomerative clustering.
   if( experiment_covariances && integrate_out_experiment_parameters )
   {
      experiment_location_shrinkages = new double[ experiment_partition_size ];
      experiment_location_mean_shrinkages = new double[ experiment_partition_size ];
      experiment_location_posterior_precisions = new double[ experiment_partition_size ];
      experiment_location_sum_posterior_precisions = new double[ experiment_partition_size ];
      for( int i = 0; i < experiment_partition_size; i++ )
      {
         int set_size = experiment_partition_sizes ? experiment_partition_sizes[i] : num_experiments;
         experiment_location_shrinkages[ i ] = 1 / ( ( set_size - 1 ) + experiment_parameter_precision/ experiment_location_precision );
         experiment_location_mean_shrinkages[ i ] = 1 / ( set_size + experiment_parameter_precision/ experiment_location_precision );
         experiment_location_posterior_precisions[ i ] = experiment_location_precision/( 1 + experiment_location_shrinkages[ i ] );
         experiment_location_sum_posterior_precisions[ i ] = 1 / ( experiment_location_precision + set_size * experiment_parameter_precision );
         copy_scaled_array( experiment_location_parameters[ i ], experiment_location_sums[ i ], experiment_location_mean_shrinkages[ i ], experiment_location_dimension );
      }
   }
}

void GLM_MVN_MCMC_Clustering::add_post_burn_in_moves()
{
//   if( experiment_shift_move_included )
//      EXPERIMENT_SIGN_CHANGE = add_move( "Expt. correlation sign change" );
}

void GLM_MVN_MCMC_Clustering::update_any_stats()
{
   initialise_mergeable_cluster_deque();
}

void GLM_MVN_MCMC_Clustering::set_starting_clusters()
{
   if( starting_configuration == 0 ) // all in one cluster, or as few as possible, to begin with
   {
      if( preserve_tree )
         calculate_single_subject_cluster_closeness_contributions();
      bic_criterion = 0;
      desired_num_clusters = 1;
      cluster_ptrs::iterator iter, iter2;
      Cluster_GLM_MVN *cl1, *cl2;
      int counter = 0, counter_max = mergeable_cluster_deque.size();
      while( counter < counter_max )
      {
         iter = mergeable_cluster_deque.begin();
         iter2 = iter;
         ++iter2;
         while( iter2 != mergeable_cluster_deque.end() )
         {
            if( deque_iterators_mergeable( iter, iter2 ) )
            {
               cl1 = *iter;
               cl2 = *iter2;
               cluster_map.erase( cl1 );
               mergeable_cluster_deque.pop_front();
               cl2->merge_cluster( cl1 );
               iter = mergeable_cluster_deque.begin();
               iter2 = iter;
               counter++;
            }
            ++iter2;
         }
         cl1 = *iter;
         mergeable_cluster_deque.pop_front();
         mergeable_cluster_deque.push_back( cl1 );
         counter++;
      }
      iter = mergeable_cluster_deque.begin();
      while( iter != mergeable_cluster_deque.end() )
      {
         (*iter)->set_closeness_contribution( calculate_closeness_contribution( *iter ) );
         ++iter;
      }
   }
   else if( starting_configuration == 1 )// all in individual clusters
      desired_num_clusters = Cluster_GLM_MVN::get_cluster_count();
}

void GLM_MVN_MCMC_Clustering::get_subject_cluster_types()
{
   if( !find_map || !changing_clustering_in_mcmc || map_allocation_sweep > 0 )
      GLM_MVN_Clustering::get_subject_cluster_types();
   else if( !Cluster_GLM_MVN::get_using_types() )
   {
      if( subject_cluster_types != best_clustering )
	 zap_vector( subject_cluster_types );
      subject_cluster_types = NULL;
   }
   else
   {
      number_of_cluster_types = best_num_clusters;
      subject_cluster_types = best_clustering;
   }
}

void GLM_MVN_MCMC_Clustering::calculate_stirling_numbers()
{
   dim_stirling_mx = num_mergeable_subjects + 1;
   if( !stirling_mx )
   {
     stirling_mx = new double* [ dim_stirling_mx ];
     for( int i = 0; i < dim_stirling_mx; i++ )
     {
        stirling_mx[ i ] = new double[ i + 1 ];
	for( int j = 0; j <= i; j++ )
	   stirling_mx[ i ][ j ] = 0;
     }
   }
   long stirl = 3;
   stirling_mx[ 1 ][ 1 ] = 1;
   if( dim_stirling_mx > 2 )
      stirling_mx[ 2 ][ 1 ] = 1;
   if( stirl == 1 || stirl == 2 )
   {
      if( stirl == 1 )//matrix to contain Stirling numbers of 1st kind
      {
         for( int i = 2; i < dim_stirling_mx; i++ )
            for( int j = 1; j <= i; j++ )
               stirling_mx[ i ][ j ] = -(double)j * stirling_mx[ i - 1][ j ] + stirling_mx[ i - 1 ][ j - 1 ];
      }
      else if( stirl == 2 )//matrix to contain Stirling numbers of 2nd kind
      {
         stirling_mx[ 2 ][ 2 ] = 1;
         for( int i = 3; i < dim_stirling_mx; i++ )
            for( int j = 1; j <= i; j++ )
               stirling_mx[ i ][ j ] = (double)j * stirling_mx[ i - 1 ][ j ] + stirling_mx[ i - 1 ][ j - 1 ];
      }
      for( int i = 1; i < dim_stirling_mx; i++ )
         for( int j = 1; j <= i; j++ )
            stirling_mx[ i ][ j ] = log( abs( stirling_mx[ i ][ j ] ) );// + gsl_sf_lnfact( j );
   }
   else if( stirl == 3 )
   {
      for( int i = 1; i < dim_stirling_mx; i++ )
         for( int j = 2; j <= i; j++ )
            stirling_mx[ i ][ j ] = log( pow( (double)j, (double)i ) - j ) - gsl_sf_lnfact( j );
   }
}

void GLM_MVN_MCMC_Clustering::move()
{
   if( move_type == MERGE || move_type == SINGLETON_MERGE )
   {
      if( number_of_clusters == 1 )
      {
         if( move_type == MERGE )
            move_type = SPLIT;
         else if( move_type == SINGLETON_MERGE )
            move_type = SINGLETON_SPLIT;
         LogProposalRatio -= LOG_TWO;
      }
      else if( number_of_clusters == 2 )
         LogProposalRatio += LOG_TWO;
   }

   if( move_type == SPLIT || move_type == SINGLETON_SPLIT )
   {
      if( number_of_clusters == (int)single_subject_clusters.size() )
      {
         if( move_type == SPLIT )
            move_type = MERGE;
         else if( move_type == SINGLETON_SPLIT )
            move_type = SINGLETON_MERGE;
         LogProposalRatio -= LOG_TWO;
      }
      else if( number_of_clusters == (int)single_subject_clusters.size() - 1 )
         LogProposalRatio += LOG_TWO;
   }

   if( move_type == SUBJECT_SHIFT || move_type == SINGLE_SUBJECT_SHIFT )
      if( number_of_clusters == 1 && ( !preserve_tree || !cascade_move_clusters ) )
      {
         if( singleton_merge_split_moves_included )
            move_type = SINGLETON_SPLIT;
         else
            move_type = SPLIT;
         LogProposalRatio -= LOG_THREE;
      }
   
   proposed_cluster = proposed_cluster2 = NULL;
   proposed_v_num = 0;
   proposed_v_num2 = 1;
   LogProposalRatio = 0;
   find_cluster_to_alter();

   if( tree_full_conditional_move_included == 1 && ( move_type == MERGE || move_type == SINGLETON_MERGE ) && proposed_iter != mergeable_cluster_deque.end() && with_probability( prob_full_conditional ) )
   {
      Cluster_GLM_MVN* cl = (*proposed_iter)->get_my_root_cluster();
      sample_subtree_above_model_nodes_from_full_conditional( cl );
   }

   bool found_a_move = 0;
   double tree_prior_change = 0;
   
   while( ( preserve_tree && shift_clusters_below_model && number_of_clusters == (int)single_subject_clusters.size() && ( move_type == SUBJECT_SHIFT || move_type == SINGLE_SUBJECT_SHIFT ) ) || ( move_type == PRIOR_V_SHIFT && !v_matrices_differ ) )
      draw_move_type();

   int i, j;
   while( !found_a_move )
   {
      if( move_type == MERGE || move_type == SINGLETON_MERGE ) // Death for a cluster (MERGE)
      {
         if( proposed_iter != mergeable_cluster_deque.end() )
         {
            iter_distinct = (*proposed_iter)->get_distinct();
            if( move_type == SINGLETON_MERGE )
               iter_splitable = (*proposed_iter)->get_splitable();
            find_mate();
            if( proposed_iter2 != mergeable_cluster_deque.end() )
               found_a_move = 1;
         }
         if( !found_a_move )
         {
            if( preserve_tree )
               random_shuffle( mergeable_cluster_deque.begin(), mergeable_cluster_deque.end() );
            else
            {
               if( move_type == MERGE )
                  move_type = SPLIT;
               else if ( move_type == SINGLETON_MERGE )
                  move_type = SINGLETON_SPLIT;
               LogProposalRatio -= LOG_TWO;
            }
            proposed_iter = mergeable_cluster_deque.begin();
         }
         else if( number_of_clusters == 2 )
            LogProposalRatio += LOG_TWO;
      }
      else if( move_type == SPLIT || move_type == SINGLETON_SPLIT ) // Birth to a new cluster (SPLIT)
      {
         while( proposed_iter != mergeable_cluster_deque.end() && !(*proposed_iter)->get_splitable() )
            ++proposed_iter;
         if( proposed_iter != mergeable_cluster_deque.end() )
            found_a_move = 1;
         if( !found_a_move )
         {
            if( move_type == SPLIT )
               move_type = MERGE;
            else if ( move_type == SINGLETON_SPLIT )
               move_type = SINGLETON_MERGE;
            LogProposalRatio -= LOG_TWO;
            if( preserve_tree )
               random_shuffle( mergeable_cluster_deque.begin(), mergeable_cluster_deque.end() );
            proposed_iter = mergeable_cluster_deque.begin();
         }
         else
         {
            if( number_of_clusters == (int)single_subject_clusters.size() - 1 )
               LogProposalRatio += LOG_TWO;
            if( agglomerative_based_split )
            {
               proposed_cluster = *proposed_iter;
               rebuild_agglomerative_tree_from_cluster( proposed_cluster );
            }
            else if( tree_full_conditional_move_included == 1 )
            {
               proposed_cluster = *proposed_iter;
               if( with_probability( prob_full_conditional ) )
                  sample_model_node_subtree_from_full_conditional( proposed_cluster );
            }
         }
      }
      else if( move_type == SUBJECT_SHIFT || move_type == SINGLE_SUBJECT_SHIFT )// Move some subject between two clusters
      {
         if( !preserve_tree || !cascade_move_clusters )
         {
            while( proposed_iter != mergeable_cluster_deque.end() && !(*proposed_iter)->get_splitable() )
               ++proposed_iter;
            if( proposed_iter != mergeable_cluster_deque.end() )
            {
               iter_distinct = (*proposed_iter)->get_distinct();
               if( !preserve_tree )
                  proposed_subject = (*proposed_iter)->find_random_subject();
               else
               {
                  holding_cluster = (*proposed_iter)->draw_random_single_subject_cluster_allocated();
		  proposed_subject = holding_cluster->get_first_subject_allocated();
               }
               find_mate();
               if( proposed_iter2 != mergeable_cluster_deque.end() )
                  found_a_move = 1;
            }
            if( !found_a_move )
            {
               if( random_number_of_clusters )
               {
                  LogProposalRatio -= log( (double)get_number_of_moves() );
                  move_type = MERGE;
               }
               else//no available move for this cluster, so pick another
                  random_shuffle( mergeable_cluster_deque.begin(), mergeable_cluster_deque.end() );
               proposed_iter = mergeable_cluster_deque.begin();
            }
         }
         else
         {
            while( !proposed_cluster )
            {
               if( move_type == SINGLE_SUBJECT_SHIFT )
               {
                  move_single_subjects_in_mcmc = true;
                  random_shuffle( single_subject_clusters.begin(), single_subject_clusters.end() );
                  proposed_cluster = *(single_subject_clusters.begin());
               }
               else
               {
                  move_single_subjects_in_mcmc = false;
                  random_shuffle( all_clusters_in_tree.begin(), all_clusters_in_tree.end() );
                  proposed_cluster = *(all_clusters_in_tree.begin());
               }
               proposed_cluster = proposed_cluster->get_parent_cluster();
               proposed_cluster2 = proposed_cluster;
               if( proposed_cluster && shift_clusters_below_model && has_child_in_model( proposed_cluster )  )
                  proposed_cluster = NULL;
            }
            if( !conditional_tree_distribution )
            {
               if( cluster_by_labels )
                  num_configurations = get_number_of_possible_clusterings_for_tree_containging_cluster( proposed_cluster );
               tree_prior_change = log( num_configurations );
            }
            found_a_move = 1;
         }
      }
      else if( move_type == V_SHIFT )// Change V_
      {
         found_a_move = 1;
         v_multiplier_store = v_multiplier;
//         v_multiplier *= exp( gsl_ran_ugaussian( rn ) * v_shift_size );
         v_multiplier = fabs( v_multiplier + gsl_ran_flat( rn, -.5, .5 ) * v_shift_size );
         double v_multiplier_ratio = v_multiplier_store / v_multiplier;
         if( V_ )
            gsl_matrix_scale( V_, 1.0 / v_multiplier_ratio );
         gsl_matrix_scale( V_inv, v_multiplier_ratio );
         if( !non_conjugate_model )
         {
            ml_const_store = ml_const;
            ml_const += 0.5 * p_y * dim_V * ( log( v_multiplier_ratio ) );
         }
//         proposed_log_auxiliary_posterior += 0.5 * ( v_multiplier_store - v_multiplier ) * experiment_location_precision;
         proposed_log_auxiliary_posterior += log_gamma_density_ratio( v_multiplier, v_multiplier_store, v_multiplier_gamma_shape_parameter, v_multiplier_gamma_scale_parameter );
      }
      else if( move_type == PRIOR_V_SHIFT )// Change log_prior_v[ . ] for 2 indices
      {
         found_a_move = 1;
         if( number_of_vs > 2 )
         {
            proposed_v_num = gsl_rng_uniform_int( rn, number_of_vs );
            proposed_v_num2 = gsl_rng_uniform_int( rn, number_of_vs - 1);
            if( proposed_v_num2 == proposed_v_num )
               proposed_v_num2 = number_of_vs - 1;
         }
         log_prior_v_store = log_prior_v[ proposed_v_num ];
         log_prior_v_store2 = log_prior_v[ proposed_v_num2 ];
         double prior_v1 = exp( log_prior_v_store );
         double prior_v_total = prior_v1 + exp( log_prior_v_store2 );
         prior_v1 += gsl_ran_flat( rn, -.5, .5 ) * prior_v_shift_size;
         double lowerb = 0;
         double upperb = prior_v_total;
         bounce( prior_v1, lowerb, upperb );
         log_prior_v[ proposed_v_num ] = log( prior_v1 );
         log_prior_v[ proposed_v_num2 ] = log( prior_v_total - prior_v1 );
      }
      
      else if( move_type == EXPERIMENT_SHIFT || move_type == EXPERIMENT_SIGN_CHANGE || move_type == EXPERIMENT_MULTIPLIER_SHIFT || move_type == EXPERIMENT_POWER_SHIFT )// Change an experiment location
      {
         found_a_move = 1;
         set_v_index( 0 );
         if( correlate_experiments && !non_conjugate_model && number_of_vs == 1 )
         {
            ml_const_store = ml_const;
            ml_const -= half_logdet_v_inv[ 0 ];
         }
         else if( correlate_columns )
         {
             lambda_prior_normalising_constant_store = lambda_prior_normalising_constant;
             lambda_prior_normalising_constant -= ( a_ + p_y - 1 ) / 2 * my_lndet( b_matrix );

         }
         if( move_type == EXPERIMENT_SHIFT || move_type == EXPERIMENT_SIGN_CHANGE || move_type == EXPERIMENT_MULTIPLIER_SHIFT )
             proposed_experiment_index = gsl_rng_uniform_int( rn, num_locations );
         if( move_type == EXPERIMENT_SHIFT )
         {
            double* mu_temp = NULL;
            if( integrate_out_experiment_parameters )
            {
               proposed_experiment_group = experiment_partition? experiment_partition[ proposed_experiment_index ]:0;
               subtract_array( experiment_location_sums[ proposed_experiment_group ], experiment_locations[ proposed_experiment_index ], experiment_location_dimension );
               copy_scaled_array( mu_temp, experiment_location_sums[ proposed_experiment_group ], experiment_location_shrinkages[ proposed_experiment_group ], experiment_location_dimension );
               proposed_experiment_variance_likelihood_diff = get_array_difference_norm( experiment_locations[ proposed_experiment_index ], mu_temp, experiment_location_dimension );
            }
            else
               proposed_experiment_variance_likelihood_diff = get_experiment_location_squared_distance_from_mean( proposed_experiment_index );
            for( j = 0; j < experiment_location_dimension; j++ )
            {
               experiment_location_store[ j ] = experiment_locations[ proposed_experiment_index ][ j ];
               if( kriging_experiment_correlation )
                  experiment_locations[ proposed_experiment_index ][ j ] += gsl_ran_gaussian( rn, experiment_shift_size );
            }
            if( !kriging_experiment_correlation )
               experiment_locations[ proposed_experiment_index ][ gsl_rng_uniform_int( rn, experiment_location_dimension ) ] += gsl_ran_gaussian( rn, experiment_shift_size );
            if( integrate_out_experiment_parameters )
            {
               proposed_experiment_variance_likelihood_diff -= get_array_difference_norm( experiment_locations[ proposed_experiment_index ], mu_temp, experiment_location_dimension );
               proposed_experiment_variance_likelihood_diff *= 0.5 * experiment_location_posterior_precisions[ proposed_experiment_group ];
               delete[] mu_temp;
            }
            else
            {
               proposed_experiment_variance_likelihood_diff -= get_experiment_location_squared_distance_from_mean( proposed_experiment_index );
               proposed_experiment_variance_likelihood_diff *= 0.5 * experiment_location_precision;
            }
            proposed_experiment_variance_likelihood = experiment_variance_likelihood + proposed_experiment_variance_likelihood_diff;
            proposed_log_auxiliary_posterior += proposed_experiment_variance_likelihood_diff;
         }
         else if( move_type == EXPERIMENT_SIGN_CHANGE )
         {
            if( !kriging_experiment_correlation )
               for( i = 0; i < num_locations; i++ )
                  proposed_experiment_variance_likelihood += 0.5 * get_experiment_location_squared_distance_from_mean( i ) * experiment_location_precision;
            for( i = 0; i < num_locations; i++ )
               experiment_signs_store[ i ] = experiment_signs[ i ];
            reverse_experiment_sign( proposed_experiment_index );
            proposed_experiment_index2 = gsl_rng_uniform_int( rn, num_locations - 1 );
            if( proposed_experiment_index2 == proposed_experiment_index )
               proposed_experiment_index2 = num_locations - 1;
            for( i = 0; i < num_locations; i++ )
               if( i != proposed_experiment_index && i != proposed_experiment_index2 && gsl_ran_bernoulli( rn, .5 ) )
                  reverse_experiment_sign( i );
            if( !kriging_experiment_correlation )
               for( i = 0; i < num_locations; i++ )
                 proposed_experiment_variance_likelihood -= 0.5 * get_experiment_location_squared_distance_from_mean( i ) * experiment_location_precision;
            if( !kriging_experiment_correlation )
               proposed_log_auxiliary_posterior += proposed_experiment_variance_likelihood - experiment_variance_likelihood;
         }
         else if( move_type == EXPERIMENT_MULTIPLIER_SHIFT )
         {
            experiment_var_multiplier_store = experiment_var_multipliers[ proposed_experiment_index ];
            experiment_var_multipliers[ proposed_experiment_index ] = fabs( experiment_var_multiplier_store + gsl_ran_flat( rn, -.5, .5 ) * v_shift_size );
//            proposed_log_auxiliary_posterior += 0.5 * ( experiment_var_multiplier_store - experiment_var_multipliers[ proposed_experiment_index ] ) * experiment_location_precision;
            proposed_experiment_variance_likelihood += log_gamma_density_ratio( experiment_var_multipliers[ proposed_experiment_index ], experiment_var_multiplier_store, v_multiplier_gamma_shape_parameter, v_multiplier_gamma_scale_parameter );
            proposed_log_auxiliary_posterior += proposed_experiment_variance_likelihood - experiment_variance_likelihood;
         }
         else if( move_type == EXPERIMENT_POWER_SHIFT )
         {
            experiment_closeness_power_store = *experiment_closeness_power;
            *experiment_closeness_power += gsl_ran_flat( rn, -.5, .5 ) * experiment_power_shift_size;
            double lowerb = 0;
            double upperb = 1;
            bounce( *experiment_closeness_power, lowerb, upperb );
         }
         if( move_type == EXPERIMENT_SHIFT || move_type == EXPERIMENT_MULTIPLIER_SHIFT )
            update_correlated_experiments_covariances( proposed_experiment_index );
         else
            update_correlated_experiments_covariances();//recalculate them all.
         update_v_arrays();
         if( correlate_experiments && !non_conjugate_model && number_of_vs == 1 )
            ml_const += half_logdet_v_inv[ 0 ];
         else if( correlate_columns )
            lambda_prior_normalising_constant += ( a_ + p_y - 1 ) / 2 * my_lndet( b_matrix );
      }
      else if( move_type == EXPERIMENT_PARAMETER_SHIFT )
      {
         found_a_move = 1;
         proposed_experiment_group = gsl_rng_uniform_int( rn, num_location_parameters );
         if( integrate_out_experiment_parameters )
         {
            set_v_index( 0 );
            if( correlate_experiments && !non_conjugate_model && number_of_vs == 1 )
            {
               ml_const_store = ml_const;
               ml_const -= half_logdet_v_inv[ 0 ];
            }
            else if( correlate_columns )
            {
               lambda_prior_normalising_constant_store = lambda_prior_normalising_constant;
               lambda_prior_normalising_constant -= ( a_ + p_y - 1 ) / 2 * my_lndet( b_matrix );
            }
//            proposed_experiment_variance_likelihood_diff = get_array_norm( experiment_location_sums[ proposed_experiment_group ], experiment_location_dimension );
            //*********************
            proposed_experiment_variance_likelihood_diff = 0;
            for( j = 0; j < experiment_location_dimension; j++ )
            {
               double sum_j_sq = experiment_location_sums[ proposed_experiment_group ][ j ] * experiment_location_sums[ proposed_experiment_group ][ j ] ;
               proposed_experiment_variance_likelihood_diff -= sum_j_sq * experiment_location_mean_shrinkages[ proposed_experiment_group ];
               for( i = 0; i < num_locations; i++ )
                  if( ( experiment_partition && experiment_partition[ i ] == proposed_experiment_group ) || ( !experiment_partition && i == proposed_experiment_group ) )
                     proposed_experiment_variance_likelihood_diff += experiment_locations[ i ][ j ] * experiment_locations[ i ][ j ];
            }
            //******************
            
            for( j = 0; j < experiment_location_dimension; j++ )
            {
               double shift_j = gsl_ran_gaussian( rn, experiment_shift_size );
               experiment_location_store[ j ] = shift_j;//use the store vector here to hold the shifts rather than the old locations, because we're moving many locations, byt the same shift, at once.
               for( i = 0; i < num_locations; i++ )
                  if( ( experiment_partition && experiment_partition[ i ] == proposed_experiment_group ) || ( !experiment_partition && i == proposed_experiment_group ) )
                     experiment_locations[ i ][ j ] += shift_j;
               if( experiment_partition_sizes )
                  shift_j *= experiment_partition_sizes[ proposed_experiment_group ];
               experiment_location_sums[ proposed_experiment_group ][ j ] += shift_j;
            }
            //*********************
            for( j = 0; j < experiment_location_dimension; j++ )
            {
               double sum_j_sq = experiment_location_sums[ proposed_experiment_group ][ j ] * experiment_location_sums[ proposed_experiment_group ][ j ];
               proposed_experiment_variance_likelihood_diff += sum_j_sq * experiment_location_mean_shrinkages[ proposed_experiment_group ];
               for( i = 0; i < num_locations; i++ )
                  if( ( experiment_partition && experiment_partition[ i ] == proposed_experiment_group ) || ( !experiment_partition && i == proposed_experiment_group ) )
                     proposed_experiment_variance_likelihood_diff -= experiment_locations[ i ][ j ] * experiment_locations[ i ][ j ];
            }
            proposed_experiment_variance_likelihood_diff *= .5 * experiment_location_precision;
            //******************
//            proposed_experiment_variance_likelihood_diff -= get_array_norm( experiment_location_sums[ proposed_experiment_group ], experiment_location_dimension );
//            proposed_experiment_variance_likelihood_diff *= 0.5 * experiment_location_sum_posterior_precisions[ proposed_experiment_group ];
            proposed_experiment_variance_likelihood = experiment_variance_likelihood + proposed_experiment_variance_likelihood_diff;
            proposed_log_auxiliary_posterior += proposed_experiment_variance_likelihood_diff;
            update_correlated_experiments_covariances();
	    update_v_arrays();
	    if( correlate_experiments && !non_conjugate_model && number_of_vs == 1 )
               ml_const += half_logdet_v_inv[ 0 ];
	    else if( correlate_columns )
	      lambda_prior_normalising_constant += ( a_ + p_y - 1 ) / 2 * my_lndet( b_matrix );
         }
         else
         {
            copy_array( experiment_location_parameter_store, experiment_location_parameters[ proposed_experiment_group ], experiment_location_dimension );
            proposed_experiment_variance_likelihood_diff = 0;
            for( i = 0; i < num_locations; i++ )
               if( ( experiment_partition && experiment_partition[ i ] == proposed_experiment_group ) || ( !experiment_partition && i == proposed_experiment_group ) )
                  proposed_experiment_variance_likelihood_diff += get_experiment_location_squared_distance_from_mean( i );
            proposed_experiment_variance_likelihood_diff2 = get_array_norm( experiment_location_parameters[ proposed_experiment_group ], experiment_location_dimension );
            experiment_location_parameters[ proposed_experiment_group ][ gsl_rng_uniform_int( rn, experiment_location_dimension ) ] += gsl_ran_gaussian( rn, 1 / sqrt( experiment_parameter_precision ) );
            for( i = 0; i < num_locations; i++ )
               if( ( experiment_partition && experiment_partition[ i ] == proposed_experiment_group ) || ( !experiment_partition && i == proposed_experiment_group ) )
                  proposed_experiment_variance_likelihood_diff -= get_experiment_location_squared_distance_from_mean( i );
            proposed_experiment_variance_likelihood_diff *= 0.5 * experiment_location_precision;
            proposed_experiment_variance_likelihood_diff2 -= get_array_norm( experiment_location_parameters[ proposed_experiment_group ], experiment_location_dimension );
            proposed_experiment_variance_likelihood_diff2 *= 0.5 * experiment_parameter_precision;
            proposed_experiment_variance_likelihood = experiment_variance_likelihood + proposed_experiment_variance_likelihood_diff + proposed_experiment_variance_likelihood_diff2;
            proposed_log_auxiliary_posterior += proposed_experiment_variance_likelihood_diff + proposed_experiment_variance_likelihood_diff2;
         }
      }
      else if( move_type == TREE_FULL_CONDITIONAL )// Change an experiment location
      {
         found_a_move = 1;
         proposed_cluster = *proposed_iter;
         sample_subtree_from_full_conditional();
      }
   }

   if( !( move_type == V_SHIFT || move_type == EXPERIMENT_SHIFT || move_type == EXPERIMENT_SIGN_CHANGE || move_type == EXPERIMENT_POWER_SHIFT || move_type == EXPERIMENT_MULTIPLIER_SHIFT || move_type == EXPERIMENT_PARAMETER_SHIFT ) )
   {
      if( !proposed_cluster )
         proposed_cluster = *proposed_iter;
      if( !proposed_cluster2 )
         proposed_cluster2 = *proposed_iter2;
   }

   if( preserve_tree && !cascade_move_clusters && ( move_type == SUBJECT_SHIFT || move_type == SINGLE_SUBJECT_SHIFT ) )
   {
      int numsubs1 = proposed_cluster->get_num_subjects_allocated();
      int numsubs2 = proposed_cluster2->get_num_subjects_allocated();
      if( move_type == SUBJECT_SHIFT )
      {
         proposed_cluster = proposed_cluster->draw_random_cluster_from_all_allocated();
         LogProposalRatio += log( 2 * numsubs1 - 2.0 ) - log( 2 * ( numsubs2 + 1 ) - 2.0 );
      }
      else
      {
         proposed_cluster = holding_cluster;//proposed_cluster->draw_random_single_subject_cluster_allocated();
         LogProposalRatio += log( (double)numsubs1 ) - log( (double)numsubs2 + 1 );
      }
      proposed_cluster = proposed_cluster->get_parent_cluster();
   }

   if( !preserve_tree && ( move_type== MERGE || move_type == SINGLETON_MERGE || move_type == SPLIT || move_type == SINGLETON_SPLIT ) )
      calculate_number_mergeable_splitable();

   propose_move();

   if( move_type == MERGE || move_type == SINGLETON_MERGE )
      proposed_iter_label = proposed_cluster2->get_label();
   else if( !( move_type == V_SHIFT || move_type == EXPERIMENT_SHIFT || move_type == EXPERIMENT_SIGN_CHANGE || move_type == EXPERIMENT_MULTIPLIER_SHIFT || move_type == EXPERIMENT_POWER_SHIFT || move_type == EXPERIMENT_PARAMETER_SHIFT || move_type == TREE_FULL_CONDITIONAL ) )
      proposed_iter_label = proposed_cluster->get_label();

   /*
     if( move_type == MERGE )
     LogProposalRatio -= log( 2 * ( pow( 2, (*proposed_iter)->get_num_subjects_allocated() + (*proposed_iter2)->get_num_subjects_allocated() ) - 2 ) );
     else if( move_type == SPLIT )
     {
     LogProposalRatio += log( 2 * ( pow( 2, (*proposed_iter)->get_num_subjects_allocated() + holding_cluster->get_num_subjects_allocated() ) - 2 ) );
     if( jump_distn )
     calculate_jump_prob();
     }
     else if( move_type == SUBJECT_SHIFT || move_type == SINGLE_SUBJECT_SHIFT )
     LogProposalRatio += log( 2 * ( pow( 2, (*proposed_iter)->get_num_subjects_allocated() + holding_cluster->get_num_subjects_allocated() ) - 2 ) ) - log( 2 * ( pow( 2, (*proposed_iter2)->get_num_subjects_allocated() + holding_cluster->get_num_subjects_allocated() ) - 2 ) );
   */

   if( !preserve_tree )
   {
      if( move_type == MERGE || move_type == SINGLETON_MERGE )
      {
         LogProposalRatio += log( (double)number_mergeable ) - LOG_TWO - log( (double)number_splitable );
         if( !jump_distn )
            LogProposalRatio += log( (double)number_mergeable - 1 );
         if( !agglomerative_based_split && !gibbs_two_mean_separation )
            LogProposalRatio -= stirling_mx[ proposed_cluster->get_num_subjects_allocated() + proposed_cluster2->get_num_subjects_allocated() ][ 2 ];
         else if( gibbs_two_mean_separation )
            LogProposalRatio += Cluster_GLM_MVN::get_gibbs_sweep_log_prob();
      }
      else if( move_type == SPLIT || move_type == SINGLETON_SPLIT )
      {
         LogProposalRatio -= log( (double)number_mergeable ) - LOG_TWO - log( (double)number_splitable );
         if( !jump_distn )
            LogProposalRatio -= log( (double)number_mergeable - 1 );
         if( !agglomerative_based_split && !gibbs_two_mean_separation )
            LogProposalRatio += stirling_mx[ proposed_cluster->get_num_subjects_allocated() + holding_cluster->get_num_subjects_allocated() ][ 2 ];
         else if( gibbs_two_mean_separation )
            LogProposalRatio -= Cluster_GLM_MVN::get_gibbs_sweep_log_prob();
         if( jump_distn )
            calculate_jump_prob();
         zap( proposed_subject_list );
      }
      else if( move_type == SUBJECT_SHIFT )
         LogProposalRatio += stirling_mx[ proposed_cluster->get_num_subjects_allocated() + holding_cluster->get_num_subjects_allocated() ][ 2 ]
            - stirling_mx[ proposed_cluster2->get_num_subjects_allocated() + holding_cluster->get_num_subjects_allocated() ][ 2 ];
      else if( move_type == SINGLE_SUBJECT_SHIFT )
         LogProposalRatio += log( (double)proposed_cluster->get_num_subjects_allocated() + holding_cluster->get_num_subjects_allocated() )
            - log( (double)proposed_cluster2->get_num_subjects_allocated() + (double)holding_cluster->get_num_subjects_allocated() );
   }
   else
   {
      if( move_type == MERGE || move_type == SINGLETON_MERGE )
      {
         if( conditional_tree_distribution == 1 )
         {
            int n1 = proposed_cluster->get_num_subjects_allocated();
            int n2 = proposed_cluster2->get_num_subjects_allocated();
            tree_prior_change = get_uniform_tree_log_prior_change( n1, n2, number_of_clusters );
         }
      }
      else if( move_type == SPLIT || move_type == SINGLETON_SPLIT )
      {
         if( conditional_tree_distribution == 1 )
         {
            int n1 = proposed_cluster->get_num_subjects_allocated();
            int n2 = holding_cluster->get_num_subjects_allocated();
            tree_prior_change = - get_uniform_tree_log_prior_change( n1, n2, number_of_clusters + 1 );
         }
      }
      else if( move_type == SUBJECT_SHIFT || move_type == SINGLE_SUBJECT_SHIFT )
      {
         if( metropolis_tree_changes && cascade_move_clusters )
            LogProposalRatio += log_path_probability - log_proposed_path_probability;
//         if( !conditional_tree_distribution )
//            tree_prior_change = get_tree_branch_log_prior( holding_cluster, proposed_cluster2 ) - get_tree_branch_log_prior( holding_cluster, proposed_cluster );
         if( !conditional_tree_distribution )
         {//temporarily bolt on holding cluster to proposed_cluster2 to count configurations.
            bool has_parent = true;
            Cluster_GLM_MVN *cl_new = new Cluster_GLM_MVN( holding_cluster, proposed_cluster2, true );
            Cluster_GLM_MVN *child1, *child2, *the_parent = proposed_cluster2->get_parent_cluster();
            if( the_parent )
            {
               the_parent->get_clusters_allocated( child1, child2 );
               if( child1 == proposed_cluster2 )
                  the_parent->set_clusters_allocated( proposed_cluster2, child2 );
               else
                  the_parent->set_clusters_allocated( child1, proposed_cluster2 );
            }
            else
            {
               has_parent = false;
               the_parent = cl_new;
            }
            proposed_num_configurations = get_number_of_possible_clusterings_for_tree_containging_cluster( cl_new );
            tree_prior_change -= log( proposed_num_configurations );
            if( has_parent )
               the_parent->set_clusters_allocated( child1, child2 );
            delete cl_new;
         }
      }
      if( !stop_cascading_at_model || !( move_type == SUBJECT_SHIFT || move_type == SINGLE_SUBJECT_SHIFT )  )
      {
         if( conditional_tree_distribution == 2 )
         {
            tree_prior_change += log( 1 + tree_prior_parameter / ( proposed_number_of_clusters + 1 ) * proposed_tree_score ) - log( 1 + tree_prior_parameter / ( number_of_clusters + 1 ) * tree_score );
            tree_prior_change += tree_log_normalising_constant - proposed_tree_log_normalising_constant;
         }
         else if( conditional_tree_distribution == 3 )
         {
            find_proposed_tree_score();
            tree_prior_change += ( proposed_tree_score - tree_score ) * log( 1 + tree_prior_parameter );
            tree_prior_change += tree_log_normalising_constant - proposed_tree_log_normalising_constant;
         }
         else if( conditional_tree_distribution == 4 )
         {
            find_proposed_tree_score();
            tree_prior_change += proposed_tree_score - tree_score;
            tree_prior_change += tree_log_normalising_constant - proposed_tree_log_normalising_constant;
         }
         LogProposalRatio += tree_prior_change;
      }
   }
   
   proposed_log_posterior = proposed_log_likelihood;

//      cout << Cluster_GLM_MVN::get_cluster_count() <<" Clusters, " << original_move_type << "  " << move_type << "  " << LogProposalRatio << "  " << proposed_log_likelihood << "  " <<  log_likelihood << endl;

}

void GLM_MVN_MCMC_Clustering::calculate_number_mergeable_splitable()
{
   number_mergeable = mergeable_cluster_deque.size();
   number_splitable = number_mergeable;
   for( cluster_ptrs::iterator iter = mergeable_cluster_deque.begin(); iter != mergeable_cluster_deque.end(); iter++ )
      if( !(*iter)->get_splitable() )
         number_splitable--;
   if( move_type == MERGE || move_type == SINGLETON_MERGE )
   {
      if( proposed_cluster->get_splitable() && proposed_cluster2->get_splitable() )
         number_splitable--;
   }
   else if( move_type == SPLIT || move_type == SINGLETON_SPLIT )
   {
      number_mergeable++;
   }
}

void GLM_MVN_MCMC_Clustering::find_mate()
{
   proposed_iter2 = proposed_iter;
//   proposed_iter2++ = proposed_iter;
   Cluster_GLM_MVN* cluster_to_mate = *proposed_iter;
   ++proposed_iter2;
   if( preserve_tree && ( !(move_type == SUBJECT_SHIFT || move_type == SINGLE_SUBJECT_SHIFT ) || cascade_move_clusters ) )
   {
      Cluster_GLM_MVN* cl_sibling = cluster_to_mate->get_sibling_cluster();
      if( cl_sibling )
      {
         //find it in mergeable_cluster_deque
         cluster_ptrs::iterator iter = mergeable_cluster_deque.begin();
         while( iter != mergeable_cluster_deque.end() && *iter != cl_sibling )
            ++iter;
         proposed_iter2 = iter;
      }
      else
         proposed_iter2 = mergeable_cluster_deque.end();
      return;
   }
   if( !jump_distn || move_type == SUBJECT_SHIFT )//|| move_type == SINGLE_SUBJECT_SHIFT )
   {
      while( proposed_iter2 != mergeable_cluster_deque.end() && !deque_iterators_mergeable() )// keep going (if the 2nd cluster is not mergeable or) if both clusters are 'distinct' or if they have labels which are different. (For move type 4, also keep going if at least one of the clusters is splitable, so has >1 subjects) !(*proposed_iter2)->get_mergeable() || <-MUST BE MERGEABLE TO BE IN THE LIST!!
         ++proposed_iter2;
   }
   else // draw merging cluster with probability proportional to cluster closeness.
   {
      double normalising_constant = 0;
      double val, val2 = 0, c = 1, highest_val = -DBL_MAX;//last value used to stop all the probs being rounded to zero.
      Cluster_GLM_MVN *temp_cluster = NULL, *temp_cluster2 = NULL;
      if( move_type == SINGLE_SUBJECT_SHIFT )
         posterior_cancels_with_proposal = false;//
      if( move_type == SINGLE_SUBJECT_SHIFT )
      {
          if( simulated_annealing && posterior_cancels_with_proposal )
             c = cooling_schedule();
          temp_cluster = create_temporary_cluster_for_subject( proposed_subject );
          temp_cluster2 = create_temporary_cluster_without_subject( proposed_subject, cluster_to_mate );
          highest_val = val2 = temp_cluster->get_cluster_closeness( temp_cluster2 ) / c;
          zap( temp_cluster2 );
          cluster_to_mate = temp_cluster;
      }

//            print_move_type();
      while( proposed_iter2 != mergeable_cluster_deque.end() )
      {
         if( deque_iterators_mergeable() )
         {
            val = cluster_to_mate->get_cluster_closeness( *proposed_iter2 ) / c;
//            if( move_type != MERGE ) cout << get_iteration() << " :  " << proposed_subject << "  " << (void*)*proposed_iter2 << "  " << val << "  " << val2 << endl;
            if( val > highest_val )
            {
                normalising_constant *= exp( highest_val - val );
                highest_val = val;
            }
            normalising_constant += exp( val - highest_val );
         }
         ++proposed_iter2;
      }
      if( move_type == SINGLE_SUBJECT_SHIFT && posterior_cancels_with_proposal )
      {
//         cout << normalising_constant << " " << normalising_constant + exp(val2-highest_val ) << "  " << normalising_constant/ (normalising_constant + exp(val2-highest_val )) << endl;
         double u = gsl_ran_flat( rn, 0, 1 ) * (normalising_constant + exp(val2-highest_val));
         if( u >= normalising_constant )
         {
            force_reject = true;
            proposed_iter2 = proposed_iter;
            proposed_iter2++;
         }
      }
      if( !force_reject && normalising_constant )
      {
          proposed_iter2 = proposed_iter;
          double  u = gsl_ran_flat( rn, 0, 1 ) * normalising_constant;
          double total = 0, log_contribution = 0;
          while( total < u && proposed_iter2 != mergeable_cluster_deque.end() )
          {
              ++proposed_iter2;
              if( proposed_iter2 != mergeable_cluster_deque.end() && deque_iterators_mergeable() )
              {
                  log_contribution = cluster_to_mate->get_cluster_closeness( *proposed_iter2 ) / c - highest_val;
                  total += exp( log_contribution );
              }
          }
          if( !posterior_cancels_with_proposal )
          {
             LogProposalRatio -= log_contribution - log( normalising_constant );// + log( (double)number_of_clusters );
             if( move_type == SINGLE_SUBJECT_SHIFT )
             {
                normalising_constant -= exp( log_contribution );
                log_contribution = val2 - highest_val;
                normalising_constant += exp( log_contribution );
                LogProposalRatio += log_contribution - log( normalising_constant );
             }
          }
      }
      zap( temp_cluster );
   }
}

void GLM_MVN_MCMC_Clustering::find_cluster_to_alter()
{
   if( move_type == V_SHIFT || move_type == EXPERIMENT_SHIFT || move_type == EXPERIMENT_SIGN_CHANGE || move_type == EXPERIMENT_MULTIPLIER_SHIFT || move_type == EXPERIMENT_POWER_SHIFT  || move_type == EXPERIMENT_PARAMETER_SHIFT )
      return;
   random_shuffle( mergeable_cluster_deque.begin(), mergeable_cluster_deque.end() );
   proposed_iter = mergeable_cluster_deque.begin();
   proposed_iter2 = proposed_iter;
   if( move_type == TREE_FULL_CONDITIONAL )
      return;
   if( preserve_tree && ( move_type == MERGE || move_type == SINGLETON_MERGE || move_type == SPLIT || move_type == SINGLETON_SPLIT ) )
   {
      //first of all, work out total of probabilities
      deque<double> merge_prob_deque, split_prob_deque;
      deque<double>::iterator iter_merge, iter_split, *prob_deque_iterator_pointer;
      double merge_prob = 0, split_prob = 0, merge_total = 0, split_total = 0;
      double max_merge_log_prob = -DBL_MAX, max_split_log_prob = -DBL_MAX;
      Cluster_GLM_MVN* cl_sibling;
      while( proposed_iter != mergeable_cluster_deque.end() )
      {
         cl_sibling = (*proposed_iter)->get_sibling_cluster();
         if( cl_sibling && is_cluster_in_model( cl_sibling ) )
            if( jump_distn )
            {
               double merge_log_prob = (*proposed_iter)->get_sibling_merge_prob( cl_sibling );
               if( jump_distn != -1 )
                  merge_prob = exp( merge_log_prob );
               else if( merge_log_prob > max_merge_log_prob )
               {
                  max_merge_log_prob = merge_log_prob;
                  if( move_type == MERGE || move_type == SINGLETON_MERGE )
                     proposed_iter2 = proposed_iter;
               }
            }
            else
               merge_prob = 1;
         else
            merge_prob = 0;
         merge_total += merge_prob;
         merge_prob_deque.push_back( merge_total );
         if( (*proposed_iter)->get_splitable() )
            if( jump_distn )
            {
               double split_log_prob = -(*proposed_iter)->get_childrens_merge_prob();
               if( jump_distn != -1 )
                  split_prob = exp( split_log_prob );
               else if( split_log_prob > max_split_log_prob )
               {
                  max_split_log_prob = split_log_prob;
                  if( move_type == SPLIT || move_type == SINGLETON_SPLIT )
                     proposed_iter2 = proposed_iter;
               }
            }
            else
               split_prob = 1;
         else
            split_prob = 0;
         split_total += split_prob;
         split_prob_deque.push_back( split_total );
         ++proposed_iter;
      }
      
      if( jump_distn == -1 )
      {
         proposed_iter = proposed_iter2;
         proposed_cluster2 = (*proposed_iter)->get_sibling_cluster();
         return;
      }
   
      double u, total_prob, previous_prob = 0, previous_merge_prob = 0, previous_split_prob = 0;
      if( move_type == MERGE || move_type == SINGLETON_MERGE )
      {
         total_prob = merge_total;
         prob_deque_iterator_pointer = &iter_merge;
      }
      else
      {
         total_prob = split_total;
         prob_deque_iterator_pointer = &iter_split;
      }
      iter_merge = merge_prob_deque.begin();
      iter_split = split_prob_deque.begin();
      if( total_prob <= 0 )
         return;
      proposed_iter = mergeable_cluster_deque.begin();
      u = gsl_ran_flat( rn, 0, total_prob );
      while( u > **prob_deque_iterator_pointer && proposed_iter != mergeable_cluster_deque.end() )
      {
         previous_prob = **prob_deque_iterator_pointer;
         previous_merge_prob = *iter_merge;
         previous_split_prob = *iter_split;
         ++iter_merge;
         ++iter_split;
         ++proposed_iter;
      }
      if( move_type == MERGE || move_type == SINGLETON_MERGE )
      {
         LogProposalRatio -= log( 2 * (**prob_deque_iterator_pointer - previous_prob) / total_prob );
         split_total -= (*iter_split - previous_split_prob);
         //------------ find cl_siblings split probability and subtract that off too
         cl_sibling = (*proposed_iter)->get_sibling_cluster();
         double siblings_previous_split_prob = 0;
         deque<double>::iterator iter_split2 = split_prob_deque.begin();
         while( proposed_iter2 != mergeable_cluster_deque.end() && (*proposed_iter2) != cl_sibling )
         {
            siblings_previous_split_prob = *iter_split2;
            iter_split2++;
            ++proposed_iter2;
         }
         split_total -= (*iter_split2 - siblings_previous_split_prob);
         //-------------
         split_total += 1 / (*iter_merge - previous_merge_prob);
         LogProposalRatio += log( 1 / (*iter_merge - previous_merge_prob) / split_total );
      }
      else
      {
         LogProposalRatio -= log( (**prob_deque_iterator_pointer - previous_prob) / total_prob );
         merge_total -= 2 * (*iter_merge - previous_merge_prob);//2*,since both the split cluster and its sibling are not available for for merger now, and both had this prob.
         merge_total += 2 / (*iter_split - previous_split_prob);
         LogProposalRatio += log( 2 / (*iter_split - previous_split_prob) / merge_total );
      }
   }
}

void GLM_MVN_MCMC_Clustering::find_proposed_tree_score()
{
   if( move_type == MERGE || move_type == SINGLETON_MERGE )
   {
      proposed_number_of_clusters = number_of_clusters - 1;
      proposed_cluster_score = proposed_cluster->get_score();
      proposed_cluster2_score = proposed_cluster2->get_score();
      int proposed_parent_score;
      Cluster_GLM_MVN* proposed_parent_cluster = proposed_cluster->get_parent_cluster();
      if( !proposed_cluster_score || !proposed_cluster2_score )
         proposed_parent_score = 0;
      else
         proposed_parent_score = rebuild_agglomerative_tree_from_cluster( proposed_parent_cluster );
      proposed_parent_cluster->set_score( proposed_parent_score );
      if( conditional_tree_distribution != 4 )
	 proposed_tree_score += proposed_parent_score - proposed_cluster_score - proposed_cluster2_score;
      if( tree_above_model_score )
         proposed_tree_above_model_score = tree_above_model_score;
      else
         proposed_tree_above_model_score = is_tree_above_model_optimal( proposed_cluster, proposed_cluster2 );
      if( conditional_tree_distribution != 4 )
	 proposed_tree_score += proposed_tree_above_model_score - tree_above_model_score;
      if( conditional_tree_distribution == 4 )
      {
         if( functional_tree_prior_parameter )
            proposed_tree_score = add_up_tree_score();
	 if( proposed_cluster_score )
	    proposed_tree_score -= get_log_one_plus_alpha_catalan_number( proposed_cluster );
	 if( proposed_cluster2_score )
	    proposed_tree_score -= get_log_one_plus_alpha_catalan_number( proposed_cluster2 );
	 if( proposed_parent_score )
	    proposed_tree_score += get_log_one_plus_alpha_catalan_number( proposed_parent_cluster );
	 if( tree_above_model_score )
	    proposed_tree_score -= get_log_one_plus_alpha_catalan_number( number_of_clusters );
	 if( proposed_tree_above_model_score )
	    proposed_tree_score += get_log_one_plus_alpha_catalan_number( proposed_number_of_clusters );
      }
      calculate_tree_score_normalising_constant( proposed_cluster, proposed_cluster2 );
   }
   else if( move_type == SPLIT || move_type == SINGLETON_SPLIT )
   {
      proposed_number_of_clusters = number_of_clusters + 1;
      Cluster_GLM_MVN* proposed_parent_cluster = proposed_cluster->get_parent_cluster();
      proposed_cluster_score = proposed_parent_cluster->get_score();
      int proposed_cluster_child1_score, proposed_cluster_child2_score;
      Cluster_GLM_MVN *proposed_cluster_child1, *proposed_cluster_child2;
      proposed_parent_cluster->get_clusters_allocated( proposed_cluster_child1, proposed_cluster_child2 );
      if( proposed_cluster_score  )
         proposed_cluster_child1_score = proposed_cluster_child2_score = proposed_cluster_score;
      else
      {
         proposed_cluster_child1_score = rebuild_agglomerative_tree_from_cluster( proposed_cluster_child1 );
         proposed_cluster_child2_score = rebuild_agglomerative_tree_from_cluster( proposed_cluster_child2 );
      }
      proposed_cluster_child1->set_score( proposed_cluster_child1_score );
      proposed_cluster_child2->set_score( proposed_cluster_child2_score );
      if( conditional_tree_distribution != 4 )
	 proposed_tree_score += proposed_cluster_child1_score + proposed_cluster_child2_score - proposed_cluster_score;
      if( !tree_above_model_score )
         proposed_tree_above_model_score = 0;
      else
         proposed_tree_above_model_score = is_tree_above_model_optimal( proposed_parent_cluster );
      if( conditional_tree_distribution != 4 )
	 proposed_tree_score += proposed_tree_above_model_score - tree_above_model_score;
      if( conditional_tree_distribution == 4 )
      {
         if( functional_tree_prior_parameter )
            proposed_tree_score = add_up_tree_score();
	 if( proposed_cluster_score )
	    proposed_tree_score -= get_log_one_plus_alpha_catalan_number( proposed_parent_cluster );
	 if( proposed_cluster_child1_score )
	    proposed_tree_score += get_log_one_plus_alpha_catalan_number( proposed_cluster_child1 );
	 if( proposed_cluster_child2_score )
	    proposed_tree_score += get_log_one_plus_alpha_catalan_number( proposed_cluster_child2 );
	 if( tree_above_model_score )
	    proposed_tree_score -= get_log_one_plus_alpha_catalan_number( number_of_clusters );
	 if( proposed_tree_above_model_score )
	    proposed_tree_score += get_log_one_plus_alpha_catalan_number( proposed_number_of_clusters );
      }
      calculate_tree_score_normalising_constant( proposed_parent_cluster );
   }
   else if( ( move_type == SUBJECT_SHIFT || move_type == SINGLE_SUBJECT_SHIFT ) && !stop_cascading_at_model && cascade_move_clusters )
   {
      //artificially create tree with subject shift effected
      proposed_number_of_clusters = number_of_clusters;
      Cluster_GLM_MVN* cl_parent2 = proposed_cluster2->get_parent_cluster();
      Cluster_GLM_MVN* cl_root2 = proposed_cluster2->get_my_root_cluster();
      Cluster_GLM_MVN* cl_temp = proposed_cluster2->merge_cluster( holding_cluster );
      Cluster_GLM_MVN *old_node_in_model_1 = NULL , *old_node_in_model_2 = NULL;
      if( node_in_model_1 == proposed_cluster2 && !cluster_in_model )
      {
         old_node_in_model_1 = node_in_model_1;
         node_in_model_1 = cl_temp;
      }
      if( node_in_model_2 == proposed_cluster2 && !cluster_in_model )
      {
         old_node_in_model_2 = node_in_model_2;
         node_in_model_2 = cl_temp;
      }
      if( node_in_model_1 )
      {
         node_in_model_1_proposed_score = rebuild_agglomerative_tree_from_cluster( node_in_model_1 );
	 if( node_in_model_1_proposed_score )
	 {
	    if( conditional_tree_distribution != 4 )
	       proposed_tree_score++;
	    else
	       proposed_tree_score += get_log_one_plus_alpha_catalan_number( node_in_model_1 );
	 }
      }
      if( node_in_model_2 && node_in_model_2 != node_in_model_1 )
      {
         node_in_model_2_proposed_score = rebuild_agglomerative_tree_from_cluster( node_in_model_2 );
	 if( node_in_model_2_proposed_score )
	 {
	    if( conditional_tree_distribution != 4 )
	       proposed_tree_score++;
	    else
	       proposed_tree_score += get_log_one_plus_alpha_catalan_number( node_in_model_2 );
	 }
      }
      proposed_tree_above_model_score = is_tree_above_model_optimal( old_node_in_model_1, old_node_in_model_2 );
      if( conditional_tree_distribution != 4 )
	 proposed_tree_score += proposed_tree_above_model_score - tree_above_model_score;
      else if( conditional_tree_distribution == 4 )
      {
	 if( tree_above_model_score )
	    proposed_tree_score -= get_log_one_plus_alpha_catalan_number( number_of_clusters );
	 if( proposed_tree_above_model_score )
	    proposed_tree_score += get_log_one_plus_alpha_catalan_number( number_of_clusters );
      }
      calculate_tree_score_normalising_constant( old_node_in_model_1, old_node_in_model_2 );
      if( node_in_model_2 == cl_temp )
         node_in_model_2 = proposed_cluster2;
      if( node_in_model_1 == cl_temp )
         node_in_model_1 = node_in_model_2;
      holding_cluster->set_parent_cluster( NULL );
      proposed_cluster2->set_parent_cluster( cl_parent2 );
      proposed_cluster2->set_my_root_cluster( cl_root2 );
      delete cl_temp;
   }
   

}

double GLM_MVN_MCMC_Clustering::calculate_tree_score()
{
   double temp_tree_score = 0;
   if( is_tree_above_model_optimal() )
   {
      if( conditional_tree_distribution != 4 )
	 temp_tree_score = 1;
      else
	 temp_tree_score = get_log_one_plus_alpha_catalan_number( number_of_clusters );
   }
   for( cluster_ptrtodbl::iterator iter = cluster_map.begin(); iter != cluster_map.end(); iter++ )
   {
      Cluster_GLM_MVN* cl = iter->first;
      if( cl->get_mergeable() && rebuild_agglomerative_tree_from_cluster( cl ) )
      {
	 if( conditional_tree_distribution != 4 )
	    temp_tree_score ++;
	 else
	    temp_tree_score += get_log_one_plus_alpha_catalan_number( cl );
      }
   }
   return temp_tree_score;
}

double GLM_MVN_MCMC_Clustering::add_up_tree_score()
{
   double temp_tree_score = 0;
   if( proposed_tree_above_model_score )
   {
      if( conditional_tree_distribution != 4 )
	 temp_tree_score = 1;
      else
	 temp_tree_score = get_log_one_plus_alpha_catalan_number( number_of_clusters );
   }
   for( cluster_ptrtodbl::iterator iter = cluster_map.begin(); iter != cluster_map.end(); iter++ )
   {
      Cluster_GLM_MVN* cl = iter->first;
      if( cl->get_mergeable() && cl->get_score() )
      {
	 if( conditional_tree_distribution != 4 )
	    temp_tree_score ++;
	 else
	    temp_tree_score += get_log_one_plus_alpha_catalan_number( cl );
      }
   }
   return temp_tree_score;
}

int GLM_MVN_MCMC_Clustering::is_tree_above_model_optimal( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 )
{
   set<Cluster_GLM_MVN*> cluster_set;
   for( cluster_ptrtodbl::iterator iter = cluster_map.begin(); iter != cluster_map.end(); iter++ )
      if( iter->first->get_mergeable() && iter->first != cl1 && iter->first != cl2 )
         cluster_set.insert( iter->first );

   if( move_type != SUBJECT_SHIFT && move_type != SINGLE_SUBJECT_SHIFT )
   {
      if( cl1 )
      {
         if( cl2 )
            cluster_set.insert( cl1->get_parent_cluster() );
         else
         {
            Cluster_GLM_MVN *cl11 = NULL, *cl12 = NULL;
            cl1->get_clusters_allocated( cl11, cl12 );
            cluster_set.insert( cl11 );
            cluster_set.insert( cl12 );
         }
      }
   }
   else
   {
      if( cl1 )
         cluster_set.insert( node_in_model_1 );
      if( cl2 )
         cluster_set.insert( node_in_model_2 );
   }
   int num_clusters_in_set = cluster_set.size();
   Cluster_GLM_MVN *cl, *cl_sibling, *cl_max;
   while( num_clusters_in_set > minimum_num_clusters + 1 && find_best_sibling_pair_in_set( cluster_set, cl, cl_sibling ) )
   {
      double d, d_max = -DBL_MAX;
      cl_max = NULL;
      for( set<Cluster_GLM_MVN*>::iterator iter = cluster_set.begin(); iter != cluster_set.end(); iter++ )
         if( (*iter)->get_mergeable() && *iter != cl )
         {
            if( storing_cluster_closenesses )
               d = cl->get_cluster_closeness( *iter );
            else
            {
               d = calculate_closeness( cl, *iter );
               if( !euclidean_or_correlation_clustering )
                  d += get_change_log_prior_cluster_size( cl, *iter, false );
            }
            if( d > d_max )
            {
               d_max = d;
               cl_max = *iter;
            }
         }
      if( cl_max != cl_sibling )
         return 0;
      else
      {
         cluster_set.erase( cl );
         cluster_set.erase( cl_sibling );
         cluster_set.insert( cl->get_parent_cluster() );
         num_clusters_in_set--;
      }
   }
   return 1;
}

void GLM_MVN_MCMC_Clustering::calculate_tree_score_normalising_constant( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 )
{
   double* catalans_all = new double[ proposed_number_of_clusters + 1 ];//array to hold the cluster sizes
   catalans_all[ 0 ] = get_log_catalan_number( proposed_number_of_clusters );
   int counter = 1;
   for( cluster_ptrtodbl::iterator iter = cluster_map.begin(); iter != cluster_map.end(); iter++ )
      if( iter->first->get_mergeable() && iter->first != cl1 && iter->first != cl2 )
         catalans_all[ counter++ ] = get_log_catalan_number( iter->first->get_num_subjects_allocated() );

   if( move_type != SUBJECT_SHIFT && move_type != SINGLE_SUBJECT_SHIFT )
   {
      if( cl1 )
      {
         if( cl2 )
            catalans_all[ counter ] = get_log_catalan_number( cl1->get_num_subjects_allocated() + cl2->get_num_subjects_allocated() );
         else
         {
            Cluster_GLM_MVN *cl11 = NULL, *cl12 = NULL;
            cl1->get_clusters_allocated( cl11, cl12 );
            catalans_all[ counter++ ] = get_log_catalan_number( cl11->get_num_subjects_allocated() );
            catalans_all[ counter ] = get_log_catalan_number( cl12->get_num_subjects_allocated() );
         }
      }
   }
   else
   {
      if( cl1 )
         catalans_all[ counter++ ] = get_log_catalan_number( node_in_model_1->get_num_subjects_allocated() );
      if( cl2 && node_in_model_2  != node_in_model_1 )
         catalans_all[ counter ] = get_log_catalan_number( node_in_model_2->get_num_subjects_allocated() );
   }

   proposed_tree_log_normalising_constant = 0;
   for( int i = 0; i < proposed_number_of_clusters + 1; i++ )
      proposed_tree_log_normalising_constant += catalans_all[ i ];
   
   if( conditional_tree_distribution == 3 || conditional_tree_distribution == 4 )
   {
      delete [] catalans_all;
      return;
   }

   if( proposed_tree_log_normalising_constant < 700 )
   {
      int min_score = 0;
      int i;
      for( i = 0; i < proposed_number_of_clusters + 1; i++ )
         if( catalans_all[ i ] <= 0 )
            min_score++;

      int size_catalans = proposed_number_of_clusters + 1 - min_score;
      double* catalans = new double[ size_catalans ];
      counter = 0;
      for( i = 0; i < proposed_number_of_clusters + 1; i++ )
         if( catalans_all[ i ] > 0 )
            catalans[ counter++ ] = exp( catalans_all[ i ] );

      double unit_score = tree_prior_parameter / proposed_number_of_clusters, scoring_mass = 0;
      for( int k = 1; k <= proposed_number_of_clusters + 1 - min_score; k++ )
         scoring_mass += ( k + min_score ) * unit_score * get_r_k( k, catalans, size_catalans );
      delete [] catalans;
      proposed_tree_log_normalising_constant = log( exp( proposed_tree_log_normalising_constant ) + scoring_mass );
   }
   delete [] catalans_all;   
}

double GLM_MVN_MCMC_Clustering::get_r_k( int k, double* catalans, int size_catalans )
{
   if( k == size_catalans )
      return 1;
   int* chosen_indices = new int[ size_catalans - k ];
   double rk =  get_r_k( k, catalans, size_catalans, chosen_indices );
   delete [] chosen_indices;
   return rk;
}

double GLM_MVN_MCMC_Clustering::get_r_k( int k, double* catalans, int size_catalans, int* chosen_indices, int index, int start_index )
{
   double r_k = 0;
   for( chosen_indices[ index ] = start_index; chosen_indices[ index ] <= k + index; chosen_indices[ index ]++ )
      if( index < size_catalans - 1 - k )
         r_k += get_r_k( k, catalans, size_catalans, chosen_indices, index + 1, chosen_indices[ index ] + 1 );
      else
      {
         double product = 1;
         for( int i = 0; i < size_catalans - k; i++ )
            product *= catalans[ chosen_indices[ i ] ] - 1;
         r_k += product;
      }
   return r_k;
}

void GLM_MVN_MCMC_Clustering::fit_optimal_tree_around_model()
{
   if( !proposed_tree_above_model_score )
   {
      Cluster_GLM_MVN* cl = proposed_cluster->get_my_root_cluster();
      rebuild_agglomerative_tree_from_cluster( cl, true );
      proposed_tree_above_model_score = 1;
   }
   cluster_ptrtodbl::iterator iter = cluster_map.begin();
   while( iter != cluster_map.end() )
      if( iter->first->get_mergeable() )
      {
         Cluster_GLM_MVN*cl = iter->first;
         if( cl->get_mergeable() && !cl->get_score() )
         {
            rebuild_agglomerative_tree_from_cluster( cl, true );
            cl->set_score( 1 );
            iter = cluster_map.begin();
         }
         else
            ++iter;
      }
      else
         ++iter;
   proposed_tree_score = add_up_tree_score();
}

void GLM_MVN_MCMC_Clustering::sample_tree_from_full_conditionals()
{
   if( number_of_clusters > 2 )
      sample_subtree_above_model_nodes_from_full_conditional( proposed_cluster->get_my_root_cluster() );
   list<Cluster_GLM_MVN*>model_clusters;
   for( cluster_ptrtodbl::iterator iter = cluster_map.begin(); iter != cluster_map.end(); ++iter )
      model_clusters.push_back( iter->first );
   for( list<Cluster_GLM_MVN*>::iterator iter2 = model_clusters.begin(); iter2 != model_clusters.end(); ++iter2 )
      sample_model_node_subtree_from_full_conditional( *iter2 );
}


void GLM_MVN_MCMC_Clustering::sample_subtree_from_full_conditional()
{
   int tree_number;
   if( number_of_clusters <= 2 )
      tree_number = (int)gsl_rng_uniform_int( rn, number_of_clusters ) + 1;
   else
      tree_number = (int)gsl_rng_uniform_int( rn, number_of_clusters + 1 );
   sample_subtree_from_full_conditional( tree_number );
}

void GLM_MVN_MCMC_Clustering::sample_subtree_from_full_conditional( int tree_number )
{
   if( tree_number == 0 )
   {
      Cluster_GLM_MVN* cl = proposed_cluster->get_my_root_cluster();
      sample_subtree_above_model_nodes_from_full_conditional( cl );
   }
   else
   {
      cluster_ptrtodbl::iterator iter = cluster_map.begin();
      int counter = 0;
      while( iter != cluster_map.end() && counter != tree_number )
      {
         if( iter->first->get_mergeable() )
            counter++;
         if( counter != tree_number )
            ++iter;
      }
      Cluster_GLM_MVN* cl = iter->first;
      sample_model_node_subtree_from_full_conditional( cl );
   }
}

void GLM_MVN_MCMC_Clustering::sample_subtree_above_model_nodes_from_full_conditional( Cluster_GLM_MVN* cl )
{
   int cl_count = get_number_of_mergeable_clusters_in_my_model( cl );
   double log_prob = get_optimal_subtree_log_prob( cl_count );
   double u = 0;
   if( log_prob < 0 )
      u = log( gsl_ran_flat( rn, 0, 1 ) );
   if( u <= log_prob )
   {
      if( !proposed_tree_above_model_score )
      {
         rebuild_agglomerative_tree_from_cluster( cl, true );
         tree_above_model_score = proposed_tree_above_model_score = 1;
         tree_score = proposed_tree_score = add_up_tree_score();
      }
   }
   else
   {
      do
      {
         rebuild_agglomerative_tree_from_cluster( cl, true, false );
      } while( is_tree_above_model_optimal() );
      tree_above_model_score = proposed_tree_above_model_score = 0;
      tree_score = proposed_tree_score = add_up_tree_score();
   }
}

bool GLM_MVN_MCMC_Clustering::find_best_sibling_pair_in_set( set<Cluster_GLM_MVN*> cluster_set, Cluster_GLM_MVN*& cl, Cluster_GLM_MVN*& cl_sibling )
{
   bool found_one = false;
   Cluster_GLM_MVN *cl1, *cl2;
   double d, d_max = -DBL_MAX;
   for( set<Cluster_GLM_MVN*>::iterator iter = cluster_set.begin(); iter != cluster_set.end(); iter++ )
   {
      cl1 = *iter;
      cl2 = cl1->get_sibling_cluster();
      if( cl2 && cluster_set.find( cl2 ) != cluster_set.end() )
      {
         found_one = true;
         if( storing_cluster_closenesses )
            d = cl1->get_cluster_closeness( cl2 );
         else
         {
            d = calculate_closeness( cl1, cl2 );
            if( !euclidean_or_correlation_clustering )
               d += get_change_log_prior_cluster_size( cl1, cl2, false );
         }
         if( d > d_max )
         {
            d_max = d;
            cl = cl1;
            cl_sibling = cl2;
         }
      }
   }
   return found_one;
}

void GLM_MVN_MCMC_Clustering::calculate_jump_prob()
{
   proposed_iter2 = mergeable_cluster_deque.begin();
   double normalising_constant = 0, prob;
   double val, highest_val = 0;
   while( proposed_iter2 != mergeable_cluster_deque.end() )
   {
      if( deque_iterators_mergeable() )
      {
         val = calculate_closeness( *proposed_iter2, holding_cluster );
         val += get_change_log_prior_cluster_size( *proposed_iter2, holding_cluster, false );
         if( !normalising_constant )
         {
            highest_val = val;
            normalising_constant = 1;
         }
         else
         {
            if( val > highest_val )
            {
               normalising_constant *= exp( highest_val - val );
               highest_val = val;
            }
            normalising_constant += exp( val - highest_val );
         }
      }
      ++proposed_iter2;
   }
   if( normalising_constant )
   {
      prob = exp( proposed_cluster->get_cluster_closeness( holding_cluster ) - highest_val ) / normalising_constant;
      LogProposalRatio += log( prob );// + log( (double)number_of_clusters );
   }
}
   

bool GLM_MVN_MCMC_Clustering::deque_iterators_mergeable()
{
   if( iter_distinct && (*proposed_iter2)->get_distinct() )
      return 0;
   if( move_type == SINGLETON_MERGE && iter_splitable && (*proposed_iter2)->get_splitable() )
      return 0;
   if( !cluster_by_labels )
      return 1;
   string* s1_ptr = (*proposed_iter)->get_label();
   if( !s1_ptr )
      return 1;
   string s1 = *s1_ptr;
   string* s2_ptr = (*proposed_iter2)->get_label();
   if( !s2_ptr )
      return 1;
   if( s1 == *s2_ptr )
      return 1;
   return 0;
}

bool GLM_MVN_MCMC_Clustering::deque_iterators_mergeable( cluster_ptrs::iterator iter, cluster_ptrs::iterator iter2 )
{
   return clusters_mergeable( *iter, *iter2 );
}

void GLM_MVN_MCMC_Clustering::propose_move()
{
   if( move_type == SINGLE_SUBJECT_SHIFT || move_type == SINGLETON_MERGE || move_type == SINGLETON_SPLIT )
      Cluster_GLM_MVN::set_split_off_one_subject( 1 );
   if( move_type == MERGE || move_type == SINGLETON_MERGE )
      propose_merge_clusters( proposed_cluster2, proposed_cluster );
   else if( move_type == SPLIT || move_type == SINGLETON_SPLIT )
      holding_cluster = propose_split_cluster( proposed_cluster, proposed_subject_list );
   else if( move_type == SUBJECT_SHIFT || move_type == SINGLE_SUBJECT_SHIFT )
   {
      kill_the_parent = 1;
      if( move_type == SINGLE_SUBJECT_SHIFT )
          holding_cluster = propose_move_subject( proposed_cluster, proposed_cluster2, proposed_subject );
      else
          holding_cluster = propose_move_subjects( proposed_cluster, proposed_cluster2 );
      if( preserve_tree )
         kill_the_parent = 0;
   }
   else if( move_type == V_SHIFT || move_type == EXPERIMENT_SHIFT || move_type == EXPERIMENT_SIGN_CHANGE || move_type == EXPERIMENT_MULTIPLIER_SHIFT || move_type == EXPERIMENT_POWER_SHIFT || ( move_type == EXPERIMENT_PARAMETER_SHIFT && integrate_out_experiment_parameters ) )
   {
      if( correlate_experiments || move_type == V_SHIFT )
         propose_new_v();
      else if( correlate_columns )
         propose_new_b();
//      if( move_type == EXPERIMENT_SHIFT )
//         cout << proposed_experiment_index << "  :  " << Cluster_GLM_MVN::get_cluster_count() <<" Clusters, " << original_move_type << "  " << move_type << "  " << LogProposalRatio << "  " << proposed_log_likelihood << "  " <<  log_likelihood << "   " << experiment_variance_likelihood << "   " <<  proposed_experiment_variance_likelihood << endl;
   }
   else if( move_type == PRIOR_V_SHIFT )
   {
      if( !(preserve_tree && mcmc_sampling) )
      {
         store_closeness_contributions();
         calculate_closeness_contributions();
      }
      calculate_log_likelihood();
   }
   if( move_type == SINGLE_SUBJECT_SHIFT || move_type == SINGLETON_MERGE || move_type == SINGLETON_SPLIT )
      Cluster_GLM_MVN::set_split_off_one_subject( 0 );
}

void GLM_MVN_MCMC_Clustering::execute_move()
{
   if( move_type == MERGE || move_type == SINGLETON_MERGE )
   {
      execute_merge_clusters( proposed_cluster2, proposed_cluster );
      mergeable_cluster_deque.pop_front();
   }
   else if( move_type == SPLIT || move_type == SINGLETON_SPLIT )
   {
      mergeable_cluster_deque.push_back( holding_cluster );
      proposed_cluster = execute_split_cluster( proposed_cluster, holding_cluster );
   }
   else if( move_type == SUBJECT_SHIFT || move_type == SINGLE_SUBJECT_SHIFT )
   {
      proposed_cluster2 = execute_move_subjects( holding_cluster, proposed_cluster, proposed_cluster2 );
      if( !conditional_tree_distribution && preserve_tree )
         num_configurations = proposed_num_configurations;
      if( preserve_tree && stop_cascading_at_model )
      {
         if( node_in_model_1 && node_in_model_1 != node_in_model_2 )
         {
            sample_subtree_above_model_nodes_from_full_conditional( node_in_model_1->get_my_root_cluster() );
            calculate_tree_score_normalising_constant();
            tree_log_normalising_constant = proposed_tree_log_normalising_constant;
         }
      }
   }
   if( move_type == V_SHIFT || ( correlate_experiments && ( move_type == EXPERIMENT_SHIFT || move_type == EXPERIMENT_MULTIPLIER_SHIFT  || move_type == EXPERIMENT_POWER_SHIFT || move_type == EXPERIMENT_SIGN_CHANGE ) ) )
   {
       execute_new_v();
       if( move_type == EXPERIMENT_SHIFT && integrate_out_experiment_parameters )
       {
          add_array( experiment_location_sums[ proposed_experiment_group ], experiment_locations[ proposed_experiment_index ], experiment_location_dimension );
          subtract_scaled_array( experiment_location_parameters[ proposed_experiment_group ], experiment_location_store, experiment_location_mean_shrinkages[ proposed_experiment_group ], experiment_location_dimension );
          add_scaled_array( experiment_location_parameters[ proposed_experiment_group ], experiment_locations[ proposed_experiment_index ], experiment_location_mean_shrinkages[ proposed_experiment_group ], experiment_location_dimension );
       }
      experiment_variance_likelihood = proposed_experiment_variance_likelihood;
   }
   else if( move_type == EXPERIMENT_PARAMETER_SHIFT && integrate_out_experiment_parameters )
   {
      execute_new_v();
      add_scaled_array( experiment_location_parameters[ proposed_experiment_group ], experiment_location_store, experiment_location_mean_shrinkages[ proposed_experiment_group ] * experiment_partition_sizes[ proposed_experiment_group ], experiment_location_dimension );
      experiment_variance_likelihood = proposed_experiment_variance_likelihood;
   }

   if( preserve_tree )
   {
      if( conditional_tree_distribution >= 2 && conditional_tree_distribution <= 4 )
      {
         bool non_mcmc = !true;
         if( non_mcmc )
            fit_optimal_tree_around_model();
         else if( move_type == V_SHIFT || move_type == EXPERIMENT_SHIFT || move_type == EXPERIMENT_SIGN_CHANGE || move_type == EXPERIMENT_MULTIPLIER_SHIFT || move_type == EXPERIMENT_POWER_SHIFT )
         {
            proposed_tree_above_model_score = 0;
            for( cluster_ptrtodbl::iterator iter = cluster_map.begin(); iter != cluster_map.end(); iter++ )
               if( iter->first->get_mergeable() )
                  iter->first->set_score( 0 );
            sample_tree_from_full_conditionals();
         }
         tree_score = proposed_tree_score;
         tree_above_model_score = proposed_tree_above_model_score;
         tree_log_normalising_constant = proposed_tree_log_normalising_constant;
      }
   }
   if( preserve_tree )
      initialise_mergeable_cluster_deque();
   proposed_number_of_clusters = number_of_clusters = mergeable_cluster_deque.size();
   if( !heteroskedastic )
      current_ytd_vstar_dty_sum = proposed_ytd_vstar_dty_sum;
   if( move_type == V_SHIFT )
      update_cluster_label_probabilities();
   else if( proposed_iter_label )
      update_cluster_label_probabilities( *proposed_iter_label );
}

void GLM_MVN_MCMC_Clustering::undo_move()
{
   if( move_type == SPLIT || move_type == SINGLETON_SPLIT )
      undo_split_cluster( proposed_cluster, holding_cluster );
   else if( move_type == SUBJECT_SHIFT || move_type == SINGLE_SUBJECT_SHIFT )
      undo_move_subjects( holding_cluster, proposed_cluster );
   else if( move_type == V_SHIFT )
   {
      if( V_ )
         gsl_matrix_scale( V_, v_multiplier_store / v_multiplier );
      gsl_matrix_scale( V_inv, v_multiplier / v_multiplier_store );
      if( !non_conjugate_model )
         ml_const = ml_const_store;
      v_multiplier = v_multiplier_store;
      undo_new_v();
   }
   else if( move_type == PRIOR_V_SHIFT )
   {
      log_prior_v[ proposed_v_num ] = log_prior_v_store;
      log_prior_v[ proposed_v_num2 ] = log_prior_v_store2;
      if( !(preserve_tree && mcmc_sampling) )
         recover_closeness_contributions();
   }
   else if( move_type == EXPERIMENT_SHIFT || move_type == EXPERIMENT_SIGN_CHANGE  || move_type == EXPERIMENT_POWER_SHIFT || move_type == EXPERIMENT_MULTIPLIER_SHIFT )
   {
      set_v_index( 0 );
      if( move_type == EXPERIMENT_SHIFT )
      {
         switch_values( experiment_locations[ proposed_experiment_index ], experiment_location_store );
//         for( int j = 0; j < experiment_location_dimension; j++ )
//            experiment_locations[ proposed_experiment_index ][ j ] = experiment_location_store[ j ];
         update_correlated_experiments_covariances( proposed_experiment_index );
         if( integrate_out_experiment_parameters )
            add_array( experiment_location_sums[ proposed_experiment_group ], experiment_locations[ proposed_experiment_index ], experiment_location_dimension );
      }
      else if( move_type == EXPERIMENT_SIGN_CHANGE )
      {
         for( int i = 0; i < num_locations; i++ )
            experiment_signs[ i ] = experiment_signs_store[ i ];
         update_correlated_experiments_covariances();
      }
      else if( move_type == EXPERIMENT_POWER_SHIFT )
      {
         *experiment_closeness_power = experiment_closeness_power_store;
         find_correlated_experiments_covariances();
      }
      else if( move_type == EXPERIMENT_MULTIPLIER_SHIFT )
      {
         experiment_var_multipliers[ proposed_experiment_index ] = experiment_var_multiplier_store;
         update_correlated_experiments_covariances( proposed_experiment_index );
      }
      if( correlate_experiments )
      {
         if( !non_conjugate_model && number_of_vs == 1 )
            ml_const = ml_const_store;
         undo_new_v();
         update_v_arrays();
      }
      else
      {
         undo_new_b();
         lambda_prior_normalising_constant = lambda_prior_normalising_constant_store;
      }
      proposed_experiment_variance_likelihood = experiment_variance_likelihood;
   }
   else if( move_type == EXPERIMENT_PARAMETER_SHIFT )
   {
      if( integrate_out_experiment_parameters )
      {
	//         subtract_array( experiment_location_sums[ proposed_experiment_group ], experiment_location_store, experiment_location_dimension );
         set_v_index( 0 );
         for( int j = 0; j < experiment_location_dimension; j++ )
         {
            double shift_j = experiment_location_store[ j ];
            for( int i = 0; i < num_locations; i++ )
               if( ( experiment_partition && experiment_partition[ i ] == proposed_experiment_group ) || ( !experiment_partition && i == proposed_experiment_group ) )
                  experiment_locations[ i ][ j ] -= shift_j;
            if( experiment_partition_sizes )
               shift_j *= experiment_partition_sizes[ proposed_experiment_group ];
	    experiment_location_sums[ proposed_experiment_group ][ j ] -= shift_j;
         }
         update_correlated_experiments_covariances();
         if( correlate_experiments )
         {
            if( !non_conjugate_model && number_of_vs == 1 )
               ml_const = ml_const_store;
            undo_new_v();
            update_v_arrays();
         }
         else
         {
            undo_new_b();
            lambda_prior_normalising_constant = lambda_prior_normalising_constant_store;
         }
      }
      else
         copy_array( experiment_location_parameters[ proposed_experiment_group ], experiment_location_parameter_store, experiment_location_dimension );
      proposed_experiment_variance_likelihood = experiment_variance_likelihood;
//      for( int j = 0; j < experiment_location_dimension; j++ )
//         experiment_location_parameters[ proposed_experiment_group ][ j ] = experiment_location_parameter_store[ j ];
   }
   if( preserve_tree )
   {
      initialise_mergeable_cluster_deque();
      if( conditional_tree_distribution >= 2 && conditional_tree_distribution <= 4 )
      {
         proposed_tree_above_model_score = tree_above_model_score;
         proposed_tree_score = tree_score;
         proposed_tree_log_normalising_constant = tree_log_normalising_constant;
      }
   }
   if( !heteroskedastic )
      proposed_ytd_vstar_dty_sum = current_ytd_vstar_dty_sum;
   proposed_number_of_clusters = number_of_clusters;
}

void GLM_MVN_MCMC_Clustering::ModelPrintOut()
{
   if( bic_criterion == 1 )
      cout << "BIC = " << BIC << endl;
   if( preserve_tree && ( conditional_tree_distribution >= 2 && conditional_tree_distribution <= 4 ) )
      cout << "Tree 'score' = " <<  ( conditional_tree_distribution !=4 ? tree_score : exp( tree_score - tree_log_normalising_constant ) ) << endl;
   cout << "Number of Clusters = " << number_of_clusters << endl;
   if( random_number_of_clusters )
      draw_histogram( cluster_probs, cluster_probs_size );
   if( prior_v_shift_move_included )
   {
      cout << "p(V_) = ";
      for( int i = 0; i < number_of_vs; i++ )
         cout << exp( log_prior_v[ i ] ) << "  ";
      cout << endl;
   }
}

void GLM_MVN_MCMC_Clustering::initialise_ergodic_averages()
{
   for( cllabtoprob::iterator iter0 = cluster_and_label_probability.begin(); iter0 != cluster_and_label_probability.end(); iter0++ )
   {
      pair< Cluster_GLM_MVN*, string > clstr_pair( (iter0->first).first, (iter0->first).second );
      mean_cluster_and_label_probability[ clstr_pair ] = 0;
   }
   storing_cluster_closenesses = false;
   zap_triangular_array( single_subject_closeness_matrix, num_subjects );//, num_subjects );
   int num_rows = unmergeable_cluster_deque.size();
   int i;
   if( num_rows && !ave_predictions_mx )
   {
      predictions_mx_rows = num_rows;
      ave_predictions_mx_cols = p_y == 1 ? 2 : 3;
      create_zeroed_array( ave_predictions_mx, predictions_mx_rows, ave_predictions_mx_cols );
   }
   if( ergodic_ave_normaliser && cluster_probs )
      zero_array( cluster_probs, cluster_probs_size );
   if( strlen( Association_Matrix_Filename ) && !association_matrix )
      create_zeroed_array( association_matrix, num_subjects, num_subjects );
   else if( run_ergodic_ave_normaliser && association_matrix )
      zero_array( association_matrix, num_subjects, num_subjects );
   if( run_ergodic_ave_normaliser && ave_prior_v_probs )
      zero_array( ave_prior_v_probs, number_of_vs ); 
   if( v_shift_move_included || experiment_shift_move_included )
   {
      vstar_maps_store = new mx_numtoptr*[ number_of_vs ];
      for( i = 0; i < number_of_vs; i++ )
	 vstar_maps_store[ i ] = new mx_numtoptr;
      vstar_map_store = vstar_maps_store[ 0 ];
      half_log_det_vstar_maps_store = new inttodbl*[ number_of_vs ];
      for( i = 0; i < number_of_vs; i++ )
         half_log_det_vstar_maps_store[ i ] = new inttodbl;
      half_log_det_vstars_store = half_log_det_vstar_maps_store[ 0 ];
   }
   copy_array( best_merger_details, merger_details, merger_details_rows, merger_details_cols );
   copy_array( best_dendrogram, dendrogram, num_subjects );
   if( multiple_runs() || ( changing_clustering_in_mcmc && !strlen( Partition_Input_Filename ) ) )
      double_size_of_merger_details_array();
   if( experiment_covariances )
   {
       if( !mean_experiment_covariances && run_ergodic_ave_normaliser && !simulated_annealing )
         mean_experiment_covariances = gsl_matrix_calloc( num_locations, num_locations );
      if( !best_experiment_covariances )
         best_experiment_covariances = gsl_matrix_calloc( num_locations, num_locations );
      gsl_matrix_memcpy( best_experiment_covariances, experiment_covariances );
      if( !best_experiment_var_multipliers )
         best_experiment_var_multipliers = new double[ num_locations ];
      copy_array( best_experiment_var_multipliers, experiment_var_multipliers, num_locations );
      copy_array( best_experiment_locations, experiment_locations, num_locations, experiment_location_dimension );
   }
   if( log_prior_v )
   {
      if( !best_log_prior_v )
         best_log_prior_v = new double[ number_of_vs ];
      copy_array( best_log_prior_v, log_prior_v, number_of_vs );
   }
   best_v_multiplier = v_multiplier;
   if( experiment_closeness_power )
      best_experiment_closeness_power = *experiment_closeness_power;

   if( changing_clustering_in_mcmc && find_map )
      copy_array( best_clustering, optimal_clustering, num_subjects );
   best_num_clusters = optimal_num_clusters;
   if( preserve_tree )
   {
      Cluster_GLM_MVN::empty_parent_data_matrices_from_top();
      store_data_throughout_tree = false;
   }
   ave_v_multiplier = 0;
   if( preserve_tree && !cluster_by_labels )
      num_configurations = get_number_of_possible_clusterings_for_tree_containging_cluster( cluster_map.begin()->first );
   proposed_tree_above_model_score = tree_above_model_score = 1;
   proposed_number_of_clusters = number_of_clusters;
   if( conditional_tree_distribution >= 2 && conditional_tree_distribution <= 4 )
   {
      calculate_tree_score_normalising_constant();
      tree_log_normalising_constant = proposed_tree_log_normalising_constant;
   }
   proposed_tree_score = tree_score = number_of_clusters + 1;
   if( conditional_tree_distribution == 4 )
   {
      tree_score = get_log_one_plus_alpha_catalan_number( number_of_clusters );
      for( cluster_ptrtodbl::iterator iter = cluster_map.begin(); iter != cluster_map.end(); iter++ )
	 if( iter->first->get_mergeable() )
	    tree_score += get_log_one_plus_alpha_catalan_number( iter->first );
      proposed_tree_score = tree_score;
   }
   log_posterior = log_likelihood;
   maximum_run_posterior = log_posterior;
}

void GLM_MVN_MCMC_Clustering::update_ergodic_averages()
{
   update_mean_label_probabilities();
   if( random_number_of_clusters )
      cluster_probs[ number_of_clusters ]++;
   ave_v_multiplier += v_multiplier;
   if( association_matrix )
      increment_association_matrix();
   if( ave_prior_v_probs )
      for( int i = 0; i < number_of_vs; i++ )
         ave_prior_v_probs[ i ] += exp( log_prior_v[ i ] );
}

void GLM_MVN_MCMC_Clustering::reset_chain()
{
   if( mean_experiment_covariances )
      gsl_matrix_set_zero( mean_experiment_covariances );
   if( ave_prior_v_probs )
      zero_array( ave_prior_v_probs, number_of_vs );
   ave_v_multiplier = 0;
   if( cluster_probs )
      zero_array( cluster_probs, cluster_probs_size );
}

void GLM_MVN_MCMC_Clustering::end_of_chain_processing()
{
   bool locations_need_updating = false;
   if( mean_experiment_covariances )
   {
      gsl_matrix_scale( mean_experiment_covariances, 1.0 / run_ergodic_ave_normaliser );
      gsl_matrix_memcpy( experiment_covariances, mean_experiment_covariances );
      locations_need_updating = true;
   }
   if( print_to_screen && experiment_covariances )
      print_gsl_matrix( experiment_covariances );
   write_experiment_locations_to_file();
   
   if( ave_prior_v_probs )
   {
      double scaling_value = 1.0 / run_ergodic_ave_normaliser; 
      scale_array( ave_prior_v_probs, scaling_value, number_of_vs );
      for( int i = 0; i < number_of_vs; i++ )
         log_prior_v[ i ] = log( ave_prior_v_probs[ i ] );
   }
   if( print_to_screen && log_prior_v )
   {
      for( int i = 0; i < number_of_vs; i++ )
         cout << exp( log_prior_v[ i ] ) << "  ";
      cout << endl;
   }
   if( variable_v )
   {
      ave_v_multiplier /= run_ergodic_ave_normaliser;
      if( !simulated_annealing )
      {
         if( V_ )
            gsl_matrix_scale( V_, v_multiplier / ave_v_multiplier );
         gsl_matrix_scale( V_inv, ave_v_multiplier / v_multiplier );
         v_multiplier = ave_v_multiplier;
      }
      if( print_to_screen )
         cout << "Mean V_ multiplier = " << ave_v_multiplier << endl;
   }
   if( find_map )
   {
      if( log_posterior > maximum_run_posterior )
      {
         maximum_run_posterior = log_posterior;
         copy_array( best_merger_details, merger_details, merger_details_rows, merger_details_cols );
         copy_array( best_dendrogram, dendrogram, num_subjects );
         if( experiment_covariances )
            gsl_matrix_memcpy( best_experiment_covariances, experiment_covariances );
         if( experiment_var_multipliers )
            copy_array( best_experiment_var_multipliers, experiment_var_multipliers, num_locations );
         if( experiment_locations )
            copy_array( best_experiment_locations, experiment_locations, num_locations, experiment_location_dimension );
         if( log_prior_v )
            copy_array( best_log_prior_v, log_prior_v, number_of_vs );
         if( changing_clustering_in_mcmc )
         {
            switch_values( best_clustering, optimal_clustering );
            switch_values( best_num_clusters, optimal_num_clusters );
            update_optimal_clustering();//actually udating best_clustering array.
            switch_values( best_clustering, optimal_clustering );
            switch_values( best_num_clusters, optimal_num_clusters );
         }
         if( variable_v )
            best_v_multiplier = v_multiplier;
         if( experiment_closeness_power )
            best_experiment_closeness_power = *experiment_closeness_power;
      }
      else
      {
         copy_array( merger_details, best_merger_details, merger_details_rows, merger_details_cols );
         copy_array( dendrogram, best_dendrogram, num_subjects );
         for( int j = 0; j < num_subjects; j++ )
            inverse_dendrogram[ dendrogram[ j ] ] = j;
      }
   }
   if( multiple_runs() && last_chain_run() && find_map )
   {
      copy_array( merger_details, best_merger_details, merger_details_rows, merger_details_cols );
      copy_array( dendrogram, best_dendrogram, num_subjects );
      for( int j = 0; j < num_subjects; j++ )
         inverse_dendrogram[ dendrogram[ j ] ] = j;
      if( experiment_covariances )
      {
         gsl_matrix_memcpy( experiment_covariances, best_experiment_covariances );
         if( mean_experiment_covariances )
            gsl_matrix_memcpy( mean_experiment_covariances, best_experiment_covariances );
//         locations_need_updating = true;
      }
      if( experiment_var_multipliers )
         copy_array( experiment_var_multipliers, best_experiment_var_multipliers, num_locations );
      if( experiment_closeness_power )
         *experiment_closeness_power = best_experiment_closeness_power;
      if( experiment_locations )
      {
         copy_array( experiment_locations, best_experiment_locations, num_locations, experiment_location_dimension );
         find_experiment_location_sums();
         update_correlated_experiments_covariances();
         reset_experiment_location_parameters();
      }

      if( log_prior_v )
      {
         copy_array( log_prior_v, best_log_prior_v, number_of_vs );
         if( ave_prior_v_probs )
            for( int i = 0; i < number_of_vs; i++ )
               ave_prior_v_probs[ i ] = exp( log_prior_v[ i ] );
      }
      if( variable_v )
      {
         if( V_ )
            gsl_matrix_scale( V_, v_multiplier / best_v_multiplier );
         gsl_matrix_scale( V_inv, best_v_multiplier / v_multiplier );
         v_multiplier = best_v_multiplier;
      }
   }
   set_v_index( 0 );
   if( !non_conjugate_model && number_of_vs == 1 )
      ml_const -= half_logdet_v_inv[ 0 ];
   if( experiment_covariances && locations_need_updating )
   {
      experiment_covariances_to_locations( experiment_covariances );
      update_correlated_experiments_covariances();
      reset_experiment_location_parameters();
   }
   update_v_arrays();
   if( !non_conjugate_model && number_of_vs == 1 )
      ml_const += half_logdet_v_inv[ 0 ];
   empty_vstar_map();
   storing_cluster_closenesses = storing_cluster_closenesses_store;
   max_log_likelihood = -DBL_MAX;
   if( last_chain_run() && !preserve_tree && changing_clustering_in_mcmc )
      Cluster_GLM_MVN::set_using_types( true );
   give_clusters_different_types();
   double** null_pointer = NULL;
   switch_values( association_matrix, null_pointer );//hide the association_matrix so it doesn't get written over.
//   if( max_map_allocation_sweeps == 0 )//otherwise this gets done later
   restart_subjects_in_own_clusters();
   merge_until_desired_num_clusters();
   switch_values( association_matrix, null_pointer );//unhide the association_matrix.
   storing_cluster_closenesses = false;
   log_posterior = log_likelihood;
}

void GLM_MVN_MCMC_Clustering::print_model_results()
{
   MCMC::print_model_results();
   if( ergodic_ave_normaliser && model_averaging )
      for( cllabtoprob::iterator iter0 = mean_cluster_and_label_probability.begin(); iter0 != mean_cluster_and_label_probability.end(); iter0++ )
         iter0->second /= ergodic_ave_normaliser;
   print_label_probabilities();
   int i;
   if( random_number_of_clusters && ergodic_ave_normaliser )
   {
      double sum_cluster_probs = 0;
      for( i = 0; i < cluster_probs_size; i++ )
         sum_cluster_probs += cluster_probs[ i ];
      double mean_num_clusters = 0;
      for( i = 0; i < cluster_probs_size; i++ )
      {
         cluster_probs[ i ] /= sum_cluster_probs;
         mean_num_clusters += i * cluster_probs[ i ];
      }
      cout << "Mean number of clusters = " << mean_num_clusters << endl;
      cout << "Posterior distribution of number of clusters:" << endl;
      cout << "#Clusters    Probability" << endl;
      cout << "------------------------" << endl;
      draw_histogram( cluster_probs, cluster_probs_size );
      cout << "------------------------" << endl;
   }
   if( variable_v && ergodic_ave_normaliser )
   {
      if( number_of_runs == 1 )
      {
         ave_v_multiplier /= run_ergodic_ave_normaliser;
         v_multiplier = ave_v_multiplier;
         if( V_ )
            gsl_matrix_scale( V_, v_multiplier / ave_v_multiplier );
         gsl_matrix_scale( V_inv, ave_v_multiplier / v_multiplier );
         if( print_to_screen )
            cout << "Mean V_ multiplier = " << ave_v_multiplier << endl;
      }
   }
   if( ave_prior_v_probs )
   {
      double sum = sum_array( ave_prior_v_probs, number_of_vs );
      cout << "p(V_) = ";
      for( int i = 0; i < number_of_vs; i++ )
         cout << ave_prior_v_probs[ i ] / sum << "  ";
      cout << endl;
   }
   if( variable_v )
      write_V_to_file();
   write_association_matrix_to_file();
   if( preserve_tree )
      write_tree_merger_details();
}

void GLM_MVN_MCMC_Clustering::write_tree_merger_details()
{
   cluster_ptrtodbl::iterator iter;
   mcmc_sampling = true;
   finished_mcmc_sampling = false;
   jump_distn = -1;
   conditional_tree_distribution = 0;//don't want to allow for induced changes in prior on tree here.
   tree_full_conditional_move_included = 0;
   proposed_cluster = proposed_cluster->get_my_root_cluster();
   sample_tree_from_full_conditionals();// with conditional_tree_distribution = 0, this will fit the optimal subtrees around the model

   int num_single_subject_clusters = single_subject_clusters.size();
   double_size_of_merger_details_array();
   update_optimal_clustering();
   initialise_mergeable_cluster_deque();
   for( iter = cluster_map.begin(); iter != cluster_map.end(); ++iter )
      iter->first->set_mergeable( false );
   move_type = SPLIT;
   while( number_of_clusters < num_single_subject_clusters )
   {
      move();
      finished_mcmc_sampling = true;
      execute_move();
      finished_mcmc_sampling = false;
      log_posterior = log_likelihood = proposed_log_likelihood;
      BIC = proposed_BIC;
   }
   mcmc_sampling = false;
   finished_mcmc_sampling = true;
   update_merger_details_first_row();
   for( iter = cluster_map.begin(); iter != cluster_map.end(); ++iter )
   {
      Cluster_GLM_MVN* iter_sib = iter->first->get_sibling_cluster();
      if( is_cluster_in_model( iter_sib ) )
      {
         double d = calculate_closeness( iter->first, iter_sib );
         if( !euclidean_or_correlation_clustering )
            d += get_change_log_prior_cluster_size( iter->first->get_num_subjects_allocated(), iter_sib->get_num_subjects_allocated(), false );
         iter->second = d;
         iter->first->set_cluster_closeness( iter_sib, d );
         iter_sib->set_cluster_closeness( iter->first, d );
         iter->first->set_closest_cluster( iter_sib );
         iter_sib->set_closest_cluster( iter->first );
      }
      else
         cluster_map[ iter->first ] = -DBL_MAX;
   }
   most_mergeable_cluster = NULL;
   max_closeness = -DBL_MAX;
   move_type = MERGE;
   while( number_of_clusters > minimum_num_clusters )
   {
      if( number_of_clusters == optimal_num_clusters )
         for( iter = cluster_map.begin(); iter != cluster_map.end(); ++iter )
            iter->first->revert_mergeable();
      find_best_merger();
      proposed_cluster2 = most_mergeable_cluster;
      proposed_cluster = most_mergeable_clusters_closest;
      propose_move();
      update_merger_details();
      execute_move();
      log_posterior = log_likelihood = proposed_log_likelihood;
      BIC = proposed_BIC;
   }
   update_dendrogram();
   split_to_optimal_clustering();
}

void GLM_MVN_MCMC_Clustering::reset_experiment_location_parameters()
{
   if( integrate_out_experiment_parameters )
      for( int i = 0; i < experiment_partition_size; i++ )
         copy_scaled_array( experiment_location_parameters[ i ], experiment_location_sums[ i ], experiment_location_mean_shrinkages[ i ], experiment_location_dimension );
   else
      Regression::reset_experiment_location_parameters();
}

double log_gamma_density_ratio( const double x_new, const double x_old, const double shape, const double scale ){
   return (shape - 1) * ( log(x_new) - log(x_old) ) - scale * ( x_new - x_old );
}
