
#if !defined( _mcmc_basis_fn_regression_h )
#define _mcmc_basis_fn_regression_h

#include "glm_mvn.h"
class MCMC_Basis_Fn_Regression : public GLM_MVN
{
  public:
   ///
   MCMC_Basis_Fn_Regression( int argc, char **argv );
   ///
   MCMC_Basis_Fn_Regression( gsl_matrix* X1, gsl_matrix* Y1 );// uses matrix,matrix constructor of base class
   ///
   MCMC_Basis_Fn_Regression( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 );// uses matrix,matrix,matrix constructor of base class
   ///
   ~MCMC_Basis_Fn_Regression();
   ///
   void initialise();

   virtual void move();

   void propose_move();// Proposes the chosen move type.

   virtual void execute_move();// Carries out the proposed move (used if move accepted).

   virtual void undo_move();// Cancels the proposed move (used if move not accepted).

   virtual void post_move_updates();// Carries out any updates required after the move, accepted or not.

   virtual void initialise_ergodic_averages();//overwritten fn initialising model averages

   virtual void update_ergodic_averages();//overwritten fn updating model averages
   
   virtual void ModelPrintOut();//overwritten fn outputting model details to screen.

   virtual void make_predictions();//calculates the predictive probs of each unmergeable cluster.
   
   virtual void rw_shift_coefficient();//randomly shift the coefficient of moved_basis.

   protected:

   int ADD_BASIS, DEL_BASIS, MOVE_BASIS, V_SHIFT, SHIFT_COEFFICIENT;//constants to be enumerated (dynamically) to indicate the different move types.

   bool random_basis_fns;//if true, have random number and locations of basis fns.

   bool v_shift_move_included;//non zero if the prior variance multiplier v_multiplier can be shifted.
   
   gsl_matrix *previous_m_star;

   gsl_matrix *previous_b_star;
   
   gsl_matrix *previous_V_star, *previous_V_star_sqrt;

   double* num_splines_probs;//array holding the probabilities of each possible number of splines.

   int num_splines_probs_length;//length of array num_splines_probs.

   int isotonic_count, antitonic_count, monotonic_count;//keep track of the number of samples satisfying these conditions.

   double basis_coeff_step_size;//the size of random walk step used on basis fn coefficients.
   
 protected:

   bool rw_on_basis_coeffs;//if true, perform random walk on basis coefficients rather than resampling from prior

};

#endif

