#include "regression.h"

unsigned long Regression::number_of_models_evaluated = 0;
int Regression::number_of_regressions = 0;
long Regression::seed = 0;
long Regression::test_grid_size = 200;

Regression::Regression( int argc, char **argv)
      : DataObj( argc, argv )
{
   preliminary_regression_initialisation();
   for( int i = 1; i < argc; i++ )
   {
      if ( strstr( argv[i], "help" ) )
      {
         cout << "parameterfile=" << ParameterFilename << endl;
         cout << "initialparameterfile=" << InitialParameterFilename << endl;
         cout << "predfile=" << PredictionFilename << "   (Name of file where any model predictions should be written to)" << endl;
         cout << "subjectpredfile=" << Subject_Predictions_Filename << "   (Name of file where predictions for each subject should be written to)" << endl;
         cout << "subjectindexfile=" << Subject_Index_Filename << "   (If different numbers of observations for different subjects, this file can give the observation number where each subject begins (starting at 0) )" << endl;
         cout << "randomeffectsmodel=" << random_effects_model << "   (If true, assume each subject has a random effects parameter.)" << endl;
         cout << "randomeffectsprecision=" << random_effects_precision << "   (If using random effects model, this is the precision of those parameters.)" << endl;
         cout << "randomeffectvariancesfile=" << Random_Effect_Variances_Filename << "   (File containing individual-specific random effect variances.)" << endl;
         cout << "numberofexperiments=" << num_experiments << "   (The number of experiments represented in the data)" << endl;
         cout << "subjectgrouping=" << subject_grouping << "   (If non-zero, every this many observations are deemed to be from the same subject and thus forced to be in the same cluster)" << endl;
         cout << "experimentnumberfile=" << Experiment_Number_Filename << "   (If multiple observations from each subject, each from varying experiments, this file can contain the experiment number to which each observation belongs)" << endl;
         cout << "equivalentexperimentsfile=" << Equivalent_Experiments_Filename << "   (If we wish to consider certain experiments as equivalent, this file should contain a set of equivalent experiment numbers to which the original ones should be mapped.)" << endl;
         cout << "separate_experiments=" << treat_experiments_separately << "   (If non-zero, fits a separate Gaussian process for each experiment)" << endl;
         cout << "differentdimensionexperiments=" << treat_experiments_as_different_dimensions << "   (If true, consider multiple experiments as observations along the axes of different dimensions of a multivariate surface, and treat as just one experiment.)" << endl;
         cout << "correlateexperiments=" << correlate_experiments << "   (If non-zero, specify a correlation between experiments)" << endl;
         cout << "startexperimentsindependent" << start_experiments_independent << "   (If non-zero, true, assume that experiments have initially no correlation)" << endl;
         cout << "krigingexperimentcorrelation=" << kriging_experiment_correlation << "   (If correlating experiments we create abstract experiment locations (say X)  and then if this parameter true, use a kriging model on X to model the correlations, otherwise use X'X as correlation matrix)" << endl;
         cout << "weightexperiments=" << weight_experiments << "   (If non-zero, does weighted regression with each experiment reweighted according to it's sample precision)" << endl;
         cout << "standardizeexperiments=" << standardize_experiments << "   (If non-zero, regression uses the targets divided by the standard deviation for the corresponding experiments)" << endl;
         cout << "experimentlocationdimension=" << experiment_location_dimension << "   (if >0, create abstract spatial locations for each experiment in this dimensional space, with closeness in locations of experiments indicating correlation between them)" << endl;
         cout << "experimentlocationprecision=" << experiment_location_precision << "   (A prior precision parameter for the locations of experiments)" << endl;
         cout << "experimentparameterprecision=" << experiment_parameter_precision << "   (A prior precision parameter for the means of the locations of experiments)" << endl;
         cout << "experimentlocationsfile=" << Experiment_Locations_Filename << "   (If experiments have abstract locations, their principal components are written to this file)" << endl;
         cout << "experimentlocationsinputfile=" << Experiment_Locations_Input_Filename << "   (If experiments have abstract locations, they are read in from this file)" << endl;
         cout << "experimentparametersfile=" << Experiment_Parameters_Filename << "   (If experiments have abstract locations, their parameters' principal components are written to this file)" << endl;
         cout << "experimentparametersinputfile=" << Experiment_Parameters_Input_Filename << "   (If experiments have abstract locations, their parameters are read in from this file)" << endl;
         cout << "experimentsignsfile=" << Experiment_Signs_Filename << "   (If experiments have abstract locations, the signs associated with them are written to this file)" << endl;
         cout << "experimentcorrelationsfile=" << Experiment_Correlations_Filename << "   (If experiments have abstract locations, their induced correlations are written to this file)" << endl;
         cout << "experimentpartitionfile=" << Experiment_Partition_Filename << "   (If experiments have abstract locations, a partition of them can be read in from this file)" << endl;
         cout << "residualsfile=" << Residuals_Filename << "   (If filename supplied, the residuals of the data are calculated and written to this file)" << endl;
         cout << "variogramfile=" << Variogram_Filename << "   (If filename supplied, the variogram of the data is calculated and written to this file)" << endl;
         cout << "calculatepairwisevariances=" << calculate_pairwise_variances << "    (If true, calculate the sample variance for every pair of subjects and write to file)" << endl;
         cout << "pairwisevariancefile=" << Pairwise_Variances_Filename << "   (If filename supplied, this replaces the default name as the file to/from which the sample variances calculated for every pair of subjects are written/read)" << endl;
         cout << "testgridsize=" << test_grid_size << "   (Number of grid points at which predictions will be made)" << endl;
	 cout << "correlfile=" << Correl_Filename << "   (If filename supplied, the correlation matrix of the data is calculated and written to this file)" << endl;
         cout << "seed=" << seed << endl;
	 cout << "normalisedatablocks" << normalise_data_blocks << "   (If non zero,normalise each data block to have mean sum of squares 1 before performing regression)" << endl;
         cout << "logtargets=" << log_targets << "   (If non zero, perform regression on the natural logarithm of the target data (+1 if logtargets=3) )" << endl;
         cout << "loginputs=" << log_inputs << "   (If non zero, perform regression using the natural logarithm of the predictor data (+1 if loginputs=3) )" << endl;
	 cout << "logistictargets=" << logistic_transform_targets << "   (If true, do y=(1+exp(-y))^-1.)" << endl;
	 cout << "logisticpredictors=" << logistic_transform_predictors << "   (If true, do x=(1+exp(-x))^-1.)" << endl;
         cout <<  "usedifferences=" << use_differences << "   (If non zero, perform regression on the first differences of the data)" << endl;
         cout <<  "subtractintercept=" << subtract_intercept << "   (If non zero, remove the intercept for each data block)" << endl;
         cout <<  "replacewithbootstrapsamples=" << replace_with_bootstrap_samples << "   (If non-zero, replace the data with this many samples with replacement from the data)" << endl;
         cout << "thindata=" << thin_data << "   (If non-zero, thin the data keeping only every other this many data points)" << endl;
         cout <<  "predictovergrid=" << predict_over_grid << "   (If true, append some test points to the data matrices ranging from the min to max target values, for predictions)" << endl;
         cout <<  "predictondata=" << predict_on_data << "   (If true, calculate predicted target values for each of the input data points)" << endl;
         cout << "gramschmidt=" << gram_schmidt <<"   (If true, use the Gram-Schmidt process to orthogonalize the design matrix.)" << endl;
         return;
       }
      if( InitialStringMatch( argv[i], "randomeffectsmodel=" ) )
         StringToBool( &random_effects_model, argv[i] );
      if( InitialStringMatch( argv[i], "randomeffectsprecision=" ) )
         StringToDouble( &random_effects_precision, argv[i] );
      if ( InitialStringMatch( argv[i], "randomeffectvariancesfile=" ) )
         StringToString( Random_Effect_Variances_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "numberofexperiments=" ) )
         StringToLong( &num_experiments, argv[i] );
      if( InitialStringMatch( argv[i], "subjectgrouping=" ) )
         StringToLong( &subject_grouping, argv[i] );
      if( InitialStringMatch( argv[i], "separate_experiments=" ) )
         StringToBool( &treat_experiments_separately, argv[i] );
      if( InitialStringMatch( argv[i], "differentdimensionexperiments=" ) )
         StringToBool( &treat_experiments_as_different_dimensions, argv[i] );
      if( InitialStringMatch( argv[i], "correlateexperiments=" ) )
         StringToBool( &correlate_experiments, argv[i] );
      if( InitialStringMatch( argv[i], "startexperimentsindependent=" ) )
         StringToBool( &start_experiments_independent, argv[i] );
      if( InitialStringMatch( argv[i], "krigingexperimentcorrelation=" ) )
         StringToBool( &kriging_experiment_correlation, argv[i] );
      if( InitialStringMatch( argv[i], "weightexperiments=" ) )
         StringToBool( &weight_experiments, argv[i] );
      if( InitialStringMatch( argv[i], "standardizeexperiments=" ) )
         StringToBool( &standardize_experiments, argv[i] );
      if( InitialStringMatch( argv[i], "experimentlocationdimension=" ) )
         StringToLong( &experiment_location_dimension, argv[i] );
      if ( InitialStringMatch( argv[i], "experimentlocationprecision=" ) )
         StringToDouble( &experiment_location_precision, argv[i] );
      if ( InitialStringMatch( argv[i], "experimentparameterprecision=" ) )
         StringToDouble( &experiment_parameter_precision, argv[i] );
      if ( InitialStringMatch( argv[i], "parameterfile=" ) )
	StringToString( ParameterFilename, argv[i] );
      if ( InitialStringMatch( argv[i], "initialparameterfile=" ) )
	StringToString( InitialParameterFilename, argv[i] );
      if ( InitialStringMatch( argv[i], "predfile=" ) )
	StringToString( PredictionFilename, argv[i] );
      if ( InitialStringMatch( argv[i], "subjectpredfile=" ) )
	StringToString( Subject_Predictions_Filename, argv[i] );
      if ( InitialStringMatch( argv[i], "testinputfile=" ) )
	StringToString( TestInputFilename, argv[i] );
      if ( InitialStringMatch( argv[i], "testtargetfile=" ) )
	StringToString( TestTargetFilename, argv[i] );
      if( InitialStringMatch( argv[i], "subjectindexfile=" ) )
         StringToString( Subject_Index_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "experimentnumberfile=" ) )
         StringToString( Experiment_Number_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "equivalentexperimentsfile=" ) )
         StringToString( Equivalent_Experiments_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "experimentlocationsfile=" ) )
         StringToString( Experiment_Locations_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "experimentlocationsinputfile=" ) )
         StringToString( Experiment_Locations_Input_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "experimentparametersfile=" ) )
         StringToString( Experiment_Parameters_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "experimentparametersinputfile=" ) )
         StringToString( Experiment_Parameters_Input_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "experimentsignsfile=" ) )
         StringToString( Experiment_Signs_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "experimentcorrelationsfile=" ) )
         StringToString( Experiment_Correlations_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "experimentpartitionfile=" ) )
         StringToString( Experiment_Partition_Filename, argv[i] );
      if ( InitialStringMatch( argv[i], "variogramfile=" ) )
	StringToString( Variogram_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "calculatepairwisevariances=" ) )
         StringToBool( &calculate_pairwise_variances, argv[i] );
      if ( InitialStringMatch( argv[i], "pairwisevariancefile=" ) )
        StringToString( Pairwise_Variances_Filename, argv[i] );
      if ( InitialStringMatch( argv[i], "residualsfile=" ) )
         StringToString( Residuals_Filename, argv[i] );
      if ( InitialStringMatch( argv[i], "testgridsize=" ) )
	StringToLong( &test_grid_size, argv[i] );
      if ( InitialStringMatch( argv[i], "correlfile=" ) )
	StringToString( Correl_Filename, argv[i] );
      if ( InitialStringMatch( argv[i], "seed=" ) )
	StringToLong( &seed, argv[i] );
      if ( InitialStringMatch( argv[i], "normalisedatablocks=" ) )
	StringToBool( &normalise_data_blocks, argv[i] );
      if ( InitialStringMatch( argv[i], "logtargets=" ) )
	StringToLong( &log_targets, argv[i] );
      if ( InitialStringMatch( argv[i], "loginputs=" ) )
         StringToLong( &log_inputs, argv[i] );
      if ( InitialStringMatch( argv[i], "logistictargets=" ) )
	StringToBool( &logistic_transform_targets, argv[i] );
      if ( InitialStringMatch( argv[i], "logisticpredictors=" ) )
	StringToBool( &logistic_transform_predictors, argv[i] );
      if ( InitialStringMatch( argv[i], "usedifferences=" ) )
	StringToBool( &use_differences, argv[i] );
      if ( InitialStringMatch( argv[i], "subtractintercept=" ) )
	StringToBool( &subtract_intercept, argv[i] );
      if ( InitialStringMatch( argv[i], "replacewithbootstrapsamples=" ) )
	StringToLong( &replace_with_bootstrap_samples, argv[i] );
      if ( InitialStringMatch( argv[i], "thindata=" ) )
         StringToLong( &thin_data, argv[i] );
      if ( InitialStringMatch( argv[i], "predictovergrid=" ) )
	StringToBool( &predict_over_grid, argv[i] );
      if ( InitialStringMatch( argv[i], "predictondata=" ) )
	StringToBool( &predict_on_data, argv[i] );
      if( InitialStringMatch( argv[i], "gramschmidt=" ) )
        StringToBool( &gram_schmidt, argv[i] );
   }
   InitialiseParameters();
}

Regression::Regression( gsl_matrix* X1, gsl_matrix* Y1 ): DataObj( X1, Y1 )
{
   preliminary_regression_initialisation();
   InitialiseParameters();
}

Regression::Regression( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 ): DataObj( X1, Y1 )
{
   preliminary_regression_initialisation();
   InitialiseParameters();
   D = gsl_matrix_alloc( D1->size1, D1->size2 );
   gsl_matrix_memcpy( D, D1 );
}

Regression::~Regression()
{
   if( D && D != X )
      gsl_matrix_free( D );
   if( Subject_index_original != Subject_index )
      zap_vector( Subject_index_original );
   zap_vector( Subject_index );
   if( experiment_numbers_original != experiment_numbers )
      zap_vector( experiment_numbers_original );
   zap_vector( experiment_numbers );
   zap_vector( equivalent_experiments );
   zap_vector( experiment_weights );
   zap_vector( experiment_sds );
   zap_vector( experiment_means );
   zap_array( experiment_locations, num_locations, experiment_location_dimension );
   zap_array( best_experiment_locations, num_locations, experiment_location_dimension );
   if( experiment_location_parameters && experiment_location_parameters != experiment_location_means )
   {
      zap_array( experiment_location_parameters, num_location_parameters, experiment_location_dimension );
      zap( experiment_location_means );//the rows are pointing to experiment_location_parameters
   }
   else
      zap_array( experiment_location_means, num_locations, experiment_location_dimension );
   zap_array( experiment_location_sums, experiment_partition_size, experiment_location_dimension );
   zap_vector( experiment_partition );
   zap_vector( experiment_partition_sizes );
   zap_vector( experiment_signs );
   zap_vector( experiment_var_multipliers );
   zap_vector( best_experiment_var_multipliers );
   zap_vector( design_point );
   zap_vector( random_effect_variances );
   zap_vector( random_effect_posterior_precisions );
   if( experiment_covariances )
      gsl_matrix_free( experiment_covariances );
   if( experiment_precisions )
      gsl_matrix_free( experiment_precisions );  
   if( mean_experiment_covariances )
      gsl_matrix_free( mean_experiment_covariances );
   if( best_experiment_covariances )
      gsl_matrix_free( best_experiment_covariances );
   if( residuals )
      gsl_matrix_free( residuals );
   if( beta )
      gsl_matrix_free( beta );
   if( w_gl )
      gsl_matrix_free( w_gl );
   if( x_gl )
      gsl_matrix_free( x_gl );
   zap_array( predictions_mx, predictions_mx_rows, predictions_mx_cols );
   if( test_grid_size )
      zap_array( test_grid_points, test_grid_size, p );
   zap_array( y_differenced_points, num_differenced_points_stored, 1 );
   zap_array( x_differenced_points, num_differenced_points_stored, 1 );
   zap_array( y_block_norms, num_data_blocks, p_y );
   zap_array( subject_exp_num_missing_values, num_subjects, num_experiments );
   gsl_rng_free( rn );
   number_of_regressions--;
   if( print_to_screen && number_of_models_evaluated )
      cout << number_of_models_evaluated << " models evaluated." << endl;
}

void Regression::preliminary_regression_initialisation()
{
   log_likelihood_constant = 0;
   normalise_data_blocks = 0;
   log_targets = 0;
   log_inputs = 0;
   logistic_transform_targets = logistic_transform_predictors = false;
   use_differences = 0;
   subtract_intercept = 0;
   gram_schmidt = false;
   y_differenced_points = 0;
   x_differenced_points = 0;
   y_block_norms = 0;
   num_differenced_points_stored = 0;
   predict_over_grid = 0;
   predict_on_data = 0;
   subject_grouping = 0;
   Subject_index = 0;
   Subject_index_original = 0;
   num_experiments_original = num_experiments = num_data_block_sizes_original;
   treat_experiments_separately = 0;
   subject_exp_num_missing_values = NULL;
   equivalent_experiments = 0;
   correlate_experiments = 0;
   correlate_columns = 0;
   kriging_experiment_correlation = 0;
   start_experiments_independent = true;
   weight_experiments = 0;
   standardize_experiments = 0;
   random_effects_model = 0;
   random_effects_precision = 1;
   random_effect_variances = NULL;
   random_effect_log_variances = NULL;
   random_effect_posterior_precisions = NULL;
   sum_log_random_effect_variances = 0;
   log_ave_experiments_sds = 0;
   experiment_numbers = 0;
   experiment_numbers_original = 0;
   experiment_weights = 0;
   experiment_sds = 0;
   experiment_means = 0;
   current_experiment = 0;
   experiment_location_dimension = 0;
   experiment_parameter_precision = experiment_location_precision = .1;
   experiment_closeness_power = NULL;
   num_location_parameters = 0;
   experiment_partition = 0;
   experiment_partition_sizes = 0;
   replace_with_bootstrap_samples = 0;
   thin_data = 0;
   treat_experiments_as_different_dimensions = false;
   seemingly_unrelated_regressions = false;
   linear_term_for_each_dimension = intercepts_for_each_dimension = false;
   active_dimension = 0;
   rn = NULL;
   rn_type = gsl_rng_env_setup();
   if( print_to_screen )
      cout << "GSL version: " << GSL_VERSION << ", " << "random generator: " << rn_type->name << endl;
   *ParameterFilename = 0;
   *PredictionFilename = 0;
   *Subject_Predictions_Filename = 0;
   *Subject_Index_Filename = 0;
   *Experiment_Number_Filename = 0;
   *Equivalent_Experiments_Filename = 0;
   *Experiment_Locations_Filename = 0;
   *Experiment_Locations_Input_Filename = 0;
   *Experiment_Parameters_Input_Filename = 0;
   *Experiment_Parameters_Filename = 0;
   *Experiment_Signs_Filename = 0;
   *Experiment_Correlations_Filename = 0;
   *Experiment_Partition_Filename = 0;
   *Variogram_Filename = 0;
   *Pairwise_Variances_Filename = 0;
   *Residuals_Filename = 0;
   *Random_Effect_Variances_Filename = 0;
   variable_to_predict_over = 0;
   second_variable_to_predict_over = 0;
   current_regression_fn_val = 0;
   calculate_pairwise_variances = false;
   LQ_pairwise_variance = median_pairwise_variance = 0;
}

void Regression::InitialiseParameters()
{
   rn = gsl_rng_alloc( rn_type );
   gsl_rng_set( rn, seed );
   srand( seed );
   BIC = 0;
   predictions_mx = 0;
   num_test_points = test_grid_size;
   test_grid_points = 0;
   unique_data_points = 0;
   num_unique_data_points = 0;
   predictions_mx_rows = predictions_mx_cols = 0;
   predictions_update_cols = 1;
   prediction_fn_value_col = 0;
   beta = 0;
   D_column_sums = 0;
   residuals = 0;
   mcmc_sampling = false;
   w_gl = x_gl = 0;
   experiment_locations = 0;
   best_experiment_locations = 0;
   experiment_location_means = 0;
   experiment_location_parameters = 0;
   experiment_location_sums = 0;
   experiment_signs = 0;
   experiment_var_multipliers = 0;
   experiment_covariances = 0;
   experiment_precisions = 0;
   mean_experiment_covariances = 0;
   best_experiment_covariances = 0;
   best_experiment_var_multipliers = 0;
   if( num_experiments > 1 )
      num_locations = num_experiments;
   else
      num_locations = p_y;
   if( !subject_grouping )
   {
      if( data_block_sizes )
         subject_grouping = num_data_block_sizes;
      else
         subject_grouping = num_experiments;
   }
//   if( !subject_grouping )
//      subject_grouping = 1;
   if( p_y > 1 && variable_to_predict_over == second_variable_to_predict_over )
   {
      if( variable_to_predict_over == 0 )
         second_variable_to_predict_over = 1;
      else
         second_variable_to_predict_over = 0;
   }
   if( !X )
   {
      p = 1;
      X = gsl_matrix_alloc( n, p );
      for( int i = 0; i < num_data_blocks/num_data_block_sizes; i++ )
      {
	int cum_dbs = 0;
	for( int ii = 0; ii < num_data_block_sizes; ii++ )
	{
	   int dbs = data_block_sizes ? data_block_sizes[ii] : data_block_size;
	   for( int i0 = 0; i0 < dbs ; i0++ )
	      for( int j = 0; j < p; j++ )
		 gsl_matrix_set( X, i * data_block_sizes_sum + ( cum_dbs + i0 ), j, i0 + 1 );
	   cum_dbs += dbs;
	}
      }
      inputs_repeated_every = data_block_sizes_sum;
      normalise_predictors = 0;
   }
   number_of_regressions++;
   if( logistic_transform_predictors )
      log_inputs = 0;
   if( logistic_transform_targets )
      log_targets = 0;
   if( log_targets || log_inputs )
      log_data();
   if( logistic_transform_targets || logistic_transform_predictors )
      logistic_transform_data();
   if( use_differences || subtract_intercept)
      difference_data( subtract_intercept );
   create_subject_indices();
   if( gram_schmidt )
      treat_experiments_as_different_dimensions = true;
   create_experiment_numbers_vector();
   if( treat_experiments_as_different_dimensions )
   {
      seemingly_unrelated_regressions = true;
      linear_term_for_each_dimension = intercepts_for_each_dimension = true;
   }
   if( missing_values )
      count_subject_experiment_missing_values();
   if( normalise_data_blocks )
      normalise_target_data_blocks();
   normalise_data_matrices( x_min==NULL );
   if( delete_point >= 0 )
      delete_points();
   if( delete_missing )
      delete_missing_points();
   if( replace_with_bootstrap_samples )
      bootstrap_data();
   if( thin_data > 0 )
      thin_the_data();
   regression_n = n;
   regression_num_data_blocks = num_data_blocks;
   regression_num_subjects = num_subjects;
   if( predict_on_data )
      find_unique_data_points();
   if( predict_over_grid )
      append_test_points();
   create_design_matrix();
   if( !start_experiments_independent )
   {
      int min_data_block_size = data_block_size;
      if( data_block_sizes )
         for( int i = 0; i < num_data_block_sizes; i++ )
            if( min_data_block_size < data_block_sizes[ i ] )
               min_data_block_size = data_block_sizes[ i ];
      if( min_data_block_size < num_design_cols )
         start_experiments_independent = true;
   }
   gauss_laguerre_multiplier = 1;
   if( !normalise_targets && y_scale[ 0 ] > 0 )
      gauss_laguerre_multiplier /= y_scale[ 0 ] * y_scale[ 0 ];
   design_point = 0;
   design_point_dim = p;
   if( !strlen( ParameterFilename ) )
   {
      char* new_file = ParameterFilename;
      strcpy( new_file, TargetFilename );
      string insert = "_parameters_";
      insert_in_string_before_dot( new_file, &insert );
   }

   if( !strlen(Pairwise_Variances_Filename) )
   {
      char* new_file = Pairwise_Variances_Filename;
      strcpy( new_file, TargetFilename );
      string insert = "_pairwise_vars";
      insert_in_string_before_dot( new_file, &insert );
   }
   if( random_effects_model )
   {
      sum_log_random_effect_variances = -regression_num_subjects * log( random_effects_precision );
      if( strlen(Random_Effect_Variances_Filename) )
      {
         scan_array_data( Random_Effect_Variances_Filename, random_effect_variances, num_subjects );
	 random_effect_posterior_precisions = new double[ num_subjects ];
         random_effect_log_variances = new double[ num_subjects ];
	 process_random_effect_variances( y_scale[ 0 ]*y_scale[ 0 ] );
      }
   }
   if( calculate_pairwise_variances )
      get_pairwise_sample_variances();
}

void Regression::process_random_effect_variances( double rescale )
{
         sum_log_random_effect_variances = 0;
         for( int i = 0; i < num_subjects; i++ )
         {
            if( normalise_targets )
               random_effect_variances[ i ] /= rescale;
            if( random_effect_variances[ i ] > 0 )
	    {
	       random_effect_log_variances[ i ] = ave_data_block_size * log( random_effect_variances[ i ] );
               sum_log_random_effect_variances += random_effect_log_variances[ i ];
	       random_effect_posterior_precisions[ i ] = 1/(1 + 1/random_effect_variances[ i ]);
	    }
	    else
            {
               random_effect_log_variances[ i ] = 0;
	       random_effect_posterior_precisions[ i ] = 0;
            }
         }
}

void Regression::log_data()
{
   int i, j;
   if( log_targets )
      for( i = 0; i < n; i++ )
         for( j = 0; j < p_y; j++ )
            gsl_matrix_set( Y, i, j, log_transform( gsl_matrix_get( Y, i, j ), log_targets ) );
   if( log_inputs )
      for( i = 0; i < n; i++ )
         for( j = 0; j < p; j++ )
            gsl_matrix_set( X, i, j, log_transform( gsl_matrix_get( X, i, j ), log_inputs ) );
}

void Regression::logistic_transform_data()
{
   int i, j;
   if( logistic_transform_targets )
      for( i = 0; i < n; i++ )
	 for( j = 0; j < p_y; j++ )
	    gsl_matrix_set( Y, i, j, logistic_transform( gsl_matrix_get( Y, i, j ) ) );
   if( logistic_transform_predictors )
      for( i = 0; i < n; i++ )
	 for( j = 0; j < p; j++ )
	    gsl_matrix_set( X, i, j, logistic_transform( gsl_matrix_get( X, i, j ) ) );
}

double log_transform( double x, int log_transform_type )
{
   if( log_transform_type == 1 )
      return log( x );
   if( log_transform_type == 2 )
      return log( log( x ) );
//   if( log_transform_type == 3 )
   return log( x + 1 );
}

double inverse_log_transform( double x, int log_transform_type )
{
   if( log_transform_type == 1 )
      return exp( x );
   if( log_transform_type == 2 )
      return exp( exp( x ) );
//   if( log_transform_type == 3 )
   return exp( x ) - 1;
}

double logistic_transform( double x )
{
   return 1/(1+exp(-x));
}

double logit_transform( double x )
{
   return log(x/(1-x));
}

void Regression::difference_data( bool from_intercept )
{
   gsl_matrix* ystore = Y;
   gsl_matrix* xstore = X;
   int i, j;
   if( !data_block_sizes && data_block_size == 1 )
   {
      if( p_y > 1 )
      {
         p_y--;
         Y = gsl_matrix_alloc( n, p_y );
         num_differenced_points_stored = n;
         create_array( y_differenced_points, n, 1 );
         for( i = 0; i < n; i++ )
         {
            for( j = 0; j < p_y; j++ )
               gsl_matrix_set( Y, i, j, gsl_matrix_get( ystore, i, j + 1 ) - gsl_matrix_get( ystore, i, from_intercept? 0:j ) );
            y_differenced_points[ i ][ 0 ] = gsl_matrix_get( ystore, i, 0 );
         }
      }
      else
      {
         n--;
         Y = gsl_matrix_alloc( n, p_y );
         X = gsl_matrix_alloc( n, p );
         num_differenced_points_stored = 1;
         create_array( y_differenced_points, 1, 1 );
         create_array( x_differenced_points, 1, 1 );
         y_differenced_points[ 0 ][ 0 ] = gsl_matrix_get( ystore, 0, 0 );
         x_differenced_points[ 0 ][ 0 ] = gsl_matrix_get( xstore, 0, 0 );
         for( i = 0; i < n; i++ )
         {
            gsl_matrix_set( Y, i, 0, gsl_matrix_get( ystore, i + 1, 0 ) - gsl_matrix_get( ystore, from_intercept? 0:i, 0 ) );
            for( j = 0; j < p; j++ )
               gsl_matrix_set( X, i, j, gsl_matrix_get( xstore, i + 1, j ) - from_intercept? 0:gsl_matrix_get( xstore, i, j ) );
         }
      }
   }
   else
   {
      if( inputs_repeated_every )
      {
         if( inputs_repeated_every == data_block_sizes_sum )
            inputs_repeated_every -= num_data_block_sizes;
         else
            inputs_repeated_every = 0;
      }
      if( data_block_sizes )
      {
	 for( i = 0; i < num_data_block_sizes; i++ )
            data_block_sizes[ i ]--;
         data_block_size = data_block_sizes[ 0 ];
         data_block_sizes_sum -= num_data_block_sizes;
         ave_data_block_size--;
      }
      else
      {
         data_block_size--;
         ave_data_block_size = data_block_sizes_sum = data_block_size;
      }
      n = num_data_blocks / num_data_block_sizes * data_block_sizes_sum;
      Y = gsl_matrix_alloc( n, p_y );
      X = gsl_matrix_alloc( n, p );
      num_differenced_points_stored = num_data_blocks;
      create_array( y_differenced_points, num_data_blocks, p_y );
      create_array( x_differenced_points, num_data_blocks, p );
      int cum_dbs = 0, cum_dbs_new = 0;
      for( int ii = 0; ii < num_data_blocks; ii++ )
      {
         get_experiment_number( ii );
         for( j = 0; j < p_y; j++ )
            y_differenced_points[ ii ][ j ] = gsl_matrix_get( ystore, cum_dbs, j );
         for( j = 0; j < p; j++ )
            x_differenced_points[ ii ][ j ] = gsl_matrix_get( xstore, cum_dbs, j );
         for( i = 0; i < data_block_size; i++ )
         {
            for( j = 0; j < p_y; j++ )
               gsl_matrix_set( Y, cum_dbs_new + i, j, gsl_matrix_get( ystore, cum_dbs + i + 1, j ) - gsl_matrix_get( ystore, cum_dbs + from_intercept? 0:i, j ) );
            for( j = 0; j < p; j++ )
	       gsl_matrix_set( X, cum_dbs_new + i, j, gsl_matrix_get( xstore, cum_dbs + i + 1, j ) - (from_intercept? 0:gsl_matrix_get( xstore, cum_dbs + i, j )) );
         }
         cum_dbs += data_block_size + 1;
	 cum_dbs_new += data_block_size;
      }
   }
   gsl_matrix_free( ystore );
   if( X != xstore )
      gsl_matrix_free( xstore );
}

void Regression::append_test_points()
{
   num_test_points = test_grid_size;
   if( variable_to_predict_over != second_variable_to_predict_over )
      num_test_points *= test_grid_size;
   double** new_ys = 0;
   create_array( new_ys, num_test_points, p_y );
   int i, j, i2;
   for( i = 0; i < test_grid_size; i++ )
      for( i2 = 0; i2 < num_test_points / test_grid_size; i2++ )
         for( j = 0; j < p_y; j++ )
         {
            if( j == variable_to_predict_over )
               if( normalise_targets )
                  new_ys[ i + i2 * test_grid_size ][ j ] = (double) i / ( test_grid_size - 1 ) * ( y_max[ j ] - y_min[ j ] ) / y_scale[ j ] + y_min[ j ] / y_scale[ j ] - y_bar[ j ] ;
               else
                  new_ys[ i + i2 * test_grid_size ][ j ] = y_min[ j ] + (double) i / ( test_grid_size - 1 ) * ( y_max[ j ] - y_min[ j ] );
            else if( j == second_variable_to_predict_over )
               if( normalise_targets )
                  new_ys[ i + i2 * test_grid_size ][ j ] = (double) i2 / ( test_grid_size - 1 ) * ( y_max[ j ] - y_min[ j ] ) / y_scale[ j ] + y_min[ j ] / y_scale[ j ] - y_bar[ j ];
               else
                  new_ys[ i + i2 * test_grid_size ][ j ] = y_min[ j ] + (double) i2 / ( test_grid_size - 1 ) * ( y_max[ j ] - y_min[ j ] );
            else
               if( normalise_targets )
                  new_ys[ i + i2 * test_grid_size ][ j ] = y_bar[ j ];
               else
                  new_ys[ i + i2 * test_grid_size ][ j ] = ( y_max[ j ] + y_min[ j ] ) / 2;
         }
   gsl_matrix* Ycopy = Y;
   Y = gsl_matrix_alloc( n + num_test_points, p_y );
   for( j = 0; j < p_y; j++ )
   {
      for( i = 0; i < n; i++ )
         gsl_matrix_set( Y, i, j, gsl_matrix_get( Ycopy, i, j ) );
      for( i = 0; i < num_test_points; i++ )
         gsl_matrix_set( Y, n + i, j, new_ys[ i ][ j ] );
   }
   zap_array( new_ys, num_test_points, p_y );
   gsl_matrix_free( Ycopy );
   if( X )
   {
      create_test_grid( variable_to_predict_over, second_variable_to_predict_over );
//       double x_val;
//       double** new_xs = 0;
//       create_array( new_xs, num_test_points, p );
//       for( i = 0; i < test_grid_size; i++ )
//          for( i2 = 0; i2 < num_test_points / test_grid_size; i2++ )
//             for( j = 0; j < p; j++ )
//             {
//                if( j == variable_to_predict_over && x_max[ j ] - x_min[ j ] )
//                {
//                   x_val = (double) i / ( test_grid_size - 1 );
//                   if( normalise_predictors == 2 || !normalise_predictors )
//                      x_val = x_min[ j ] + ( x_max[ j ] - x_min[ j ] ) * x_val;
//                }
//                else if( j == second_variable_to_predict_over && x_max[ j ] - x_min[ j ] )
//                {
//                   x_val = (double) i2 / ( test_grid_size - 1 );
//                   if( normalise_predictors == 2 || !normalise_predictors )
//                      x_val = x_min[ j ] + ( x_max[ j ] - x_min[ j ] ) * x_val;
//                }
//                else
//                   x_val = x_bar[ j ];
//                new_xs[ i + i2 * test_grid_size ][ j ] = x_val;
//             }
      gsl_matrix* Xcopy = X;
      X = gsl_matrix_alloc( n + num_test_points, p );
      for( j = 0; j < p; j++ )
      {
         for( i = 0; i < n; i++ )
            gsl_matrix_set( X, i, j, gsl_matrix_get( Xcopy, i, j ) );
         for( i = 0; i < num_test_points; i++ )
            gsl_matrix_set( X, n + i, j, test_grid_points[ i ][ j ] );
      }
//      zap_array( new_xs, num_test_points, p );
      gsl_matrix_free( Xcopy );
   }
   n += num_test_points;
   num_data_blocks = n / data_block_sizes_sum * num_data_block_sizes;
  
   if( Subject_index )
   {
      int* sub_copy = new int[ num_subjects + 1 ];
      for( i = 0; i <= num_subjects; i++ )
         sub_copy[ i ] = Subject_index[ i ];
      delete [] Subject_index;
      Subject_index = new int[ num_subjects + 1 + num_test_points ];
      for( i = 0; i < num_subjects ; i++ )
         Subject_index[ i ] = sub_copy[ i ];
      for( i = 0; i < num_test_points; i++ )
         Subject_index[ num_subjects + i ] = Subject_index[ num_subjects + i - 1 ] + 1;
      num_subjects += num_test_points;
      Subject_index[ num_subjects ] = num_data_blocks;
      delete [] sub_copy;
   }
   else
      num_subjects += num_test_points;
}

void Regression::normalise_target_data_blocks()
{
   if( !y_block_norms )
   {
      y_block_norms = new double*[ num_data_blocks ];
      for( int i = 0; i < num_data_blocks; i++ )
	 y_block_norms[ i ] = new double[ p_y ];
   }
   int i, row_counter = 0;
   for( int ii = 0; ii < num_subjects; ii++ )
   {
      int start = get_subject_index( ii );
      int end = get_subject_index( ii + 1 );
      for( int i0 = start; i0 < end; i0++ )
      {
         get_experiment_number( i0 );
	 for( int j = 0; j < p_y; j++ )
	 {
	   double norm_j = 0;
	   for( i = row_counter; i < row_counter + data_block_size; i++ )
	   {
	      double y_ij = gsl_matrix_get( Y, i, j );
	      norm_j += y_ij * y_ij;
	   }
	   norm_j = sqrt( norm_j / data_block_size );
	   for( i = row_counter; i < row_counter + data_block_size; i++ )
	      *gsl_matrix_ptr( Y, i, j ) /= norm_j;
	   y_block_norms[ i0 ][ j ] = norm_j;
	 }
	 row_counter += data_block_size;
      }
   }

}


void Regression::create_design_matrix()
{
   D = X;
}

void Regression::create_subject_indices()
{
   if( strlen(Subject_Index_Filename) )
   {
      ifstream SubjectIndexFile( Subject_Index_Filename, ios::in );
      if( SubjectIndexFile )
      {
         gsl_matrix_int* dummy = scan_integer_data( SubjectIndexFile );
         int dum_rows = dummy->size1;
         int dum_cols = dummy->size2;
         num_subjects = dum_rows * dum_cols;
         Subject_index = new int[ num_subjects + 1 ];
         int i;
         for( i = 0; i < dum_rows; i++ )
            for( int j = 0; j < dum_cols; j++ )
               Subject_index[ i * dum_cols + j ] = gsl_matrix_int_get( dummy, i, j );
         gsl_matrix_int_free( dummy );
         if( Subject_index[ 0 ] == 1 )//assume accidentally not counting from zero
            for( i = 0; i < num_subjects; i++ )
               Subject_index[ i ]--;
         Subject_index[ num_subjects ] = num_data_blocks;
      }
   }
   if( !Subject_index )
   {
      if( num_data_blocks % subject_grouping )
      {
         cout << "Error, number of observations  (" << num_data_blocks << ") not a multiple of the number of grouped observations (" << subject_grouping << ")." << endl;
         exit(1);
      }
      else
         num_subjects = num_data_blocks / subject_grouping;
   }
}

int Regression::get_subject_index( int sub_number )
{
   if( Subject_index )
      return Subject_index[ sub_number ];
   return sub_number * subject_grouping;
}

void Regression::get_subject_row_range( int subject_num, int& start_row, int& end_row )
{
   start_row = get_subject_index( subject_num ) * data_block_sizes_sum / num_data_block_sizes;
   end_row = get_subject_index( subject_num + 1 ) * data_block_sizes_sum / num_data_block_sizes - 1;
}

gsl_matrix* Regression::get_targets( list<int> subject_list, int* exp_number )
{
   if( subject_list.empty() )
      return NULL;
   gsl_matrix* ys = 0;
   int counter = 0;
   list<int>::iterator iter;
   for( iter = subject_list.begin(); iter != subject_list.end(); iter++ )
   {
      int subject_num = *iter;
      int start = get_subject_index( subject_num );
      int end = get_subject_index( subject_num + 1 );
      for( int i0 = start; i0 < end; i0++ )
         if( !exp_number || get_experiment_number( i0 ) == *exp_number )
            counter += data_block_size;
   }
   ys = gsl_matrix_alloc( counter, p_y );
   counter = 0;
   for( iter = subject_list.begin(); iter != subject_list.end(); iter++ )
   {
      int subject_num = *iter;
      int start = get_subject_index( subject_num );
      int end = get_subject_index( subject_num + 1 );
      for( int i0 = start; i0 < end; i0++ )
         if( !exp_number || get_experiment_number( i0 ) == *exp_number )
         {
            for( int j = 0; j < p_y; j++ )
               for( int i = 0; i < data_block_size; i++ )
                  gsl_matrix_set( ys, counter + i, j, gsl_matrix_get( Y, i0 * data_block_size + i, j ) );
            counter += data_block_size;
         }
   }
   return ys;
}

double Regression::get_first_target( list<int> subject_list, int vbl )
{
   int subject_num = *subject_list.begin();
   int start = get_subject_index( subject_num );
   return gsl_matrix_get( Y, start * data_block_sizes_sum / num_data_block_sizes, vbl );
}

gsl_matrix* Regression::get_target_averages( list<int> subject_list, int* exp_number )
{
   if( data_block_sizes && exp_number )
      data_block_size = data_block_sizes[ *exp_number ];
   gsl_matrix* ys = gsl_matrix_calloc( data_block_size, p_y );
   int counter = 0;
   for( list<int>::iterator iter = subject_list.begin(); iter != subject_list.end(); iter++ )
   {
      int subject_num = *iter;
      int start = get_subject_index( subject_num );
      int end = get_subject_index( subject_num + 1 );
      int row_start = start * data_block_sizes_sum / num_data_block_sizes;
      for( int i0 = start; i0 < end; i0++ )
         if( !exp_number || get_experiment_number( i0 ) == *exp_number )
         {
            for( int i = row_start; i < ( i0 + 1 ) * data_block_size; i++ )
               for( int j = 0; j < p_y; j++ )
                  *gsl_matrix_ptr( ys, i - row_start, j ) += gsl_matrix_get( Y, i, j );
            counter++;
	    row_start += data_block_size;
         }
   }
   if( counter )
      gsl_matrix_scale( ys, 1.0 / (double)counter );
   return ys;
}

gsl_matrix* Regression::get_target_covariances( bool corr, bool residual_covariances )
{
   list<int> list_of_subjects;
   for( int i = 0; i < num_subjects; i++ )
      list_of_subjects.push_back( i );
   return get_target_covariances( list_of_subjects, corr, residual_covariances );
}

gsl_matrix* Regression::get_target_covariances( list<int> subject_list, bool corr, bool residual_covariances )
{
   int num_subs = subject_list.size();
   int ys_cols = p_y * data_block_sizes_sum * num_experiments / num_data_block_sizes;
   gsl_matrix* ys = gsl_matrix_alloc( num_subs, ys_cols );
   bool **miss_vals = 0;
   if( missing_values )
      create_array( miss_vals, num_subs, ys_cols );
   int row_num = 0;
   for( list<int>::iterator iter = subject_list.begin(); iter != subject_list.end(); iter++ )
   {
      int subject_num = *iter;
      int start = get_subject_index( subject_num );
      int end = get_subject_index( subject_num + 1 );
      int row_counter = start * data_block_sizes_sum / num_data_block_sizes;
      int counter = 0;
      for( int i0 = start; i0 < end; i0++ )
      {
         get_experiment_number( i0 );
         for( int i = 0; i < data_block_size; i++ )
            for( int j = 0; j < p_y; j++ )
            {
               gsl_matrix_set( ys, row_num, (counter + i) * p_y + j, gsl_matrix_get( residual_covariances ? residuals : Y, row_counter + i, j ) );
               if( missing_values )
                  miss_vals[ row_num ][ (counter + i) * p_y + j ] =  missing_values[ row_counter + i ][ j ];
            }
	 row_counter += data_block_size;
	 counter += data_block_size;
      }
      row_num++;
   }
   gsl_matrix* target_covariances = NULL;
   if( !corr )
      target_covariances = calculate_covariances( ys, miss_vals );
   else
      target_covariances = calculate_correlations( ys, miss_vals );
   gsl_matrix_free( ys );
   if( missing_values )
      zap_array( miss_vals, num_subs, p_y * data_block_sizes_sum );
   return target_covariances;
}


void Regression::calculate_design_matrix_column_sums()
{
   if( D_column_sums && (int)D_column_sums->size2 != num_design_cols )
   {
      gsl_matrix_free( D_column_sums );
      D_column_sums = 0;
   }
   if( !D_column_sums )
      D_column_sums = gsl_matrix_alloc( 1, num_design_cols );
   for( int j = 0; j < num_design_cols; j++ )
   {
      double dj = 0;
      for( int i = 0; i < ( inputs_repeated_every ? inputs_repeated_every : regression_n ); i++ )
         dj += gsl_matrix_get( D, i, j );
      gsl_matrix_set( D_column_sums, 0, j, dj );
   }
}

void Regression::create_experiment_numbers_vector()
{
   int i, j;
   if( strlen( Experiment_Number_Filename ) )
   {
      ifstream ExperimentNumberFile( Experiment_Number_Filename, ios::in );
      if( ExperimentNumberFile )
      {
         experiment_numbers = new int[ num_data_blocks ];
         for( i = 0; i < num_data_blocks; i++ )
            experiment_numbers[ i ] = 0;
         gsl_matrix_int* dummy = scan_integer_data( ExperimentNumberFile );
         int dum_rows = dummy->size1;
         int dum_cols = dummy->size2;
         for( i = 0; i < dum_rows; i++ )
            for( j = 0; j < dum_cols; j ++ )
            {
               if( i * dum_cols + j < num_data_blocks )//precaution, in case experimentnumberfile too big.
               {
                  experiment_numbers[ i * dum_cols + j ] = gsl_matrix_int_get( dummy, i, j );
                  if( num_experiments < experiment_numbers[ i * dum_cols + j ] + 1 )
                     num_experiments = experiment_numbers[ i * dum_cols + j ] + 1;
               }
            }
         gsl_matrix_int_free( dummy );
      }
      if( num_experiments > 1 )
         num_locations = num_experiments;
      else
         num_locations = p_y;
   }

   num_experiments_original = num_experiments;

   if( treat_experiments_as_different_dimensions && num_experiments == 1 )
       treat_experiments_as_different_dimensions = false;

   if( treat_experiments_as_different_dimensions )
   {
      X_original = X;
      int new_p = p * num_experiments;
      X = gsl_matrix_calloc( n, new_p );
      x_min = new double[ new_p ];
      x_max = new double[ new_p ];
      global_x_min = DBL_MAX;
      global_x_max = -DBL_MAX;
      for( i = 0; i < n; i++ )
	 for( j = 0; j < p; j++ )
	 {
	    double x_ij = gsl_matrix_get( X_original, i, j );
	    if( x_ij < global_x_min )
	       global_x_min = x_ij;
	    if( x_ij > global_x_max )
	       global_x_max = x_ij;
	 }
      for ( j = 0; j < new_p; j++ )
      {
	 for( i = 0; i < n; i++ )
	    gsl_matrix_set( X, i, j, global_x_min - 1 );
         x_min[ j ] = DBL_MAX;
         x_max[ j ] = -DBL_MAX;
      }
      int row_counter = 0;
      for( int i0 = 0; i0 < num_data_blocks; i0++ )
      {
         int experiment = get_experiment_number( i0 );
         int shift = experiment * p;
         for( i = row_counter; i < row_counter + data_block_size; i++ )
         {
            for( j = 0; j < p; j++ )
            {
               double x_ij = gsl_matrix_get( X_original, i, j );
               int j_new = shift + j;
               gsl_matrix_set( X, i, j_new , x_ij ); 
               if( x_ij < x_min[ j_new ] )
                  x_min[ j_new ] = x_ij;
               else if( x_ij > x_max[ j_new ] )
                  x_max[ j_new ] = x_ij;
            }
         }
         row_counter += data_block_size;
      }
      //      for ( j = 0; j < new_p; j++ )
      //     {
      //         if( x_min[ j ] < global_x_min )
      //            global_x_min = x_min[ j ];
      //         if( x_max[ j ] > global_x_max )
      //            global_x_max = x_max[ j ];
      //      }
      global_x_range = global_x_max - global_x_min;
      p = new_p;
      int min_inputs_repeated_every = data_block_sizes_sum;
      if( !data_block_sizes )
         min_inputs_repeated_every *= p;
      if( inputs_repeated_every && inputs_repeated_every < min_inputs_repeated_every )
         inputs_repeated_every = min_inputs_repeated_every;
   }
   if( random_effects_model || data_block_sizes )
      correlate_experiments = 1;
   if( num_experiments <= 1 || treat_experiments_as_different_dimensions )
   {
      weight_experiments = 0;
      treat_experiments_separately = 0;
      if( correlate_experiments && p_y > 1 )
         correlate_columns = true;
      correlate_experiments = 0;
   }
   if( correlate_experiments )
      treat_experiments_separately = 0;
   else if( !correlate_columns )
      experiment_location_dimension = 0;

   experiment_partition_size = 1;
   if( strlen( Experiment_Partition_Filename ) )
   {
      gsl_matrix_int* partition_mx = scan_integer_data( Experiment_Partition_Filename );
      int partition_size = partition_mx->size1;
      experiment_partition = new int[ partition_size ];
      experiment_partition_size = -INT_MAX;
      int min_set_number = INT_MAX;
      for( i = 0; i < partition_size; i++ )
      {
         int set_number = gsl_matrix_int_get( partition_mx, i, 0 );
         experiment_partition[ i ] = set_number;
         if( set_number > experiment_partition_size )
            experiment_partition_size = set_number;
         if( set_number < min_set_number )
            min_set_number = set_number;
      }
      gsl_matrix_int_free( partition_mx );
      for( i = 0; i < partition_size; i++ )
         experiment_partition[ i ] -= min_set_number;
      experiment_partition_size -= min_set_number - 1;
      create_zeroed_array( experiment_partition_sizes, experiment_partition_size );
      for( i = 0; i < partition_size; i++ )
	 experiment_partition_sizes[ experiment_partition[ i ] ]++;
      num_location_parameters = experiment_partition_size;
      if( experiment_location_dimension < num_location_parameters )
         experiment_location_dimension = num_location_parameters;
   }
   if( num_experiments > 1 && !treat_experiments_as_different_dimensions && strlen( Equivalent_Experiments_Filename ) )
   {
      ifstream EquivalentExpsFile( Equivalent_Experiments_Filename, ios::in );
      if( EquivalentExpsFile )
      {
         gsl_matrix_int* dummy = scan_integer_data( EquivalentExpsFile );
         int dum_rows = dummy->size1;
         int dum_cols = dummy->size2;
         equivalent_experiments = new int[ num_experiments ];
         for( i = 0; i < dum_rows; i++ )
            for( j = 0; j < dum_cols; j ++ )
               if( i * dum_cols + j < num_experiments )//precaution, in case equivalent_experimentsfile too big.
                  equivalent_experiments[ i * dum_cols + j ] = gsl_matrix_int_get( dummy, i, j );
         gsl_matrix_int_free( dummy );
      }
   }
   if( correlate_experiments && experiment_location_dimension == 0 )
      experiment_location_dimension = num_experiments;
   if( experiment_location_dimension > 0 )
   {
      gsl_matrix* input_locations = NULL;
      if( strlen( Experiment_Locations_Input_Filename ) )
      {
         input_locations = scan_data( Experiment_Locations_Input_Filename );
	 experiment_location_dimension = input_locations->size2;
      }
      gsl_matrix* input_parameters = NULL;
      if( strlen( Experiment_Parameters_Input_Filename ) )
      {
         input_parameters = scan_data( Experiment_Parameters_Input_Filename );
	 experiment_location_dimension = input_parameters->size2;
      }
      experiment_locations = new double*[ num_locations ];
      experiment_location_means = new double*[ num_locations ];
      if( num_location_parameters )
      {
         create_zeroed_array( experiment_location_parameters, num_location_parameters, experiment_location_dimension );
      }
      if( !experiment_location_parameters )
         experiment_location_parameters = experiment_location_means;
      create_zeroed_array( experiment_location_sums, experiment_partition_size, experiment_location_dimension );
      experiment_covariances = gsl_matrix_calloc( num_locations, num_locations );
      experiment_precisions = gsl_matrix_calloc( num_locations, num_locations );
      if( experiment_partition )
         for( i = 0; i < num_location_parameters; i++ )
            if( !input_parameters )
               experiment_location_parameters[ i ][ i ] = 1;
            else
               for( j = 0; j < experiment_location_dimension; j++ )
                  experiment_location_parameters[ i ][ j ] = gsl_matrix_get( input_parameters, i, j );
      for( i = 0; i < num_locations; i++ )
      {
         experiment_locations[ i ] = new double[ experiment_location_dimension ];
         if( !experiment_location_parameters || experiment_location_parameters == experiment_location_means )
            experiment_location_means[ i ] = new double[ experiment_location_dimension ];
         else
            experiment_location_means[ i ] = experiment_location_parameters[ experiment_partition[ i ] ];
         for( j = 0; j < experiment_location_dimension; j++ )
         {
            if( input_locations )
	    {
               if( j < (int)input_locations->size2 )
                  experiment_locations[ i ][ j ] = gsl_matrix_get( input_locations, i, j );
               else
                  experiment_locations[ i ][ j ] = 0;
	    }
            if( !input_locations )
            {
               if( !experiment_partition )
                  if( i != j )
                     experiment_locations[ i ][ j ] = 0;
                  else
                     experiment_locations[ i ][ j ] = 1;
               else
		  experiment_locations[ i ][ j ] = experiment_location_means[ i ][ j ];
            }
            if( !experiment_partition )
               experiment_location_means[ i ][ j ] = 0;
         }
	 if( !input_locations && experiment_location_dimension < num_locations && ( i >= experiment_location_dimension || experiment_partition ) )
	    experiment_locations[ i ][ 0 ] += (double)( i - experiment_location_dimension + 1 ) / ( num_locations - experiment_location_dimension + 1 );
      }
      if( input_locations )
         gsl_matrix_free( input_locations );
      if( input_parameters )
         gsl_matrix_free( input_parameters );
      experiment_signs = new int[ num_locations ];
      for( i = 0; i < num_locations; i++ )
         experiment_signs[ i ] = 1;
      experiment_var_multipliers = new double[ num_locations ];
      for( i = 0; i < num_locations; i++ )
         experiment_var_multipliers[ i ] = 1;
      if( experiment_location_parameters )
         sample_locations();
      find_experiment_location_sums();
   }
}

void Regression::find_experiment_sds()
{
   double** exp_means;
   if( !data_block_sizes )
      create_array( exp_means, num_experiments, data_block_size );
   else
      create_ragged_array( exp_means, num_experiments, data_block_sizes );
   if( !experiment_sds )
      experiment_sds = new double[ num_experiments ];
   if( !experiment_means )
      experiment_means = new double[ num_experiments ];
   int* exp_counts = new int[ num_experiments ];
   int i, j, counter = 0;
   for( i = 0; i < num_experiments; i ++ )
   {
      if( data_block_sizes )
         data_block_size = data_block_sizes[ i ];
      for( j = 0; j < data_block_size; j++ )
         exp_means[ i ][ j ] = 0;
      experiment_sds[ i ] = 0;
      experiment_means[ i ] = 0;
      exp_counts[ i ] = 0;
   }
   for( int i0 = 0; i0 < num_data_blocks; i0++ )
   {
      int experiment = get_experiment_number( i0 );
      if( equivalent_experiments )
	 experiment = equivalent_experiments[ experiment ];
      exp_counts[ experiment ]++;
      for( i = counter; i < counter + data_block_size; i++ )
         for( j = 0; j < p_y; j++ )
         {
            double y_ij = gsl_matrix_get( Y, i, j );
            exp_means[ experiment ][ i - counter ] += y_ij;
            experiment_sds[ experiment ] += y_ij * y_ij;
         }
      counter += data_block_size;
   }
   log_ave_experiments_sds = 0;
   for( i = 0; i < num_experiments; i++ )
      if( exp_counts[ i ] )
      {
         double n_i = exp_counts[ i ] * p_y;
         if( data_block_sizes )
            data_block_size = data_block_sizes[ i ];
         for( j = 0; j < data_block_size; j++ )
         {
            double exp_mean_ij = exp_means[ i ][ j ] / n_i;
            experiment_sds[ i ] -= n_i * exp_mean_ij * exp_mean_ij;
            experiment_means[ i ] += exp_means[ i ][ j ];
         }
         experiment_sds[ i ] /= ( n_i * (double)data_block_size - 1 );
         experiment_sds[ i ] = sqrt( experiment_sds[ i ] );
         experiment_means[ i ] /= ( n_i * (double)data_block_size - 1 );
         if( experiment_sds[ i ] > 0 )
            log_ave_experiments_sds += log( experiment_sds[ i ] );
      }
   log_ave_experiments_sds /= num_experiments;
   zap_array( exp_means, num_experiments, data_block_size );
   delete [] exp_counts;
}

void Regression::standardize_targets_by_experiment()
{
   int counter = 0;
   for( int i0 = 0; i0 < num_data_blocks; i0++ )
   {
      int experiment = get_experiment_number( i0 );
      if( equivalent_experiments )
	 experiment = equivalent_experiments[ experiment ];
      for( int i = counter; i < counter + data_block_size; i++ )
         for( int j = 0; j < p_y; j++ )
         {
            double y_ij = gsl_matrix_get( Y, i, j );
            gsl_matrix_set( Y, i, j, (y_ij - experiment_means[ experiment ] ) / experiment_sds[ experiment ] );
         }
      counter += data_block_size;
   }
}

int Regression::get_experiment_number( int obsn_num )
{
   int exp_num = 0;
   if( experiment_numbers )
      exp_num = experiment_numbers[ obsn_num ];
   else if( num_experiments > 1 )
      exp_num = obsn_num % num_experiments;
   if( data_block_sizes )
      data_block_size = data_block_sizes[ exp_num ];
   return exp_num;
}

double Regression::get_experiment_distance( int e1, int e2 )
{
   double d_j, d = 0;
   for( int j = 0; j < experiment_location_dimension; j++ )
   {
      d_j = experiment_locations[ e1 ][ j ] - experiment_locations[ e2 ][ j ];
      d += d_j * d_j;
   }
   return d;
}

double Regression::get_experiment_dot_product( int e1, int e2 )
{
   double d = 0;
   for( int j = 0; j < experiment_location_dimension; j++ )
      d += experiment_locations[ e1 ][ j ] * experiment_locations[ e2 ][ j ];
   return d;
}

double Regression::get_experiment_location_norm( int e1 )
{
   return get_experiment_dot_product( e1, e1 );
}

double Regression::get_experiment_location_squared_distance_from_mean( int e1 )
{
   if( !experiment_location_means )
      return get_experiment_location_norm( e1 );
   double d_j, d = 0;
   for( int j = 0; j < experiment_location_dimension; j++ )
   {
      d_j = experiment_locations[ e1 ][ j ];
      if( !kriging_experiment_correlation )
         d_j *= experiment_signs[ e1 ];
      d_j -= experiment_location_means[ e1 ][ j ];
      d += d_j * d_j;
   }
   return d;
}

double Regression::get_experiment_correlation( int e1, int e2 )
{
   if( e1 == e2 )
   {
      gsl_matrix_set( experiment_covariances, e1, e1, 1 );
      return 1;
   }
   double corr = 0, dist = 0;
   int sign = experiment_signs[ e1 ] * experiment_signs[ e2 ];
   if( kriging_experiment_correlation )
   {
      dist = get_experiment_distance( e1, e2 );
      if( experiment_closeness_power )
	 dist = pow( dist , *experiment_closeness_power );
      corr = exp( -dist ) * sign;
   }
   else
      corr = get_experiment_dot_product( e1, e2 ) / sqrt( get_experiment_location_norm( e1 ) * get_experiment_location_norm( e2 ) ) * sign;
   gsl_matrix_set( experiment_covariances, e1, e2, corr );
   gsl_matrix_set( experiment_covariances, e2, e1, corr );
   return corr;
}

void Regression::get_experiment_correlations()
{
   get_experiment_covariances();
   covariances_to_correlations( experiment_covariances );
}

double Regression::get_experiment_covariance( int e1, int e2 )
{
   double cov;
   if( kriging_experiment_correlation )
      if( e1 == e2 )
          cov = experiment_var_multipliers[ e1 ];
      else
          cov = get_experiment_correlation( e1, e2 ) * sqrt( experiment_var_multipliers[ e1 ] * experiment_var_multipliers[ e2 ] );
   else //assuming Wishart V_, if inverse Wishart shouldnt' be calling this
      cov = get_experiment_dot_product( e1, e2 );
   gsl_matrix_set( experiment_covariances, e1, e2, cov );
   if( e1 != e2 )
      gsl_matrix_set( experiment_covariances, e2, e1, cov );
   return cov;
}

void Regression::get_experiment_precisions()
{
   double prec_ij = 0;
   int i, j;
   for( i = 0; i < num_locations; i++ )
      for( j = 0; j <= i; j++ )
      {
         prec_ij = get_experiment_dot_product( i, j );
         gsl_matrix_set( experiment_precisions, i, j, prec_ij );
         if( i != j )
            gsl_matrix_set( experiment_precisions, j, i, prec_ij );
      }
   my_invert( experiment_precisions, experiment_covariances );
}

void Regression::update_experiment_precisions( int e1 )
{
   double prec_ij = 0;
   int i;
   for( i = 0; i < num_locations; i++ )
   {
      prec_ij = get_experiment_dot_product( i, e1 );
      gsl_matrix_set( experiment_precisions, i, e1, prec_ij );
      if( i != e1 )
         gsl_matrix_set( experiment_precisions, e1, i, prec_ij );
   }
   my_invert( experiment_precisions, experiment_covariances );
//   gsl_matrix_memcpy( experiment_covariances, experiment_precisions );
}

void Regression::get_experiment_covariances()
{
   if( !kriging_experiment_correlation )
   {
      get_experiment_precisions();
      return;
   }
   for( int i = 0; i < num_locations; i++ )
      for( int j = 0; j <= i; j++ )
         get_experiment_covariance( i, j );
}

void Regression::reverse_experiment_sign( int e1 )
{
   if( e1 > 0 )
      experiment_signs[ e1 ] = -experiment_signs[ e1 ];
   else
      for( int i = 1; i < num_locations; i++ )
         experiment_signs[ i ] = -experiment_signs[ i ];
}

gsl_matrix* Regression::calculate_correlation_by_experiment( gsl_matrix* mx )
{
   int experiment, mx_rows = mx->size1, mx_cols = mx->size2;
   int reshaped_mx_rows = mx_rows / num_locations * mx_cols;
   int* counter = new int[ num_locations ];
   int i;
   for( i = 0; i < num_locations; i++ )
      counter[ i ] = 0;
   gsl_matrix* reshaped_mx = gsl_matrix_alloc( reshaped_mx_rows, num_locations );
   for( i = 0; i < mx_rows; i++ )
   {
      if( num_experiments > 1 || !correlate_columns )
         experiment = get_experiment_number( i );
      else
         experiment = i % p_y;
      for( int j = 0; j < mx_cols; j++ )
	 if( counter[ experiment ] + j < reshaped_mx_rows )
	    gsl_matrix_set( reshaped_mx, counter[ experiment ] + j , experiment, gsl_matrix_get( mx, i, j ) );
      counter[ experiment ] += mx_cols;
   }
   delete [] counter;
   gsl_matrix* corr_mx = calculate_correlations( reshaped_mx );
   gsl_matrix_free( reshaped_mx );
   return corr_mx;
}

void Regression::experiment_covariances_to_distances( gsl_matrix* mx )
{
   if( kriging_experiment_correlation )
      for( int i = 0; i < num_locations; i++ )
      {
         gsl_matrix_set( mx, i, i, 0 );
         for( int j = 0; j < i; j++ )
         {
            double dist = -log( fabs( gsl_matrix_get( mx, i, j ) ) / sqrt( experiment_var_multipliers[ i ] * experiment_var_multipliers[ j ] ) );
	    if( experiment_closeness_power )
	      dist = pow( dist , 1 / *experiment_closeness_power );
            gsl_matrix_set( mx, i, j, dist );
            gsl_matrix_set( mx, j, i, dist );
         }
      }
   else
      gsl_matrix_scale( mx, -1 );
}

void Regression::experiment_covariances_to_locations( gsl_matrix* mx )
{
   if( kriging_experiment_correlation )
      experiment_mds_correlations_to_locations( mx );
   else
   {
      gsl_matrix* mx_inv = gsl_matrix_alloc( mx->size1, mx->size2 );
      my_invert( mx, mx_inv );
      experiment_cholesky_correlations_to_locations( mx_inv );
      gsl_matrix_free( mx_inv );
   }
   find_experiment_location_sums();
   reset_experiment_location_parameters();
}

void Regression::experiment_mds_correlations_to_locations( gsl_matrix* mx )
{
   int i, j;
   if( experiment_signs )
   {
      for( i = 0; i < num_locations; i++ )//compare with the middle experiment, assuming this is not the control with perhaps low correlation with the other experiments..
         experiment_signs[ i ] = gsl_matrix_get( mx, num_locations/2, i ) < 0 ? -1 : 1;
   }
   if( experiment_var_multipliers )
      for( i = 0; i < num_locations; i++ )
         experiment_var_multipliers[ i ] = gsl_matrix_get( mx, i, i );

   experiment_covariances_to_distances( mx );
   double* row_sums = new double[ num_locations ];
   double total_sum = 0;
   for( i = 0; i < num_locations; i++ )
   {
      row_sums[ i ] = 0;
      for( int j = 0; j < num_locations; j++ )
         row_sums[ i ] += gsl_matrix_get( mx, i, j );
      row_sums[ i ] /= num_locations;
      total_sum += row_sums[ i ];
   }
   total_sum /= num_locations;
   for( i = 0; i < num_locations; i++ )
      for( j = 0; j < num_locations; j++ )
         gsl_matrix_set( mx, i, j, -.5 * ( gsl_matrix_get( mx, i, j ) - row_sums[ i ] - row_sums[ j ] + total_sum ) );
   delete [] row_sums;
   singular_value_decomposition( mx );
   for( j = 0; j < num_locations; j++ )
   {
      double rt_eval = sqrt( fabs( gsl_vector_get( eigen_values, j ) ) );
      for( i = 0; i < num_locations; i++ )
         gsl_matrix_set( mx, i, j, gsl_matrix_get( eigen_vectors, i, j ) * rt_eval );
   }
   for( i = 0; i < num_locations; i++ )
      for( j = 0; j < experiment_location_dimension; j++ )
         if( j < (int)mx->size2 )
            experiment_locations[ i ][ j ] = gsl_matrix_get( mx, i, j );
         else
            experiment_locations[ i ][ j ] = 0;
}

void Regression::reset_experiment_location_parameters()
{
   for( int i = 0; i < experiment_partition_size; i++ )
   {
      int set_size = experiment_partition_sizes ? experiment_partition_sizes[i] : num_experiments;
      double scaling_value = 1.0/set_size;
      copy_scaled_array( experiment_location_parameters[ i ], experiment_location_sums[ i ], scaling_value, experiment_location_dimension );
   }
}

void Regression::experiment_cholesky_correlations_to_locations( gsl_matrix* mx )
{
   gsl_linalg_cholesky_decomp( mx );
   for( int i = 0; i < num_locations; i++ )
      for( int j = 0; j < experiment_location_dimension; j++ )
         if( j <=i )
            experiment_locations[ i ][ j ] = gsl_matrix_get( mx, i, j );
         else
            experiment_locations[ i ][ j ] = 0;
}

void Regression::sample_locations()
{
    for( int i = 0; i < num_locations; i++ )
       for( int j = 0; j < experiment_location_dimension; j++ )
          experiment_locations[ i ][ j ] = experiment_location_means[ i ][ j ] + gsl_ran_gaussian( rn, 1 / sqrt( experiment_parameter_precision ) );
}

void Regression::find_experiment_location_sums()
{
  zero_array( experiment_location_sums, experiment_partition_size, experiment_location_dimension );
  for( int i = 0; i < num_locations; i++ )
     add_array( experiment_location_sums[ (experiment_partition_size > 1 ) ? experiment_partition[ i ]: 0 ], experiment_locations[ i ], experiment_location_dimension );
}

void Regression::delete_missing_points()
{
   if( !num_obsns_missing_values )
      return;
   if( print_to_screen )
     cout << "Deleting all " << num_obsns_missing_values << " observations with missing values" << endl;
   int i, j;
   if( !deleted_missing_points && num_obsns_missing_values )
   {
      deleted_missing_points = new double*[ num_obsns_missing_values ];
      for( i = 0; i < num_obsns_missing_values; i++ )
	 deleted_missing_points[ i ] = new double[ p ];
   }
   X_original = X;
   Y_original = Y;
   Subject_index_original = Subject_index;
   experiment_numbers_original = experiment_numbers;
   int* temp_exp_numbers = 0;
   if( num_experiments > 1 )
   {
      int cum_dbs = 0;
      temp_exp_numbers = new int[ n ];
      for( int ii = 0; ii < num_data_blocks; ii++ )
      {
         int exp_num = get_experiment_number( ii );
         for( i = 0; i < data_block_size; i++ )
	    temp_exp_numbers[ cum_dbs + i ] = exp_num;
         cum_dbs += data_block_size;
      }
   }

   n_original = n;
   n -= num_obsns_missing_values;
   Y = gsl_matrix_alloc( n, p_y );
   X = gsl_matrix_alloc( n, p );
   if( !Subject_index )
   {
      Subject_index = new int[ num_subjects + 1 ];
      for( int i = 0; i <= num_subjects; i++ )
	 Subject_index[ i ] = i * data_block_sizes_sum;
   }
   else
   {
      Subject_index = new int[ num_subjects + 1 ];
      copy_array( Subject_index, Subject_index_original, num_subjects + 1 );
   }
   if( num_experiments > 1 )
      experiment_numbers = new int[ n ];
   int counter = 0, missing_counter = 0;
   int subject = 0;
   for( i = 0; i < n_original; i++ )
   {
      bool missing_i = false;
      j = 0;
      while( !missing_i && j < p_y )
	 if( missing_values[ i ][ j++ ] )
	   missing_i = true;
      if( !missing_i )
      {
         for( j = 0; j < p_y; j++ )
            gsl_matrix_set( Y, counter, j, gsl_matrix_get( Y_original, i, j ) );
         for( j = 0; j < p; j++ )
            gsl_matrix_set( X, counter, j, gsl_matrix_get( X_original, i, j ) );
	 if( experiment_numbers )
	    experiment_numbers[ counter ] = temp_exp_numbers[ i ];
	 counter++;
      }
      else
      {	
	 for( j = 0; j < p; j++ )
	    deleted_missing_points[ missing_counter ][ j ] = gsl_matrix_get( X_original, i, j );
	 missing_counter++;
      }
      if( subject <= num_subjects && i == Subject_index[ subject + 1 ] )
      {
	 subject++;
	 Subject_index[ subject ] -= missing_counter;
      }
   }
   Subject_index[ num_subjects ] -= missing_counter;

//   for( i = 0; i < n_original; i++ )
//     delete [] missing_values[ i ];
//   delete [] missing_values;
//   missing_values = 0;
   inputs_repeated_every = 0;
   data_block_size_original = data_block_size;
   if( data_block_sizes )
   {
      data_block_sizes_original = data_block_sizes;
      data_block_sizes = 0;
   }
   num_data_block_sizes_original = num_data_block_sizes;
   num_data_block_sizes = 1;
   data_block_sizes_sum_original = data_block_sizes_sum;
   data_block_size = data_block_sizes_sum = 1;
   ave_data_block_size = 1;
   num_data_blocks = n;
   if( num_experiments > 1 )
      delete [] temp_exp_numbers;
}

void Regression::bootstrap_data()
{
   gsl_matrix *Xcopy = X, *Ycopy = Y, *xs, *ys;
   double *random_effect_variances_copy = random_effect_variances;
   if( random_effect_variances )
      random_effect_variances = new double [replace_with_bootstrap_samples];
   double *random_effect_posterior_precisions_copy = random_effect_posterior_precisions;
   if( random_effect_posterior_precisions )
      random_effect_posterior_precisions = new double [replace_with_bootstrap_samples];
   int* new_sub_ind = 0;
   if( Subject_index )
   {
      new_sub_ind = new int[ replace_with_bootstrap_samples + 1 ];
      new_sub_ind[ 0 ] = 0;
   }
   int sample_sub = 0, sub_start = 0, sub_end = 0;
   int i, j;
   for( int sample = 0; sample < replace_with_bootstrap_samples; sample++ )
   {
      sample_sub = rand() % num_subjects;//static_cast<int>( gsl_ran_flat( rn, 0, num_subjects ) );
      if( random_effect_variances )
         random_effect_variances[ sample ] = random_effect_variances_copy[ sample_sub ];
      if( random_effect_posterior_precisions )
	 random_effect_posterior_precisions[ sample ] = random_effect_posterior_precisions_copy[ sample_sub ];
      get_subject_row_range( sample_sub, sub_start, sub_end );
      int num_rows = sub_end - sub_start + 1;
      xs = gsl_matrix_alloc( num_rows, p );
      ys = gsl_matrix_alloc( num_rows, p_y );
      for( i = 0; i < num_rows; i++ )
      {
         for( j = 0; j < p; j++ )
            gsl_matrix_set( xs, i, j, gsl_matrix_get( Xcopy, sub_start + i, j ) );
         for( j = 0; j < p_y; j++ )
            gsl_matrix_set( ys, i, j, gsl_matrix_get( Ycopy, sub_start + i, j ) );
      }
      if( sample == 0 )
      {
         X = xs;
         Y = ys;
      }
      else
      {
         matrix_join( X, xs, 0 );
         matrix_join( Y, ys, 0 );
         gsl_matrix_free( xs );
         gsl_matrix_free( ys );
      }
      if( Subject_index )
         new_sub_ind[ sample + 1 ] = new_sub_ind[ sample ] + Subject_index[ sample_sub + 1 ] - Subject_index[ sample_sub ];
   }
   gsl_matrix_free( Xcopy );
   gsl_matrix_free( Ycopy );
   regression_n = n = X->size1;
   regression_num_data_blocks = num_data_blocks = n / data_block_sizes_sum * num_data_block_sizes;
   regression_num_subjects = num_subjects = replace_with_bootstrap_samples;
   if( Subject_index )
   {
      switch_values( Subject_index, new_sub_ind );
      delete [] new_sub_ind;
   }
   zap_vector( random_effect_variances_copy );
   zap_vector( random_effect_posterior_precisions_copy );
}

void Regression::thin_the_data()
{
   for( int i = 0; i < num_subjects; i++ )
   {
      int start_row, end_row;
      get_subject_row_range( i, start_row, end_row );
      for( int j = start_row; j <= end_row; j ++ )
         if( ( j - start_row ) % thin_data )
            delete_row_range( j, j );
   }
}

void Regression::switch_subject_and_expt_structures()
{
    switch_values( Subject_index_original, Subject_index );      
    switch_values( data_block_size_original, data_block_size );
    switch_values( data_block_sizes_original, data_block_sizes );
    switch_values( num_data_block_sizes_original, num_data_block_sizes );
    switch_values( data_block_sizes_sum_original, data_block_sizes_sum );
    switch_values( num_experiments_original, num_experiments );
    switch_values( experiment_numbers_original, experiment_numbers );
}

int Regression::get_number_of_design_variables()
{
  return D->size2;
}

double Regression::calculate_log_prior()
{
  log_prior = 0;
  return log_prior;
}

double Regression::calculate_log_posterior()
{
  log_posterior = calculate_log_likelihood() + calculate_log_prior();
  return log_posterior;
}

double* Regression::get_x_min()
{
  return x_min;
}

double* Regression::get_x_max()
{
  return x_max;
}

double* Regression::get_x_bar()
{
  return x_bar;
}

double Regression::function_evaluate( double* v, int dim_v )
{
   return v[ 0 ];
}

double Regression::integral_function_evaluate( double* v, int integral_order, int integral_component )
{
   return v[ 0 ] * v[ 0 ] / 2.0;
}

double Regression::function_variance_evaluate( double* v )
{
   double fn_var_val = 0;
   for( int j = 0; j < p; j++ )
      fn_var_val += v[ j ] * v[ j ];
   return fn_var_val;
}

void Regression::find_unique_data_points()
{
   bool component_at_a_time = true;
   list<int> unique_list;
   list<int> unique_component_list;
   list<int>::iterator iter, iter_comp;
   int i, j, j_start = 0, j_end = p;
   unique_data_points = new double*[ p ];
   create_zeroed_array( num_unique_data_points, p );
   int total_num_unique_data_points = 0;
   double xval;
   if( component_at_a_time ) 
     j_end = 1;
   while( j_end <= p )
   {

      int row_counter = 0;
      for( int ii = 0; ii < num_subjects; ii++ )
      {
	 int start = get_subject_index( ii );
	 int end = get_subject_index( ii + 1 );
	 for( int i0 = start; i0 < end; i0++ )
	 {
	    int exp_num = get_experiment_number( i0 );
	    for( i = row_counter; i < row_counter + data_block_size; i++ )
	    {
	      j = j_start;
	      set_active_dimension( j );
	    if( !treat_experiments_as_different_dimensions || j == exp_num )
	    {
	      bool replica = false;
	      do{
		replica = false;
		if( i < regression_n )
		  xval = gsl_matrix_get( X, i, j );
		else
                  xval = deleted_x_point[ j ];
	       //              bool not_smallest = normalise_predictors == 1 ? xval > 0 : ( xval - x_min[ j ] ) > 0;
	       //               bool not_largest = normalise_predictors == 1 ? ( 1 - xval ) != 0 : ( x_max[ j ] - xval ) != 0;
	       //               bool not_null = normalise_predictors == 1 ? xval >= 0 : ( xval - x_min[ j ] ) >= 0;
	       //               if( not_null && ( power_max < 0 || two_sided_splines || ( ( power_max > 0 || not_smallest ) && ( !power_min || ( ( full_1way_model_sign != -1 && not_largest ) || ( full_1way_model_sign == -1 && not_smallest ) ) ) ) ) )
		{
		   iter = unique_list.begin();
		   iter_comp = unique_component_list.begin();
		   while( iter != unique_list.end() && !replica )
		   {
		     if( gsl_matrix_get( X, *iter, j ) == xval  && *iter_comp == j )
		        replica = true;
		     ++iter;
		     ++iter_comp;
		   }
		}
	       //               else
	       //                  replica = true;
	      }while( replica && ++j < j_end );
         
	      if( !replica )
	      {
		 num_unique_data_points[ j_start ]++;
		 unique_list.push_back( i );
		 unique_component_list.push_back( j );
	      }
	    }
	    }
	    row_counter += data_block_size;
	 }
      }
      j_start++;
      j_end++;
   }
      
   iter = unique_list.begin();
   for( j = 0; j < p; j++ )
   {
      total_num_unique_data_points += num_unique_data_points[ j ];
      unique_data_points[ j ] = new double[ num_unique_data_points[ j ] ];
      for( i = 0; i < num_unique_data_points[ j ]; i++ )
      {
	 int row_num = *iter;
	 if( row_num < regression_n )
	    unique_data_points[ j ][ i ] = gsl_matrix_get( X, row_num, j );
	 else
	    unique_data_points[ j ][ i ] = deleted_x_point[ j ];
	 //	 cout << j << " " << i << " "<< unique_data_points[ j ][ i ] << endl;
	 ++iter;
      }
   }
}

void Regression::create_test_grid( int col1, int col2 )
{
   double x_val;
   double x_min1 = 0, x_min2 = 0, x_max1 = 1, x_max2 = 1;
   int i1, i2, j;
   zap_array( test_grid_points, num_test_points, p );
   if( predict_on_data )
   {
      num_test_points = test_grid_size = num_unique_data_points[ col1 ];
      if( col2 > 0 && col1 != col2 )
	 num_test_points *= num_unique_data_points[ col2 ];
   }
   else
   {
      num_test_points = test_grid_size;
      if( col2 > 0 && col1 != col2 )
	 num_test_points *= test_grid_size;

      if( normalise_predictors == 2 || !normalise_predictors )
	{
	  if( treat_experiments_as_different_dimensions )
	    {
	      x_min1 = x_min2 = global_x_min;
	      x_max1 = x_max2 = global_x_max;
	    }
	  else
	    {
	      x_min1 = x_min[ col1 ];
	      x_max1 = x_max[ col1 ];
	      x_min2 = x_min[ col2 ];
	      x_max2 = x_max[ col2 ];
	    }
	}
      if( log_inputs )
      {
         x_min1 = inverse_log_transform( x_min1, log_inputs );
         x_max1 = inverse_log_transform( x_max1, log_inputs );
         x_min2 = inverse_log_transform( x_min2, log_inputs );
         x_max2 = inverse_log_transform( x_max2, log_inputs );
      }
      else if( logistic_transform_predictors )
      {
	 x_min1 = logit_transform( x_min1 + 0.001 );
	 x_max1 = logit_transform( x_max1 - 0.001 );
	 x_min2 = logit_transform( x_min2 );
	 x_max2 = logit_transform( x_max2 );
      }
   }

   create_array( test_grid_points, num_test_points, p );
   for( i1 = 0; i1 < test_grid_size; i1++ )
      for( i2 = 0; i2 < num_test_points / test_grid_size; i2++ )
         for( j = 0; j < p; j++ )
         {
            if( j == col1 && x_max1 > x_min1 )
            {
	       if( predict_on_data )
		  x_val = unique_data_points[ j ][ i1 ];
	       else
	       {
		  x_val = (double) i1 / ( test_grid_size - 1 );
		  if( normalise_predictors == 2 || !normalise_predictors || logistic_transform_predictors )
		     x_val = x_min1 + ( x_max1 - x_min1 ) * x_val;
		  if( log_inputs )
                     x_val = log_transform( x_val, log_inputs );
		  else if( logistic_transform_predictors )
		     x_val = logistic_transform( x_val );
	       }
            }
            else if( j == col2 && x_max2 > x_min2 )
            {
  	       if( predict_on_data )
		  x_val = unique_data_points[ j ][ i2 ];
	       else
	       {
		  x_val = (double) i2 / ( test_grid_size - 1 );
		  if( normalise_predictors == 2 || !normalise_predictors )
		     x_val = x_min2 + ( x_max2 - x_min2 ) * x_val;
		  if( log_inputs )
		     x_val = log_transform( x_val, log_inputs );
		  else if( logistic_transform_predictors )
		     x_val = logistic_transform( x_val );
	       }
            }
            else
               x_val = 0;//x_bar[ j ];
            test_grid_points[ i1 + i2 * test_grid_size ][ j ] = x_val;
         }
//    int col = col1;
//    if( p > 1 )
//       for( int j = 0; j < p; j++ )
//          if( j != col )
//             for( int i = 0; i < test_grid_size; i++ )
//                test_grid_points[ i ][ j ] = x_bar[ j ];
//    double x_col_min = 0, x_col_max = 1, x_col_range = 1;
//    if( normalise_predictors == 2 || !normalise_predictors )
//    {
//       x_col_min = x_min[ col ];
//       x_col_max = x_max[ col ];
//    }
//    if( log_inputs )
//    {
//       x_col_min = exp( x_col_min );
//       x_col_max = exp( x_col_max );
//    }
//    x_col_range = x_col_max - x_col_min;
//    for( int i = 0; i < test_grid_size; i++ )
//    {
//       double x_i = (double)( i )/(double)( test_grid_size - 1);
//       test_grid_points[ i ][ col ] = x_col_min + x_col_range * x_i;
//       if( log_inputs )
//          test_grid_points[ i ][ col ] = log( test_grid_points[ i ][ col ] );
//    }
}

void Regression::make_grid_predictions( int col, int ycol )
{
   if( test_grid_size <= 0 )
      return;
//   if( !test_grid_points )
   set_active_dimension( col );
   create_test_grid( col );
   if( predictions_mx_rows != test_grid_size || predictions_mx_cols != 3 )
   {
      zap_array( predictions_mx, predictions_mx_rows, predictions_mx_cols );
      predictions_mx_rows = test_grid_size;
      predictions_mx_cols = 3;
   }
   predictions_update_cols = predictions_mx_cols;
   prediction_fn_value_col = 1;
   if( !predictions_mx )
      create_array( predictions_mx, predictions_mx_rows, predictions_mx_cols );
   current_regression_fn_val = 0;
   double fn_sd = 0;
   for( int i = 0; i < test_grid_size; i++ )
   {
      double* xx = test_grid_points[ i ];
      current_regression_fn_val = function_evaluate( xx, p );
      fn_sd = sqrt( function_variance_evaluate( xx ) );
      if( standardize_experiments && experiment_means && experiment_sds )
      {
         current_regression_fn_val = current_regression_fn_val * experiment_sds[ current_experiment ] + experiment_means[ current_experiment ];
         fn_sd *= experiment_sds[ current_experiment ];
      }
      if( normalise_targets )
      {
         current_regression_fn_val = ( current_regression_fn_val + y_bar[ ycol ] ) * y_scale[ ycol ];//then multiply by scale and add on subtracted mean value (which needs to be rescaled).
         fn_sd *= y_scale[ ycol ];
      }
      predictions_mx[ i ][ 0 ] = current_regression_fn_val - 2 * fn_sd;
      predictions_mx[ i ][ 1 ] = current_regression_fn_val;
      predictions_mx[ i ][ 2 ] = current_regression_fn_val + 2 * fn_sd;
   }
}

void Regression::write_predictions( ofstream *any_stream )
{
   if( !predictions_mx )
   {
      make_predictions();
      if( !predictions_mx )
         return;
   }
   if( !any_stream )
   {
      if( strlen( PredictionFilename ) )
      {
         ofstream PredStream( PredictionFilename, ios::out );
         Regression::write_predictions( &PredStream );
      }
      return;
   }
   for( int i = 0; i < predictions_mx_rows; i++ )
   {
      for( int j = 0; j < predictions_mx_cols; j++ )
         *any_stream << predictions_mx[ i ][ j ] << " ";
      *any_stream << endl;
   }
}

void Regression::write_residuals( ofstream *any_stream )
{
   if( !residuals )
      return;
   if( !any_stream )
   {
      if( strlen( Residuals_Filename ) )
      {
         ofstream ResidStream( Residuals_Filename, ios::out );
         write_residuals( &ResidStream );
      }
      return;
   }
   for( int i = 0; i < n; i++ )
   {
      for( int j = 0; j < p_y; j++ )
         *any_stream << gsl_matrix_get( residuals, i, j ) << " ";
      *any_stream << endl;
   }
}

void Regression::write_experiment_locations_to_file( bool principal_components )
{
//    if( !kriging_experiment_correlation && experiment_locations )
//    {
//       gsl_matrix* exp_correls = get_experiment_correlations();
//       experiment_mds_correlations_to_locations( exp_correls );
//       gsl_matrix_free( exp_correls );
//    }
   if( !experiment_covariances )
      return;
   if( !mean_experiment_covariances )
      get_experiment_covariances();
   else
      gsl_matrix_memcpy( experiment_covariances, mean_experiment_covariances );

   if( !strlen( Experiment_Correlations_Filename ) )
   {
      char* new_file = Experiment_Correlations_Filename;
      strcpy( new_file, TargetFilename );
      string insert = "_exp_correlations_";
      insert_in_string_before_dot( new_file, &insert );
   }

   int i, j;
   if( strlen( Experiment_Correlations_Filename ) )
   {
      ofstream ExpCorStream( Experiment_Correlations_Filename, ios::out );
      if( ExpCorStream )
         for( i = 0; i < num_locations; i++ )
         {
            double ev_i =  gsl_matrix_get( experiment_covariances, i, i );
            for( j = 0; j < num_locations; j++ )
            {
               double ev_j =  gsl_matrix_get( experiment_covariances, j, j );
	       ExpCorStream << gsl_matrix_get( experiment_covariances, i, j ) / sqrt( ev_i * ev_j ) << " ";
            }
            ExpCorStream << endl;
         }
   }
   if( !experiment_locations )
      return;

   if( !strlen( Experiment_Locations_Filename ) )
   {
      char* new_file = Experiment_Locations_Filename;
      strcpy( new_file, TargetFilename );
      string insert = "_exp_locations_";
      insert_in_string_before_dot( new_file, &insert );
   }
   
   if( experiment_location_dimension <= 2 || experiment_location_dimension > num_locations )
      principal_components = false;
   if( !kriging_experiment_correlation )
      for( i = 0; i < num_locations; i++ )
      {
         for( j = 0; j < experiment_location_dimension; j++ )
            experiment_locations[ i ][ j ] *= experiment_signs[ i ];
         experiment_signs[ i ] = 1;
      }
   if( print_to_screen )
   {
      cout << "\nExperiment `locations':" << endl;
      for( i = 0; i < num_locations; i++ )
      {
         for( j = 0; j < experiment_location_dimension; j++ )
            cout << experiment_locations[ i ][ j ] << " ";
         cout << endl;
      }
      if( experiment_location_sums )
      {
         cout << "\nExperiment `location' sums:" << endl;
         for( i = 0; i < experiment_partition_size; i++ )
         {
            for( j = 0; j < experiment_location_dimension; j++ )
               cout << experiment_location_sums[ i ][ j ] << " ";
            cout << endl;
         }
      }
      if( experiment_var_multipliers && kriging_experiment_correlation )
      {
         cout << "\nExperiment variances:" << endl;
         for( i = 0; i < num_locations; i++ )
            cout << experiment_var_multipliers[ i ]<< " ";
         cout << endl;
      }
      if( experiment_location_parameters )
      {
         cout << "\nExperiment location parameters:" << endl;
         print_array( experiment_location_parameters, num_location_parameters, experiment_location_dimension );
      }
      cout << "\nExperiment correlations:" << endl;
      for( i = 0; i < num_locations; i++ )
      {
         double ev_i =  gsl_matrix_get( experiment_covariances, i, i );
         for( j = 0; j < num_locations; j++ )
         {
            double ev_j =  gsl_matrix_get( experiment_covariances, j, j );
            cout << gsl_matrix_get( experiment_covariances, i, j ) / sqrt( ev_i * ev_j ) << " ";
         }
         cout << endl;
      }
 
//      print_gsl_matrix( experiment_covariances );
   }

   if( strlen( Experiment_Locations_Filename ) )
   {
      ofstream ExpLocStream( Experiment_Locations_Filename, ios::out );
      if( ExpLocStream )
      {
         double* exp_normalisers = new double[ num_locations ];
         if( kriging_experiment_correlation )
            for( i = 0; i < num_locations; i++ )
               exp_normalisers[ i ] = 1;
         if( !principal_components )
         {
            if( !kriging_experiment_correlation )
               for( i = 0; i < num_locations; i++ )
                  exp_normalisers[ i ] = sqrt( get_experiment_location_norm( i ) );
            for( i = 0; i < num_locations; i++ )
            {
               for( j = 0; j < experiment_location_dimension; j++ )
                  ExpLocStream << experiment_locations[ i ][ j ] / exp_normalisers[ i ] << " ";
               ExpLocStream << endl;
            }
         }
         else
         {
            gsl_matrix* exp_locs = gsl_matrix_alloc( num_locations, experiment_location_dimension );
            for( i = 0; i < num_locations; i++ )
               for( j = 0; j < experiment_location_dimension; j++ )
                  gsl_matrix_set( exp_locs, i, j, experiment_locations[ i ][ j ] );
            gsl_matrix* exp_locs_pcs = calculate_principal_components( exp_locs, false );
            gsl_matrix_free( exp_locs );
            if( !kriging_experiment_correlation )
            {
               for( i = 0; i < num_locations; i++ )
               {
                  double sqd_exp_normalisers_i = 0;
                  for( j = 0; j < experiment_location_dimension; j++ )
                  {
                     double loc_j = gsl_matrix_get( exp_locs_pcs, i, j );
                     sqd_exp_normalisers_i += loc_j * loc_j;
                  }
                  exp_normalisers[ i ] = sqrt( sqd_exp_normalisers_i );
               }
            }
            for( i = 0; i < num_locations; i++ )
            {
               for( j = 0; j < experiment_location_dimension; j++ )
                  ExpLocStream << gsl_matrix_get( exp_locs_pcs, i, j ) / exp_normalisers[ i ] << " ";
               ExpLocStream << endl;
            }
         }
         delete [] exp_normalisers;
      }
   }
   if( strlen( Experiment_Signs_Filename ) )
   {
      ofstream ExpSignsStream( Experiment_Signs_Filename, ios::out );
      if( ExpSignsStream )
         for( i = 0; i < num_locations; i++ )
            ExpSignsStream << ( ( experiment_signs && kriging_experiment_correlation ) ? experiment_signs[ i ] : 1 ) << endl;
   }
   if( strlen( Experiment_Parameters_Filename ) && experiment_location_parameters )
   {
      gsl_matrix* exp_loc_pars_pcs = NULL;
      gsl_matrix* exp_loc_pars = gsl_matrix_alloc( num_location_parameters, experiment_location_dimension );
      for( i = 0; i < num_location_parameters; i++ )
         for( j = 0; j < experiment_location_dimension; j++ )
	    gsl_matrix_set( exp_loc_pars, i, j, experiment_location_parameters[ i ][ j ] );
      if( principal_components )
         gsl_matrix_multiply( exp_loc_pars_pcs, exp_loc_pars, eigen_vectors );
      ofstream ExpParsStream( Experiment_Parameters_Filename, ios::out );
      if( ExpParsStream )
         for( i = 0; i < num_location_parameters; i++ )
	 {
	    for( j = 0; j < experiment_location_dimension; j++ )
	       ExpParsStream << gsl_matrix_get( principal_components ? exp_loc_pars_pcs : exp_loc_pars, i, j ) << " ";
	    ExpParsStream << endl;
	 }
     gsl_matrix_free( exp_loc_pars );
     if( exp_loc_pars_pcs )
        gsl_matrix_free( exp_loc_pars_pcs );
   }
}

void Regression::create_residuals_matrix()
{
   if( !residuals )
      residuals = gsl_matrix_alloc( n, p_y );
}

void Regression::calculate_residuals()
{
   create_residuals_matrix();
   int num_fn_values = inputs_repeated_every ? inputs_repeated_every : regression_n;
   double* fn_values = new double[ num_fn_values ];
   double* x_row = new double[ p ];
   double ssq = 0, rssq = 0;
   int i, j;
   for( i = 0; i < num_fn_values; i ++ )
   {
      for( j = 0; j < p; j++ )
         x_row[ j ] = gsl_matrix_get( X, i, j );
      if( predict_on_data && predictions_mx )
         fn_values[ i ] = predictions_mx[ i ][ prediction_fn_value_col ];
      else
         fn_values[ i ] = function_evaluate( x_row, p );
   }
   delete [] x_row;
   for( i = 0; i < regression_n; i ++ )
   {
      current_regression_fn_val = fn_values[ inputs_repeated_every ? i % inputs_repeated_every : i ] ;
      for( j = 0; j < p_y; j++ )
      {
         double y_ij = gsl_matrix_get( Y, i, j );
         double residual_ij = y_ij - current_regression_fn_val;
         gsl_matrix_set( residuals, i, j, residual_ij );
         ssq += ( y_ij - y_bar[ j ] ) * ( y_ij - y_bar[ j ] );
         rssq += residual_ij * residual_ij;
      }
   }
   delete [] fn_values;
   r_squared = 1 - rssq / ssq;
}

void Regression::calculate_obsn_residual( int obsn )
{
   if( !residuals )
      create_residuals_matrix();
   double* x_row = new double[ p ];
   int j;
   for( j = 0; j < p; j++ )
      x_row[ j ] = gsl_matrix_get( X, obsn, j );
   current_regression_fn_val = function_evaluate( x_row, p );
   delete [] x_row;
   for( j = 0; j < p_y; j++ )
      gsl_matrix_set( residuals, obsn, j, gsl_matrix_get( Y, obsn, j ) - current_regression_fn_val );
}

gsl_matrix* Regression::get_residuals()
{
   if( !residuals )
      calculate_residuals();
   return residuals;
}

double Regression::get_r_squared()
{
   if( !residuals )
      calculate_residuals();
   return r_squared;
}

void Regression::calculate_correlation_matrix()
{
   if ( !strlen( Correl_Filename ) )
      return;
   gsl_matrix* temp2 = NULL;
   if( data_block_size > 1 )
   {
      int num_blocks = n / data_block_sizes_sum;
      gsl_matrix* temp = gsl_matrix_alloc( num_blocks, p_y * data_block_sizes_sum );
      {
         for( int i = 0; i < num_data_blocks; i++ )
	    for( int j = 0; j< data_block_sizes_sum; j++ )
               for( int j2 = 0; j2 < p_y; j2++ )
                  gsl_matrix_set( temp, i, j + j2 * data_block_sizes_sum, gsl_matrix_get( Y, i * data_block_sizes_sum + j, j2 ) );
      }
      temp2 = calculate_correlations( temp );
      gsl_matrix_free( temp );
   }
   else
      if( p_y > 1 )
         temp2 = calculate_correlations( Y );

   if( !temp2 )
      return;
   ofstream CorrelStream( Correl_Filename, ios::out );
   for( unsigned int i = 0; i < temp2->size1; i++ )
   {
      for( unsigned int j = 0; j < temp2->size2; j++)
         CorrelStream << gsl_matrix_get( temp2, i, j ) << "  ";
      CorrelStream << endl;
   }
   gsl_matrix_free( temp2 );
}

void Regression::calculate_residual_variogram()
{
   if ( !strlen( Variogram_Filename ) )
      return;
   if( !residuals )
      calculate_residuals();
   int counter;
   double temp;
   int num_blocks = n / data_block_size;
   int num_block_comparisons = ( data_block_size * ( data_block_size - 1 ) ) / 2;
   int num_comparisons = num_block_comparisons * num_blocks;
   gsl_matrix* x_diffs = gsl_matrix_alloc( num_comparisons, 1 );
   gsl_matrix* residual_diffs = gsl_matrix_alloc( num_comparisons, p_y );
   double *x1 = new double[ p ];
   double *x2 = new double[ p ];
   int j;
   for( int block = 0; block < num_blocks; block++ )
   {
      counter = 0;
      for( int i1 = 0; i1 < data_block_size - 1; i1++ )
      {
         for( j = 0; j < p; j++ )
            x1[ j ] = gsl_matrix_get( X, block * data_block_size + i1, j );
         for( int i2 = i1 + 1; i2 < data_block_size; i2++ )
         {
            for( j = 0; j < p; j++ )
               x2[ j ] = gsl_matrix_get( X, block * data_block_size + i2, j );
            temp = 0;
            for( j = 0; j < p; j++ )
               temp += ( x1[ j ] - x2[ j ] )* ( x1[ j ] - x2[ j ] );
            gsl_matrix_set( x_diffs, block * num_block_comparisons + counter, 0, temp );
            for( j = 0; j < p_y; j++ )
            {
               temp = gsl_matrix_get( residuals, block * data_block_size + i1, j ) - gsl_matrix_get( residuals, block * data_block_size + i2, j );
               gsl_matrix_set( residual_diffs, block * num_block_comparisons + counter, j, temp * temp );
            }
            counter++;
         }
      }
   }
   delete [] x1;
   delete [] x2;

   output_residual_variogram( x_diffs, residual_diffs );// If this regression is a GLM_MVN, Caclulates a dummy GLM to interpolate these variogram values, write its predictions to the Variogram_Filename.
   gsl_matrix_free( x_diffs );
   gsl_matrix_free( residual_diffs );
}

void Regression::output_residual_variogram( gsl_matrix* x_diffs, gsl_matrix* residual_diffs )
{
   ofstream VarioStream( Variogram_Filename, ios::out );
   if ( !VarioStream )
   {
      cout << "ERROR: Couldn't open residual variogram output file"<< endl;
      exit( 1 );
   }
   int num_comparisons = x_diffs->size1;
   for( int i = 0; i < num_comparisons; i++ )
   {
      double x_val = gsl_matrix_get( x_diffs, i, 0 );
      if( normalise_predictors )
         x_val *= x_scale[ 0 ];
      for( int j = 0; j < p_y; j++ )
         VarioStream << x_val << "  " << gsl_matrix_get( residual_diffs, i, j ) << endl;
   }
}

void Regression::create_gauss_laguerre_matrices()
{
   x_gl = scan_data( "gauss-laguerre_abscissae.txt" );
   gsl_matrix_scale( x_gl, gauss_laguerre_multiplier );
   w_gl = scan_data( "gauss-laguerre_weights.txt" );
   gsl_matrix_scale( w_gl, gauss_laguerre_multiplier );
   dim_gl = x_gl->size1;
}

double Regression::gauss_laguerre_integrate( double (*integrand)( double ) )
{
   double integral = 0;
   for( int i = 0; i < dim_gl; i++ )
      integral += gsl_matrix_get( w_gl, i, 0 ) * (*integrand)( gsl_matrix_get( x_gl, i, 0 ) );
   return integral;
}

double Regression::gauss_laguerre_integrate( double (*integrand)( int ) )
{
   double integral = 0;
   for( int i = 0; i < dim_gl; i++ )
      integral += gsl_matrix_get( w_gl, i, 0 ) * (*integrand)( i );
   return integral;
}

double Regression::gauss_laguerre_integrate()
{
   double integral = 0;
   for( int i = 0; i < dim_gl; i++ )
      integral += gsl_matrix_get( w_gl, i, 0 ) * gauss_laguerre_integrand( i );
   return integral;
}

bool Regression::with_probability( double d )
{
   if( d > 1 || gsl_ran_flat( rn, 0, 1 ) < d )
      return true;
   return false;
}

void Regression::set_active_dimension( int activ_dim ){
   active_dimension = activ_dim;
}

void Regression::count_subject_experiment_missing_values()
{
   zap_array( subject_exp_num_missing_values, num_subjects, num_experiments );
   create_zeroed_array( subject_exp_num_missing_values, num_subjects, num_experiments );
   for( int ii = 0; ii < num_subjects; ii++ )
   {
      int start = get_subject_index( ii );
      int end = get_subject_index( ii + 1 );
      int row_counter = start * data_block_sizes_sum / num_data_block_sizes;
      for( int i0 = start; i0 < end; i0++ )
      {
         int exp_num = get_experiment_number( i0 );
         for( int i = row_counter; i < row_counter + data_block_size; i++ )
            for( int j = 0; j < p_y; j++ )
               if( missing_values[ i ][ j ] )
                  subject_exp_num_missing_values[ ii ][ exp_num ]++;
         row_counter += data_block_size;
      }
   }
}

void Regression::get_pairwise_sample_variances()
{
    if( print_to_screen )
        cout << "Calculating all " << num_subjects*(num_subjects - 1)*num_experiments/2 << " subject pairwise variances..." << endl;
    ofstream PairwiseVariancesStream;
    if( strlen(Pairwise_Variances_Filename) )
       PairwiseVariancesStream.open( Pairwise_Variances_Filename, ios::out );
//    int num_pairs = num_data_blocks * (num_subjects - 1)/ 2;
    vector<double> pairwise_variances;//( num_pairs );
//    int pair_count = 0;
    for( int subj1 = 0; subj1 < num_subjects - 1; subj1++ )
    {
        int start_sub1 = get_subject_index( subj1 );
        int end_sub1 = get_subject_index( subj1 + 1 );
        int row_start1 = start_sub1 * data_block_sizes_sum / num_data_block_sizes;
        for( int subj2 = subj1 + 1; subj2 < num_subjects; subj2++ )
        {
            int start_sub2 = get_subject_index( subj2 );
            int end_sub2 = get_subject_index( subj2 + 1 );
            int row_start2 = start_sub2 * data_block_sizes_sum / num_data_block_sizes;
            int i01 = start_sub1, i02 = start_sub2;
            int counter = 0;
            while( (i01 < end_sub1) && (i02 < end_sub2) )
            {
                double ssq = 0;
                int expnum1 = get_experiment_number( i01 );
                if( expnum1 == get_experiment_number( i02 ) )
                {
                    int missing_count = 0;
                    for( int i = 0; i < data_block_size; i++ )
                    {
                if( !missing_values || (!missing_values[ row_start1 + counter + i ][ 0 ] && !missing_values[ row_start2 + counter + i ][ 0 ]) )
                        {
                            double diff = (*gsl_matrix_ptr( Y, row_start1 + counter + i, 0 ) - *gsl_matrix_ptr( Y, row_start2 + counter + i, 0 ) ) / 2;
                            ssq += diff * diff;
                        }
                        else
                            missing_count++;
                     }
                     counter += data_block_size;
                     if( missing_count != data_block_size )
                     {
                         ssq /= data_block_size-missing_count;
                         pairwise_variances.push_back( ssq );//[ pair_count ] = ssq;
//                         pair_count++;
                         if( PairwiseVariancesStream.is_open() )
                            PairwiseVariancesStream << ssq << endl;
                     }
                }
                i01++;
                i02++;
            }
        }
    }
    PairwiseVariancesStream.close();
    sort(pairwise_variances.begin(),pairwise_variances.end());
    LQ_pairwise_variance = pairwise_variances[ (pairwise_variances.size()+1)/4 - 1 ];
    median_pairwise_variance = pairwise_variances[ (pairwise_variances.size()+1)/2 - 1 ];
}
               
gsl_matrix* matrix_join( gsl_matrix*& m1, gsl_matrix* m2, bool side, bool new_matrix )
{
   int m2_rows = m2->size1;
   int m2_cols = m2->size2;
   int m1_rows = 0;
   int m1_cols = 0;
   if( m1 )
   {
      m1_rows = m1->size1;
      m1_cols = m1->size2;
   }
   else
      if( side )
         m1_rows = m2_rows;
      else
         m1_cols = m2_cols;
   gsl_matrix* new_m1 = NULL;
   int i, j;
   if( side )
   {
      if( m1_rows != m2_rows )
      {
         cout << " Cant join matrices together with different numbers of rows side to side" << endl;
         exit(1);
      }
      new_m1 = gsl_matrix_alloc( m1_rows, m1_cols + m2_cols );
      for( i = 0; i < m1_rows; i++ )
      {
         for( j = 0; j < m1_cols; j++ )
            gsl_matrix_set( new_m1, i, j, gsl_matrix_get( m1, i, j ) );
         for( j = 0; j < m2_cols; j++ )
            gsl_matrix_set( new_m1, i, m1_cols + j, gsl_matrix_get( m2, i, j ) );
      }
   }
   else
   {
      if( m1_cols != m2_cols )
      {
         cout << " Cant join matrices together with different numbers of columns top to bottom" << endl;
         exit(1);
      }
      new_m1 = gsl_matrix_alloc( m1_rows + m2_rows, m1_cols );
      for( j = 0; j < m1_cols; j++ )
      {
         for( i = 0; i < m1_rows; i++ )
            gsl_matrix_set( new_m1, i, j, gsl_matrix_get( m1, i, j ) );
         for( i = 0; i < m2_rows; i++ )
            gsl_matrix_set( new_m1, m1_rows + i, j, gsl_matrix_get( m2, i, j ) );
      }
   }
   if( new_matrix )
      return new_m1;
   switch_values( new_m1, m1 );
   if( new_m1 )
      gsl_matrix_free( new_m1 );
   return NULL;
}

void matrix_delete_column( gsl_matrix*& m, int col )
{
   int m_new_rows = m->size1;
   int m_new_cols = m->size2 - 1;
   if( m_new_cols < 1 || col >= m_new_cols + 1 )
   {
      cout << "Can't delete column " << col << " , matrix doesn't have that many columns"<< endl;
      return;
   }
   gsl_matrix* mcopy = m;
   m = gsl_matrix_alloc( m_new_rows, m_new_cols );
   int j;
   for( int i = 0; i < m_new_rows; i++ )
   {
      for( j = 0; j < col; j++ )
         gsl_matrix_set( m, i, j, gsl_matrix_get( mcopy, i, j ) );
      for( j = col; j < m_new_cols; j++ )
         gsl_matrix_set( m, i, j, gsl_matrix_get( mcopy, i, j + 1 ) );
   }
   gsl_matrix_free( mcopy );
}

gsl_matrix* matrix_block( gsl_matrix*& m, int n_times, bool overwrite )
{
   int m_rows = m->size1;
   int m_cols = m->size2;
   gsl_matrix* m_block = gsl_matrix_calloc( m_rows * n_times, m_cols * n_times );
   for( int i0 = 0; i0 < n_times; i0++ )
      for( int i = 0; i < m_rows; i++ )
         for( int j = 0; j < m_cols; j++ )
            gsl_matrix_set( m_block, i0 * m_rows + i, i0 * m_cols + j, gsl_matrix_get( m, i, j ) );
   if( overwrite )
   {
      gsl_matrix_free( m );
      m = m_block;
   }
   return m_block;
}

gsl_matrix* matrix_block( gsl_matrix*& m1, gsl_matrix*& m2, bool overwrite_first, bool delete_second )
{
   int m1_rows = 0, m1_cols = 0, m2_rows = 0, m2_cols = 0;
   if( m1 )
   {
      m1_rows = m1->size1;
      m1_cols = m1->size2;
   }
   if( m2 )
   {
      m2_rows = m2->size1;
      m2_cols = m2->size2;
   }
   int i, j;
   gsl_matrix* m_block = gsl_matrix_calloc( m1_rows + m2_rows, m1_cols +m2_cols );
   for( i = 0; i < m1_rows; i++ )
      for( j = 0; j < m1_cols; j++ )
         gsl_matrix_set( m_block, i, j, gsl_matrix_get( m1, i, j ) );
   for( i = 0; i < m2_rows; i++ )
      for( j = 0; j < m2_cols; j++ )
         gsl_matrix_set( m_block, m1_rows + i, m1_cols + j, gsl_matrix_get( m2, i, j ) );
  if( overwrite_first )
   {
      if( m1 )
	 gsl_matrix_free( m1 );
      m1 = m_block;
   }
  if( delete_second )
   {
      if( m2 )
	 gsl_matrix_free( m2 );
      m2 = NULL;
   }
   return m_block;
}

gsl_matrix* matrix_block( list<gsl_matrix*>& m_list, bool matrices_same_size )
{
   gsl_matrix* m_block = NULL;
   int list_size = m_list.size();
   int row_counter = 0, col_counter = 0;
   int m_rows = 0, m_cols = 0;
   for( list<gsl_matrix*>::iterator iter = m_list.begin(); iter != m_list.end(); ++iter )
   {
      gsl_matrix* m = *iter;
      if( !matrices_same_size )
         matrix_block( m_block, m, true, false );
      else
      {
         if( iter == m_list.begin() )
         {
            m_rows = m->size1;
            m_cols = m->size2;
            m_block = gsl_matrix_calloc( m_rows * list_size, m_cols * list_size );
         }
         for( int i = 0; i < m_rows; i++ )
            for( int j = 0; j < m_cols; j++ )
               gsl_matrix_set( m_block, row_counter + i, col_counter + j, gsl_matrix_get( m, i, j ) );
         row_counter += m_rows;
         col_counter += m_cols;
      }
   }
   return m_block;
}

void get_rows( gsl_matrix* m_src, int row_start, int row_end, gsl_matrix*& m_tgt )
{
   int m_src_rows = m_src->size1;
   if( m_src_rows <= row_end )
      row_end = m_src_rows - 1;
   int m_src_cols = m_src->size2;
   int new_rows = row_end - row_start + 1;
   if( m_tgt && ( (int)m_tgt->size1 != new_rows || (int)m_tgt->size2 != m_src_cols ) )
      gsl_matrix_free( m_tgt );
   if( !m_tgt )
      m_tgt = gsl_matrix_alloc( new_rows, m_src_cols );
   for( int i = 0; i < new_rows; i++ )
         for( int j = 0; j < m_src_cols; j++ )
            gsl_matrix_set( m_tgt, i, j, gsl_matrix_get( m_src, row_start + i, j ) );
}

void get_submatrix( gsl_matrix* m_src, bool* row_included, bool* col_included, gsl_matrix*& m_tgt, bool negate_row, bool negate_col )
{
   int m_src_rows = m_src->size1;
   int m_src_cols = m_src->size2;
   int m_tgt_rows = 0, m_tgt_cols = 0;
   int i, j;
   if( m_tgt ){
      m_tgt_rows = m_tgt->size1;
      m_tgt_cols = m_tgt->size2;
   }
   else{
      for( i = 0; i < m_src_rows; i++ )
         if( row_included[ i ] != negate_row )
            m_tgt_rows++;
      for( j = 0; j < m_src_cols; j++ )
         if( col_included[ j ] != negate_col )
            m_tgt_cols++;
       m_tgt = gsl_matrix_calloc( m_tgt_rows, m_tgt_cols );
   }

   int row_counter = 0, col_counter;
   for( i = 0; i < m_src_rows; i++ )
   {
      if( row_included[ i ] != negate_row )
      {
         col_counter = 0;
         for( j = 0; j < m_src_cols; j++ )
            if( col_included[ j ] != negate_col )
               gsl_matrix_set( m_tgt, row_counter, col_counter++, gsl_matrix_get( m_src, i, j ) );
         row_counter++;
      }          
   }
}

void draw_histogram( double* v, int length_v )
{
   double vsum = 0;
   int non_zero_count = 0;
   int i, j;
   for( i = 0; i < length_v; i++ )
   {
      vsum += v[ i ];
      if( v[ i ] > 0 )
	 non_zero_count++;
   }
   if( !vsum )
      return;
   int gran = 50;
   int bin_count_threshold = 10;
   double prob_threshold = 0.01;//don't show any probs less than this if there are lots of bins (more than bin_count_threshold).
   double threshold = 0;
   if( non_zero_count > bin_count_threshold )
     threshold = prob_threshold * vsum;
   int number_of_figures = (int)floor( log10( (double)length_v ) );
   for( i = 0; i < length_v; i++ )
      if( v[ i ] > threshold )
      {
         double v_i = v[ i ] / vsum;
         for( j = 1; j <= number_of_figures; j++ )
            if( i < pow( 10, (double)j ) )
               cout << " ";
         cout << i << "          ";
         j = 1;
         while( j <= gran && v_i > (double)j / (double)( gran + 1 ) )
         {
               cout << "*";
               j++;
         }
         if( j > 1 )
            cout << "  ";
         cout << "(" << v_i << ")" << endl;
      }
   cout << endl;
}

void my_invert( gsl_matrix m, gsl_matrix* m_inv )
{
   gsl_permutation* perm = gsl_permutation_calloc( m.size1 );
   int sign;
   gsl_linalg_LU_decomp( &m, perm, &sign );
   if( m_inv->size1 != m.size1 || m_inv->size2 != m.size2 )
   {
      cout << "Matrices for inversion not the same dimension, exiting" << endl;
      exit( 1 );
   }
   gsl_linalg_LU_invert( &m, perm, m_inv );
   gsl_permutation_free( perm );
}

void my_invert( gsl_matrix* m, gsl_matrix* m_inv )
{
   gsl_permutation* perm = gsl_permutation_calloc( m->size1 );
   int sign;
   gsl_matrix* mcopy = gsl_matrix_alloc( m->size1, m->size2 );
   gsl_matrix_memcpy( mcopy, m );
   gsl_linalg_LU_decomp( mcopy, perm, &sign );
   if( m_inv->size1 != m->size1 || m_inv->size2 != m->size2 )
   {
      cout << "Matrices for inversion not the same dimension, exiting" << endl;
      exit( 1 );
   }
   gsl_linalg_LU_invert( mcopy, perm, m_inv );
   gsl_matrix_free( mcopy );
   gsl_permutation_free( perm );
}

double my_lndet( gsl_matrix* m )
{
   gsl_permutation* perm;// = new gsl_permutation;
   int sign;
   perm = gsl_permutation_calloc( m->size1 );
   gsl_matrix* m_copy = gsl_matrix_alloc( m->size1, m->size2 );
   gsl_matrix_memcpy( m_copy, m );
   gsl_linalg_LU_decomp( m_copy, perm, &sign );
   gsl_permutation_free( perm );
   double lndet = gsl_linalg_LU_lndet( m_copy );
   gsl_matrix_free( m_copy );
   return lndet;
}

bool diagonalise( gsl_matrix* m, double threshold )
{
   if( threshold > 0 )
   {
      for( unsigned int i = 0; i < m->size1; i++ )
         for( unsigned int j = 0; j < m->size2; j++ )
            if( i != j && fabs( gsl_matrix_get( m, i, j ) ) > threshold )
               return false;
   }
   for( unsigned int i = 0; i < m->size1; i++ )
      for( unsigned int j = 0; j < m->size2; j++ )
         if( i != j )
            gsl_matrix_set( m, i, j, 0 );
   return true;
}

void square_matrix_stripe( gsl_matrix* m, int offset, double val )
{
   int m_rows = m->size1;
   if( offset > m_rows )
   {
      cout << "cannot stripe matrix starting at this offset, outside range" << endl;
      exit(1);
   }
   for( int i = 0; i < m_rows - offset; i++ )
   {
      gsl_matrix_set( m, i + offset, i, val );
      gsl_matrix_set( m, i, i + offset, val );
   }
}

void matrix_square_symmetric_strip( gsl_matrix* m, int row, int col, int length, double val )
{
   int m_rows = m->size1;
   if( row + length > m_rows || col + length > m_rows )
   {
      cout << "cannot strip matrix starting at this (row,col), will end up outside range" << endl;
      exit(1);
   }
   for( int i = 0; i < length; i++ )
   {
      gsl_matrix_set( m, row + i, col + i, val );
      gsl_matrix_set( m, col + i, row + i, val );
   }
}

bool gsl_matrices_differ( const gsl_matrix* m1, const gsl_matrix* m2 )
{
   if( !m1 )
   {
      if( !m2 )
         return false;
   }
   else if( !m2 )
      return true;
   if( m1->size1 != m2->size1 || m1->size2 != m2->size2 )
      return true;
   for( unsigned int i = 0; i < m1->size1; i++ )
      for( unsigned int j = 0; j < m1->size2; j++ )
         if( gsl_matrix_get( m1, i, j ) != gsl_matrix_get( m2, i, j ) )
            return true;
   return false;
}

bool matrix_cols_equal_up_to_pairs( const gsl_matrix* m, int col1, int col2 )
{
   for( unsigned int i = 0; i < (m->size1 - 1); i += 2 )
      if( gsl_matrix_get( m, i, col1 ) != gsl_matrix_get( m, i, col2 ) )
      {
	if( gsl_matrix_get( m, i + 1, col1 ) != gsl_matrix_get( m, i, col2 ) || gsl_matrix_get( m, i, col1 ) != gsl_matrix_get( m, i + 1, col2 ))
          return false;
      }
      else if( gsl_matrix_get( m, i + 1, col1 ) != gsl_matrix_get( m, i + 1, col2 ) )
	 return false;
   return true;
}

bool matrix_cols_almost_equal( const gsl_matrix* m, int col1, int col2, double tolerance )
{
   double tolerance_max = tolerance * m->size1;
   int difference_count = 0;
   for( unsigned int i = 0; i < m->size1; i++ )
      if( gsl_matrix_get( m, i, col1 ) != gsl_matrix_get( m, i, col2 ) )
      {
         difference_count++;
         if( difference_count > tolerance_max )
            return false;
      }
   return true;
}

void gram_schmidt_process( gsl_matrix* M, double** coeff_matrix, bool orthonormalise, unsigned int gram_rows, unsigned int num_blocks, unsigned int* blocks, int orthonormlised_blocks )
{
   unsigned int rows = M->size1;
   unsigned int cols = M->size2;
   if( !gram_rows )
      gram_rows = rows;
   unsigned int i, j, k;
   for( i = 0; i < cols; i++ )
      for( j = 0; j < cols; j++ )
         if( orthonormalise || i != j )
            coeff_matrix[ i ][ j ] = 0;
         else
            coeff_matrix[ i ][ j ] = 1;
//    double norm_i = 0;
//    for( k = 0; k < gram_rows; k++ )
//    {
//       double Mk0 = gsl_matrix_get( M, k, 0 );
//       norm_i += Mk0 * Mk0;
//    }
//    norm_i = sqrt( norm_i );
//    for( k = 0; k < rows; k++ )
//       gsl_matrix_set( M, k, 0, gsl_matrix_get( M, k, 0 ) / norm_i );
//    coeff_matrix[ 0 ][ 0 ] = 1.0 / norm_i;
//    for( k = 1; k < cols; k++ )
//       coeff_matrix[ 0 ][ k ] = 0;//first column of M just rescaled.

   double* square_norms = new double[ cols ];
   i = 0;
   for( unsigned int bk = 0; bk < num_blocks; bk++ )
   {
      unsigned int block_start = i;
      for( unsigned int counter = 0; counter < (blocks ? blocks[ bk ]:cols/num_blocks); counter++ )
      {
         double norm_i = 0;
         for( j = block_start; j < i; j++ )
         {
            double dot_product_ij = 0;
            for( k = 0; k < gram_rows; k++ )
               dot_product_ij += gsl_matrix_get( M, k, i ) * gsl_matrix_get( M, k, j );
            dot_product_ij /= square_norms[ j ];
            coeff_matrix[ i ][ j ] = -dot_product_ij;
         }
         for( k = 0; k < rows; k++ )
         {
            double Mki= gsl_matrix_get( M, k, i );
            if( i > block_start )
            {
               for( j = block_start; j < i; j++ )
                  Mki+= coeff_matrix[ i ][ j ] * gsl_matrix_get( M, k, j );
               gsl_matrix_set( M, k, i, Mki );
            }
            if( k < gram_rows && orthonormalise )
               norm_i += Mki * Mki;
         }
         square_norms[ i ] = norm_i;
         if( orthonormalise )
         {
            double block_scaler = orthonormlised_blocks * gram_rows / rows;
            square_norms[ i ] = block_scaler;
            norm_i = sqrt( norm_i / block_scaler );
            for( k = 0; k < rows; k++ )
               *gsl_matrix_ptr( M, k, i ) /= norm_i;
//               gsl_matrix_set( M, k, i, gsl_matrix_get( M, k, i ) / norm_i );
            for( j = block_start; j < i; j++ )
               coeff_matrix[ i ][ j ] /= norm_i;
            coeff_matrix[ i ][ i ] = 1.0 / norm_i;
         }
//      for( j = (i+1); j < cols; j++ )
//         coeff_matrix[ i ][ j ] = 0;
         for( j = block_start; j < i; j++ )//go back and correct the coefficients (from us to vs).
         {
            double coeff_ij = 0;
            for( k = j; k <=i-1; k++ )
               coeff_ij += coeff_matrix[ k ][ j ] * coeff_matrix[ i ][ k ];
            coeff_matrix[ i ][ j ] = coeff_ij;
         }
         i++;
      }
   }
   delete [] square_norms;
}

double sum_gsl_matrix( gsl_matrix*& m )
{
   double m_sum = 0;
   for( unsigned int i = 0; i < m->size1; i++ )
      for( unsigned int j = 0; j < m->size2; j++ )
	 m_sum += gsl_matrix_get( m, i, j );
   return m_sum;
}
